/*
**   IRC - Internet Relay Chat, source/access.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1995-1996 Philippe Levan
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "access.h"
#include "cidr.h"

RCSID("$Id: access.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

typedef struct _findstruct	FINDSTRUCT;
struct _findstruct
{
	CONFENTRY	*ceptr;
	CONFPTR		*matching;
	CONFPTR		*current;
	int			num_matched;
};

static char	_conf_empty[] = "";

static int _access_sortable(char *hostname);
static char *_access_reverse(char *hostname);
static void _access_entry_free(CONFHOSTIPENTRY *entry);

static int _access_add_host_linear(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);
static int _access_add_host_recurs(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);
static int _access_add_host(ACCESS *access,
			CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);

static int _access_match_host_linear(FINDSTRUCT *findstruct, char *pattern,
		CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched);
static int _access_match_host(FINDSTRUCT *findstruct, char *pattern,
		CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched);

#if 0
static int _access_add_ip_linear(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);
#endif
static int _access_add_ip_sorted(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);
static int _access_add_ipaddr(ACCESS *access,
			CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr);

#if 0
static int _access_match_ip_linear(FINDSTRUCT *findstruct,
	char *ipstring, CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched);
#endif
static int _access_match_ip(FINDSTRUCT *findstruct, ircd_uint32_t ipaddr,
		CONFHOSTIPENTRY **entryptr, u_int flags);

static void _access_clear_host_linear(ACCESS *access, CONFENTRY *ce, u_int flags);
static void _access_clear_host_sorted(ACCESS *access, CONFENTRY *ce, u_int flags);
#if 0
static void _access_clear_ip_linear(ACCESS *access, CONFENTRY *ce, u_int flags);
#endif
static void _access_clear_ip_sorted(ACCESS *access, CONFENTRY *ce, u_int flags);

static void _access_flush_host_linear(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags);
static void _access_flush_host_sorted(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags);
#if 0
static void _access_flush_ip_linear(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags);
#endif
static void _access_flush_ip_sorted(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags);

static void _access_add_to_list(CONFHOSTIPENTRY **bucket, CONFHOSTIPENTRY *entry);
static void _access_del_from_list(CONFHOSTIPENTRY **bucket, CONFHOSTIPENTRY *entry);


/*
**
** returns:
**    0 if not sortable
**    1 if sortable as-is (forward)
**   -1 if sortable if reversed..
**
** This function ripped from Dianora
*/
static int _access_sortable(char *hostname)
{
	int		state = 0;
	char	*p = hostname;

	if (!p)
		return 0;
    for(;;)
    {
		if (*p == '?') /* no '?'s in sortable ? */
			return 0;
		switch(state)
		{
			case 0:
				if (*p == '*')
					state = 1;
				else
					state = 4;
				break;
			case 1:
				if (!*p) /* only wildcards...hm... */
					return 0;
				else if (*p != '*')
					state = 2;
				break;
			case 2:
				if (!*p)
					return -1; /* "*whatever" */
				else if (*p == '*')
					return 0;
				break;
			case 4:
				if (*p == '*')
					state = 5;
				else if (!*p)
					return 1; /* no wildcards */
                break;
            case 5:
				if (!*p)
					return 1; /* "whatever*" */
   				else if (*p != '*')
					return 0; /* "whatever*s" */
                break;
            default:
				state = 0;
				break;
		}
		p++;
	}
	/* NOT REACHED */
	return 0;
}

static char *_access_reverse(char *hostname)
{
	char	*ptr = hostname;
	char	*ptr2;
	char	*str;

	if (!*hostname)
		return MyStrdup(hostname);
	while(*ptr)
		ptr++;
	ptr2 = str = (char *)MyMalloc(ptr-hostname+1);
	do
	{
		*ptr2++ = *--ptr;
	} while (ptr > hostname);
	*ptr2 = '\0';
	return str;
}

static void _access_entry_free(CONFHOSTIPENTRY *entry)
{
	if (!entry)
		return;
	if (entry->pattern)
		MyFree(entry->pattern);
	if (entry->password)
		MyFree(entry->password);
	if (entry->comment)
		MyFree(entry->comment);
	if (entry->username)
		MyFree(entry->username);
	if (entry->hostname)
		MyFree(entry->hostname);
	if (entry->ipstring)
		MyFree(entry->ipstring);
	if (entry->data)
		MyFree(entry->data);
	if (entry->classname)
		MyFree(entry->classname);
	MyFree(entry);
}

static int _access_add_host_linear(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	CONFPTR			*cptr;
	CONFHOSTIPENTRY	*e;

	if (flags & ACCESS_FLAGS_CHECKEXISTS)
	{
		for(cptr=ce->table;cptr;cptr=cptr->next)
		{
			e = cptr->entry;
			if (!match(e->hostname, entry->hostname))
			{
				if (!e->username)
				{
					if (existptr)
						*existptr = e;
					return -EEXIST;
				}
				if (entry->username && !match(e->username, entry->username))
				{
					if (existptr)
						*existptr = e;
					return -EEXIST;
				}
			}
		}
	}
	cptr = (CONFPTR *)MyMalloc(sizeof(CONFPTR));
	cptr->entry = entry;
	cptr->sub = NULL; /* not used for linear */
	cptr->next = ce->table;
	ce->table = cptr;
	ce->num++;
	return 0;
}

static int _access_add_host_recurs(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	int			lower;
	int			upper;
	int			compare;
	int			pos = -1;

	if (!ce->num)
	{
		if (!ce->size)
		{
			ce->size = 20;
			ce->table = (CONFPTR *)MyMalloc(sizeof(CONFPTR)*ce->size);
			BZERO(ce->table, sizeof(CONFPTR)*ce->size);
		}
		ce->table[0].entry = entry;
		ce->num++;
		return 0;
	}

	lower = 0;
	upper = ce->num-1;
	compare = upper/2;
	while(lower != upper)
	{
		pos = strcasecmp(entry->pattern, ce->table[compare].entry->pattern);
		if (!pos||(!match(entry->pattern, ce->table[compare].entry->pattern) ||
				!match(ce->table[compare].entry->pattern, entry->pattern)))
		{
			lower = upper = compare;
			break;
		}
		else
		{
			if (pos < 0)
				upper = compare;
			else
			{
				if (lower == compare)
					lower = compare+1;
				else
					lower = compare;
			}
		}
		compare = (lower+upper)/2;
	}

	if (pos)
		pos = strcasecmp(entry->pattern, ce->table[compare].entry->pattern);

	if (!pos)
	{
		CONFPTR	*cptr;

		if (flags & ACCESS_FLAGS_CHECKEXISTS)
		{
			if (!ce->table[compare].entry->username)
				return -ENOENT;
			if (entry->username &&
					!match(ce->table[compare].entry->username,
							entry->username))
				return -ENOENT;
		}
		cptr = (CONFPTR *)MyMalloc(sizeof(CONFPTR));
		cptr->entry = entry;
		cptr->next = ce->table[compare].next;
		cptr->sub = NULL;
		ce->table[compare].next = cptr;
		return 0;
	}

	if (!match(ce->table[compare].entry->pattern, entry->pattern))
	{
		if (!ce->table[compare].sub)
		{
			ce->table[compare].sub = (CONFENTRY *)MyMalloc(sizeof(CONFENTRY));
			BZERO(ce->table[compare].sub, sizeof(CONFENTRY));
		}
		return _access_add_host_recurs(ce->table[compare].sub,
					entry, flags, existptr);
	}
	if (!match(entry->pattern, ce->table[compare].entry->pattern))
	{
		int			bottom = compare;
		int			top = compare;
		CONFENTRY	*newce;

		while ((bottom > 0) &&
					!match(entry->pattern, ce->table[bottom-1].entry->pattern))
			bottom--;
		while ((top < ce->num-1) &&
					!match(entry->pattern, ce->table[top+1].entry->pattern))
			top++;

		newce = (CONFENTRY *)MyMalloc(sizeof(CONFENTRY));
		newce->num = top-bottom+1;
		newce->size = newce->num + 5;
		newce->table = (CONFPTR *)MyMalloc(sizeof(CONFPTR)*newce->size);
		BCOPY(&(ce->table[bottom]), newce->table,
				newce->num*sizeof(CONFPTR));
		BCOPY(&(ce->table[top+1]),
				&(ce->table[bottom+1]),
				(ce->num-top-1) * sizeof(CONFPTR));
		ce->table[bottom].entry = entry;
		ce->table[bottom].sub = newce;
		ce->table[bottom].next = NULL;
		ce->num -= (top-bottom);
		return 0;
	}

	if ((ce->size - ce->num) < 5)
	{
		ce->size += 10;
		ce->table = MyRealloc(ce->table, sizeof(CONFPTR)*ce->size);
		BZERO((ce->table+ce->size-10), sizeof(CONFPTR)*10);
	}

	if (pos > 0)
	{
		BCOPY(&(ce->table[compare+1]), &(ce->table[compare+2]),
				sizeof(CONFPTR)*(ce->num-1-compare));
		ce->table[compare+1].entry = entry;
		ce->table[compare+1].next = NULL;
		ce->table[compare+1].sub = NULL;
	}
	else
	{
		BCOPY(&(ce->table[compare]), &(ce->table[compare+1]),
				sizeof(CONFPTR)*(ce->num-compare));
		ce->table[compare].entry = entry;
		ce->table[compare].next = NULL;
		ce->table[compare].sub = NULL;
	}
	ce->num++;
	return 0;
}

static int _access_add_host(ACCESS *access,
			CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	CONFHOST	*ch = access->conf_host;
	CONFENTRY	*ce = &(ch->rev_table);
	int			ret;

	switch(_access_sortable(entry->hostname))
	{
		case 0:	/* can't sort! */
			entry->pattern = NULL;
			return _access_add_host_linear(&(ch->lin_list), entry,
					flags, existptr);
		case 1:		/* forward */
			entry->pattern = MyStrdup(entry->hostname);
			ce = &(ch->for_table);
			break;
		case -1:	/* reverse */
			entry->pattern = _access_reverse(entry->hostname);
			break;
	}
	ret = _access_add_host_recurs(ce, entry, flags, existptr);
	if (ret)
	{
		MyFree(entry->pattern);
		entry->pattern = NULL;
	}
	return ret;
}

static int _access_match_host_linear(FINDSTRUCT *findstruct,
		char *pattern, CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched)
{
	CONFENTRY	*ce;
	int			temp;

	if (!(ce = findstruct->ceptr) || !ce->num)
		return -ENOENT;
	if (findstruct->current == NULL)
		findstruct->current = ce->table;
	else
		findstruct->current = findstruct->current->next;
	for(;findstruct->current;findstruct->current=findstruct->current->next)
	{
		temp = matches(findstruct->current->entry->hostname,
					pattern);
		if (!temp)
			continue;
		*entryptr = findstruct->current->entry;
		if (num_matched)
			*num_matched = temp;
		return 0;
	}
	if (num_matched)
		*num_matched = 0;
	return -ENOENT;
}

static int _access_match_host(FINDSTRUCT *findstruct,
		char *pattern, CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched)
{
	CONFENTRY	*ce;
	int			lower;
	int			upper;
	int			compare;
	int			pos;
	int			temp;

	if (!(ce = findstruct->ceptr) || !ce->num)
		return -ENOENT;

	if (findstruct->current == NULL)
	{
		findstruct->matching = NULL;
		if (num_matched)
			*num_matched = 0;
		lower = 0;
		upper = ce->num-1;
		compare = (lower+upper)/2;

		while(lower != upper)
		{
			if ((temp=matches(ce->table[compare].entry->pattern,
							pattern)))
			{
				findstruct->matching =
						findstruct->current = &(ce->table[compare]);
				findstruct->num_matched = temp;
				if (num_matched)
					*num_matched = temp;
				*entryptr = ce->table[compare].entry;
				return 0;
			}

			pos = strcasecmp(pattern, ce->table[compare].entry->pattern);
			if (pos < 0)
				upper = compare;
			else
			{
				if (lower == compare)
					lower = compare + 1;
				else
					lower = compare;
			}
			compare = (lower+upper)/2;
		}
		if ((temp=matches(ce->table[compare].entry->pattern, pattern)))
		{
			findstruct->matching =
					findstruct->current = &(ce->table[compare]);
			findstruct->num_matched = temp;
			if (num_matched)
				*num_matched = temp;
			*entryptr = ce->table[compare].entry;
			return 0;
		}
		findstruct->num_matched = 0;
		return -ENOENT;
	}
	if (!findstruct->matching)
	{
		if (num_matched)
			*num_matched = 0;
		return -ENOENT;
	}
	if (findstruct->current->next)
	{
		findstruct->current = findstruct->current->next;
		if (num_matched)
			*num_matched = findstruct->num_matched;
		*entryptr = findstruct->current->entry;
		return 0;
	}
	if (!findstruct->matching->sub)
	{
		if (num_matched)
			*num_matched = 0;
		return -ENOENT;
	}
	findstruct->ceptr = findstruct->matching->sub;
	findstruct->current = NULL;
	return _access_match_host(findstruct,
					pattern, entryptr, flags, num_matched);
}

#if 0
static int _access_add_ip_linear(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	CONFPTR			*cptr;
	CONFHOSTIPENTRY	*e;

	if (flags & ACCESS_FLAGS_CHECKEXISTS)
	{
		for(cptr=ce->table;cptr;cptr=cptr->next)
		{
			e = cptr->entry;
			if (!match(e->ipstring, entry->ipstring))
			{
				if (!e->username)
				{
					if (existptr)
						*existptr = e;
					return -EEXIST;
				}
				if (entry->username && !match(e->username, entry->username))
				{
					if (existptr)
						*existptr = e;
					return -EEXIST;
				}
			}
		}
	}
	cptr = (CONFPTR *)MyMalloc(sizeof(CONFPTR));
	cptr->entry = entry;
	cptr->sub = NULL; /* not used for linear */
	cptr->next = ce->table;
	ce->table = cptr;
	ce->num++;
	return 0;
}
#endif

static int _access_add_ip_sorted(CONFENTRY *ce,
		CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	int			lower;
	int			upper;
	int			compare;

	if (!ce->num)
	{
		if (!ce->size)
		{
			ce->size = 20;
			ce->table = (CONFPTR *)MyMalloc(sizeof(CONFPTR)*ce->size);
			BZERO(ce->table, sizeof(CONFPTR)*ce->size);
		}
		ce->table[0].entry = entry;
		ce->num++;
		return 0;
	}

	lower = 0;
	upper = ce->num-1;
	compare = upper/2;
	while(lower != upper)
	{
		if (ce->table[compare].entry->ip_net == entry->ip_net)
		{
			lower = upper = compare;
			break;
		}
		else
		{
			if (entry->ip_net < ce->table[compare].entry->ip_net)
				upper = compare;
			else
			{
				if (lower == compare)
					lower = compare+1;
				else
					lower = compare;
			}
		}
		compare = (lower+upper)/2;
	}

	if (ce->table[compare].entry->ip_net == entry->ip_net)
	{
		CONFPTR	*cptr;

		if (flags & ACCESS_FLAGS_CHECKEXISTS)
		{
			if (!ce->table[compare].entry->username)
				return -ENOENT;
			if (entry->username &&
					!match(ce->table[compare].entry->username,
						entry->username))
				return -ENOENT;
		}
		cptr = (CONFPTR *)MyMalloc(sizeof(CONFPTR));
		cptr->entry = entry;
		cptr->next = ce->table[compare].next;
		cptr->sub = NULL;
		ce->table[compare].next = cptr;
		return 0;
	}

	if ((ce->size - ce->num) < 5)
	{
		ce->size += 10;
		ce->table = MyRealloc(ce->table, sizeof(CONFPTR)*ce->size);
		BZERO((ce->table+ce->size-10), sizeof(CONFPTR)*10);
	}

	if (entry->ip_net > ce->table[compare].entry->ip_net)
	{
		BCOPY(&(ce->table[compare+1]), &(ce->table[compare+2]),
				sizeof(CONFPTR)*(ce->num-1-compare));
		ce->table[compare+1].entry = entry;
		ce->table[compare+1].next = NULL;
		ce->table[compare+1].sub = NULL;
	}
	else
	{
		BCOPY(&(ce->table[compare]), &(ce->table[compare+1]),
				sizeof(CONFPTR)*(ce->num-compare));
		ce->table[compare].entry = entry;
		ce->table[compare].next = NULL;
		ce->table[compare].sub = NULL;
	}
	ce->num++;
	return 0;
}

static int _access_add_ipaddr(ACCESS *access,
			CONFHOSTIPENTRY *entry, u_int flags, CONFHOSTIPENTRY **existptr)
{
	CONFIP		*ci = access->conf_ip;
	int			pl;


	if (entry->ip_net == -1)
#if 1
		return -EINVAL;
#else
		return _access_add_ip_linear(&(ci->lin_list), entry, flags, existptr);
#endif

	pl = cidr_netmask_to_plength(entry->ip_mask);
	if (pl == -1)
		return -EFAULT;
	return _access_add_ip_sorted(&(ci->sort_table[entry->ip_net >> 24][pl]), entry, flags, existptr);
}

#if 0
static int _access_match_ip_linear(FINDSTRUCT *findstruct,
		char *ipstring, CONFHOSTIPENTRY **entryptr, u_int flags, int *num_matched)
{
	CONFENTRY	*ce;
	int			temp;

	if (!(ce = findstruct->ceptr) || !ce->num)
		return -ENOENT;
	if (findstruct->current == NULL)
		findstruct->current = ce->table;
	else
		findstruct->current = findstruct->current->next;
	for(;findstruct->current;findstruct->current=findstruct->current->next)
	{
		temp = matches(findstruct->current->entry->ipstring, ipstring);
		if (!temp)
			continue;
		*entryptr = findstruct->current->entry;
		if (num_matched)
			*num_matched = temp;
		return 0;
	}
	if (num_matched)
		*num_matched = 0;
	return -ENOENT;
}
#endif

static int _access_match_ip(FINDSTRUCT *findstruct, ircd_uint32_t ipaddr,
		CONFHOSTIPENTRY **entryptr, u_int flags)
{
	int			lower;
	int			upper;
	int			compare;
	int			pos = -1;
	CONFENTRY	*ce;

	if (!(ce = findstruct->ceptr) || !ce->num)
		return -ENOENT;

	if (!findstruct->current)
	{
		findstruct->matching = NULL;
		lower = 0;
		upper = ce->num-1;
		compare = (lower+upper)/2;

		while(lower != upper)
		{
			pos = cidr_match(ce->table[compare].entry->ip_net,
					ce->table[compare].entry->ip_mask, ipaddr);
			if (!pos)
			{
				lower = upper = compare;
				break;
			}
			pos = ipaddr < ce->table[compare].entry->ip_net ? -1 : 1;
			if (pos < 0)
				upper = compare;
			else
			{
				if (lower == compare)
					lower = compare + 1;
				else
					lower = compare;
			}
			compare = (lower+upper)/2;
		}
		if (pos) /* check again and below.. */
		{
			pos = cidr_match(ce->table[compare].entry->ip_net,
					ce->table[compare].entry->ip_mask, ipaddr);
			while (pos && compare)
			{
				compare--;
				if (ce->table[compare].entry->ip_net > ipaddr)
					break;
				pos = cidr_match(ce->table[compare].entry->ip_net,
					ce->table[compare].entry->ip_mask, ipaddr);
			}
			if (pos)
				return -ENOENT;
		}
		compare++;
		while(compare<ce->num)
		{
			if (cidr_match(ce->table[compare].entry->ip_net,
				ce->table[compare].entry->ip_mask, ipaddr))
				break;
			compare++;
		}
		compare--;
		findstruct->current = &(ce->table[compare]);
		findstruct->matching = findstruct->current;
		*entryptr = ce->table[compare].entry;
		return 0;
	}
	if (!findstruct->matching)
		return -ENOENT;
	if (findstruct->current->next)
	{
		findstruct->current = findstruct->current->next;
		*entryptr = findstruct->current->entry;
		return 0;
	}
	while (findstruct->matching > &(ce->table[0]))
	{
		findstruct->current = --findstruct->matching;
		if (findstruct->current->entry->ip_net > ipaddr)
			return -ENOENT;
		pos = cidr_match(findstruct->current->entry->ip_net,
					findstruct->current->entry->ip_mask, ipaddr);
		if (!pos)
		{
			*entryptr = findstruct->current->entry;
			return 0;
		}
	}
	return -ENOENT;
}

static void _access_clear_host_linear(ACCESS *access, CONFENTRY *ce, u_int flags)
{
	CONFPTR	*nextcptr;

	for(;ce->table;ce->table=nextcptr)
	{
		nextcptr = ce->table->next;
		if (ce->table->entry->ip_net == -1)
		{
			_access_del_from_list(&(access->conf_all), ce->table->entry);
			_access_entry_free(ce->table->entry);
		}
		MyFree(ce->table);
		ce->table = NULL;
	}
	ce->num = ce->size = 0;
	return;
}

static void _access_clear_host_sorted(ACCESS *access, CONFENTRY *ce, u_int flags)
{
	int		i;

	for(i=0;i<ce->num;i++)
	{
		if (ce->table[i].sub)
		{
			_access_clear_host_sorted(access, ce->table[i].sub, flags);
			MyFree(ce->table[i].sub);
			ce->table[i].sub = NULL;
		}
		while(ce->table[i].next)
		{
			CONFPTR	*n = ce->table[i].next->next;
			CONFHOSTIPENTRY	*entry = ce->table[i].next->entry;

			/*
			** no need to check ->sub...next's don't have them
			**  also, only free ->entry if it isn't in the IP table
			*/
			if (entry->ip_net == -1)
			{
				_access_del_from_list(&(access->conf_all), entry);
				_access_entry_free(entry);
			}
			MyFree(ce->table[i].next);
			ce->table[i].next = n;
		}
		if (ce->table[i].entry->ip_net == -1)
		{
			_access_del_from_list(&(access->conf_all), ce->table[i].entry);
			_access_entry_free(ce->table[i].entry);
		}
	}
	if (ce->table)
	{
		MyFree(ce->table);
		ce->table = NULL;
	}
	ce->num = 0;
	ce->size = 0;
	return;
}

#if 0
static void _access_clear_ip_linear(ACCESS *access, CONFENTRY *ce, u_int flags)
{
	CONFPTR	*nextcptr;

	for(;ce->table;ce->table=nextcptr)
	{
		nextcptr = ce->table->next;
		_access_del_from_list(&(access->conf_all), ce->table->entry);
		_access_entry_free(ce->table->entry);
		MyFree(ce->table);
		ce->table = NULL;
	}
	ce->num = ce->size = 0;
	return;
}
#endif

static void _access_clear_ip_sorted(ACCESS *access, CONFENTRY *ce, u_int flags)
{
	int		i;

	for(i=0;i<ce->num;i++)
	{
		while(ce->table[i].next)
		{
			CONFPTR			*n = ce->table[i].next->next;
			CONFHOSTIPENTRY	*entry = ce->table[i].next->entry;

			_access_del_from_list(&(access->conf_all), entry);
			_access_entry_free(entry);
			MyFree(ce->table[i].next);
			ce->table[i].next = n;
		}
		_access_del_from_list(&(access->conf_all), ce->table[i].entry);
		_access_entry_free(ce->table[i].entry);
	}
	if (ce->table)
	{
		MyFree(ce->table);
		ce->table = NULL;
	}
	ce->num = ce->size = 0;
	return;
}

static void _access_flush_host_linear(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags)
{
	CONFPTR	*lastone = NULL;
	CONFPTR	*cptr;
	CONFPTR	*nextcptr;
	
	for(cptr=ce->table;cptr;cptr=nextcptr)
	{
		nextcptr = cptr->next;
		if ((cptr->entry == chentry) ||
			(cptr->entry->expires && (cptr->entry->expires < Now)))
		{
			if (lastone == NULL)
				ce->table = nextcptr;
			else
				lastone->next = nextcptr;
			if (cptr->entry->ip_net == -1)
			{
				_access_del_from_list(&(access->conf_all), cptr->entry);
				_access_entry_free(cptr->entry);
			}
			ce->num--;
			MyFree(cptr);
			continue;
		}
		lastone = cptr;
	}
}

static void _access_flush_host_sorted(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags)
{
	CONFPTR	**tmp;
	CONFPTR	*tmp2;
	int		i;

	for(i=0;i<ce->num;i++)
	{
		if (ce->table[i].sub)
		{
			_access_flush_host_sorted(access, ce->table[i].sub, chentry, flags);
			if (!ce->table[i].sub->num)
			{
				if (ce->table[i].sub) /* should be true */
				{
					if (ce->table[i].sub->table) /* should not be true */
						MyFree(ce->table[i].sub->table);
					MyFree(ce->table[i].sub);
				}
				ce->table[i].sub = NULL;
			}
		}
		for(tmp=&(ce->table[i].next);*tmp;tmp=&((*tmp)->next))
		{
			if ((chentry == (*tmp)->entry) ||
				((*tmp)->entry->expires && ((*tmp)->entry->expires < Now)))
			{
				tmp2 = *tmp;
				*tmp = tmp2->next;
				if (tmp2->entry->ip_net == -1)
				{
					_access_del_from_list(&(access->conf_all), tmp2->entry);
					_access_entry_free(tmp2->entry);
				}
				MyFree(tmp2);
				if (!*tmp)
					break;
			}
		}
		if ((ce->table[i].entry == chentry) ||
			(ce->table[i].entry->expires&&(ce->table[i].entry->expires < Now)))
		{
			if (ce->table[i].entry->ip_net == -1)
			{
				_access_del_from_list(&(access->conf_all), ce->table[i].entry);
				_access_entry_free(ce->table[i].entry);
			}
			if (ce->table[i].next)
			{
				CONFPTR *n = ce->table[i].next;

				n->sub = ce->table[i].sub;
				BCOPY(n, &(ce->table[i]), sizeof(CONFPTR));
				MyFree(n);
			}
			else if (ce->table[i].sub)
			{
				CONFENTRY	*tmp = ce->table[i].sub;

				assert(tmp->num > 0);
				ce->size += tmp->num+1;
				ce->table = (CONFPTR *)MyRealloc(ce->table, ce->size*sizeof(CONFPTR));
				BZERO(ce->table+ce->size-(tmp->num+1), sizeof(CONFPTR)*(tmp->num+1));
				if (tmp->num != 1)
					BCOPY((ce->table+i+1), (ce->table+i+tmp->num),
						(tmp->num-1)*sizeof(CONFPTR));
				BCOPY(tmp->table, (ce->table+i), sizeof(CONFPTR)*tmp->num);
				ce->num += tmp->num-1;
				MyFree(tmp->table);
				MyFree(tmp);
			}
			else
			{
				if (ce->num-i-1)
				{
					BCOPY(ce->table+i+1, ce->table+i, sizeof(CONFPTR) * (ce->num-i-1));
					BZERO((ce->table+ce->num-1), sizeof(CONFPTR));
				}
				else
					BZERO(ce->table+i, sizeof(CONFPTR));
				i--; /* for the continue */
				ce->num--;
			}
		}
	}
	if (!ce->num)
	{
		if (ce->table)
		{
			MyFree(ce->table);
			ce->table = NULL;
		}
		ce->size = 0;
	}
	return;
}

#if 0
static void _access_flush_ip_linear(ACCESS *access, CONFENTRY *ce, u_int flags)
{
	CONFPTR	*lastone = NULL;
	CONFPTR	*cptr;
	CONFPTR	*nextcptr;
	
	for(cptr=ce->table;cptr;cptr=nextcptr)
	{
		nextcptr = cptr->next;
		if (cptr->entry->expires && (cptr->entry->expires < Now))
		{
			if (lastone == NULL)
				ce->table = nextcptr;
			else
				lastone->next = nextcptr;
			_access_del_from_list(&(access->conf_all), cptr->entry);
			_access_entry_free(cptr->entry);
			MyFree(cptr);
			continue;
		}
		lastone = cptr;
	}
}
#endif

static void _access_flush_ip_sorted(ACCESS *access, CONFENTRY *ce, CONFHOSTIPENTRY *chentry, u_int flags)
{
	CONFPTR	**tmp;
	CONFPTR	*tmp2;
	int		i;

	for(i=0;i<ce->num;i++)
	{
        if (ce->table[i].sub)
        {
			_access_flush_ip_sorted(access, ce->table[i].sub, chentry, flags);
			if (!ce->table[i].sub->num)
			{
				if (ce->table[i].sub) /* should be true */
				{
					if (ce->table[i].sub->table) /* should not be true */
						MyFree(ce->table[i].sub->table);
					MyFree(ce->table[i].sub);
				}
				ce->table[i].sub = NULL;
			}
		}
		for(tmp=&(ce->table[i].next);*tmp;tmp=&((*tmp)->next))
		{
			if ((chentry == (*tmp)->entry) ||
				((*tmp)->entry->expires && ((*tmp)->entry->expires < Now)))
			{
				tmp2 = *tmp;
				*tmp = tmp2->next;
				_access_del_from_list(&(access->conf_all), tmp2->entry);
				_access_entry_free(tmp2->entry);
				MyFree(tmp2);
				if (!*tmp)
					break;
			}
		}
		if ((chentry == ce->table[i].entry) ||
			(ce->table[i].entry->expires&&(ce->table[i].entry->expires < Now)))
		{
			_access_del_from_list(&(access->conf_all), ce->table[i].entry);
			_access_entry_free(ce->table[i].entry);
			if (ce->table[i].next)
			{
				CONFPTR *n = ce->table[i].next;

				n->sub = ce->table[i].sub;
				BCOPY(n, &(ce->table[i]), sizeof(CONFPTR));
				MyFree(n);
			}
			else if (ce->table[i].sub)
			{
				CONFENTRY	*tmp = ce->table[i].sub;

				ce->size += tmp->size+1;
				ce->table = (CONFPTR *)MyRealloc(ce->table, ce->size*sizeof(CONFPTR));
				BZERO((ce->table+ce->size-(tmp->size+1)), sizeof(CONFPTR)*(tmp->size+1));
				if (tmp->size != 1)
					BCOPY((ce->table+i+1), (ce->table+i+tmp->size),
						(tmp->size-1)*sizeof(CONFPTR));
				BCOPY(tmp->table, (ce->table+i), sizeof(CONFPTR)*tmp->size);
				ce->num += tmp->size-1;
				MyFree(tmp->table);
				MyFree(tmp);
			}
			else
			{
				if (ce->num-i-1)
				{
					BCOPY(ce->table+i+1, ce->table+i, sizeof(CONFPTR) * (ce->num-i-1));
					BZERO((ce->table+ce->num-1), sizeof(CONFPTR));
				}
				else
					BZERO(ce->table+i, sizeof(CONFPTR));
				i--; /* for the continue */
				ce->num--;
			}
		}
	}
	if (!ce->num)
	{
		if (ce->table)
		{
			MyFree(ce->table);
			ce->table = NULL;
		}
		ce->size = 0;
	}
	return;
}

static void _access_add_to_list(CONFHOSTIPENTRY **bucket, CONFHOSTIPENTRY *entry)
{
	entry->ch_prev = NULL;
	if ((entry->ch_next = *bucket) != NULL)
		entry->ch_next->ch_prev = entry;
	*bucket = entry;
}

static void _access_del_from_list(CONFHOSTIPENTRY **bucket, CONFHOSTIPENTRY *entry)
{
	if (entry->ch_prev)
		entry->ch_prev->ch_next = entry->ch_next;
	else
		*bucket = entry->ch_next;
	if (entry->ch_next)
		entry->ch_next->ch_prev = entry->ch_prev;
}

int access_init(ACCESS *access)
{
	access->conf_host = (CONFHOST *)MyMalloc(sizeof(CONFHOST));
	access->conf_ip = (CONFIP *)MyMalloc(sizeof(CONFIP));
	BZERO(access->conf_host, sizeof(CONFHOST));
	BZERO(access->conf_ip, sizeof(CONFIP));
	return 0;
}

int access_match_hostname(ACCESS *access,
		char *username, char *hostname,
		u_int flags, CONFHOSTIPENTRY **entryptr)
{
	FINDSTRUCT		findstruct;
	char			*tomatch = NULL;
	int				num_matched;
	int				num2;
	int				i;
	int				ret;
	int				savenum = 0;
	CONFHOSTIPENTRY	*eptr;
	CONFHOSTIPENTRY	*saveentry = NULL;

	if (!username)
		username = _conf_empty;

	for(i=0;i<3;i++)
	{
		switch(i)
		{
			case 0:
				findstruct.ceptr = &(access->conf_host->rev_table);
				tomatch = _access_reverse(hostname);
				break;
			case 1:
				MyFree(tomatch);
				findstruct.ceptr = &(access->conf_host->for_table);
				tomatch = hostname;
				break;
			case 2:
				findstruct.ceptr = &(access->conf_host->lin_list);
				/* tomatch already pointing to hostname */
				break;
		}
		findstruct.current = NULL;
		for(;;)
		{
			if (i == 2)
				ret = _access_match_host_linear(&findstruct, tomatch,
							&eptr, flags, &num_matched);
			else
				ret = _access_match_host(&findstruct, tomatch, &eptr, 
							flags, &num_matched);
			if (ret)
				break;
			if (eptr->expires && (eptr->expires < Now))
				continue;
			if (flags & ACCESS_FLAGS_BESTMATCH)
			{
				if (!eptr->username)
				{
					if (num_matched > savenum)
					{
						savenum = num_matched;
						saveentry = eptr;
					}
					continue;
				}
				if ((num2 = matches(eptr->username, username)) != 0)
	            {
					num_matched += num2;
					if (num_matched > savenum)
					{
						savenum = num_matched;
						saveentry = eptr;
					}
				}
				continue;
			}
			if (!eptr->username || !match(eptr->username, username))			
				break;
		}
		if (!ret)
		{
			if (!i)
				MyFree(tomatch);
			break;
		}
	}
	if (saveentry)
	{
		if (entryptr)
			*entryptr = saveentry;
		return 0;
	}
	/* ret is 0 if we have a match...and eptr pointing to the struct */
	if (ret)
		return -ENOENT;
	if (entryptr)
		*entryptr = eptr;
	return 0;
}

int access_match_ipaddr(ACCESS *access,
		char *username, char *ipstring,
		ircd_uint32_t ipaddr, u_int flags, CONFHOSTIPENTRY **entryptr)
{
	FINDSTRUCT		findstruct;
	int				i;
	int				ret;
	CONFHOSTIPENTRY	*eptr;
	CONFHOSTIPENTRY	*saveeptr = NULL;
	int				num_matched = -1;
	int				num;
#if 0
	int				num2;
#endif
	int				fo;

	if (!username)
		username = _conf_empty;

	fo = ipaddr >> 24;
	for(i=32;i>=0;i--)
	{
		findstruct.ceptr = &(access->conf_ip->sort_table[fo][i]);
		findstruct.current = NULL;

		while((ret = _access_match_ip(&findstruct, ipaddr, &eptr,
					flags)) == 0)
		{
			if (eptr->expires && (eptr->expires < Now))
				continue;
			num = 0;
			if (!eptr->username || (num=matches(eptr->username, username)))
			{
				if (!(flags & ACCESS_FLAGS_BESTMATCH))
					break;
				if (num > num_matched)
				{
					num_matched = num;
					saveeptr = eptr;
				}
			}
		}
		if (saveeptr)
		{
			if (entryptr)
				*entryptr = saveeptr;
			return 0;
		}
		if (!ret)
		{
			if (entryptr)
				*entryptr = eptr;
			return 0;
		}
	}

#if 0
	findstruct.ceptr = &(access->conf_ip->lin_list);
	findstruct.current = NULL;
	while(!(ret = _access_match_ip_linear(&findstruct,
		ipstring, &eptr, flags, &num)))
	{
		if (eptr->expires && (eptr->expires < Now))
			continue;
		num2 = 0;
		if (!eptr->username || (num2=match(eptr->username, username)))
		{
			if (!(flags & ACCESS_FLAGS_BESTMATCH))
				break;
			num+=num2;
			if (num > num_matched)
			{
				num_matched = num;
				saveeptr = eptr;
			}
		}
	}
	if (saveeptr)
	{
		if (entryptr)
			*entryptr = saveeptr;
		return 0;
	}
	if (!ret)
	{
		if (entryptr)
			*entryptr = eptr;
		return 0;
	}
#endif
	return -ENOENT;
}

int access_match_ipaddr_default(ACCESS *access,
		char *username, char *ipstring,
		ircd_uint32_t ipaddr, u_int flags, CONFHOSTIPENTRY **entryptr)
{
	FINDSTRUCT		findstruct;
	int				ret;
	CONFHOSTIPENTRY	*eptr;
	CONFHOSTIPENTRY	*saveeptr = NULL;
	int				num_matched = -1;
	int				num;

	if (!username)
		username = _conf_empty;

	findstruct.ceptr = &(access->conf_ip->sort_table[0][0]);
	findstruct.current = NULL;

	while((ret = _access_match_ip(&findstruct, ipaddr, &eptr, flags)) == 0)
	{
		if (eptr->expires && (eptr->expires < Now))
			continue;
		num = 0;
		if (!eptr->username || (num=matches(eptr->username, username)))
		{
			if (!(flags & ACCESS_FLAGS_BESTMATCH))
				break;
			if (num > num_matched)
			{
				num_matched = num;
				saveeptr = eptr;
			}
		}
	}
	if (saveeptr)
	{
		if (entryptr)
			*entryptr = saveeptr;
		return 0;
	}
	if (!ret)
	{
		if (entryptr)
			*entryptr = eptr;
		return 0;
	}
	return -ENOENT;
}

int access_add(ACCESS *access, char *username, char *hostname,
		char *ipstring, char *class, char *data, int expires,
		u_int flags, int aflags, CONFHOSTIPENTRY **existptr,
		char *comment)
{
	CONFHOSTIPENTRY		*entry;
	int					ret;

	if (!access || (!hostname && !ipstring))
		return -EINVAL;

	entry = (CONFHOSTIPENTRY *)MyMalloc(sizeof(CONFHOSTIPENTRY));
	BZERO(entry, sizeof(CONFHOSTIPENTRY));

	if (username)
		entry->username = MyStrdup(username);
	if (hostname)
		entry->hostname = MyStrdup(hostname);
	if (comment)
		entry->comment = MyStrdup(comment);
	entry->ip_net = -1;
	if (ipstring)
	{
		ircd_uint32_t	ipnet;
		ircd_uint32_t	ipmask;

		entry->ipstring = MyStrdup(ipstring);
		if (1 == cidr_ip2netmask(ipstring, &ipnet, &ipmask))
		{
			entry->ip_net = ipnet;
			entry->ip_mask = ipmask;
		}
	}
	if (class)
	{
		int sl = strlen(class);
		int m = sl < MAXCLASSNAME ? sl : MAXCLASSNAME;

		if (entry->classname)
			MyFree(entry->classname);
		entry->classname = MyMalloc(m+1);
		support_strncpy(entry->classname, class, m+1);
	}
	else
		entry->classname = MyMalloc(1);
	entry->flags = flags;
	if (data)
		entry->data = MyStrdup(data);
	if (expires > 0)
		entry->expires = expires;

	if (!entry->hostname && (entry->ip_net == -1))
	{
		_access_entry_free(entry);
		return -EINVAL;
	}

	ret = 0;
	if (entry->hostname)
		ret = _access_add_host(access, entry, aflags, existptr);
	if (!ret && (entry->ip_net != -1))
	{
		ret = _access_add_ipaddr(access, entry, aflags, existptr);
		if (ret && entry->hostname)
		{
			entry->ip_net = -1;
			access_del(access, entry, 0);
			return ret;
		}
	}
	if (ret)
		_access_entry_free(entry);
	else
		_access_add_to_list(&(access->conf_all), entry);
	return ret;
}

int access_clear(ACCESS *access, u_int flags)
{
	int	i;
	int	ii;

	_access_clear_host_sorted(access, &(access->conf_host->rev_table), flags);
	_access_clear_host_sorted(access, &(access->conf_host->for_table), flags);
	_access_clear_host_linear(access, &(access->conf_host->lin_list), flags);
	for(i=0;i<256;i++)
		for(ii=0;ii<33;ii++)
		_access_clear_ip_sorted(access, &(access->conf_ip->sort_table[i][ii]), flags);
	assert(access->conf_all == NULL);
#if 0
	_access_clear_ip_linear(access, &(access->conf_ip->lin_list), flags);
#endif
	return 0;
}

int access_flush(ACCESS *access, u_int flags)
{
	int	i;
	int ii;

	_access_flush_host_sorted(access, &(access->conf_host->rev_table),
		NULL, flags);
	_access_flush_host_sorted(access, &(access->conf_host->for_table),
		NULL, flags);
	_access_flush_host_linear(access, &(access->conf_host->lin_list),
		NULL, flags);
	for(i=0;i<256;i++)
		for(ii=0;ii<33;ii++)
		_access_flush_ip_sorted(access, &(access->conf_ip->sort_table[i][ii]),
				NULL, flags);
#if 0
	_access_flush_ip_linear(access, &(access->conf_ip->lin_list), flags);
#endif
	return 0;
}

int access_del(ACCESS *access, CONFHOSTIPENTRY *chentry, u_int flags)
{
	int	i;
	int ii;

	_access_flush_host_sorted(access, &(access->conf_host->rev_table),
		chentry, flags);
	_access_flush_host_sorted(access, &(access->conf_host->for_table),
		chentry, flags);
	_access_flush_host_linear(access, &(access->conf_host->lin_list),
		chentry, flags);
	for(i=0;i<256;i++)
		for(ii=0;ii<33;ii++)
		_access_flush_ip_sorted(access, &(access->conf_ip->sort_table[i][ii]),
				chentry, flags);
#if 0
	_access_flush_ip_linear(access, &(access->conf_ip->lin_list),chentry,flags);
#endif
	return 0;
}
