/*
**   IRC - Internet Relay Chat, source/cidr.c
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "cidr.h"

RCSID("$Id: cidr.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

/*
** i dunno...is this quicker than 0xffffffff << (32-prefix_length) ?
**  oh well..i need this table for the reverse anyway, since I
**  can't seem to figure out how to go netmask -> prefix_length
*/
static ircd_uint32_t _cidr_masks[33] = { 0x00000000,
	0x80000000, 0xc0000000, 0xe0000000, 0xf0000000, 0xf8000000,
	0xfc000000, 0xfe000000, 0xff000000, 0xff800000, 0xffc00000,
	0xffe00000, 0xfff00000, 0xfff80000, 0xfffc0000, 0xfffe0000,
	0xffff0000, 0xffff8000, 0xffffc000, 0xffffe000, 0xfffff000,
	0xfffff800, 0xfffffc00, 0xfffffe00, 0xffffff00, 0xffffff80,
	0xffffffc0, 0xffffffe0, 0xfffffff0, 0xfffffff8, 0xfffffffc,
	0xfffffffe, 0xffffffff };

int cidr_netmask_to_plength(ircd_uint32_t netmask)
{
	int		lower = 0;
	int		upper = (sizeof(_cidr_masks)/sizeof(_cidr_masks[0])) - 1;
	int		compare;

	compare = (lower+upper)/2;
	while(lower != upper)
	{
		if (_cidr_masks[compare] == netmask)
			return compare;
		if (netmask < _cidr_masks[compare])
			upper = compare;
		else
		{
			if (lower == compare)
				lower = compare+1;
			else
				lower = compare;
		}
		compare = (lower+upper)/2;
	}
	if (_cidr_masks[compare] == netmask)
		return compare;
	return -1;
}


/*
** check to see if ip is really an ip...
** valid formats:
** 1) 206.173.*
** 2) 206.173/16
** 3) 206.173/255.255.0.0
** 4) 206.173.1.1
*/
int cidr_valid(char *ip)
{
	int		gotstar = 0;
	int		gotslash = 0;
	int		dots = 0;
	int		dots2 = 0;
	int		curnum = 0;
	char	*ptr;
	char	*lastchar = NULL;

	for(ptr=ip;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case '.':
				if (!gotslash && ++dots > 3)
					return 0;
				if (gotslash && ++dots2 > 3)
					return 0;
				if (!lastchar || *lastchar == '.')
					return 0;
				if (curnum > 255)
					return 0;
				curnum = 0;
				break;
			case '*':
				if (gotslash)
					return 0;
				gotstar++;
				break;
			case '/':
				if (gotstar)
					return 0;
				if (gotslash++) /* already got one ? */
					return 0;
				if (*(ptr+1) == '\0')
					return 0;
				if (curnum > 255)
					return 0;
				curnum = 0;
				break;
			default:
				if (!isdigit((int)*ptr))
					return 0;
				curnum = (curnum*10) + *ptr - '0';
		}
		lastchar = ptr;
	}
	if (dots2)
	{
		if (dots2 != 3)
			return 0;
	}
	else if (gotslash)
	{
		if (curnum > 32)
			return 0;
	}
	return -1;
}

/* convert ip into net and mask */
int cidr_ip2netmask(char *ip, ircd_uint32_t *net, ircd_uint32_t *mask)
{
	char		*ptr;
	char		*lastchar;
	int			dots = 0;
	int			dots2 = 0;
	ircd_uint32_t	mynet = 0;
	ircd_uint32_t	mymask = 0;
	int			curnum = 0;
	int			gotslash = 0;
	int			gotstar = 0;

	if (*ip == '\0')
		return 0;
	else if (*ip == '*')
	{
		if (*(ip+1) == '\0')
		{
			*net = mynet;
			*mask = mymask;
			return 1;
		}
		return 0;
	}
	lastchar = ip;
	curnum = *ip-'0';
	for(ptr=ip+1;*ptr;ptr++)
	{
		if (*ptr == '*')
		{
			if (gotslash)
				return 0;
			if (*lastchar != '.')
				/* we have "10.1*.." */
				return 0;
			if ((*(ptr+1) != '\0'))
				return 0;
			mymask = 0xffffff00;
			gotstar = 1;
			while(dots < 3)
			{
				mynet <<= 8;
				mymask <<= 8;
				dots++;
			}
			break;
		}
		else if (*ptr == '/')
		{
			gotslash++;
			mymask = 0;
			mynet += curnum;
			curnum = 0;
		}
		else if (*ptr == '.')
		{
			if (gotslash)
			{
				mymask = (mymask + curnum) << 8;
				dots2++;
			}
			else
			{
				mymask = (mymask + 255) << 8;
				mynet = (mynet + curnum) << 8;
				dots++;
			}
			curnum = 0;
		}
		else
		{
			if (!isdigit((int)*ptr))
				return 0;
			curnum = (curnum*10) + *ptr - '0';
		}
		lastchar = ptr;
	}
	if (!gotslash)
		mynet += curnum;
	while(dots < 3)
	{
		mynet <<= 8;
		dots++;
	}
	if (gotslash)
	{
		if (!dots2)
		{
			if (curnum > 32)
				return 0;
			mymask = _cidr_masks[curnum];
		}
		else
			mymask += curnum;
	}
	else if (!gotstar)
		mymask = _cidr_masks[32];

	mynet &= mymask;

	*net = mynet;
	*mask = mymask;

	return 1;
}

/* convert 206.173.1.1 into ip bitmask */
ircd_uint32_t cidr_string2ip(char *ipstr)
{
	int			curnum = 0;
	ircd_uint32_t	ret = 0;
	char		*ptr;

	for(ptr=ipstr;*ptr;ptr++)
	{
		if (*ptr == '.')
		{
			ret <<= 8;
			ret += curnum;
			curnum = 0;
		}
		else
			curnum = (curnum*10) + *ptr - '0';
	}
	return (ret << 8) + curnum;
}

int cidr_match(ircd_uint32_t net, ircd_uint32_t mask, ircd_uint32_t ip)
{
	ircd_uint32_t	mynet = (ip & mask);

	return (mynet == net) ? 0 : (mynet < net ? 1 : -1);
}

