/*
**   IRC - Internet Relay Chat, source/fdlist.c
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "fdlist.h"

RCSID("$Id: fdlist.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

int fdlist_add_fd(FDLIST *fdlist, int fd, int flags)
{
	int	n;

	if (!fdlist || (fd < 0) ||
			!(flags & (FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE)))
	{
		return -EINVAL;
	}
	if ((n=fdlist->pfd_num) == fdlist->pfd_size)
	{
		void 	*ptr;

		fdlist->pfd_size += 16;
		ptr = realloc(fdlist->pfd_array,
					fdlist->pfd_size * sizeof(FDLIST_POLLFD));
		if (!ptr)
		{
			fdlist->pfd_size -= 16;
			return -ENOMEM;
		}
		fdlist->pfd_array = (FDLIST_POLLFD *)ptr;
	}
	fdlist->pfd_array[n].fd = fd;
	fdlist->pfd_array[n].revents = 0;
	fdlist->pfd_array[n].events = 0;
	if (flags & FDLIST_FLAGS_READ)
		fdlist->pfd_array[n].events |= POLLIN;
	if (flags & FDLIST_FLAGS_WRITE)
		fdlist->pfd_array[n].events |= POLLOUT;
	return fdlist->pfd_num++;
}

int fdlist_del_fd(FDLIST *fdlist, int pos, int *mfd)
{
	int		n;

	if (!fdlist || (pos > (n = fdlist->pfd_num)) || (pos < 0))
	{
		return -EINVAL;
	}
	if (mfd)
		*mfd = -1;
	if ((--fdlist->pfd_num) && (n-1 != pos))
	{
		BCOPY(&(fdlist->pfd_array[n-1]), &(fdlist->pfd_array[pos]),
			sizeof(FDLIST_POLLFD));
		if (mfd)
			*mfd = fdlist->pfd_array[pos].fd;
	}
	return fdlist->pfd_num;
}

int fdlist_add_event(FDLIST *fdlist, int pos, int flags)
{
	if (!fdlist || (pos > fdlist->pfd_num) || (pos < 0))
		return -EINVAL;
	if (flags & FDLIST_FLAGS_READ)
		fdlist->pfd_array[pos].events |= POLLIN;
	if (flags & FDLIST_FLAGS_WRITE)
		fdlist->pfd_array[pos].events |= POLLOUT;
	return fdlist->pfd_array[pos].events;
}

int fdlist_del_event(FDLIST *fdlist, int pos, int flags)
{
	if (!fdlist || (pos > fdlist->pfd_num) || (pos < 0))
		return -EINVAL;
	if (flags & FDLIST_FLAGS_READ)
		fdlist->pfd_array[pos].events &= ~POLLIN;
	if (flags & FDLIST_FLAGS_WRITE)
		fdlist->pfd_array[pos].events &= ~POLLOUT;
	return fdlist->pfd_array[pos].events;
}

int fdlist_set_events(FDLIST *fdlist, int pos, int flags)
{
	if (!fdlist || (pos > fdlist->pfd_num) || (pos < 0) || !flags)
		return -EINVAL;
	fdlist->pfd_array[pos].events = 0;
	if (flags & FDLIST_FLAGS_READ)
		fdlist->pfd_array[pos].events |= POLLIN;
	if (flags & FDLIST_FLAGS_WRITE)
		fdlist->pfd_array[pos].events |= POLLOUT;
	return fdlist->pfd_array[pos].events;
}

int fdlist_get_events(FDLIST *fdlist, int pos)
{
	int flags = 0;

	if (!fdlist || (pos > fdlist->pfd_num) || (pos < 0))
		return -EINVAL;
	if (fdlist->pfd_array[pos].events & POLLIN)
		flags |= FDLIST_FLAGS_READ;
	if (fdlist->pfd_array[pos].events & POLLOUT)
		flags |= FDLIST_FLAGS_WRITE;
	return flags;
}


#if 0

int fdlist_add_fd(FDLIST *fdlist, int fd, int flags)
{
	if (!fdlist || (fd < 0) ||
			!(flags & (FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE)))
		return -EINVAL;
	if (fd < fdlist->lowest_fd)
		fdlist->lowest_fd = fd;
	if (flags & FDLIST_FLAGS_READ)
		FD_SET(fd, &fdlist->read_fds);
	if (flags & FDLIST_FLAGS_WRITE)
		FD_SET(fd, &fdlist->write_fds);
	fdlist->pfd_num++;
	return fd;
}

int fdlist_del_fd(FDLIST *fdlist, int fd, int *mfd)
{
	if (!fdlist || (fd < 0))
		return -EINVAL;
	if (mfd)
		*mfd = -1;
	FD_CLR(fd, &fdlist->read_fds);
	FD_CLR(fd, &fdlist->write_fds);
	fdlist->pfd_num--;
	return 0;
}

int fdlist_add_event(FDLIST *fdlist, int fd, int flags)
{
	if (!fdlist || (fd < 0))
		return -EINVAL;
	if (flags & FDLIST_FLAGS_READ)
		FD_SET(fd, &fdlist->read_fds);
	if (flags & FDLIST_FLAGS_WRITE)
		FD_SET(fd, &fdlist->write_fds);
	return 0;
}

int fdlist_del_event(FDLIST *fdlist, int fd, int flags)
{
	if (!fdlist || (fd < 0))
		return -EINVAL;
	if (flags & FDLIST_FLAGS_READ)
		FD_CLR(fd, &fdlist->read_fds);
	if (flags & FDLIST_FLAGS_WRITE)
		FD_CLR(fd, &fdlist->write_fds);
	return 0;
}

int fdlist_set_events(FDLIST *fdlist, int fd, int flags)
{
	if (!fdlist || (fd < 0) || !flags)
		return -EINVAL;
	FD_CLR(fd, &fdlist->read_fds);
	FD_CLR(fd, &fdlist->write_fds);
	if (flags & FDLIST_FLAGS_READ)
		FD_SET(fd, &fdlist->read_fds);
	if (flags & FDLIST_FLAGS_WRITE)
		FD_SET(fd, &fdlist->write_fds);
	return 0;
}

int fdlist_get_events(FDLIST *fdlist, int fd)
{
	int flags = 0;

	if (!fdlist || (fd < 0))
		return -EINVAL;
	if (FD_ISSET(fd, &fdlist->read_fds))
		flags |= FDLIST_FLAGS_READ;
	if (FD_ISSET(fd, &fdlist->write_fds))
		flags |= FDLIST_FLAGS_WRITE;
	return flags;
}

#endif /* 0 */
