/*
**   IRC - Internet Relay Chat, source/fileio.c
**   Copyright (C) 1998 Thomas Helvey <tomh@inxpress.net>
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "fileio.h"

RCSID("$Id: fileio.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

static int fbfill(aFileBuf *fb);

aFileBuf *fbopen(const char *filename, const char *mode)
{
	int			openmode = 0;
	int			pmode = 0;
	aFileBuf	*fb = NULL;
	int			fd;

	if ((filename == NULL) || (mode == NULL))
		return NULL;

	while (*mode)
	{
		switch (*mode)
		{
			case 'r':
				openmode = O_RDONLY;
				break;
			case 'w':
				openmode = O_WRONLY | O_CREAT | O_TRUNC;
				pmode = S_IREAD | S_IWRITE;
				break;
			case 'a':
				openmode = O_WRONLY | O_CREAT | O_APPEND;
				pmode = S_IREAD | S_IWRITE;
				break;
			case '+':
				openmode &= ~(O_RDONLY | O_WRONLY);
				openmode |= O_RDWR;
				break;
			default:
				break;
		}
		++mode;
	}
#ifdef _WIN32
	openmode |= O_BINARY;
#endif /* _WIN32 */

	if ((fd = open(filename, openmode, pmode)) == -1)
		return NULL;

	fb = (aFileBuf *)MyMalloc(sizeof(aFileBuf));
	fb->inbufsize = 1024;
	fb->buf = (char *)MyMalloc(fb->inbufsize);
	if (fb->buf == NULL)
	{
		MyFree(fb);
		close(fd);
		return NULL;
	}
	fb->outbufsize = 4096;
	fb->outbufpos = 0;
	fb->outbuf = (char *)MyMalloc(fb->outbufsize);
	if (fb->outbuf == NULL)
	{
		MyFree(fb->buf);
		MyFree(fb);
		close(fd);
		return NULL;
	}
	fb->ptr = fb->endp = fb->buf;
	fb->fd = fd;
	fb->pid = 0;
	fb->flags = 0;
	return fb;
}

int fbsetbuf(aFileBuf *fb, int which, int size)
{
	char *ptr = NULL;

	if ((fb == NULL) || (size < 0))
		return -1;
	if (which == FB_INBUFFER)
	{
		if (size < 1024)
			return -1;
		if ((fb->ptr != fb->buf) || (fb->endp != fb->buf))
			return -1;
	}
	else if (which != FB_OUTBUFFER)
		return -1;
	if (size) /* will always be if INBUFFER */
	{
		ptr = (char *)MyMalloc(size);
		if (ptr == NULL)
			return -1;
	}
	if (which == FB_INBUFFER)
	{
		MyFree(fb->buf);
		fb->ptr = fb->endp = fb->buf = ptr;
		fb->inbufsize = size;
	}
	else
	{
		if (size)
		{
			fbflush(fb);
			MyFree(fb->outbuf);
			fb->outbuf = ptr;
		}
		fb->outbufsize = size;
	}
	return 0;
}

/*
 * fbopenp - opens a pipe, forks,
 * calls execlp(ename, ename, arg1, arg2, 0),
 * creates a FileBuf from the pipe descriptor, and returns it.
 */
aFileBuf *fbopenp(const char *ename, const char *arg1, const char *arg2)
{
#ifndef _WIN32
	int			pi[2];
	int			i = 0;
	aFileBuf	*fb;
	pid_t		pid;

	if (ename == NULL)
		return NULL;
	if (pipe(pi) == -1)
		return NULL;
	pid = fork();
	switch (pid)
	{
		case -1:
			return NULL;
		case 0:
			close(pi[0]);
			if (pi[1] != 1)
			{
				dup2(pi[1], 1);
				close(pi[1]);
			}
			dup2(1, 2);
			for (i = 3; i < 4098; i++)
				close(i);
			execlp(ename, ename, arg1, arg2, 0);
			exit(-1);
		default:
			close(pi[1]);
			fb = (aFileBuf *)MyMalloc(sizeof(aFileBuf));
			fb->inbufsize = 1024;
			fb->buf = (char *)MyMalloc(fb->inbufsize);
			if (fb->buf == NULL)
			{
				MyFree(fb);
				close(pi[0]);
				waitpid(pid, NULL, 0);
				return NULL;
			}
			fb->outbufsize = 4096;
			fb->outbufpos = 0;
			fb->outbuf = (char *)MyMalloc(fb->outbufsize);
			if (fb->outbuf == NULL)
			{
				MyFree(fb->buf);
				MyFree(fb);
				close(pi[0]);
				waitpid(pid, NULL, 0);
				return NULL;
			}
			fb->ptr = fb->endp = fb->buf;
			fb->pid = pid;
			fb->fd = pi[0];
			return fb;
	}
#endif /* !_WIN32 */
	return NULL; 
}

void fbclose(aFileBuf *fb)
{
	if (fb != NULL)
	{
		close(fb->fd);
		if (fb->pid)
			waitpid(fb->pid, NULL, 0);
		MyFree((void *)fb->outbuf);
		MyFree((void *)fb->buf);
		MyFree((void *)fb);
	}
}

static int fbfill(aFileBuf *fb)
{
	int n = read(fb->fd, fb->buf, fb->inbufsize-1);
	if (n < 0)
	{
		fb->flags |= FB_FAIL;
		return -1;
	}
	if (0 == n)
		fb->flags |= FB_EOF;
	fb->ptr = fb->buf;
	fb->endp = fb->buf + n;
	return n;
}

int fbgetc(aFileBuf *fb)
{
	if (fb == NULL)
		return EOF;
	if (fb->ptr < fb->endp)
		return *fb->ptr++;
	if (fb->flags & FB_EOF)
		return EOF;
	return (fbfill(fb) > 0) ? *fb->ptr++ : EOF;
}

char *fbgets(char *buf, size_t len, aFileBuf *fb)
{
	char *p = buf;

	if ((buf == NULL) || (fb == NULL) || (len < 0))
		return NULL;

	if (fb->ptr == fb->endp)
	{
		/* buffer empty */
		if (fbfill(fb) <= 0)
			return NULL;
	}
	if (fb->ptr < fb->endp)
	{
		while (--len)
		{
			*p = *fb->ptr++;
			/* check for newline */
			if ('\n' == *p)
			{
				*++p = 0;
				return buf;
			}
			/* check for cr */
			else if ('\r' == *p)
			{
				/* skip newline if there */
				if ((fb->ptr < fb->endp || fbfill(fb) > 0)
					&& '\n' == *fb->ptr)
					++fb->ptr;
				/* translate to newline */
				*p = '\n';
				*++p = 0;
				return buf;
			}
			++p;
			if (fb->ptr < fb->endp || fbfill(fb) > 0)
				continue;
			*p = 0;
			return buf;
		}
	}
	return NULL;
}

int fbflush(aFileBuf *fb)
{
	if (fb->outbufpos)
	{
		if (write(fb->fd, fb->outbuf, fb->outbufpos) != fb->outbufpos)
		{
			fb->outbufpos = 0;
			return -1;
		}
		fb->outbufpos = 0;
	}
	return 0;
}

int fbwrite(aFileBuf *fb, const void *buffer, int size)
{
	if (fb->outbufsize)
	{
		if (fb->outbufpos + size > fb->outbufsize)
		{
			if (fbflush(fb) == -1)
				return -1;
			if (fb->outbufpos + size > fb->outbufsize) /* doh */
				return write(fb->fd, buffer, size);
		}
		BCOPY(buffer, fb->outbuf+fb->outbufpos, size);
		fb->outbufpos += size;
	}
	else
		return write(fb->fd, buffer, size);
	return size;
}

int fbputs(const char *str, aFileBuf *fb)
{
	if ((str == NULL) || (fb == NULL))
		return -1;
	return fbwrite(fb, str, strlen(str));
}

int fbstat(struct stat *sb, aFileBuf *fb)
{
	if ((sb == NULL) || (fb == NULL))
		return -1;
	return fstat(fb->fd, sb);
}

pid_t fbpid(aFileBuf *fb)
{
	return fb->pid;
}
