/*
**   IRC - Internet Relay Chat, source/hash.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1991 Darren Reed
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "hash.h"

RCSID("$Id: hash.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

static IRCD_CLIENT		**CLIENT_HASH;
static IRCD_CHANNEL		**CHANNEL_HASH;
static unsigned int		*_hashtab = NULL;

static unsigned int hash_compute_client(char *nname, unsigned int *hashv);
static unsigned int hash_compute_channel(char *cname, unsigned int *hashv);
static void add_client_to_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client);
static void del_client_from_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client);
static void add_channel_to_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void del_channel_from_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void add_channel_to_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);
static void del_channel_from_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel);

void hash_init(void)
{
	int i;

	_hashtab = (u_int *) MyMalloc(256 * sizeof(unsigned int));
	for (i = 0; i < 256; i++)
		_hashtab[i] = tolower((char)i) * 109;
	CLIENT_HASH = (IRCD_CLIENT **)MyMalloc(sizeof(IRCD_CLIENT *)*(HASH_CLIENT_SIZE+1));
	CHANNEL_HASH = (IRCD_CHANNEL **)MyMalloc(sizeof(IRCD_CHANNEL *)*HASH_CHANNEL_SIZE);
}

static unsigned int hash_compute_client(char *nname, unsigned int *hashv)
{
	unsigned char	*name = (unsigned char *)nname;
	unsigned int	hash = 1;

	for(;*name;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name];
	}
	*hashv = hash;
	if (*nname == '*')
		return HASH_CLIENT_SIZE;
	return hash % HASH_CLIENT_SIZE;
}

unsigned int hash_compute_whowas(char *nname)
{
	unsigned char	*name = (unsigned char *)nname;
	unsigned int	hash = 1;

	for(;*name;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name];
	}
	return hash;
}

static unsigned int hash_compute_channel(char *cname, unsigned int *hashv)
{
	unsigned char	*name = (unsigned char *)cname;
	unsigned int	hash = 1;
	int				i = 30;

	for(;*name && --i;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name] + (i << 1);
	}
	*hashv = hash;
	return hash % HASH_CHANNEL_SIZE;
}


IRCD_CLIENT *hash_find_client(char *name)
{
	unsigned int	hashv;
	unsigned int	hv;
	IRCD_CLIENT		*cptr;

	hv = hash_compute_client(name, &hashv);
	for(cptr=CLIENT_HASH[hv];cptr;cptr=cptr->c_h_next)
		if ((cptr->c_hashv == hashv) && !mycmp(name, cptr->name))
			break;
	return cptr;
}

IRCD_CLIENT *hash_find_chasing(char *name)
{
	IRCD_CLIENT	*cptr;

	cptr = hash_find_client(name);
	if (!cptr)
		cptr = whowas_get_history(name, 60);
	return cptr;
}

IRCD_CLIENT *hash_find_client_incl_hm(char *name)
{
	IRCD_CLIENT		*cptr;
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_client(name, &hashv);
	for(cptr=CLIENT_HASH[hv];cptr;cptr=cptr->c_h_next)
		if ((cptr->c_hashv == hashv) && !mycmp(name, cptr->name))
			break;
	if (!cptr && (hv != HASH_CLIENT_SIZE))
	{
		for(cptr=CLIENT_HASH[HASH_CLIENT_SIZE];cptr;cptr=cptr->c_h_next)
			if (!match(cptr->name, name))
				break;
	}
	return cptr;
}

IRCD_CHANNEL *hash_find_channel(char *name)
{
	IRCD_CHANNEL	*chptr;
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_channel(name, &hashv);
	for(chptr=CHANNEL_HASH[hv];chptr;chptr=chptr->h_next)
		if ((chptr->hashv == hashv) && !mycmp(name, chptr->name))
			break;
	return chptr;
}

int hash_add_client(IRCD_CLIENT *cptr)
{
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_client(cptr->name, &hashv);
	if (hv == HASH_CLIENT_SIZE)
		cptr->c_flags |= CLIENT_FLAGS_HOSTMASK;
	cptr->c_hashv = hashv;
	add_client_to_hlist(&CLIENT_HASH[hv], cptr);
	return 0;
}

int hash_del_client(IRCD_CLIENT *cptr)
{
	del_client_from_hlist(&CLIENT_HASH[ (cptr->c_flags & CLIENT_FLAGS_HOSTMASK) ?
				HASH_CLIENT_SIZE : cptr->c_hashv % HASH_CLIENT_SIZE ], cptr);
	return 0;
}

int hash_add_channel(IRCD_CHANNEL *chptr)
{
	unsigned int	hashv;
	unsigned int	hv;

	hv = hash_compute_channel(chptr->name, &hashv);
	chptr->hashv = hashv;
	add_channel_to_hlist(&CHANNEL_HASH[hv], chptr);
	add_channel_to_list(&Channels, chptr);
	return 0;
}

int hash_del_channel(IRCD_CHANNEL *chptr)
{
	del_channel_from_hlist(&CHANNEL_HASH[chptr->hashv % HASH_CHANNEL_SIZE],
			chptr);
	del_channel_from_list(&Channels, chptr);
	return 0;
}

static void add_channel_to_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	channel->h_prev = NULL;
	if ((channel->h_next = *bucket) != NULL)
		channel->h_next->h_prev = channel;
	*bucket = channel;
}

static void del_channel_from_hlist(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	if (channel->h_prev)
		channel->h_prev->h_next = channel->h_next;
	else
		*bucket = channel->h_next;
	if (channel->h_next)
		channel->h_next->h_prev = channel->h_prev;
}

static void add_client_to_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_h_prev = NULL;
	if ((client->c_h_next = *bucket) != NULL)
		client->c_h_next->c_h_prev = client;
	*bucket = client;
}

static void del_client_from_hlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_h_prev)
		client->c_h_prev->c_h_next = client->c_h_next;
	else
		*bucket = client->c_h_next;
	if (client->c_h_next)
		client->c_h_next->c_h_prev = client->c_h_prev;
}

void add_auth_to_alist(IRCD_AUTH **bucket, IRCD_AUTH *auth)
{
	auth->a_prev = NULL;
	if ((auth->a_next = *bucket) != NULL)
		auth->a_next->a_prev = auth;
	*bucket = auth;
}

void del_auth_from_alist(IRCD_AUTH **bucket, IRCD_AUTH *auth)
{
	if (auth->a_prev)
		auth->a_prev->a_next = auth->a_next;
	else
		*bucket = auth->a_next;
	if (auth->a_next)
		auth->a_next->a_prev = auth->a_prev;
}

void add_rejecthold_to_rhlist(IRCD_REJECTHOLD **bucket, IRCD_REJECTHOLD *rejecthold)
{
	rejecthold->rh_prev = NULL;
	if ((rejecthold->rh_next = *bucket) != NULL)
		rejecthold->rh_next->rh_prev = rejecthold;
	*bucket = rejecthold;
}

void del_rejecthold_from_rhlist(IRCD_REJECTHOLD **bucket, IRCD_REJECTHOLD *rejecthold)
{
	if (rejecthold->rh_prev)
		rejecthold->rh_prev->rh_next = rejecthold->rh_next;
	else
		*bucket = rejecthold->rh_next;
	if (rejecthold->rh_next)
		rejecthold->rh_next->rh_prev = rejecthold->rh_prev;
}

void add_listener_to_llist(IRCD_LISTENER **bucket, IRCD_LISTENER *listener)
{
	listener->l_prev = NULL;
	if ((listener->l_next = *bucket) != NULL)
		listener->l_next->l_prev = listener;
	*bucket = listener;
}

void del_listener_from_llist(IRCD_LISTENER **bucket, IRCD_LISTENER *listener)
{
	if (listener->l_prev)
		listener->l_prev->l_next = listener->l_next;
	else
		*bucket = listener->l_next;
	if (listener->l_next)
		listener->l_next->l_prev = listener->l_prev;
}

void add_client_to_llist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_l_prev = NULL;
	if ((client->c_l_next = *bucket) != NULL)
		client->c_l_next->c_l_prev = client;
	*bucket = client;
}

void del_client_from_llist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_l_prev)
		client->c_l_prev->c_l_next = client->c_l_next;
	else
		*bucket = client->c_l_next;
	if (client->c_l_next)
		client->c_l_next->c_l_prev = client->c_l_prev;
}

void add_client_to_lolist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_lo_prev = NULL;
	if ((client->c_lo_next = *bucket) != NULL)
		client->c_lo_next->c_lo_prev = client;
	*bucket = client;
}

void del_client_from_lolist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_lo_prev)
		client->c_lo_prev->c_lo_next = client->c_lo_next;
	else
		*bucket = client->c_lo_next;
	if (client->c_lo_next)
		client->c_lo_next->c_lo_prev = client->c_lo_prev;
}

void add_client_to_cllist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_cl_prev = NULL;
	if ((client->c_cl_next = *bucket) != NULL)
		client->c_cl_next->c_cl_prev = client;
	*bucket = client;
}

void del_client_from_cllist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_cl_prev)
		client->c_cl_prev->c_cl_next = client->c_cl_next;
	else
		*bucket = client->c_cl_next;
	if (client->c_cl_next)
		client->c_cl_next->c_cl_prev = client->c_cl_prev;
}

void add_client_to_ilist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_i_prev = NULL;
	if ((client->c_i_next = *bucket) != NULL)
		client->c_i_next->c_i_prev = client;
	*bucket = client;
}

void del_client_from_ilist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_i_prev)
		client->c_i_prev->c_i_next = client->c_i_next;
	else
		*bucket = client->c_i_next;
	if (client->c_i_next)
		client->c_i_next->c_i_prev = client->c_i_prev;
}

void add_client_to_oslist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_oslist = bucket;
	client->c_os_prev = NULL;
	if ((client->c_os_next = *bucket) != NULL)
		client->c_os_next->c_os_prev = client;
	*bucket = client;
}

void add_client_to_nlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_n_prev = NULL;
	if ((client->c_n_next = *bucket) != NULL)
		client->c_n_next->c_n_prev = client;
	*bucket = client;
}

void del_client_from_nlist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_n_prev)
		client->c_n_prev->c_n_next = client->c_n_next;
	else
		*bucket = client->c_n_next;
	if (client->c_n_next)
		client->c_n_next->c_n_prev = client->c_n_prev;
}

void del_client_from_oslist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_oslist = NULL;
	if (client->c_os_prev)
		client->c_os_prev->c_os_next = client->c_os_next;
	else
		*bucket = client->c_os_next;
	if (client->c_os_next)
		client->c_os_next->c_os_prev = client->c_os_prev;
}

void add_client_to_ulist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_u_prev = NULL;
	if ((client->c_u_next = *bucket) != NULL)
		client->c_u_next->c_u_prev = client;
	*bucket = client;
}

void del_client_from_ulist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_u_prev)
		client->c_u_prev->c_u_next = client->c_u_next;
	else
		*bucket = client->c_u_next;
	if (client->c_u_next)
		client->c_u_next->c_u_prev = client->c_u_prev;
}

#if 0
void add_client_to_slist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->s_prev = NULL;
	if ((client->s_next = *bucket) != NULL)
		client->s_next->s_prev = client;
	*bucket = client;
}

void del_client_from_slist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->s_prev)
		client->s_prev->s_next = client->s_next;
	else
		*bucket = client->s_next;
	if (client->s_next)
		client->s_next->s_prev = client->s_prev;
}
#endif

void add_client_to_flist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	client->c_f_prev = NULL;
	if ((client->c_f_next = *bucket) != NULL)
		client->c_f_next->c_f_prev = client;
	*bucket = client;
}

void del_client_from_flist(IRCD_CLIENT **bucket, IRCD_CLIENT *client)
{
	if (client->c_f_prev)
		client->c_f_prev->c_f_next = client->c_f_next;
	else
		*bucket = client->c_f_next;
	if (client->c_f_next)
		client->c_f_next->c_f_prev = client->c_f_prev;
}

void add_clientlist_to_list(IRCD_CLIENTLIST **bucket, IRCD_CLIENTLIST *clientlist)
{
	clientlist->prev = NULL;
	if ((clientlist->next = *bucket) != NULL)
		clientlist->next->prev = clientlist;
	*bucket = clientlist;
}

void del_clientlist_from_list(IRCD_CLIENTLIST **bucket, IRCD_CLIENTLIST *clientlist)
{
	if (clientlist->prev)
		clientlist->prev->next = clientlist->next;
	else
		*bucket = clientlist->next;
	if (clientlist->next)
		clientlist->next->prev = clientlist->prev;
}

void add_channellist_to_list(IRCD_CHANNELLIST **bucket, IRCD_CHANNELLIST *channellist)
{
	channellist->prev = NULL;
	if ((channellist->next = *bucket) != NULL)
		channellist->next->prev = channellist;
	*bucket = channellist;
}

void del_channellist_from_list(IRCD_CHANNELLIST **bucket, IRCD_CHANNELLIST *channellist)
{
	if (channellist->prev)
		channellist->prev->next = channellist->next;
	else
		*bucket = channellist->next;
	if (channellist->next)
		channellist->next->prev = channellist->prev;
}

void add_invite_to_list(IRCD_INVITE **bucket, IRCD_INVITE *invite)
{
	invite->prev = NULL;
	if ((invite->next = *bucket) != NULL)
		invite->next->prev = invite;
	*bucket = invite;
}

void del_invite_from_list(IRCD_INVITE **bucket, IRCD_INVITE *invite)
{
	if (invite->prev)
		invite->prev->next = invite->next;
	else
		*bucket = invite->next;
	if (invite->next)
		invite->next->prev = invite->prev;
}

void add_ban_to_list(IRCD_BAN **bucket, IRCD_BAN *ban)
{
	ban->prev = NULL;
	if ((ban->next = *bucket) != NULL)
		ban->next->prev = ban;
	*bucket = ban;
}

void del_ban_from_list(IRCD_BAN **bucket, IRCD_BAN *ban)
{
	if (ban->prev)
		ban->prev->next = ban->next;
	else
		*bucket = ban->next;
	if (ban->next)
		ban->next->prev = ban->prev;
}

static void add_channel_to_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	channel->prev = NULL;
	if ((channel->next = *bucket) != NULL)
		channel->next->prev = channel;
	*bucket = channel;
}

static void del_channel_from_list(IRCD_CHANNEL **bucket, IRCD_CHANNEL *channel)
{
	if (channel->prev)
		channel->prev->next = channel->next;
	else
		*bucket = channel->next;
	if (channel->next)
		channel->next->prev = channel->prev;
}

void add_reqlist_to_list(IRCDRES_REQLIST **bucket, IRCDRES_REQLIST *reqlist)
{
	reqlist->rl_prev = NULL;
	if ((reqlist->rl_next = *bucket) != NULL)
		reqlist->rl_next->rl_prev = reqlist;
	*bucket = reqlist;
}

void del_reqlist_from_list(IRCDRES_REQLIST **bucket, IRCDRES_REQLIST *reqlist)
{
	if (reqlist->rl_prev)
		reqlist->rl_prev->rl_next = reqlist->rl_next;
	else
		*bucket = reqlist->rl_next;
	if (reqlist->rl_next)
		reqlist->rl_next->rl_prev = reqlist->rl_prev;
}

int hash_init_reload(void **client_hash, int client_hash_size, void **channel_hash, int channel_hash_size)
{
	int				i;
	IRCD_CLIENT		*cptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CLIENT		*nextcptr;
	IRCD_CHANNEL	*nextchptr;

	if (client_hash_size == HASH_CLIENT_SIZE)
	{
		CLIENT_HASH = (IRCD_CLIENT **)client_hash;
	}
	else
	{
		CLIENT_HASH = (IRCD_CLIENT **)MyMalloc(sizeof(IRCD_CLIENT *)*(HASH_CLIENT_SIZE+1));
		for(i=0;i<client_hash_size;i++)
		{
			for(cptr=((IRCD_CLIENT **)client_hash)[i];cptr;cptr=nextcptr)
			{
				nextcptr = cptr->c_h_next;
				add_client_to_hlist(&CLIENT_HASH[(cptr->c_flags &
					CLIENT_FLAGS_HOSTMASK) ? HASH_CLIENT_SIZE :
					cptr->c_hashv % HASH_CLIENT_SIZE], cptr);
			}
		}
		MyFree(client_hash);
	}
	if (channel_hash_size == HASH_CHANNEL_SIZE)
	{
		CHANNEL_HASH = (IRCD_CHANNEL **)channel_hash;
	}
	else
	{
		CHANNEL_HASH = (IRCD_CHANNEL **)MyMalloc(sizeof(IRCD_CHANNEL *)*HASH_CHANNEL_SIZE);
		for(i=0;i<channel_hash_size;i++)
		{
			for(chptr=((IRCD_CHANNEL **)channel_hash)[i];chptr;chptr=nextchptr)
			{
				nextchptr = chptr->h_next;
				add_channel_to_hlist(&CHANNEL_HASH[chptr->hashv %
						HASH_CHANNEL_SIZE], chptr);
			}
		}
		MyFree(channel_hash);
	}
	return 0;
}

int hash_deinit_reload(void ***client_hash, int *client_hash_size, void ***channel_hash, int *channel_hash_size)
{
	MyFree(_hashtab);
	if (client_hash)
		*client_hash = (void **)CLIENT_HASH;
	if (client_hash_size)
		*client_hash_size = HASH_CLIENT_SIZE;
	if (channel_hash)
		*channel_hash = (void **)CHANNEL_HASH;
	if (channel_hash_size)
		*channel_hash_size = HASH_CHANNEL_SIZE;
	return 0;
}

