/*
**   IRC - Internet Relay Chat, source/iphash.c
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "iphash.h"

RCSID("$Id: iphash.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

static IRCD_NETHASH *Clones = NULL;
static IRCD_NETHASH	*Throttles = NULL;
static IRCD_NETHASH	**NET_HASH[256];

#define iphash_bucket_get(__x) (NET_HASH[(__x) >> 24][(((__x) >> 8) & 0xffff) % HASH_IP_SIZE])

static void _nethash_check_free(IRCD_NETHASH *nhptr);
static IRCD_NETHASH *nethash_find(ircd_uint32_t ipaddr);
static IRCD_NETHASH *nethash_create(ircd_uint32_t ipaddr);
static IRCD_IPENTRY *ipentry_create(IRCD_NETHASH *nhptr, ircd_uint32_t ipaddr);
static void add_nethash_to_hlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void del_nethash_from_hlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void add_nethash_to_clist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void del_nethash_from_clist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void add_nethash_to_tlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void del_nethash_from_tlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash);
static void add_ipentry_to_list(IRCD_IPENTRY **bucket, IRCD_IPENTRY *ipentry);
static void del_ipentry_from_list(IRCD_IPENTRY **bucket, IRCD_IPENTRY *ipentry);

static void _nethash_check_free(IRCD_NETHASH *nhptr)
{
	assert(!nhptr->ipaddrs);
	if (nhptr->flags &
			(IRCD_NETHASH_FLAGS_ATHROTTLED|IRCD_NETHASH_FLAGS_CTHROTTLED))
		return;
	else if ((nhptr->accept_rate_time + nhptr->accept_rate_maxtime < Now) &&
		(nhptr->client_rate_time + nhptr->client_rate_maxtime < Now))
	{
		del_nethash_from_hlist(&iphash_bucket_get(nhptr->ipaddr), nhptr);
		MyFree(nhptr);
	}   
}

IRCD_IPENTRY *ipentry_find(IRCD_NETHASH *nhptr, ircd_uint32_t ipaddr)
{
	IRCD_IPENTRY	*ihptr;

	if (!nhptr)
	{
		nhptr = nethash_find(ipaddr);
		if (!nhptr)
			return NULL;
	}
	for(ihptr=nhptr->ipaddrs;ihptr;ihptr=ihptr->next)
		if (ihptr->ipaddr == ipaddr)
			break;
	return ihptr;
}

static IRCD_NETHASH *nethash_find(ircd_uint32_t ipaddr)
{
	ircd_uint32_t	netaddr = ipaddr & 0xffffff00;
	IRCD_NETHASH	*nhptr;

	for(nhptr=iphash_bucket_get(ipaddr);nhptr;nhptr=nhptr->h_next)
		if (nhptr->ipaddr == netaddr)
			break;
	return nhptr;
}

static IRCD_NETHASH *nethash_create(ircd_uint32_t ipaddr)
{
	IRCD_NETHASH	*nhptr = (IRCD_NETHASH *)MyMalloc(sizeof(IRCD_NETHASH));

	nhptr->ipaddr = ipaddr & 0xffffff00;
	nhptr->accept_rate_time = nhptr->client_rate_time = Now;

	add_nethash_to_hlist(&iphash_bucket_get(ipaddr), nhptr);
	return nhptr;
}

static IRCD_IPENTRY *ipentry_create(IRCD_NETHASH *nhptr, ircd_uint32_t ipaddr)
{
	IRCD_IPENTRY	*iptr = (IRCD_IPENTRY *)MyMalloc(sizeof(IRCD_IPENTRY));

	iptr->ipaddr = ipaddr;

	add_ipentry_to_list(&(nhptr->ipaddrs), iptr);
	iptr->nhptr = nhptr;
	return iptr;
}

static void add_nethash_to_hlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    nethash->h_prev = NULL;
    if ((nethash->h_next = *bucket) != NULL)
        nethash->h_next->h_prev = nethash;
    *bucket = nethash;
}

static void del_nethash_from_hlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    if (nethash->h_prev)
        nethash->h_prev->h_next = nethash->h_next;
    else
        *bucket = nethash->h_next;
    if (nethash->h_next)
        nethash->h_next->h_prev = nethash->h_prev;
}

static void add_nethash_to_clist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    nethash->c_prev = NULL;
    if ((nethash->c_next = *bucket) != NULL)
        nethash->c_next->c_prev = nethash;
    *bucket = nethash;
}

static void del_nethash_from_clist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    if (nethash->c_prev)
        nethash->c_prev->c_next = nethash->c_next;
    else
        *bucket = nethash->c_next;
    if (nethash->c_next)
        nethash->c_next->c_prev = nethash->c_prev;
}

static void add_nethash_to_tlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    nethash->t_prev = NULL;
    if ((nethash->t_next = *bucket) != NULL)
        nethash->t_next->t_prev = nethash;
    *bucket = nethash;
}

static void del_nethash_from_tlist(IRCD_NETHASH **bucket, IRCD_NETHASH *nethash)
{
    if (nethash->t_prev)
        nethash->t_prev->t_next = nethash->t_next;
    else
        *bucket = nethash->t_next;
    if (nethash->t_next)
        nethash->t_next->t_prev = nethash->t_prev;
}

static void add_ipentry_to_list(IRCD_IPENTRY **bucket, IRCD_IPENTRY *ipentry)
{
    ipentry->prev = NULL;
    if ((ipentry->next = *bucket) != NULL)
        ipentry->next->prev = ipentry;
    *bucket = ipentry;
}

static void del_ipentry_from_list(IRCD_IPENTRY **bucket, IRCD_IPENTRY *ipentry)
{
    if (ipentry->prev)
        ipentry->prev->next = ipentry->next;
    else
        *bucket = ipentry->next;
    if (ipentry->next)
        ipentry->next->prev = ipentry->prev;
}

int iphash_init(void)
{
	int i;

	for(i=0;i<256;i++)
	{
		NET_HASH[i] = (IRCD_NETHASH **)MyMalloc(
								sizeof(IRCD_NETHASH *)*HASH_IP_SIZE);
	}
	return 0;
}

int iphash_handle_accept(ircd_uint32_t ipaddr)
{
	int	rate_num = (int)vars_value_get_bynum(IRCD_VAR_IPADDR_ACCEPT_RATE_NUM);
	int rate_time = (int)vars_value_get_bynum(IRCD_VAR_IPADDR_ACCEPT_RATE_TIME);
	IRCD_NETHASH	*nhptr;

	if (!rate_num || !rate_time)
		return 0;
	nhptr = nethash_find(ipaddr);
	if (!nhptr)
	{
		nhptr = nethash_create(ipaddr);
		nhptr->accept_rate_num = 1;
		nhptr->accept_rate_time = Now;
		return 0;
	}
	if (rate_time > nhptr->accept_rate_maxtime)
		nhptr->accept_rate_maxtime = rate_time;
	if (nhptr->accept_rate_time + rate_time < Now)
	{
		nhptr->accept_rate_persec = nhptr->accept_rate_num/(Now - nhptr->accept_rate_time);
		if ((nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED) &&
				(nhptr->accept_rate_num <= rate_num))
		{
			int nipaddr = htonl(nhptr->ipaddr);

			nhptr->flags &= ~IRCD_NETHASH_FLAGS_ATHROTTLED;
			if (!(nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED))
				del_nethash_from_tlist(&Throttles, nhptr);
			send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
				"Unthrottling accepts from %s",
				inet_ntoa(*((struct in_addr *)&nipaddr)));
		}
		nhptr->accept_rate_time = Now;
		nhptr->accept_rate_num = 0;
	}
	nhptr->accept_rate_num++;
	if (nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED)
		return 1;
	if (nhptr->accept_rate_num > rate_num)
	{
		nhptr->flags |= IRCD_NETHASH_FLAGS_ATHROTTLED;
		if (!(nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED))
			add_nethash_to_tlist(&Throttles, nhptr);
		if (nhptr->accept_rate_num == rate_num + 1)
		{
			int nipaddr = htonl(nhptr->ipaddr);

#if 0
			if (nhptr->accept_throttle_notice + 900 < Now)
			{
#endif
			send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
				"Throttling accepts from %s: (%i > %i) within %is",
				inet_ntoa(*((struct in_addr *)&nipaddr)),
				nhptr->accept_rate_num, rate_num, rate_time);
			nhptr->accept_throttle_notice = Now;

			if (vars_value_get_bynum(IRCD_VAR_CLONES_LOG))
				log_printf(_ircd_conf.log_clones, "%s Throttling accepts from %s: (%i > %i) within %is\n",
					support_ctime(Now),
					inet_ntoa(*((struct in_addr *)&nipaddr)),
					nhptr->accept_rate_num, rate_num, rate_time);

#if 0
			}
#endif
		}
		return 1;
	}
	return 0;
}

int iphash_handle_client(IRCD_CLIENT *pptr)
{
	IRCD_NETHASH	*nhptr;
	IRCD_IPENTRY	*ieptr;

	if (pptr->c_rem_ipaddr == -1) /* we don't know their IP */
		return 0;

	nhptr = nethash_find(pptr->c_rem_ipaddr);
	if (!nhptr)
		nhptr = nethash_create(pptr->c_rem_ipaddr);
	if (nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED)
	{
		client_del(meptr, pptr, "Accepts from your network are throttled");
		return 1; /* these got by the accept() check */
	}
	if (IsLocal(pptr))
	{
		int rate_time = pptr->c_class->ipaddr_rate_time;
		int rate_num = pptr->c_class->ipaddr_rate_num;

		if (!rate_num)
			rate_num = (int)vars_value_get_bynum(IRCD_VAR_IPADDR_CLIENT_RATE_NUM);
		if (!rate_time)
			rate_time = (int)vars_value_get_bynum(IRCD_VAR_IPADDR_CLIENT_RATE_TIME);
		if (rate_time && rate_num)
		{
			if (rate_time > nhptr->client_rate_maxtime)
				nhptr->client_rate_maxtime = rate_time;
			if (nhptr->client_rate_time + rate_time < Now)
			{
				if ((nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED) &&
							(nhptr->accept_rate_num <= rate_num))
				{
					ircd_uint32_t nipaddr = htonl(nhptr->ipaddr);

					nhptr->flags &= ~IRCD_NETHASH_FLAGS_CTHROTTLED;
					if (!(nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED))
						del_nethash_from_tlist(&Throttles, nhptr);
					send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
                        "Unthrottling clients from %s",
						inet_ntoa(*((struct in_addr *)&nipaddr)));
				}
				nhptr->client_rate_num = 0;
				nhptr->client_rate_time = Now;
			}
			nhptr->client_rate_num++;
			if (nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED)
			{
				client_del(meptr, pptr, "Clients from your network are throttled");
				return 1;
			}
			if (nhptr->client_rate_num > rate_num)
			{
				nhptr->flags |= IRCD_NETHASH_FLAGS_CTHROTTLED;
				if (!(nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED))
					add_nethash_to_tlist(&Throttles, nhptr);
				if (nhptr->client_rate_num == rate_num+1)
				{
					ircd_uint32_t nipaddr = htonl(nhptr->ipaddr);

					send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
				"Throttling client connections from %s: (%i > %i) within %is",
					inet_ntoa(*((struct in_addr *)&nipaddr)),
					nhptr->client_rate_num, rate_num, rate_time);
				if (vars_value_get_bynum(IRCD_VAR_CLONES_LOG))
					log_printf(_ircd_conf.log_clones,
				"%s Throttling client connections from %s: (%i > %i) within %is\n",
					support_ctime(Now),
					inet_ntoa(*((struct in_addr *)&nipaddr)),
					nhptr->client_rate_num, rate_num, rate_time);

				}
				client_del(meptr, pptr, "Clients from your network are throttled");
				return 1;
			}
		}
	}
	else
		nhptr->client_rate_num++;
	ieptr = ipentry_find(nhptr, pptr->c_rem_ipaddr);
	if (!ieptr)
		ieptr = ipentry_create(nhptr, pptr->c_rem_ipaddr);
	if (IsLocal(pptr))
	{
		nhptr->local_connect_num++;
		nhptr->local_connect_lasttime = Now;
		ieptr->local_connect_num++;
		ieptr->local_connect_lasttime = Now;
	}
	else
	{
		nhptr->remote_connect_num++;
		nhptr->remote_connect_lasttime = Now;
		ieptr->remote_connect_num++;
		ieptr->remote_connect_lasttime = Now;
	}
	if ((nhptr->local_connect_num + nhptr->remote_connect_num) == 3)
		add_nethash_to_clist(&Clones, nhptr);
	add_client_to_ilist(&(ieptr->clients), pptr);
	add_client_to_nlist(&(nhptr->clients), pptr);
	pptr->c_ipaddr_ptr = ieptr;
	return 0;
}

int iphash_handle_exit(IRCD_CLIENT *pptr)
{
	IRCD_IPENTRY	*iptr = pptr->c_ipaddr_ptr;
	IRCD_NETHASH	*nptr = iptr->nhptr;

	del_client_from_ilist(&(iptr->clients), pptr);
	del_client_from_nlist(&(nptr->clients), pptr);
	pptr->c_ipaddr_ptr = NULL;

	if (IsLocal(pptr))
	{
		iptr->local_connect_num--;
		nptr->local_connect_num--;
	}
	else
	{
		iptr->remote_connect_num--;
		nptr->remote_connect_num--;
	}
	if ((nptr->local_connect_num + nptr->remote_connect_num) == 2)
		del_nethash_from_clist(&Clones, nptr);
	if (!iptr->local_connect_num && !iptr->remote_connect_num)
	{
		del_ipentry_from_list(&(nptr->ipaddrs), iptr);
		MyFree(iptr);
		if (!nptr->local_connect_num && !nptr->remote_connect_num)
			_nethash_check_free(nptr);
	}
	return 0;
}

int iphash_cleanup(void)
{
	static int		which = 0;
	IRCD_NETHASH	*nhptr;
	IRCD_NETHASH	*nexth;
	int				i;
	int				ii;
	ircd_uint32_t	nipaddr;

	which = (which + 1) % 6;

	for(nhptr=Throttles;nhptr;nhptr=nexth)
	{
		nexth = nhptr->t_next;

		nipaddr = htonl(nhptr->ipaddr);
		if ((nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED) &&
				(nhptr->accept_rate_time + (2*nhptr->accept_rate_maxtime) < Now))
		{
			nhptr->flags &= ~IRCD_NETHASH_FLAGS_ATHROTTLED;
			send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
				"Unthrottling accepts from %s",
				inet_ntoa(*((struct in_addr *)&nipaddr)));
		}
		if ((nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED) &&
				(nhptr->client_rate_time + (2*nhptr->client_rate_maxtime) < Now))
		{
			nhptr->flags &= ~IRCD_NETHASH_FLAGS_CTHROTTLED;
			send_to_flag(CLIENT_NFLAGS_THROTTLES, NULL,
				"Unthrottling clients from %s",
				inet_ntoa(*((struct in_addr *)&nipaddr)));
		}
		if (!(nhptr->flags & (IRCD_NETHASH_FLAGS_ATHROTTLED|IRCD_NETHASH_FLAGS_CTHROTTLED)))
			del_nethash_from_tlist(&Throttles, nhptr);
	}
	if (!which)
	{
		for(i=0;i<256;i++)
		for(ii=0;ii<HASH_IP_SIZE;ii++)
		for(nhptr = NET_HASH[i][ii];nhptr;nhptr=nexth)
		{
			nexth = nhptr->h_next;
			if (nhptr->local_connect_num || nhptr->remote_connect_num)
				continue;
			_nethash_check_free(nhptr);
		}
	}
	return 0;
}

int iphash_command(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *args)
{
	IRCD_IPENTRY	*ieptr;
	IRCD_NETHASH	*nhptr;
	char			*command = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!command || !*command)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Usage: IPHASH lookup|throttles|clones",
				pptr->name);
		return 3;
	}

	if (!strcasecmp(command, "lookup") || !strcasecmp(command, "show"))
	{
		char			*ipstring = ircd_next_subarg_space(&args);
		ircd_uint32_t	ipaddr;
		ircd_uint32_t	nipaddr;

		if (!ipstring || !*ipstring)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Usage: IPHASH lookup ip-address", pptr->name);
			return 3;
		}
		ipaddr = inet_addr(ipstring);
		if (ipaddr == -1)
		{
			send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Invalid IP address specified", pptr->name);
			return 3;
		}
		ipaddr = ntohl(ipaddr);
		nhptr = nethash_find(ipaddr);
		if (!nhptr)
		{
			send_prefix_to_client(meptr, pptr,
					"NOTICE %s :No information for \"%s\" found",
					pptr->name, ipstring);
			return 3;
		}
		nipaddr = htonl(nhptr->ipaddr);
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Information for \"%s\":", pptr->name,
				inet_ntoa(*(struct in_addr *)&nipaddr));
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s : Current local connections: %i  Last at: %s",
				pptr->name, nhptr->local_connect_num,
				nhptr->local_connect_lasttime ?
					support_ctime(nhptr->local_connect_lasttime) : "Never");
		for(acptr=nhptr->clients;acptr;acptr=acptr->c_n_next)
		{
			if (!IsLocal(acptr))
				continue;

			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :  .%i -- %s!%s@%s on %s",
					pptr->name, acptr->c_rem_ipaddr & 0xff, acptr->name,
					acptr->username, acptr->hostname, acptr->servername);
		}
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current remote connections: %i  Last at: %s",
				pptr->name, nhptr->remote_connect_num,
				nhptr->remote_connect_lasttime ?
					support_ctime(nhptr->remote_connect_lasttime) : "Never");
		for(acptr=nhptr->clients;acptr;acptr=acptr->c_n_next)
		{
			if (IsLocal(acptr))
				continue;

			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :  .%i -- %s!%s@%s on %s",
					pptr->name, acptr->c_rem_ipaddr & 0xff, acptr->name,
					acptr->username, acptr->hostname, acptr->servername);
		}
		if (nhptr->accept_rate_persec)
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Accept rate info: %i attempt%s in last %is.  %i per second average", 
					pptr->name, nhptr->accept_rate_num,
					nhptr->accept_rate_num == 1 ? "" : "s",
					Now - nhptr->accept_rate_time, nhptr->accept_rate_persec);
		else
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Accept rate info: %i attempt%s in last %is.", 
					pptr->name, nhptr->accept_rate_num,
					nhptr->accept_rate_num == 1 ? "" : "s",
					Now - nhptr->accept_rate_time);
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Client rate info: %i attempt%s in last %is.", 
					pptr->name, nhptr->client_rate_num,
					nhptr->client_rate_num == 1 ? "" : "s",
					Now - nhptr->client_rate_time);
		ieptr = ipentry_find(nhptr, ipaddr);
		if (ieptr)
		{
			nipaddr = htonl(ieptr->ipaddr);
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Information for \"%s\":", pptr->name,
				inet_ntoa(*(struct in_addr *)&(nipaddr)));
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s : Current local connections: %i  Last at: %s",
				pptr->name, ieptr->local_connect_num,
				ieptr->local_connect_lasttime ?
					support_ctime(ieptr->local_connect_lasttime) : "Never");
			for(acptr=ieptr->clients;acptr;acptr=acptr->c_i_next)
			{
				if (!IsLocal(acptr))
					continue;
	
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :  %s!%s@%s on %s",
						pptr->name, acptr->name,
						acptr->username, acptr->hostname, acptr->servername);
			}
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current remote connections: %i  Last at: %s",
				pptr->name, ieptr->remote_connect_num,
				ieptr->remote_connect_lasttime ?
					support_ctime(ieptr->remote_connect_lasttime) : "Never");
			for(acptr=ieptr->clients;acptr;acptr=acptr->c_i_next)
			{
				if (IsLocal(acptr))
					continue;
	
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :  %s!%s@%s on %s",
						pptr->name, acptr->name,
						acptr->username, acptr->hostname, acptr->servername);
			}
		}
	}
	else if (!strcasecmp(command, "clones"))
	{
		char			*numsame = ircd_next_subarg_space(&args);
		int				shown = 0;
		int				num = 5;
		int				ipaddr;
		IRCD_NETHASH	*nhptr;
		char			which = 'l';

		if (numsame && *numsame)
			num = atoi(numsame);
		for(nhptr=Clones;nhptr;nhptr=nhptr->c_next)
		{
			if ((nhptr->local_connect_num >= num) ||
					((which == 'r') && (nhptr->remote_connect_num >= num)) ||
					((which == 'a') &&
				((nhptr->remote_connect_num+nhptr->local_connect_num) >= num)))
			{
				ipaddr = htonl(nhptr->ipaddr);
				shown++;
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :%s -- %i local and %i remote connection%s",
					pptr->name, inet_ntoa(*((struct in_addr *)&ipaddr)),
					nhptr->local_connect_num,
					nhptr->remote_connect_num,
					nhptr->remote_connect_num == 1 ? "" : "s");
				for(acptr=nhptr->clients;acptr;acptr=acptr->c_n_next)
				{
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :  .%i -- %s!%s@%s on %s",
						pptr->name, acptr->c_rem_ipaddr & 0xff, acptr->name,
						acptr->username, acptr->hostname, acptr->servername);
				}
			}
		}
		if (!shown)
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :No clients found for num >= %i.", pptr->name,
				num);
	}
	else if (!strcasecmp(command, "throttles"))
	{
		ircd_uint32_t	ipaddr;

		for(nhptr=Throttles;nhptr;nhptr=nhptr->t_next)
		{
			ipaddr = htonl(nhptr->ipaddr);

			if (nhptr->flags & IRCD_NETHASH_FLAGS_ATHROTTLED)
			{
				if (nhptr->accept_rate_persec)
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Accepts throttled for %s: %i attempt%s in last %is.  %i per second average", 
						pptr->name, inet_ntoa(*((struct in_addr *)&ipaddr)),
						nhptr->accept_rate_num,
						nhptr->accept_rate_num == 1 ? "" : "s",
						Now - nhptr->accept_rate_time,
						nhptr->accept_rate_persec);
				else
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Accepts throttled for %s: %i attempt%s in last %is.",
						pptr->name, inet_ntoa(*((struct in_addr *)&ipaddr)),
						nhptr->accept_rate_num,
						nhptr->accept_rate_num == 1 ? "" : "s",
						Now - nhptr->accept_rate_time);
			}
			if (nhptr->flags & IRCD_NETHASH_FLAGS_CTHROTTLED)
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Clients throttled for %s: %i attempt%s in last %is.", 
						pptr->name, inet_ntoa(*((struct in_addr *)&ipaddr)),
						nhptr->client_rate_num,
						nhptr->client_rate_num == 1 ? "" : "s",
						Now - nhptr->client_rate_time);
			}
		}
		if (!Throttles)
			send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Nothing is being throttled.", pptr->name);
	}
	else
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Unknown IPHASH command.", pptr->name);
	}
	return 3;
}

int iphash_count_memory(int *num_nethash, int *num_ips, int *size_nethash)
{
	int				ih_num_nethash = 0;
	int				ih_num_ips = 0;
	int				i;
	int				ii;
	IRCD_NETHASH	*nhptr;
	IRCD_IPENTRY	*ieptr;

	for(i=0;i<256;i++)
	for(ii=0;ii<HASH_IP_SIZE;ii++)
	for(nhptr=NET_HASH[i][ii];nhptr;nhptr=nhptr->h_next)
	{
		ih_num_nethash++;
		for(ieptr=nhptr->ipaddrs;ieptr;ieptr=ieptr->next)
			ih_num_ips++;
	}

	*num_nethash = ih_num_nethash;
	*num_ips = ih_num_ips;
	*size_nethash = sizeof(NET_HASH);

	return 0;
}

int iphash_init_reload(void ***nethash, int nethash_size, void *clones, void *throttles)
{
	int				i;
	int				ii;
	IRCD_NETHASH	*nhptr;
	IRCD_NETHASH	*nextnhptr;

	Clones = (IRCD_NETHASH *)clones;
	Throttles = (IRCD_NETHASH *)throttles;
	if (nethash_size == HASH_IP_SIZE)
	{
		for(i=0;i<256;i++)
			NET_HASH[i] = ((IRCD_NETHASH ***)nethash)[i];
		MyFree(nethash);
		return 0;
	}

	for(i=0;i<256;i++)
	{
		for(ii=0;ii<nethash_size;ii++)
		{
			for(nhptr=((IRCD_NETHASH ***)nethash)[i][ii];nhptr;nhptr=nextnhptr)
			{
				nextnhptr = nhptr->h_next;

				add_nethash_to_hlist(&iphash_bucket_get(nhptr->ipaddr), nhptr);
			}
		}
	}

	return 0;
}

int iphash_deinit_reload(void ****nethash, int *nethash_size, void **clones, void **throttles)
{
	if (nethash)
	{
		int i;

		*nethash = (void ***)MyMalloc(sizeof(IRCD_NETHASH **) * 256);

		for(i=0;i<256;i++)
			(*nethash)[i] = (void **)NET_HASH[i];
	}
	if (nethash_size)
		*nethash_size = HASH_IP_SIZE;
	if (clones)
		*clones = (void *)Clones;
	if (throttles)
		*throttles = (void *)Throttles;
	return 0;
}

