/*
**   IRC - Internet Relay Chat, source/match.c
**   Copyright (C) 1990 Jarkko Oikarinen
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "match.h"

RCSID("$Id: match.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

static int _match(const char *mask, const char *name);

/*
**  Compare if a given string (name) matches the given
**  mask (which can contain wild cards: '*' - match any
**  number of chars, '?' - match any single character.
**
**	return	0, if match
**		1, if no match
*/

/*
** _match()
** Iterative matching function, rather than recursive.
** Written by Douglas A Lewis (dalewis@acsu.buffalo.edu)
*/

static int _match(const char *mask, const char *name)
{
	const u_char	*m = (u_char *)mask;
	const u_char	*n = (u_char *)name;
	const char		*ma = mask;
	const char		*na = name;
	int		num_matched = 1;
	int		wild = 0;
	int		q = 0;
	int		blah;

	while (1)
	{
		if (*m == '*')
		{
			while (*m == '*')
				m++;
			wild = 1;
			ma = (char *)m;
			na = (char *)n;
		}

		if (!*m)
		{
	  		if (!*n)
				return num_matched;
	  		for (m--; (m > (u_char *)mask) && (*m == '?'); m--)
				;
			if ((*m == '*') && (m > (u_char *)mask) &&
				(m[-1] != '\\'))
				return num_matched;
			if (!wild) 
				return 0;
			m = (u_char *)ma;
			n = (u_char *)++na;
		}
		else if (!*n)
		{
			while(*m == '*')
				m++;
			return (*m == 0) ? num_matched : 0;
		}
		if ((*m == '\\') && ((m[1] == '*') || (m[1] == '?')))
		{
			m++;
			q = 1;
		}
		else
			q = 0;

		if ((blah=(tolower(*m) != tolower(*n))) && ((*m != '?') || q))
		{
			if (!wild)
				return 0;
			m = (u_char *)ma;
			n = (u_char *)++na;
		}
		else
		{
			if (!blah)
				num_matched++;
			if (*m)
				m++;
			if (*n)
				n++;
		}
	}
}

/*
** External interfaces to the above matching routine.
*/
int	match(const char *ma, const char *na)
{
	return (_match(ma, na) == 0);
}

int	matches(const char *ma, const char *na)
{
	return _match(ma, na);
}

/*
** collapse a pattern string into minimal components.
** This particular version is "in place", so that it changes the pattern
** which is to be reduced to a "minimal" size.
*/
char *collapse(char *pattern)
{
	char	*s = pattern;
	char	*s1;
	char	*t;

	if (!pattern)
		return pattern;
	/*
	 * Collapse all \** into \*, \*[?]+\** into \*[?]+
	 */
	for (; *s; s++)
		if (*s == '\\')
		{
			if (!*(s + 1))
				break;
			else
				s++;
		}
		else if (*s == '*')
		{
			if (*(t = s1 = s + 1) == '*')
				while (*t == '*')
					t++;
			else if (*t == '?')
				for (t++, s1++; *t == '*' || *t == '?'; t++)
					if (*t == '?')
						*s1++ = *t;
			while ((*s1++ = *t++))
				;
		}
	return pattern;
}


/*
**  Case insensitive comparison of two NULL terminated strings.
**
**	returns	 0, if s1 equal to s2
**		<0, if s1 lexicographically less than s2
**		>0, if s1 lexicographically greater than s2
*/
int	mycmp(const char *s1, const char *s2)
{
	register const unsigned char	*str1 = (unsigned char *)s1;
	register const unsigned char	*str2 = (unsigned char *)s2;
	register int					res;

	while ((res = toupper(*str1) - toupper(*str2)) == 0)
	{
		if (*str1 == '\0')
			return 0;
		str1++;
		str2++;
	}
	return (res);
}

int	myncmp(const char *str1, const char *str2, int n)
{
	register const unsigned char	*s1 = (unsigned char *)str1;
	register const unsigned char	*s2 = (unsigned char *)str2;
	register int					res;

	while ((res = toupper(*s1) - toupper(*s2)) == 0)
	{
		s1++;
		s2++;
		n--;
		if (n == 0 || (*s1 == '\0' && *s2 == '\0'))
			return 0;
	}
	return (res);
}

unsigned char tolowertab[] =
		{ 0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
		  0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
		  0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
		  0x1e, 0x1f,
		  ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
		  '*', '+', ',', '-', '.', '/',
		  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		  ':', ';', '<', '=', '>', '?',
		  '@', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
		  'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
		  't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
		  '_',
		  '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
		  'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
		  't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
		  0x7f,
		  0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
		  0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
		  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
		  0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
		  0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
		  0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
		  0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
		  0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
		  0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
		  0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
		  0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
		  0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
		  0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
		  0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
		  0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
		  0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff };

unsigned char touppertab[] =
		{ 0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
		  0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
		  0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
		  0x1e, 0x1f,
		  ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
		  '*', '+', ',', '-', '.', '/',
		  '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
		  ':', ';', '<', '=', '>', '?',
		  '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
		  'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
		  'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
		  0x5f,
		  '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
		  'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
		  'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
		  0x7f,
		  0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
		  0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
		  0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
		  0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
		  0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
		  0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
		  0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
		  0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
		  0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
		  0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
		  0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
		  0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
		  0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
		  0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
		  0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
		  0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff };

unsigned char char_atribs[] = {
/* 0-7 */	MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL,
/* 8-12 */	MATCH_CNTRL, MATCH_CNTRL|MATCH_SPACE, MATCH_CNTRL|MATCH_SPACE, MATCH_CNTRL|MATCH_SPACE, MATCH_CNTRL|MATCH_SPACE,
/* 13-15 */	MATCH_CNTRL|MATCH_SPACE, MATCH_CNTRL, MATCH_CNTRL,
/* 16-23 */	MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL,
/* 24-31 */	MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL, MATCH_CNTRL,
/* space */	MATCH_PRINT|MATCH_SPACE,
/* !"#$%&'( */	MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT,
/* )*+,-./ */	MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT,
/* 0123 */	MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT,
/* 4567 */	MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT,
/* 89:; */	MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT|MATCH_DIGIT, MATCH_PRINT, MATCH_PRINT,
/* <=>? */	MATCH_PRINT, MATCH_PRINT, MATCH_PRINT, MATCH_PRINT,
/* @ */		MATCH_PRINT,
/* ABC */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* DEF */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* GHI */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* JKL */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* MNO */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* PQR */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* STU */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* VWX */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* YZ[ */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* \]^ */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* _   */	MATCH_PRINT,
/* abc */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* def */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* ghi */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* jkl */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* mno */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* pqr */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* stu */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* vwx */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* yz{ */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* \}~ */	MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA, MATCH_PRINT|MATCH_ALPHA,
/* del */	0,
/* 80-8f */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* 90-9f */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* a0-af */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* b0-bf */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* c0-cf */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* d0-df */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* e0-ef */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
/* f0-ff */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
		};

