/*
**   IRC - Internet Relay Chat, source/misc.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
#include "misc.h"

RCSID("$Id: misc.c,v 1.19 2001/09/30 11:02:14 cbehrens Exp $");

typedef struct _command			COMMAND;

struct _command
{
	char			*command;
	commandfunc_t	function;
};

typedef struct _commandentry	COMMANDENTRY;

struct _commandentry
{
	COMMAND			*mptr;
	COMMANDENTRY	*next[27];
};

static COMMANDENTRY *_commandtree[27];
static COMMAND	_Commands[] =
{
	{	"SET",		m_set		},
	{	"PRIVMSG",	m_privmsg	},
	{	"WHO",		m_who		},
	{	"WHOIS",	m_whois		},
	{	"WHOWAS",	m_whowas	},
	{	"USER",		m_user		},
	{	"NICK",		m_nick		},
	{	"SERVER",	m_server	},
	{	"LIST",		m_list		},
	{	"TOPIC",	m_topic		},
	{	"INVITE",	m_invite	},
	{	"VERSION",	m_version	},
	{	"QUIT",		m_quit		},
	{	"SQUIT",	m_squit		},
	{	"KILL",		m_kill		},
	{	"INFO",		m_info		},
	{	"LINKS",	m_links		},
	{	"STATS",	m_stats		},
	{	"HELP",		m_help		},
	{	"ERROR",	m_error		},
	{	"AWAY",		m_away		},
	{	"CONNECT",	m_connect	},
	{	"PING",		m_ping		},
	{	"PONG",		m_pong		},
	{	"OPER",		m_oper		},
	{	"PASS",		m_pass		},
	{	"WALLOPS",	m_wallops	},
	{	"TIME",		m_time		},
	{	"NAMES",	m_names		},
	{	"ADMIN",	m_admin		},
	{	"TRACE",	m_trace		},
	{	"NOTICE",	m_notice	},
	{	"JOIN",		m_join		},
	{	"PART",		m_part		},
	{	"LUSERS",	m_lusers	},
	{	"MOTD",		m_motd		},
	{	"MODE",		m_mode		},
	{	"KICK",		m_kick		},
	{	"USERHOST",	m_userhost	},
	{	"ISON",		m_ison		},
	{	"NOTE",		m_note		},
#ifdef USE_SERVICES
	{	"SERVICE",	m_service	},
	{	"SQUERY",	m_squery	},
	{	"SERVLIST",	m_servlist	},
	{	"SERVSET",	m_servset	},
#endif
	{	"REHASH",	m_rehash	},
	{	"RESTART",	m_restart	},
	{	"CLOSE",	m_close		},
	{	"SVINFO",	m_svinfo	},
	{	"SJOIN",	m_sjoin		},
	{	"DIE",		m_die		},
	{	"DNS",		m_dns		},
	{	"GLINE",	m_gline		},
	{	"EXT",		m_ext		},
	{	"OPERWALL",	m_operwall	},
	{	"KLINE",	m_kline		},
	{	"QUEUE",	m_queue		},
	{	"FLAGS",	m_flags		},
	{	"IPHASH",	m_iphash	},
	{	"CAPAB",	m_capab		},
	{	"REJECTHOLD",	m_rejecthold	},
	{	"LOCOPS",	m_locops	},
	{	"UNKLINE",	m_unkline	},
	{	NULL,		NULL		}
};

static void _command_tree_add(char *name, COMMAND *cptr);
static COMMAND *_command_tree_find(char *name);
static int _client_del_remote(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason, int squits, char *clireason);
static int _client_del_local(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason);
static int _client_res_callback1(IRCDRES *ircdres, IRCDRES_RESULT *result, IRCD_CLIENT *cptr);
static int _client_res_callback2(IRCDRES *ircdres, IRCDRES_RESULT *result, IRCD_CLIENT *cptr);

static void _command_tree_free(COMMANDENTRY *ce)
{
	int				i;

	if (!ce)
		return;

	for(i=0;i<27;i++)
	{
		if (ce->next[i])
		{
			_command_tree_free(ce->next[i]);
			MyFree(ce->next[i]);
			ce->next[i] = NULL;
		}
	}
}

static void _command_tree_add(char *name, COMMAND *cptr)
{
	char			*ptr = name;
	COMMANDENTRY	*lastentry = NULL;
	COMMANDENTRY	*entry;
	unsigned int	num;

	if (!*ptr)
		return;
	do
	{
		if (!*ptr)
			num = 0;
		else if (islower(*ptr))
			num = toupper(*ptr) - 'A' + 1;
		else
			num = *ptr - 'A' + 1;
		if (num >= 27)
			return;
		if (!lastentry)
			entry = _commandtree[num];
		else
			entry = lastentry->next[num];
		if (!entry)
   		{
			entry = (COMMANDENTRY *)MyMalloc(sizeof(COMMANDENTRY));
			BZERO(entry, sizeof(COMMANDENTRY));
			if (!lastentry)
				_commandtree[num] = entry;
			else  
				lastentry->next[num] = entry;
		}
		lastentry = entry;
	} while(*ptr++); 
	if (entry->mptr)
		return;
	entry->mptr = cptr;
}

static COMMAND *_command_tree_find(char *name)
{
	char			*ptr = name;
	COMMANDENTRY	*lastentry = NULL;
	COMMANDENTRY	*entry;
	unsigned int	num;

	if (!*ptr)
		return NULL;
	do
	{
		if (!*ptr)
			num = 0;
		else if (islower(*ptr))
			num = toupper(*ptr) - 'A' + 1;
        else
            num = *ptr - 'A' + 1;
		if (num >= 27)
			return NULL;
		if (!lastentry)
			entry = _commandtree[num];
		else
			entry = lastentry->next[num];
		lastentry = entry;
		if (!entry)
			return NULL;
	} while(*ptr++);
	return entry->mptr;
}

void command_tree_init(void)
{
	COMMAND	*cptr = _Commands;
	int		i;

	for(i=0;i<27;i++)
		_commandtree[i] = NULL;
	for(;cptr->command;cptr++)
		_command_tree_add(cptr->command, cptr);
}

int ircd_open_fd(int fd, int flags, IRCD_CONN *ptr, FDLIST *fdlistptr)
{
	FDLISTINFO	*fdlistinfo;

	if (fd < 0)
		return -EINVAL;
	if (fd >= Fds_size)
	{
		Fds_size = fd + 1024;
		Fds = (IRCD_CONN **)MyRealloc(Fds, sizeof(IRCD_CONN *) * Fds_size);
	}
	if (fd >= Sentalong_size)
	{
		Sentalong_size = fd + 1024;
		Sentalong = (int *)MyRealloc(Sentalong, sizeof(int) * Sentalong_size);
		Sentalong_end = Sentalong + Sentalong_size;
		Sentalong_marker = 0; /* next "send" call will 0 Sentalong out */
	}
	SET_FDS(fd, ptr);
	fdlistinfo = &(ptr->fdlistinfo);
	fdlistinfo->fd = fd;
	fdlistinfo->fdlist = fdlistptr ? fdlistptr : Fdlist_main;
	if (flags)
	{
		fdlistinfo->fdlist_pos = fdlist_add_fd(fdlistinfo->fdlist, fd, flags);
		return fdlistinfo->fdlist_pos;
	}
	fdlistinfo->fdlist_pos = -1;
	return 0;
}

int ircd_close_fd(FDLISTINFO *fdlistinfo)
{
	int		ret;
	int		mfd;

	if (!fdlistinfo || fdlistinfo->fd < 0)
		return -EINVAL;
	close(fdlistinfo->fd);
	SET_FDS(fdlistinfo->fd, NULL);
	if (fdlistinfo->fdlist && fdlistinfo->fdlist_pos >= 0)
	{
		ret = fdlist_del_fd(fdlistinfo->fdlist, fdlistinfo->fdlist_pos, &mfd);
		if ((ret > 0) && (mfd >= 0)) /* update moved fds's position in fdlist */
		{
			if (GET_FDS(mfd))
				GET_FDS(mfd)->fdlistinfo.fdlist_pos = fdlistinfo->fdlist_pos;
		}
	}
	return 0;
}

int client_add_rejecthold(int fd, struct sockaddr_in *rem_sin)
{
	IRCD_REJECTHOLD	*rhptr;
	int				ret;

	if ((ret = mysocket_set_nonblocking(fd)) < 0)
		return ret;
	rhptr = (IRCD_REJECTHOLD *)MyMalloc(sizeof(IRCD_REJECTHOLD));
	if (!rhptr)
		return -ENOMEM;
	rhptr->rh_connptr = (IRCD_CONN *)MyMalloc(sizeof(IRCD_CONN));
	if (!rhptr->rh_connptr)
	{
        MyFree(rhptr->rh_connptr);
		MyFree(rhptr);
		return -ENOMEM;
	}
	rhptr->rh_connptr->conn_type = IRCD_CONN_TYPE_REJECTHOLD;
	rhptr->rh_connptr->conn_ptr = (void *)rhptr;
    rhptr->rh_rem_ipaddr = ntohl(rem_sin->sin_addr.s_addr);
    rhptr->rh_rem_port = ntohs(rem_sin->sin_port);
	rhptr->rh_lasttime = rhptr->rh_firsttime = Now;
    ret = ircd_open_fd(fd, FDLIST_FLAGS_READ, rhptr->rh_connptr, NULL);
    if (ret < 0)
    {
        MyFree(rhptr->rh_connptr);
        MyFree(rhptr);
        return ret;
    }
	add_rejecthold_to_rhlist(&Rejectholds, rhptr);
	return 0;
}

static int _client_res_callback1(IRCDRES *ircdres, IRCDRES_RESULT *result, IRCD_CLIENT *cptr)
{
	IRCDRES_HOSTENT	*heptr;
	int				sh;

	sh = (int)vars_value_get_bynum(IRCD_VAR_SHOW_HEADERS);

	cptr->c_res_req->rl_req = NULL;
	cptr->c_flags &= ~CLIENT_FLAGS_DOINGDNS;
	if (result)
	{
		int ret;

		if (sh)
		{
			send_to_client(cptr, "NOTICE * :*** Reverse DNS reply received.");
			send_to_client(cptr, "NOTICE * :*** Doing forward DNS lookup...");
		}

		cptr->c_res_req->rl_reverse = result->res_he;
		ret = ircdres_gethostbyname((IRCDRES *)(Ircd_resconn->conn_ptr), result->res_he->he_he.h_name,
				(ircdres_func_t)_client_res_callback2, (void *)cptr,
				&heptr, &cptr->c_res_req->rl_req);
		if ((ret >= 0) && heptr)
		{
			if (sh)
				send_to_client(cptr, "NOTICE * :*** Forward DNS reply received.");
			cptr->c_res_req->rl_forward = heptr;
			del_reqlist_from_list(&Reqlist, cptr->c_res_req);
			assert(0); /* FIX if we finish dns cache */
		}
		else if (ret == -EINPROGRESS)
		{
			cptr->c_flags |= CLIENT_FLAGS_DOINGDNS;
			cptr->c_res_req->rl_firsttime = Now;
			cptr->c_res_req->rl_lasttime = Now;
		}
		else
		{
			if (sh)
				send_to_client(cptr, "NOTICE * :*** Forward DNS request failed.");
			del_reqlist_from_list(&Reqlist, cptr->c_res_req);
		}
	}
	else
	{
		if (sh)
			send_to_client(cptr, "NOTICE * :*** Reverse DNS request failed.");
		del_reqlist_from_list(&Reqlist, cptr->c_res_req);
	}
	if (!(cptr->c_flags & (CLIENT_FLAGS_DOINGAUTH|CLIENT_FLAGS_DOINGDNS)))
	{
		ircd_parse(cptr);
	}
	return 0;
}

static int _client_res_callback2(IRCDRES *ircdres, IRCDRES_RESULT *result, IRCD_CLIENT *cptr)
{
	int		sh = (int)vars_value_get_bynum(IRCD_VAR_SHOW_HEADERS);

	cptr->c_res_req->rl_req = NULL;
	cptr->c_flags &= ~CLIENT_FLAGS_DOINGDNS;
	del_reqlist_from_list(&Reqlist, cptr->c_res_req);
	if (result)
	{
		struct in_addr	inaddr;

		if (sh)
			send_to_client(cptr, "NOTICE * :*** Forward DNS reply received.");

		cptr->c_res_req->rl_forward = result->res_he;

		inaddr.s_addr = htonl(cptr->c_rem_ipaddr);
		if (cptr->c_res_req->rl_forward->he_he.h_addr_list[0] &&
			(cptr->c_res_req->rl_forward->he_he.h_length ==
						sizeof(struct in_addr)) &&
			!memcmp(cptr->c_res_req->rl_forward->he_he.h_addr_list[0],
				&inaddr, sizeof(struct in_addr)))
		{
			if (strlen(cptr->c_res_req->rl_reverse->he_he.h_name) < HOSTLEN)
			{
				strcpy(cptr->hostname, cptr->c_res_req->rl_reverse->he_he.h_name);
			}
		}
	}
	else
	{
		if (sh)
			send_to_client(cptr, "NOTICE * :*** Forward DNS request failed.");
	}
	if (!(cptr->c_flags & CLIENT_FLAGS_DOINGAUTH))
	{
		ircd_parse(cptr);
	}
	return 0;
}

int client_add_local(IRCD_LISTENER *listener, int fd, struct sockaddr_in *rem_sin)
{
    int                 ret;
    struct sockaddr_in  loc_sin;
    int                 sin_size = sizeof(loc_sin);
    IRCD_CLIENT         *cptr;
	IRCD_AUTH			*authptr;
	int					afd;
	IRCDRES_HOSTENT		*heptr;
	struct in_addr		inaddr;
	int					sh;

	ret = getsockname(fd, (struct sockaddr *)(&loc_sin), &sin_size);
	if (ret < 0)
		return -errno;
	if ((ret = mysocket_set_nonblocking(fd)) < 0)
		return ret;
	cptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	if (!cptr)
		return -ENOMEM;
	cptr->c_connptr = (IRCD_CONN *)MyMalloc(sizeof(IRCD_CONN));
	if (!cptr->c_connptr)
	{
		MyFree(cptr);
		return -ENOMEM;
	}
	cptr->c_connptr->conn_type = IRCD_CONN_TYPE_CLIENT;
	cptr->c_connptr->conn_ptr = (void *)cptr;
	cptr->c_maxsendq = 8192;
	cptr->c_from = cptr;
    cptr->c_rem_ipaddr = ntohl(rem_sin->sin_addr.s_addr);
    cptr->c_rem_port = ntohs(rem_sin->sin_port);
    cptr->c_loc_ipaddr = ntohl(loc_sin.sin_addr.s_addr);
    cptr->c_loc_port = ntohs(loc_sin.sin_port);
  	strcpy(cptr->c_ipstring, inet_ntoa(rem_sin->sin_addr));
	cptr->c_listener = listener;
	cptr->c_lasttime = cptr->c_firsttime = Now;
	cptr->c_idletime = Now;

	strcpy(cptr->username, "unknown");
	support_strncpy(cptr->hostname, inet_ntoa(rem_sin->sin_addr),
				sizeof(cptr->hostname));

    ret = queue_init(&(cptr->c_queue_in), QUEUE_SIZE, QUEUEINIT_FLAGS);
    if (ret < 0)
    {
        MyFree(cptr->c_connptr);
        MyFree(cptr);
        return ret;
    }
    ret = queue_init(&(cptr->c_queue_out), QUEUE_SIZE, QUEUEINIT_FLAGS);
    if (ret < 0)
    {
        queue_delete(&(cptr->c_queue_in));
        MyFree(cptr->c_connptr);
        MyFree(cptr);
        return ret;
    }
    ret = ircd_open_fd(fd, FDLIST_FLAGS_READ, cptr->c_connptr, NULL);
    if (ret < 0)
    {
        queue_delete(&(cptr->c_queue_in));
        queue_delete(&(cptr->c_queue_out));
        MyFree(cptr->c_connptr);
        MyFree(cptr);
        return ret;
    }
	add_client_to_lolist(&Locals, cptr);
	cptr->c_flags |= CLIENT_FLAGS_LOCAL;
	add_client_to_llist(&(listener->l_clients), cptr);
	listener->l_num_on++;

	cptr->c_res_req = (IRCDRES_REQLIST *)MyMalloc(sizeof(IRCDRES_REQLIST));
	cptr->c_res_req->rl_req = NULL;
	cptr->c_res_req->rl_firsttime = cptr->c_res_req->rl_lasttime = Now;

	inaddr.s_addr = htonl(cptr->c_rem_ipaddr);

	sh = (int)vars_value_get_bynum(IRCD_VAR_SHOW_HEADERS);

	if (sh)
		send_to_client(cptr, "NOTICE * :*** Doing reverse DNS lookup...");

	ret = ircdres_gethostbyaddr((IRCDRES *)(Ircd_resconn->conn_ptr), &inaddr,
				(ircdres_func_t)_client_res_callback1, (void *)cptr,
				&heptr, &cptr->c_res_req->rl_req);
	if ((ret >= 0) && heptr)
	{
		if (sh)
		{
			send_to_client(cptr, "NOTICE * :*** Reverse DNS reply received.");
			send_to_client(cptr, "NOTICE * :*** Doing forward DNS lookup...");
		}
		cptr->c_res_req->rl_reverse = heptr;
		ret = ircdres_gethostbyname((IRCDRES *)(Ircd_resconn->conn_ptr),
				heptr->he_he.h_name,
				(ircdres_func_t)_client_res_callback2, (void *)cptr,
				&heptr, &cptr->c_res_req->rl_req);
		if ((ret >= 0) && heptr)
		{
			if (sh)
				send_to_client(cptr, "NOTICE * :*** Forward DNS reply received.");
			cptr->c_res_req->rl_forward = heptr;
		}
		else if (ret == -EINPROGRESS)
		{
			cptr->c_flags |= CLIENT_FLAGS_DOINGDNS;
			add_reqlist_to_list(&Reqlist, cptr->c_res_req);
		}
	}
	else if (ret == -EINPROGRESS)
	{
		cptr->c_flags |= CLIENT_FLAGS_DOINGDNS;
		add_reqlist_to_list(&Reqlist, cptr->c_res_req);
	}

	if (sh)
		send_to_client(cptr, "NOTICE * :*** Connecting for ident request...");

	afd = mysocket_connectip(cptr->c_rem_ipaddr, 113,
		cptr->c_loc_ipaddr, MYSOCKET_FLAGS_REUSEADDR|MYSOCKET_FLAGS_NONBLOCKING);
	if (afd >= 0)
	{
		authptr = cptr->c_auth_ptr = (IRCD_AUTH *)MyMalloc(sizeof(IRCD_AUTH));
		authptr->a_connptr = (IRCD_CONN *)MyMalloc(sizeof(IRCD_CONN));
		authptr->a_connptr->conn_type = IRCD_CONN_TYPE_AUTH;
		authptr->a_connptr->conn_ptr = (void *)authptr;
		authptr->a_clptr = cptr;
		authptr->a_firsttime = Now;
		cptr->c_flags |= CLIENT_FLAGS_DOINGAUTH;
	    ret = queue_init(&(authptr->a_queue_in), QUEUE_SIZE, QUEUEINIT_FLAGS);
	    if (ret < 0)
	    {
			if (sh)
				send_to_client(cptr, "NOTICE * :*** Ident request failed: in_queue init failed");
			close(afd);
	        MyFree(authptr->a_connptr);
	        MyFree(authptr);
			cptr->c_auth_ptr = NULL;
			cptr->c_flags &= ~CLIENT_FLAGS_DOINGAUTH;
	        return 0;
	    }
		if (ircd_open_fd(afd, FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE, authptr->a_connptr, NULL) < 0)
		{
			if (sh)
				send_to_client(cptr, "NOTICE * :*** Ident request failed: fd init failed");
			queue_delete(&(authptr->a_queue_in));
			close(afd);
	        MyFree(authptr->a_connptr);
			MyFree(authptr);
			cptr->c_auth_ptr = NULL;
			cptr->c_flags &= ~CLIENT_FLAGS_DOINGAUTH;
			return 0;
		}
		add_auth_to_alist(&Auths, authptr);
	}
	else
	{
		if (sh)
			send_to_client(cptr, "NOTICE * :*** Ident request failed.");
	}
    return 0;
}

int rejecthold_del(IRCD_CLIENT *from, IRCD_REJECTHOLD *rptr, char *reason)
{
	ircd_close_fd(&(rptr->rh_connptr->fdlistinfo));
	del_rejecthold_from_rhlist(&Rejectholds, rptr);
	rptr->rh_connptr->conn_ptr = NULL;
	MyFree(rptr->rh_connptr);
	MyFree(rptr);
	return 0;
}

int auth_del(IRCD_CLIENT *from, IRCD_AUTH *cptr, char *reason)
{
	ircd_close_fd(&(cptr->a_connptr->fdlistinfo));
    queue_delete(&(cptr->a_queue_in));
	del_auth_from_alist(&Auths, cptr);
	Auths_did_delete = 1;
	if (cptr->a_clptr)
	{
		cptr->a_clptr->c_flags &= ~CLIENT_FLAGS_DOINGAUTH;
		cptr->a_clptr->c_auth_ptr = NULL;
		cptr->a_clptr = NULL;
	}
	cptr->a_connptr->conn_ptr = NULL;
	MyFree(cptr->a_connptr);
	MyFree(cptr);
	return 0;
}

int conn_del(IRCD_CLIENT *from, IRCD_CONN *cptr, char *reason)
{
	int ret = 0;

	if (IsAuth(cptr))
	{
		ret = auth_del(from, (IRCD_AUTH *)cptr->conn_ptr, reason);
	}
	else if (IsCClient(cptr))
	{
		ret = client_del(from, (IRCD_CLIENT *)cptr->conn_ptr, reason);
	}
	else if (IsRejecthold(cptr))
	{
		ret = rejecthold_del(from, (IRCD_REJECTHOLD *)cptr->conn_ptr, reason);
	}
	return ret;
}


int client_del(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason)
{
	if (IsLocal(cptr))
	{
		char	cnamebuf1[CLIENTNAMESIZE];

		if (!(cptr->c_flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			if (from != NULL && from != cptr)
				send_to_client(cptr, "ERROR :Closing Link: %s %s (%s)",
	                   ircd_clientname(cptr,0,cnamebuf1),
	                   from->name, reason);
			else
				send_to_client(cptr, "ERROR :Closing Link: %s (%s)",
	                   ircd_clientname(cptr,0,cnamebuf1), reason);
		}
		_client_del_local(from, cptr, reason);
		return -1;
	}
	if (IsServer(cptr))
	{
		char	comment[HOSTLEN*2+2];

		strcpy(comment, cptr->c_servptr->name);
		strcat(comment, " ");
		strcat(comment, cptr->name);
		if (from->c_from != cptr->c_from)
			send_to_client(cptr, ":%s SQUIT %s :%s", from->name, cptr->name, reason);
#ifndef REMOVE_CLIENTS_IMMEDIATELY
		else
#endif
			_client_del_remote(from, cptr, reason, 1, comment);
	}
	else
		_client_del_remote(from, cptr, reason, 1, NULL);
	return 0;
}

static int _client_del_remote(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason, int squits, char *clireason)
{
	/* remotes are either servers or clients*/
	IRCD_CLIENT	**servbucket;
	IRCD_CLIENT	**userbucket;
	IRCD_CLIENT	*tempuser;
	IRCD_CLIENT	*nextuser;
	IRCD_CLIENT	*acptr;
	char		cnamebuf1[CLIENTNAMESIZE];

	assert(!IsLocal(cptr) || IsServer(cptr));
	if (!clireason)
		clireason = reason;
	servbucket = &(cptr->c_servers);
	userbucket = &(cptr->c_users);
	if (cptr->c_oslist)
	{
		del_client_from_oslist(cptr->c_oslist, cptr);
		cptr->c_oslist = NULL;
	}
	if (cptr->c_ipaddr_ptr)
	{
		iphash_handle_exit(cptr);
	}
	if (IsServer(cptr))
	{
		if (CliSentEOB(cptr) && !CliGotEOB(cptr))
		{
			Globalinfo->g_serversbursting--;
		}
		if (IsLocal(cptr))
			send_to_flag(CLIENT_NFLAGS_SQUITS, NULL,
				"%s has split: %s",
				cptr->name, reason);
		else
			send_to_flag(CLIENT_NFLAGS_SQUITS, NULL,
				"%s has split from %s: %s",
				cptr->name, cptr->c_servptr->name, reason);
			
		for(tempuser=*servbucket;tempuser;tempuser=nextuser)
		{
			nextuser = tempuser->c_u_next;
			_client_del_remote(from, tempuser, reason, 0, clireason);
		}
		for(tempuser = *userbucket;tempuser;tempuser=nextuser)
		{
			nextuser = tempuser->c_u_next;
			assert(tempuser->c_servptr == cptr);
			_client_del_remote(from, tempuser, reason, 0, clireason);
		}
		for(acptr=meptr->c_servers;acptr;acptr=acptr->c_u_next)
		{
			if (acptr == cptr)
				continue;
			if (!match(conf_myname_for_link(meptr->name,
						acptr->c_hostmask, cnamebuf1), cptr->name))
				continue;
			if (cptr->c_from == acptr)
				continue; /* used to have :prefix squit .. here */
			else
			{
				if (squits ||
						!IsCapable(acptr, IRCD_CAPAB_VALUE_NOQUITSTORM))
				send_to_client(acptr, "SQUIT %s :%s",
					cptr->name, reason);
			}
		}
		*servbucket = NULL;
		*userbucket = NULL;
		hash_del_client(cptr);
		Globalinfo->g_servers--;
		if (cptr->c_servptr) /* has to be true ? */
		{
			del_client_from_ulist(&(cptr->c_servptr->c_servers), cptr);
			cptr->c_servptr = NULL;
		}
		MyFree(cptr);
	}
	else
	{
		whowas_add_history(cptr, 0);
		whowas_off_history(cptr);
		if (!(cptr->c_flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			IRCD_CLIENT	*ssptr;
			for(ssptr=meptr->c_servers;ssptr;ssptr=ssptr->c_u_next)
			{
				if (cptr->c_from == ssptr)
					continue;
				if (!IsCapable(ssptr, IRCD_CAPAB_VALUE_NOQUITSTORM) ||
					(ssptr->c_hostmask &&
					!match(conf_myname_for_link(meptr->name,
							ssptr->c_hostmask, cnamebuf1), cptr->servername)))
				send_to_client(ssptr, ":%s QUIT :%s",
					cptr->name, clireason);
			}
		}
		send_to_common_channels(cptr, "QUIT :%s", clireason);
		if (cptr->c_umodes & CLIENT_UMODES_OPER)
			Globalinfo->g_opers--;
		if (cptr->c_umodes & CLIENT_UMODES_IMODE)
			Globalinfo->g_invis--;
		Globalinfo->g_clients--;
		hash_del_client(cptr);
		_user_del_channels(cptr);
		_channel_invite_del(NULL, cptr);
		if (cptr->c_servptr) /* has to be true ? */
			del_client_from_ulist(&(cptr->c_servptr->c_users), cptr);
		if (cptr->c_away)
			MyFree(cptr->c_away);
		MyFree(cptr);
	}
	return 0;
}

/* called for local connections that are unknown or auth */
static int _client_del_local(IRCD_CLIENT *from, IRCD_CLIENT *cptr, char *reason)
{
	assert(IsLocal(cptr));
	if (cptr->c_connectby)
	{
		MyFree(cptr->c_connectby);
		cptr->c_connectby = NULL;
	}
	client_queue_flush(cptr, NULL);
	ircd_close_fd(&(cptr->c_connptr->fdlistinfo));
	cptr->c_connptr->conn_ptr = NULL;
	MyFree(cptr->c_connptr);
	cptr->c_connptr = NULL;
    queue_delete(&(cptr->c_queue_in));
    queue_delete(&(cptr->c_queue_out));
	cptr->c_flags |= CLIENT_FLAGS_DEAD;
	assert(IsLocal(cptr));
	del_client_from_lolist(&Locals, cptr);
	{
		SERVLINK	*s;
		SERVLINK	*nexts;
		SERVLINK	*lasts = NULL;
		SERVLINK	*lastp = NULL;

		if (IsServer(cptr))
		{
			int connected = Now - cptr->c_firsttime;

			if (connected > 15*60)
			{
				Autoconncheck = 1;

				for(s=_ircd_conf.servers;s;lastp=s,s=s->next)
					continue;
				for(s=_ircd_conf.servers;s;lasts=s,s=nexts)
				{
					nexts = s->next;
					if (!strcmp(s->servername, cptr->name))
					{
						s->tryconnecttime = 0;

						if (!lasts)
							_ircd_conf.servers = s->next;
						else
							lasts->next = s->next;

						s->next = _ircd_conf.servers;
						_ircd_conf.servers = s;
					}
					if (s == lastp)
						break;
				}
			}
			else if (cptr->c_class)
				cptr->c_class->lastconnect = Now;

			send_to_opers("Link with %s closed: %s", cptr->name, reason);
			send_to_opers("%s had been connected for %d day%s, %02d:%02d:%02d",
				cptr->name, connected/86400,
				(connected/86400 == 1) ? "" : "s",
				(connected % 86400) / 3600,
				(connected % 3600) / 60,
				(connected % 60));

		}
		else if (cptr->c_flags & CLIENT_FLAGS_SCONNECT)
		{

			send_to_opers("Link with %s could not be established: %s", cptr->name, reason);
			for(s=_ircd_conf.servers;s;lastp=s,s=s->next)
				continue;
			for(s=_ircd_conf.servers;s;lasts=s,s=nexts)
			{
				nexts = s->next;
				if (!strcasecmp(s->servername, cptr->name))
				{
					s->tryconnecttime = 0;
					if (_ircd_conf.servers->next && (Now - s->lastconnect < 600))
					{
						if (!lasts)
							_ircd_conf.servers = s->next;
						else
							lasts->next = s->next;
						s->next = lastp->next;
						lastp->next = s;
					}
				}
				if (s == lastp)
					break;
			}
		}
	}
	if (cptr->c_res_req)
	{
		if (cptr->c_res_req->rl_req)
		{
			del_reqlist_from_list(&Reqlist, cptr->c_res_req);
			ircdres_request_del(cptr->c_res_req->rl_req, 0);
		}
		if (cptr->c_res_req->rl_reverse)
			MyFree(cptr->c_res_req->rl_reverse);
		if (cptr->c_res_req->rl_forward)
			MyFree(cptr->c_res_req->rl_forward);
		MyFree(cptr->c_res_req);
		cptr->c_res_req = NULL;
	}
	if (cptr->c_auth_ptr)
	{
		cptr->c_auth_ptr->a_clptr = NULL;
		auth_del(from, cptr->c_auth_ptr, NULL);
		cptr->c_auth_ptr = NULL;
    }
	if (cptr->c_away)
	{
		MyFree(cptr->c_away);
		cptr->c_away = NULL;
	}
	if (cptr->c_flood_list)
	{
		del_client_from_flist(cptr->c_flood_list, cptr);
		cptr->c_flood_list = NULL;
		Flooders_did_delete = 1;
	}
	if (IsReg(cptr))
	{
		whowas_add_history(cptr, 0);
		whowas_off_history(cptr);
		if (cptr->c_servptr)
		{
			del_client_from_ulist(&(cptr->c_servptr->c_users), cptr);
			cptr->c_servptr = NULL;
		}
		if (cptr->c_umodes & CLIENT_UMODES_OPER)
		{
			Globalinfo->m_opers--;
			Globalinfo->g_opers--;
		}
		if (cptr->c_umodes & CLIENT_UMODES_IMODE)
		{
			Globalinfo->m_invis--;
			Globalinfo->g_invis--;
		}
		Globalinfo->m_clients--;
		Globalinfo->g_clients--;
		if (IsClient(cptr))
		{
			int on_for = Now - cptr->c_firsttime;

			if (vars_value_get_bynum(IRCD_VAR_USERS_LOG))
				log_printf(_ircd_conf.log_users, "%s (%3d:%02d:%02d): %s!%s@%s [%s] [%s]\n",
					support_ctime(cptr->c_firsttime), on_for/3600,
					(on_for % 3600) / 60, on_for % 60,
					cptr->name, cptr->username, cptr->hostname,
					cptr->c_authusername, cptr->c_ipstring);
		}
	}
	else if (IsServer(cptr))
	{
		assert(cptr->c_servptr == meptr);
		if (cptr->c_servptr)
		{
			del_client_from_ulist(&(cptr->c_servptr->c_servers), cptr);
			cptr->c_servptr = NULL;
		}
		Globalinfo->m_servers--;
#if 0
		/* done in del_remote */
		Globalinfo->g_servers--;
#endif
	}
	if (cptr->c_listener)
	{
		del_client_from_llist(&(cptr->c_listener->l_clients), cptr);
		if (!--cptr->c_listener->l_num_on &&
					(cptr->c_listener->l_flags & IRCD_LISTENER_FLAGS_ILLEGAL))
		{
			assert(cptr->c_listener->l_clients == NULL);
			del_listener_from_llist(&(_ircd_conf.listeners), cptr->c_listener);
			ircd_close_fd(&(cptr->c_listener->l_connptr->fdlistinfo));
			cptr->c_listener->l_connptr->conn_ptr = NULL;
			MyFree(cptr->c_listener->l_connptr);
            MyFree(cptr->c_listener);
		}
		cptr->c_listener = NULL;
	}
	if (*cptr->name && (IsUnknown(cptr) || IsClient(cptr)))
		hash_del_client(cptr);
	if (cptr->c_oslist)
	{
		del_client_from_oslist(cptr->c_oslist, cptr);
		cptr->c_oslist = NULL;
	}
	if (cptr->c_ipaddr_ptr)
	{
		iphash_handle_exit(cptr);
	}
	if (cptr->c_class)
	{
		del_client_from_cllist(&(cptr->c_class->clients), cptr);
		if (!--cptr->c_class->num_on &&
					(cptr->c_class->flags & CLASS_FLAGS_ILLEGAL))
		{
			class_del(&(_ircd_conf.classes), cptr->c_class);
		}
	}
	if (IsServer(cptr))
	{
		char	comment[HOSTLEN*2+2];

		strcpy(comment, meptr->name);
		strcat(comment, " ");
		strcat(comment, cptr->name);
		_client_del_remote(from, cptr, comment, 1, NULL);
		/* cptr is gone */
		return 0;
	}
	if (IsClient(cptr))
	{
		if (!(cptr->c_flags & CLIENT_FLAGS_NOSENDQUIT))
		{
			send_to_servers(cptr->c_from, ":%s QUIT :%s",
				cptr->name, reason);
		}
		if (vars_value_get_bynum(IRCD_VAR_CLIENTCONNECT))
		{
			if (!(cptr->c_flags & CLIENT_FLAGS_KILLED))
				send_to_flag(CLIENT_NFLAGS_CLIDISCONNECTS, NULL,
					"Client exiting: %s (%s@%s) [Client Quit]",
					cptr->name, cptr->username, cptr->hostname);
			else
				send_to_flag(CLIENT_NFLAGS_CLIDISCONNECTS, NULL,
					"Client exiting: %s (%s@%s) [%s]",
					cptr->name, cptr->username, cptr->hostname, reason);
		}
		send_to_common_channels(cptr, "QUIT :%s", reason);
		_user_del_channels(cptr);
		_channel_invite_del(NULL, cptr);
	}
    MyFree(cptr);
    return 0;
}

int ircd_parse(IRCD_CLIENT *cptr)
{
	int				len;
	char			buf[1024];
	IRCD_CLIENT		*from;
	char			*ptr;
	char			*sender;
	char			*command;
	int				numeric;
	COMMAND			*mptr;
	int				ret;

	if (!cptr)
		return -EINVAL;
	if (!cptr->c_queue_in.queue_size)
	{
		return 0;
	}
	cptr->c_flags &= ~CLIENT_FLAGS_PINGSENT;
	cptr->c_lasttime = Now;
	for(;;)
	{
		if (cptr->c_rate_msg_limit && !(IsServer(cptr) ||
			(IsAnOper(cptr) && (cptr->c_oline_flags & OLINE_FLAGS_CANFLOOD))))
		{
			if (Now > cptr->c_rate_time)
			{
				cptr->c_rate_time = Now + cptr->c_rate_time_limit;
				cptr->c_rate_msgs = 0;
				if (cptr->c_flood_list)
				{
					del_client_from_flist(cptr->c_flood_list, cptr);
					cptr->c_flood_list = NULL;
					Flooders_did_delete = 1;
				}
			}
			if (cptr->c_rate_msgs >= cptr->c_rate_msg_limit)
			{
				if (Now > cptr->c_rate_lastflood + 300)
					cptr->c_rate_exit_num = 0;
				cptr->c_rate_lastflood = Now;
				if (!cptr->c_flood_list)
				{
					cptr->c_flood_list = Flooders;
					add_client_to_flist(cptr->c_flood_list, cptr);
					cptr->c_rate_exit_num++;
					if (cptr->c_rate_exit_num_limit &&
						(cptr->c_rate_exit_num >= cptr->c_rate_exit_num_limit))
					{
						return client_del(cptr, cptr, "Too much flooding");
					}
				}
				return 0;
			}
		}

		len = queue_data_get_nl(&(cptr->c_queue_in), buf, sizeof(buf)-1, 0);
		if (len <= 0)
			break;

		buf[512] = '\0';

		cptr->c_recv_msg++;

		for(ptr=buf;*ptr==' ';ptr++)
			continue;

		from = NULL;
		sender = NULL;

		if (*ptr == ':')
		{
			sender = ++ptr;
			while(*ptr && (*ptr != ' '))
				ptr++;
			if (!*ptr)
				continue; /* ":<prefix>\0" */
			if (*sender && IsServer(cptr))
			{
				*ptr++ = '\0';
				from = hash_find_client_incl_hm(sender);
				if (!from)
				{
					/* FIX -- add error notice */
					continue;
				}
				if (from->c_from != cptr)
				{
					/* FIX -- add error notice */
					continue;
				}
			}
			else
				sender = cptr->name;
			while(*ptr == ' ')
				ptr++;
		}
		else
			sender = cptr->name;
		if (!*ptr)
			continue;
		command = ptr;
		while(*ptr && (*ptr != ' '))
			ptr++;
		len = ptr - command;
		numeric = 0;
		if ((len == 3) && *ptr &&
				isdigit(*command) &&
				isdigit(*(command+1)) &&
				isdigit(*(command+2)))
		{
			numeric = ((*command - '0')*100)+
						((*(command + 1)-'0')*10)+(*(command + 2)-'0');
			mptr = NULL;
		}
		else
		{
			if (*ptr)
				*ptr++ = '\0';
			while(*ptr && (*ptr == ' '))
				ptr++;
			mptr = _command_tree_find(command);
			if (!mptr)
			{
				if (IsClient(cptr))
					send_prefix_to_client(meptr, cptr,
						"%d %s %s :Unknown command", ERR_UNKNOWNCOMMAND,
						cptr->name, command);
				cptr->c_rate_msgs++;
				continue;
			}
		}
		if (!from)
			from = cptr;
		if (mptr)
			ret = mptr->function(cptr, from, sender, ptr);
		else if (numeric && IsServer(cptr))
		{
	       /* numeric */
			if (IsServer(from) && CliSentEOB(from) && !CliGotEOB(from))
			{
				if (cptr == from)
				{
					send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
   					"End-of-Burst acknowledged from %s",
					cptr->name);
				}
	            Globalinfo->g_serversbursting--;
	            CliSetGotEOB(from);
	        }
			ret = do_numeric(cptr, from, numeric, sender, ptr); /* do numeric */
		}
		else
			ret = 0;
		if (ret < 0)
			return ret;
		cptr->c_rate_msgs += ret;
	}
	return 0;
}

char *ircd_nextarg(char **src, int flags)
{
	char	*tok;

	if (!src || !*src || !**src)
		return NULL;
	while(**src && **src == ' ')
		(*src)++;
	if (**src)
		tok = *src;
	else
		return NULL;
	if (*tok == ':')
	{
		tok++;
		*src = NULL;
	}
	else if (!(flags & NEXTARG_FLAGS_LASTARG))
	{
		while(**src && **src != ' ')
			(*src)++;
		if (**src)
		{
			**src = '\0';
			(*src)++;
			while(**src && **src == ' ')
				(*src)++;
		}
	}
	else
		*src = NULL;
	return tok;
}

char *ircd_next_subarg_space(char **src)
{
	char	*sptr;
	char	*tok;

	if (!src || !(sptr = *src))
		return NULL;
	while(*sptr == ' ')
		sptr++;
	if (!*sptr)
		return NULL;
	tok = sptr;
	while(*sptr && *sptr != ' ')
		sptr++;
	if (*sptr)
	{
		*sptr++ = '\0';
		while(*sptr == ' ')
			sptr++;
		*src = sptr;
	}
	else
		*src = NULL;
	return tok;
}

char *ircd_next_subarg_comma(char **src)
{
	char	*sptr;
	char	*tok;

	if (!src || !(sptr = *src))
		return NULL;
	while(*sptr == ',')
		sptr++;
	if (!*sptr)
		return NULL;
	tok = sptr;
	while(*sptr && *sptr != ',')
		sptr++;
	if (*sptr)
	{
		*sptr++ = '\0';
		while(*sptr == ',')
			sptr++;
		*src = sptr;
	}
	else
		*src = NULL;
	return tok;
}

char *ircd_clientname(IRCD_CLIENT *sptr, int showip, char *buffer)
{
	if (IsLocal(sptr))
	{
		if (showip)
		{
			sprintf(buffer, "%s[%s%s%s]",
				sptr->name,
				*sptr->c_authusername ? sptr->c_authusername : "",
				*sptr->c_authusername ? "@" : "",
				IsServer(sptr) ? "127.0.0.1" : sptr->c_ipstring);
		}
		else
		{
			if (IsServer(sptr))
			{
				sprintf(buffer, "%s[%s%s127.0.0.1]",
					sptr->name, *sptr->username ? sptr->username : "",
					*sptr->username ? "@" : "");
			}
			else if (*sptr->hostname)
				sprintf(buffer, "%s[%s%s%s]",
					sptr->name, *sptr->username ? sptr->username : "",
					*sptr->username ? "@" : "", sptr->hostname);
			else
				sprintf(buffer, "%s[%s%s%s]",
					sptr->name, *sptr->username ? sptr->username : "",
					*sptr->username ? "@" : "",
					*sptr->servername ? sptr->servername : "UNKNOWN");
		}
	}
	else
		strcpy(buffer, sptr->name);
	return buffer;
}

int client_queue_flush(IRCD_CLIENT *cptr, char *exitbuf)
{
	int ret;

	for(;cptr->c_queue_out.queue_size;)
	{
		char	*ptr;
		int		len;

		ret = queue_data_get(&(cptr->c_queue_out), &ptr, &len);
		assert(ret >= 0);
		ret = write(cptr->c_connptr->fdlistinfo.fd, ptr, len);
		if (ret > 0)
		{
			if (IsClient(cptr))
				Globalinfo->bw_c_outgoing_bytes += ret;
			else if (IsServer(cptr))
				Globalinfo->bw_s_outgoing_bytes += ret;
			cptr->c_writetries = 0;
			queue_data_del(&(cptr->c_queue_out), ret);
			if (ret == len)
				continue;
		}
		else if (!ret)
		{
			cptr->c_flags |= CLIENT_FLAGS_DEAD;
			if (exitbuf)
				strcpy(exitbuf, "Write error: EOF of client");
			return -1;
		}
		else if (ret < 0)
		{
			if ((errno != EAGAIN) && (errno != EINTR) &&
					(errno != EWOULDBLOCK))
			{
				cptr->c_flags |= CLIENT_FLAGS_DEAD;
				if (exitbuf)
					sprintf(exitbuf, "Write error: %i", errno);
				return -1;
			}
		}
		break;
	}
	if (!(cptr->c_lastqsize = cptr->c_queue_out.queue_size))
		fdlist_del_event(cptr->c_connptr->fdlistinfo.fdlist,
			cptr->c_connptr->fdlistinfo.fdlist_pos, FDLIST_FLAGS_WRITE);
	return 0;
}

int ircd_ratelimit(IRCD_CLIENT *pptr, int *numptr, int *timeptr, int numlimit, int timelimit)
{
	if (!numlimit || !timelimit || !IsClient(pptr))
		return 0;
	if (Now - *timeptr > timelimit)
	{
		*numptr = 1;
		*timeptr = Now;
		return 0;
	}
	if (++(*numptr) <= numlimit)
		return 0;
	if (*numptr == (numlimit + 1))
		*timeptr = Now; /* wait 'timelimit' seconds before allowing another*/
	return 1;
}

int debug_locals(void)
{
	IRCD_CLIENT *lptr;
	char buffer[1024];
	int fd;

	fd = open("locals.dbg", O_CREAT|O_WRONLY|O_APPEND, 0660);
	if (fd >= 0)
	{
		for(lptr=Locals;lptr;lptr=lptr->c_lo_next)
		{
			sprintf(buffer, "%x: %s\n", (int)lptr, lptr->name);
			write(fd, buffer, strlen(buffer));
			fsync(fd);
		}
		close(fd);
	}
	return 0;
}

void servlink_rotate(SERVLINK *s)
{
	SERVLINK	*sptr;
	SERVLINK	*lastp = NULL;
	SERVLINK	*lasts = NULL;
	int			found = 0;
	IRCD_CLASSENTRY	*classptr;
	int				ret;

	for(sptr=_ircd_conf.servers;sptr;lastp=sptr,sptr=sptr->next)
	{
		if (s == sptr)
		{
			found = 1;
			lasts = lastp;
		}
	}
	if (!found || (s == lastp))
		return;
	if (!lasts)
		_ircd_conf.servers = s->next;
	else
		lasts->next = s->next;
	s->next = NULL;
	lastp->next = s;
	if (s->classptr)
		classptr = (IRCD_CLASSENTRY *)s->classptr;
	else
	{
		ret = class_find(&(_ircd_conf.classes), s->classname, &classptr);
		if (ret < 0) /* hrm...*/
			return;
		s->classptr = (void *)classptr;
	}
}

void command_tree_free(void)
{
	int i;

	for(i=0;i<27;i++)
	{
		if (_commandtree[i])
		{
			_command_tree_free(_commandtree[i]);
			MyFree(_commandtree[i]);
			_commandtree[i] = NULL;
		}
	}
}

void stringlist_add(IRCD_STRINGLIST **stlist, char *string)
{
	IRCD_STRINGLIST		*sptr;

	if (!stlist)
		return;

	sptr = (IRCD_STRINGLIST *)MyMalloc(sizeof(IRCD_STRINGLIST));
	sptr->string = MyStrdup(string);
	sptr->prev = NULL;
	if ((sptr->next = *stlist) != NULL)
		(*stlist)->prev = sptr;
	*stlist = sptr;
}

void stringlist_del(IRCD_STRINGLIST **stlist, IRCD_STRINGLIST *sptr)
{
	if (sptr->prev)
		sptr->prev->next = sptr->next;
	else
		*stlist = sptr->next;
	if (sptr->next)
		sptr->next->prev = sptr;
	if (sptr->string)
		MyFree(sptr->string);
	MyFree(sptr);
}

IRCD_STRINGLIST *stringlist_find(IRCD_STRINGLIST *stlist, char *string)
{
	IRCD_STRINGLIST		*sptr;

	if (!string)
		return NULL;

	for(sptr=stlist;sptr;sptr=sptr->next)
	{
		if (sptr->string && !mycmp(sptr->string, string))
			return sptr;
	}
	return NULL;
}

IRCD_STRINGLIST *stringlist_match(IRCD_STRINGLIST *stlist, char *string)
{
	IRCD_STRINGLIST	*sptr;

	if (!string)
		return NULL;

	for(sptr=stlist;sptr;sptr=sptr->next)
	{
		if (sptr->string && !match(sptr->string, string))
			return sptr;
	}
	return NULL;
}

void stringlist_free(IRCD_STRINGLIST **stlist)
{
	IRCD_STRINGLIST *n;

	while(*stlist)
	{
		n = (*stlist)->next;
		if ((*stlist)->string)
			MyFree((*stlist)->string);
		MyFree(*stlist);
		*stlist = n;
	}
}

