/*
**   IRC - Internet Relay Chat, source/mysocket.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "mysocket.h"

RCSID("$Id: mysocket.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");

int mysocket_set_nonblocking(int fd)
{
	int		res;

#if 1 && defined(FIONBIO)
	res = 1;
	if (ioctl(fd, FIONBIO, &res) < 0)
		return -errno;
#else
	int		nonb = 0;

	nonb = O_NDELAY;
	if ((res = fcntl(fd, F_GETFL, 0)) == -1)
		return -errno;
	else if (fcntl(fd, F_SETFL, res | nonb) == -1)
		return -errno;
#endif
	return 0;
}

int mysocket_listenport(ircd_uint32_t ipaddr, ircd_uint16_t port, int flags)
{
	struct sockaddr_in	sai;
	int					sock;

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return -errno;
	if (flags & MYSOCKET_FLAGS_REUSEADDR)
	{
		int on = 1;
		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
	}
	BZERO(&sai, sizeof(sai));
	sai.sin_family = AF_INET;
	if (ipaddr != (ircd_uint32_t)-1)
		sai.sin_addr.s_addr = htonl(ipaddr);
	else
		sai.sin_addr.s_addr = INADDR_ANY;
	sai.sin_port = htons(port);
	if (bind(sock, (struct sockaddr *)&sai, sizeof(sai)) < 0)
	{
		int e = errno;
		close(sock);
		return -e;
	}
	if (listen(sock, 1024) < 0)
	{
		int e = errno;
		close(sock);
		return -e;
	}
	if (flags & MYSOCKET_FLAGS_NONBLOCKING)
		mysocket_set_nonblocking(sock);
	return sock;
}

int mysocket_connectip(ircd_uint32_t ip, ircd_uint16_t port, ircd_uint32_t bindip, int flags)
{
	int					s;
	struct sockaddr_in	server;

	if ((ip == -1) || (port == 0))
		return -EINVAL;
	BZERO(&server, sizeof(server));
	server.sin_addr.s_addr = htonl(ip);
	server.sin_family = AF_INET;
	server.sin_port = htons(port);
	if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return -errno;
	if (flags & MYSOCKET_FLAGS_REUSEADDR)
	{
		int on = 1;
		setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on));
	}
	if (flags & MYSOCKET_FLAGS_KEEPALIVE)
	{
		int on = 1;
		setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *)&on, sizeof(on));
	}
	if (bindip != (ircd_uint32_t)-1)
	{
		struct sockaddr_in sock;

		BZERO(&sock, sizeof(sock));
	    sock.sin_port = htonl(0);
	    sock.sin_family = AF_INET;
	    sock.sin_addr.s_addr = htonl(bindip);
	    bind(s, (struct sockaddr *) &sock, sizeof(sock));
	}
	if (flags & MYSOCKET_FLAGS_NONBLOCKING)
	{
		int		ret;
		if ((ret = mysocket_set_nonblocking(s)) < 0)
		{
			close(s);
			return ret;
		}
	}
	if (connect(s, (struct sockaddr *)&server, sizeof(server)) < 0)
	{
		int e = errno;

		if (e != EINPROGRESS)
		{
		    close(s);
			return -e;
		}
	    return s;
	}
	return s;
}

