/*
**   IRC - Internet Relay Chat, source/ircd.c
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "queue.h"

RCSID("$Id: queue.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");

#define QUEUE_FREELIST_HASHSIZE		10

static int		_queue_fl_size = 0;
static IQENTRY *_queue_freelist[QUEUE_FREELIST_HASHSIZE] = { NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL };
static int		_queue_inited = 0;
static int		_queue_size = 0;
static int		_queue_malloced_size = 0;

static IQENTRY *_queue_freelist_get(IQUEUE *queue);
static void *_queue_mmap(int size);
static IQENTRY *_queue_iqentry_MyMalloc(IQUEUE *queue);
static int _queue_iqentry_free(IQENTRY *qe);
static int _queue_iqentry_realfree(IQENTRY *qe);
static int _queue_data_get(IQUEUE *queue, char **data,
	int *datalen, IQENTRY **qenext);

static int _queue_freelist_del(void)
{
	IQENTRY	*nextqe;
	int		i;

	for(i=0;i<QUEUE_FREELIST_HASHSIZE;i++)
	{
		for(;_queue_freelist[i];_queue_freelist[i]=nextqe)
		{
			nextqe = _queue_freelist[i]->next;
			_queue_iqentry_realfree(_queue_freelist[i]);
		}
	}
	assert(!_queue_fl_size);
	return 0;
}

static IQENTRY *_queue_freelist_get(IQUEUE *queue)
{
	IQENTRY	*qe;
	IQENTRY	**qeptr;
	int		num = 5;
	int		hash = queue->entrysize % QUEUE_FREELIST_HASHSIZE;

	qeptr = &_queue_freelist[hash];
	for(qe=*qeptr;qe && num;--num,qeptr=&(qe->next),qe=qe->next)
		if (queue->entrysize == qe->bufsize)
		{
			*qeptr = qe->next;
			_queue_fl_size -= qe->bufsize;
			break;
		}
	if (!num)
		return NULL;
	return qe;
}

static void *_queue_mmap(int size)
{
	void	*ptr;

#ifdef MAP_ANON
	ptr = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_ANON, -1, 0);
#else
	ptr = mmap(NULL, size, PROT_READ|PROT_WRITE, MAP_PRIVATE, Ircd_zero_fd, 0);
#endif

#ifdef MAP_FAILED
	if (ptr == MAP_FAILED)
#else
	if (ptr == (void *)-1)
#endif
		return (void *)-1;
	return ptr;
}

static IQENTRY *_queue_iqentry_MyMalloc(IQUEUE *queue)
{
	IQENTRY	*qe;

	if (!_queue_inited)
	{
		int i;
		_queue_inited = 1;
		for(i=0;i<QUEUE_FREELIST_HASHSIZE;i++)
			_queue_freelist[i] = NULL;
	}
	qe = _queue_freelist_get(queue);
	if (!qe)
	{
		if (!(queue->flags & IQUEUE_FLAGS_USEMMAP))
		{
			qe = (IQENTRY *)MyMalloc(sizeof(IQENTRY)+queue->entrysize);
			qe->flags = 0;
		}
		else
		{
			if ((qe = _queue_mmap(
					sizeof(IQENTRY)+queue->entrysize)) == (void *)-1)
			{
				qe = (IQENTRY *)MyMalloc(sizeof(IQENTRY)+queue->entrysize);
				qe->flags = 0;
			}
			else
				qe->flags = IQUEUE_FLAGS_USEMMAP;
		}
		qe->buffer = (char *)(qe+1);
		qe->bufsize = queue->entrysize;
		_queue_malloced_size += queue->entrysize;
	}
	qe->curpos = 0;
	qe->buflen = 0;
	qe->next = NULL;
	return qe;
}

static int _queue_iqentry_realfree(IQENTRY *qe)
{
	if (!qe)
		return -EINVAL;
	_queue_fl_size -= qe->bufsize;
	_queue_malloced_size -= qe->bufsize;
	if (qe->flags & IQUEUE_FLAGS_USEMMAP)
		munmap((void *)qe, sizeof(IQENTRY)+qe->bufsize);
	else
		MyFree(qe);
	return 0;
}

static int _queue_iqentry_free(IQENTRY *qe)
{
	int	hash = qe->bufsize % QUEUE_FREELIST_HASHSIZE;
	qe->next = _queue_freelist[hash];
	_queue_freelist[hash] = qe;
	_queue_fl_size += qe->bufsize;
	return 0;
}

int queue_init(IQUEUE *queue, int entrysize, int flags)
{
	int	diff;

	if (!queue || (entrysize <= 0))
		return -EINVAL;
	queue->entrysize = entrysize;
	diff = (sizeof(IQENTRY)+queue->entrysize) % ircd_pagesize;
	if (diff)
		queue->entrysize += (ircd_pagesize-diff);
	queue->flags = flags;
	queue->queue_size = 0;
	queue->entries = queue->last_entry = NULL;
	return 0;
}

int queue_data_add(IQUEUE *queue, char *data, int datalen)
{
	IQENTRY		*qe;
	int			curpos = 0;

	if (!queue || !data)
		return -EINVAL;
	if (!queue->entries)
		qe = queue->entries = queue->last_entry = _queue_iqentry_MyMalloc(queue);
	else
		qe = queue->last_entry;
	while(datalen)
	{

		if (datalen <= (qe->bufsize-qe->buflen))
		{
			BCOPY(data+curpos, qe->buffer+qe->buflen, datalen);
			qe->buflen += datalen;
			queue->queue_size += datalen;
			_queue_size += datalen;
			return 0;
		}

		queue->queue_size += (qe->bufsize - qe->buflen);
		_queue_size += (qe->bufsize - qe->buflen);
		BCOPY(data+curpos, qe->buffer+qe->buflen, qe->bufsize - qe->buflen);
		datalen -= (qe->bufsize - qe->buflen);
		curpos += (qe->bufsize - qe->buflen);
		qe->buflen = qe->bufsize;
		qe = _queue_iqentry_MyMalloc(queue);
		queue->last_entry->next = qe;
		queue->last_entry = qe;
	}
	return 0; /* likely not reached */
}

static int _queue_data_get(IQUEUE *queue, char **data,
		int *datalen, IQENTRY **qenext)
{
	IQENTRY	*qe;

	if (!queue)
		return -EINVAL;
	if (!qenext || !*qenext)
	{
		if (!(qe=queue->entries))
			return -ENOENT;
	}
	else
	{
		if (!(qe = (*qenext)->next))
			return -ENOENT;
	}
	if (qenext)
		*qenext = qe;
	if (!(qe->buflen-qe->curpos))
		return -ENOENT;
	if (data)
		*data = (qe->buffer+qe->curpos);
	if (datalen)
		*datalen = (qe->buflen-qe->curpos);
	return 0;
}

int queue_data_get(IQUEUE *queue, char **data, int *datalen)
{
	return _queue_data_get(queue, data, datalen, NULL);
}

int queue_data_get_nl(IQUEUE *queue, char *buffer, int buflen, int flags)
{
	char	*buf;
	char	*ptr;
	char	*bptr = buffer;
	int		bleft = buflen-1;
	int		len;
	int		i;
	int		skip = 0;
	int		numskipped = 0;
	IQENTRY	*qeptr = NULL;

	*bptr = '\0';
	for(;;)
	{
		if (_queue_data_get(queue, &buf, &len, &qeptr))
		{
			*bptr = '\0';
			return 0;
		}
		for(ptr=buf,i=len;i;--i,ptr++)
		{
			if ((*ptr == '\r') || (*ptr == '\n'))
			{
				if (bptr == buffer)
				{
					queue_data_del(queue, numskipped+1);
					qeptr = NULL;
					skip = 0;
					numskipped = 0;
					break;
				}
				queue_data_del(queue, bptr - buffer + 1);
				*bptr = '\0';
				return bptr - buffer;
			}
			if (skip)
			{
				numskipped++;
				continue;
			}
			*bptr++ = *ptr;
			if (!--bleft)
			{
				/*
				** ouch...well...this means that we have a long line
				**  hrm...let's skip to next nl I guess..
				*/
				skip = 1;
				numskipped = bptr - buffer;
				bptr = buffer;
				bleft = buflen-1;
			}
		}
	}
	return 0; /* not reached */
}

int queue_delete(IQUEUE *queue)
{
	IQENTRY	*qe;

	if (!queue)
		return -EINVAL;

	for(;queue->entries;queue->entries=qe)
	{
		qe = queue->entries->next;
		_queue_size -= (queue->entries->buflen - queue->entries->curpos);
		_queue_iqentry_free(queue->entries);
	}
	queue->queue_size = 0;
	queue->last_entry = NULL;
	return 0;
}

int queue_data_del(IQUEUE *queue, int datalen)
{
	IQENTRY	*qe;

	if (!queue || (datalen < 0))
		return -EINVAL;
	if (!(qe=queue->entries))
		return -EFAULT;

	while(datalen)
	{
		int diff = (qe->buflen - qe->curpos);

		if (datalen >= diff)
		{
			datalen -= diff;
			queue->queue_size -= diff;
			_queue_size -= diff;
#if 0
			if (qe->next)
			{
				queue->entries = qe->next;
				_queue_iqentry_free(qe);
				qe = queue->entries;
			}
			else
			{
				qe->curpos = qe->buflen = 0;
				if (datalen)
					return -ENOSPC; /* asked to delete more than we have */
				else
					return 0;
			}
#else
			if ((queue->entries = qe->next) == NULL)
				queue->last_entry = NULL;
			_queue_iqentry_free(qe);
			qe = queue->entries;
			if (!qe)
			{
				if (datalen)
					return -ENOSPC; /* asked to delete more than we have */
				else
					return 0;
			}
#endif
		}
		else
		{
			qe->curpos += datalen;
			queue->queue_size -= datalen;
			_queue_size -= datalen;
			return 0;
		}
	}
	return 0; /* likely never reach */
}

int queue_cleanup(void)
{
	(void)_queue_freelist_del();
	return 0;
}

int queue_resize(IQUEUE *queue, int newsize)
{
	IQUEUE	newqueue;
	char	*dptr;
	int		dlen;

	if (queue->entrysize == newsize)
		return 0;

	queue_init(&newqueue, newsize, queue->flags);

	while(!queue_data_get(queue, &dptr, &dlen))
	{
		if (queue_data_add(&newqueue, dptr, dlen) < 0)
		{
			queue_delete(&newqueue);
			return -1;
		}
		queue_data_del(queue, dlen);
	}
	queue_delete(queue);
	BCOPY(&newqueue, queue, sizeof(IQUEUE));
	return 0;
}

int queue_get_stats(int *queue_sizes, int *queue_fl_sizes, int *queue_malloced_sizes)
{
	if (queue_sizes)
		*queue_sizes = _queue_size;
	if (queue_fl_sizes)
		*queue_fl_sizes = _queue_fl_size;
	if (queue_malloced_sizes)
		*queue_malloced_sizes = _queue_malloced_size;
	return 0;
}

int queue_set_stats(int queue_sizes, int queue_malloced_size)
{
	_queue_size = queue_sizes;
	_queue_malloced_size = queue_malloced_size;
	return 0;
}

