/*
**   IRC - Internet Relay Chat, source/res.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1992, 1993, 1994 Darren Reed. 
**
**   Also contains code from Bleep (Thomas Helvey <tomh@inxpress.net>)
**      from July 1999
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include "ircd_includes.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <errno.h>

#include "res.h"


#define MAXPACKET				1024

#define IRCDRES_REQ_HASHSIZE		1024
#define IRCDRES_MAXADDRS			35
#define IRCDRES_MAXALIASES		35

#define ADDRSBLEN ((IRCDRES_MAXADDRS * sizeof(char *)) + sizeof(char *))
#define ADDRSDLEN (IRCDRES_MAXADDRS * sizeof(struct in_addr))
#define ALIASBLEN ((IRCDRES_MAXALIASES * sizeof(char *)) + sizeof(char *))
#define MAXGETHOSTLEN (ADDRSBLEN + ADDRSDLEN + ALIASBLEN + MAXPACKET)

#define IRCDRES_GETSHORT(__x)  ((unsigned short)((unsigned char *)(__x))[0]<<8)| \
                            ((unsigned short)((unsigned char *)(__x))[1])

#define IRCDRES_GETLONG(__x)   ((unsigned int)((unsigned char *)(__x))[0]<<24)| \
                            ((unsigned int)((unsigned char *)(__x))[1]<<16)| \
                            ((unsigned int)((unsigned char *)(__x))[2]<<8)| \
                            ((unsigned int)((unsigned char *)(__x))[3])


#define AR_TTL 600

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif
#ifndef INT16SZ
#define INT16SZ 2
#endif
#ifndef INT32SZ
#define INT32SZ 4
#endif

static IRCDRES_REQUEST	*Requests = NULL;
static int				NumRequests = 0;
static IRCDRES_REQUEST	*HRequests[IRCDRES_REQ_HASHSIZE];
static int _req_init = 0;

static void add_request_to_reqlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request);
static void del_request_from_reqlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request);
static void add_request_to_reqhlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request);
static void del_request_from_reqhlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request);
static IRCDRES_REQUEST *_ircdres_request_find_id(u_short id);
static int _ircdres_send(IRCDRES *resptr, char *buf, int buflen, u_int max_ns);
static int _ircdres_send_query(IRCDRES *resptr, IRCDRES_REQUEST *reqptr);
static void _ircdres_request_del(IRCDRES_REQUEST *request, int call_callback);
static IRCDRES_HOSTENT *_ircdres_hostent_init(void);
static int _ircdres_proc_answer(IRCDRES *resptr, IRCDRES_REQUEST *reqptr, char *buffer, char *ending);


static void add_request_to_reqlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request)
{
	request->req_prev = NULL;
	if ((request->req_next = *bucket) != NULL)
		request->req_next->req_prev = request;
	*bucket = request;
}

static void del_request_from_reqlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request)
{		   
	if (request->req_prev)
		request->req_prev->req_next = request->req_next;
	else
		*bucket = request->req_next;
	if (request->req_next)
		request->req_next->req_prev = request->req_prev;
}

static void add_request_to_reqhlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request)
{
	request->req_h_prev = NULL;
	if ((request->req_h_next = *bucket) != NULL)
		request->req_h_next->req_h_prev = request;
	*bucket = request;
}

static void del_request_from_reqhlist(IRCDRES_REQUEST **bucket, IRCDRES_REQUEST *request)
{		   
	if (request->req_h_prev)
		request->req_h_prev->req_h_next = request->req_h_next;
	else
		*bucket = request->req_h_next;
	if (request->req_h_next)
		request->req_h_next->req_h_prev = request->req_h_prev;
}

static IRCDRES_REQUEST *_ircdres_request_find_id(u_short id)
{
	register IRCDRES_REQUEST *rptr = HRequests[id % IRCDRES_REQ_HASHSIZE];

	for(;rptr;rptr=rptr->req_h_next)
	{
		if (rptr->req_id == id)
			break;
	}
	return rptr;
}

static int _ircdres_send(IRCDRES *resptr, char *buf, int buflen, u_int max_ns)
{
	int		num = 0;
	int		sends = 0;

	if (resptr->res_ns_num < max_ns)
		max_ns = resptr->res_ns_num;

	for(;num<max_ns;num++)
	{
		if (sendto(resptr->res_fd, buf, buflen, 0,
				(struct sockaddr *)&(resptr->res_ns_ipaddrs[num]),
				sizeof(struct sockaddr_in)) == buflen)
			sends++;
	}
	return sends ? sends : -1;
}

static int _ircdres_send_query(IRCDRES *resptr, IRCDRES_REQUEST *reqptr)
{
	char			buffer[MAXPACKET];
	HEADER			*hptr;
	int				ret;

	ret = res_mkquery(QUERY, reqptr->req_name, C_IN, reqptr->req_qtype,
				NULL, 0, NULL, (u_char *)buffer, sizeof(buffer));
	if (ret < 0)
		return -ENOSPC;
	hptr = (HEADER *)buffer;
	hptr->id = ntohs(hptr->id) + (u_int)hptr;
	while(_ircdres_request_find_id(hptr->id) != NULL)
		hptr->id++;
	reqptr->req_id = hptr->id;
	hptr->id = htons(hptr->id);
	reqptr->req_sends++;
	ret = _ircdres_send(resptr, buffer, ret, reqptr->req_sends);
	if (ret < 0)
		return -EFAULT;
	return 0;
}

int ircdres_deinit(IRCDRES *resptr)
{
	if (!resptr)
		return -EINVAL;
	if (resptr->res_fd >= 0)
		close(resptr->res_fd);
	if (resptr->res_ns_ipaddrs)
		MyFree(resptr->res_ns_ipaddrs);
	MyFree(resptr);
	return 0;
}

int ircdres_init(IRCDRES **rptr)
{
	IRCDRES	*resptr;

	if (!rptr)
		return -EINVAL;
	if (!_req_init)
	{
		BZERO(HRequests, sizeof(HRequests));
		_req_init = 1;
		res_init();
	}
	resptr = (IRCDRES *)MyMalloc(sizeof(IRCDRES));
	if (!resptr)
		return -ENOMEM;
	resptr->res_fd = socket(AF_INET, SOCK_DGRAM, 0);
	if (resptr->res_fd < 0)
	{
		int e = -errno;
		MyFree(resptr);
		return e;
	}
	mysocket_set_nonblocking(resptr->res_fd);
	*rptr = resptr;
	return 0;
}

int ircdres_set_ns(IRCDRES *resptr, struct sockaddr_in *nameservers, int num)
{
	int i;

	if (resptr->res_ns_ipaddrs)
	{
		MyFree(resptr->res_ns_ipaddrs);
		resptr->res_ns_ipaddrs = NULL;
		resptr->res_ns_num = 0;
	}
	resptr->res_ns_ipaddrs = (struct sockaddr_in *)MyMalloc(sizeof(struct sockaddr_in) * num);
	if (!resptr->res_ns_ipaddrs)
	{
		return -ENOMEM;
	}
	for(i=0;i<num;i++)
	{
		resptr->res_ns_ipaddrs[i].sin_family = AF_INET;
		resptr->res_ns_ipaddrs[i].sin_port = nameservers[i].sin_port;
		resptr->res_ns_ipaddrs[i].sin_addr.s_addr =
						nameservers[i].sin_addr.s_addr;
	}
	resptr->res_ns_num = num;
	return 0;
}
static void _ircdres_request_del(IRCDRES_REQUEST *request, int call_callback)
{
	request->req_flags |= REQ_FLAGS_CLOSED;

	del_request_from_reqhlist(&HRequests[request->req_id%IRCDRES_REQ_HASHSIZE],
			request);
	del_request_from_reqlist(&Requests, request);
	NumRequests--;
	if (call_callback)	/* failed req */
		request->req_callback(request->req_res, NULL, request->req_callback_arg);
}

int ircdres_request_del(IRCDRES_REQUEST *request, int call_callback)
{
	assert(!(request->req_flags & REQ_FLAGS_CLOSED));
	_ircdres_request_del(request, call_callback);
	MyFree(request->req_name);
	MyFree(request);
	return 0;
}

int ircdres_gethostbyname(IRCDRES *resptr, char *name, ircdres_func_t callback, void *callback_arg, IRCDRES_HOSTENT **result, IRCDRES_REQUEST **request)
{
	IRCDRES_REQUEST	*reqptr;
	int				ret;

	if (!resptr || !name || !callback)
		return -EINVAL;

	/* check cache */

	if (!request)
		return -EINVAL;
	reqptr = (IRCDRES_REQUEST *)MyMalloc(sizeof(IRCDRES_REQUEST));
	if (!reqptr)
		return -ENOMEM;
	reqptr->req_name = MyStrdup(name);
	if (!reqptr->req_name)
	{
		MyFree(reqptr);
		return -ENOMEM;
	}

	reqptr->req_callback = callback;
	reqptr->req_callback_arg = callback_arg;
	reqptr->req_qtype = T_A;
	reqptr->req_res = resptr;

	ret = _ircdres_send_query(resptr, reqptr);
	if (ret < 0)
	{
		MyFree(reqptr->req_name);
		MyFree(reqptr);
		return ret;
	}

	add_request_to_reqhlist(&HRequests[reqptr->req_id%IRCDRES_REQ_HASHSIZE],
			reqptr);
	add_request_to_reqlist(&Requests, reqptr);
	NumRequests++;

	*request = reqptr;
	return -EINPROGRESS;
}

int ircdres_gethostbyaddr(IRCDRES *resptr, struct in_addr *addr, ircdres_func_t callback, void *callback_arg, IRCDRES_HOSTENT **result, IRCDRES_REQUEST **request)
{
	IRCDRES_REQUEST	*reqptr;
	int				ret;
	u_char			*cp;

	if (!resptr || !addr || !callback)
		return -EINVAL;

	cp = (u_char *)&addr->s_addr;

	/* check cache */

	if (!request)
		return -EINVAL;
	reqptr = (IRCDRES_REQUEST *)MyMalloc(sizeof(IRCDRES_REQUEST));
	if (!reqptr)
		return -ENOMEM;
	reqptr->req_name = (char *)MyMalloc(32);
	if (!reqptr->req_name)
	{
		MyFree(reqptr);
		return -ENOMEM;
	}

	reqptr->req_flags = 0;
	reqptr->req_callback = callback;
	reqptr->req_callback_arg = callback_arg;
	reqptr->req_qtype = T_PTR;
	reqptr->req_inaddr = *addr;
	reqptr->req_res = resptr;

	sprintf(reqptr->req_name, "%u.%u.%u.%u.in-addr.arpa.",
			(u_int)cp[3], (u_int)cp[2], (u_int)cp[1],
			(u_int)cp[0]);

	ret = _ircdres_send_query(resptr, reqptr);
	if (ret < 0)
	{
		MyFree(reqptr->req_name);
		MyFree(reqptr);
		return ret;
	}

	add_request_to_reqhlist(&HRequests[reqptr->req_id%IRCDRES_REQ_HASHSIZE],
			reqptr);
	add_request_to_reqlist(&Requests, reqptr);
	NumRequests++;

	*request = reqptr;
	return -EINPROGRESS;
}

static IRCDRES_HOSTENT *_ircdres_hostent_init(void)
{
	IRCDRES_HOSTENT	*hptr;

	hptr = (IRCDRES_HOSTENT *)MyMalloc(sizeof(IRCDRES_HOSTENT) + MAXGETHOSTLEN);
	hptr->he_hbuf = (char *)(hptr + 1);
	hptr->he_he.h_aliases = (char **)hptr->he_hbuf;
	hptr->he_he.h_addr_list = (char **)(((char *)hptr->he_he.h_aliases) + ALIASBLEN);
	return hptr;
}

static int _ircdres_proc_answer(IRCDRES *resptr, IRCDRES_REQUEST *reqptr, char *buffer, char *ending)
{
	IRCDRES_RESULT	*reptr;
	int				ret;
	HEADER			*hptr = (HEADER *)buffer;
	char			*cp = buffer + sizeof(HEADER);
	char			hostbuf[512];
	int				gotdot;
	register char	*sptr;
	int				len;
	int				addr_count = 0;
	char			**addr;
	int				alias_count = 0;
	char			**alias;
	u_short			addr_class;
	u_short			dlen;
	int				ans = 0;
	char			*ad_fr;
	char			*al_fr;
	char			*endp;
	struct hostent	*hp;

	reptr = (IRCDRES_RESULT *)MyMalloc(sizeof(IRCDRES_RESULT));
	if (!reptr)
		return -ENOMEM;
	reptr->res_he = _ircdres_hostent_init();
	if (!reptr->res_he)
	{
		MyFree(reptr);
		return -ENOMEM;
	}
	hp = &(reptr->res_he->he_he);
	addr = hp->h_addr_list;
	alias = hp->h_aliases;

	ad_fr = ((char *)hp->h_addr_list) + ADDRSBLEN;
	al_fr = ((char *)hp->h_addr_list) + ADDRSBLEN + ADDRSDLEN;
	endp = reptr->res_he->he_hbuf + MAXGETHOSTLEN;

#if 0
	if (reqptr->req_type == T_PTR)
	{
		BCOPY(&reqptr->req_inaddr, ad_fr, sizeof(struct in_addr));
		*hp->h_addr_list = ad_fr;
		addr_count++;
		ad_fr += sizeof(struct in_addr);
	}
#endif
	
	for(;hptr->qdcount > 0;hptr->qdcount--)
	{
		if ((ret = dn_skipname((u_char *)cp, (u_char *)ending)) == -1)
			break;
		else
			cp += (ret + QFIXEDSZ);
	}

	if (!hptr->ancount)
	{
		MyFree(reptr->res_he);
		MyFree(reptr);
		return -EFAULT; /* force retry */
	}

	hostbuf[sizeof(hostbuf)-1] = '\0';

	while(cp && (hptr->ancount-- > 0) && (cp < ending))
	{
		ret = dn_expand((u_char *)buffer, (u_char *)ending,
					(u_char *)cp, hostbuf, sizeof(hostbuf)-1);
		if (ret < 0)
			break;
		cp += ret;
		if (cp + INT16SZ + INT16SZ + INT32SZ + INT16SZ >= ending)
			break;
		reptr->res_type = IRCDRES_GETSHORT(cp);
		cp += INT16SZ;
		addr_class = IRCDRES_GETSHORT(cp);
		cp += INT16SZ;
		reptr->res_ttl = IRCDRES_GETLONG(cp);
		cp += INT32SZ;
		dlen = IRCDRES_GETSHORT(cp);
		cp += INT16SZ;

		if ((cp + dlen) > ending)
			break;

		gotdot = 0;
		for(sptr=hostbuf;*sptr;sptr++)
			if (*sptr == '.')
				gotdot++;
		if (!gotdot)
			break;
		len = sptr - hostbuf;

		switch(reptr->res_type)
		{
			case T_A:
				if ((dlen != sizeof(struct in_addr)) ||
						(++addr_count >= IRCDRES_MAXADDRS))
				{
					cp += dlen;
					break;
				}
				hp->h_length = dlen;
				if (ans == 0)
					hp->h_addrtype = (addr_class == C_IN) ? 
											AF_INET : AF_UNSPEC;
				if (!hp->h_name)
				{
					if (len+1 >= (endp - al_fr))
					{
						cp = NULL;
						break;
					}
					strcpy(al_fr, hostbuf);
					hp->h_name = al_fr;
					al_fr += len + 1;
				}
				BCOPY(cp, ad_fr, sizeof(struct in_addr));
				*addr++ = ad_fr;
				*addr = 0;
				cp += dlen;
				ad_fr += sizeof(struct in_addr);
				ans++;
				break;
			case T_PTR:
				if ((ret = dn_expand((u_char *)buffer, (u_char *)ending,
							(u_char *)cp, hostbuf, sizeof(hostbuf)-1)) < 0)
				{
					cp = NULL; /* break out of loop */
					break;
				}
				cp += ret;
				len = strlen(hostbuf);
				if (len+1 >= (endp - al_fr))
				{
					cp = NULL;
					break;
				}
				if (hp->h_name)
				{
					if (++alias_count >= IRCDRES_MAXALIASES)
						break;
					strcpy(al_fr, hostbuf);
					*alias++ = al_fr; 
				}
				else
				{
					strcpy(al_fr, hostbuf);
					hp->h_name = al_fr;
				}
				al_fr += len + 1;
				ans++;
				break;
			case T_CNAME:
				cp += dlen;
				if (++alias_count >= IRCDRES_MAXALIASES)
					break;
				/* len is already strlen(hostbuf) */
				if (len+1 >= (endp - al_fr))
				{
					cp = NULL;
					break;
				}
				strcpy(al_fr, hostbuf);
				*alias++ = al_fr;
				*alias = NULL;
				al_fr += len + 1;
				ans++;
				break;
		}
	}

	if (!ans ||
		((reqptr->req_qtype == T_A) && !reptr->res_he->he_he.h_addr_list[0]) ||
		((reqptr->req_qtype == T_PTR) && !reptr->res_he->he_he.h_name))
	{
		MyFree(reptr->res_he);
		MyFree(reptr);
		reqptr->req_callback(resptr, NULL, reqptr->req_callback_arg);
		return 0;
	}

	reqptr->req_callback(resptr, reptr, reqptr->req_callback_arg);
	/* caller must free reptr->res_he */
	MyFree(reptr);

	return 1;
}

int ircdres_get_answer(IRCDRES *resptr)
{
	struct sockaddr_in	sin;
	int					len;
	int					r_errno = 0;
	IRCDRES_REQUEST		*req;
	int					rc;
	int					ret;
	char				buffer[sizeof(HEADER) + MAXPACKET];
	HEADER				*hptr;
	int					i;
	int					max;

	for(;;)		/* fake loop to avoid goto */
	{
		len = sizeof(sin);
		rc = recvfrom(resptr->res_fd, buffer, sizeof(buffer),
					0, (struct sockaddr *) &sin, &len);
		if (rc <= sizeof(HEADER))
			return -EFAULT;
		hptr = (HEADER *)buffer;
		hptr->id = ntohs(hptr->id);
		hptr->ancount = ntohs(hptr->ancount);
		hptr->qdcount = ntohs(hptr->qdcount);
		hptr->nscount = ntohs(hptr->nscount);
		hptr->arcount = ntohs(hptr->arcount);

        if ((req = _ircdres_request_find_id(hptr->id)) == NULL)
            return -EFAULT;

		/* req->req_res should equal resptr! */

		max = (resptr->res_ns_num < req->req_sends) ?
				resptr->res_ns_num : req->req_sends;

        for (i=0;i<max;i++)
        {
			if (!BCMP((char *)&sin.sin_addr,
                    (char *)&resptr->res_ns_ipaddrs[i].sin_addr,
					sizeof(struct in_addr)))
				break;
        }
        if (i == max)
        {
			return -EFAULT;
		}

		if ((hptr->rcode != NOERROR) || (hptr->ancount == 0))
		{
			switch(hptr->rcode)
			{
				case NOERROR:
					r_errno = NO_DATA;
					break;
				case SERVFAIL:
				case NXDOMAIN:
					r_errno = TRY_AGAIN;
					break;
				case FORMERR:
				case NOTIMP:
				case REFUSED:
				default:
					r_errno = NO_RECOVERY;
					break;
			}
			break;
		}
		ret = _ircdres_proc_answer(req->req_res, req, buffer, buffer + rc);
		if (ret < 0)
		{
			return ret;
		}
		ircdres_request_del(req, 0);
		return 0;
	}
	if ((r_errno == TRY_AGAIN) || (r_errno == NO_DATA))
	{
		ircdres_request_del(req, 1);
		return -EFAULT;
	}
	/* ignore, and let caller's timeout re-send */
	return -EFAULT;
}

int ircdres_deinit_reload(IRCDRES_REQUEST **reqptr)
{
	IRCDRES_REQUEST	*next;

	if (!reqptr)
		return -EINVAL;
/* hack for now */
	while(Requests)
	{
		next = Requests->req_next;
		ircdres_request_del(Requests, 1);
		Requests = next;
	}
	*reqptr = NULL;
	return 0;
};

int ircdres_init_reload(IRCDRES_REQUEST *reqptr)
{
	IRCDRES_REQUEST	*rptr;

	if (!_req_init)	/* should be true */
	{
		BZERO(HRequests, sizeof(HRequests));
		_req_init = 1;
	}

	for(rptr=reqptr;rptr;rptr=rptr->req_next)
	{
		add_request_to_reqhlist(&HRequests[rptr->req_id%IRCDRES_REQ_HASHSIZE],
			rptr);
		NumRequests++;
	}
	return 0;
}

int ircdres_requests_outstanding(void)
{
	return NumRequests;
}

int ircdres_request_resend(IRCDRES_REQUEST *reqptr)
{
	int ret;

	assert(!(reqptr->req_flags & REQ_FLAGS_CLOSED));
	_ircdres_request_del(reqptr, 0);
	ret = _ircdres_send_query(reqptr->req_res, reqptr);
	if (ret < 0)
	{
		reqptr->req_callback(reqptr->req_res, NULL, reqptr->req_callback_arg);
		MyFree(reqptr->req_name);
		MyFree(reqptr);
		return ret;
	}
	reqptr->req_flags &= ~REQ_FLAGS_CLOSED;
	add_request_to_reqhlist(&HRequests[reqptr->req_id%IRCDRES_REQ_HASHSIZE],
		reqptr);
	add_request_to_reqlist(&Requests, reqptr);
	NumRequests++;
	return ret;
}

