/*
**   IRC - Internet Relay Chat, source/send.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "send.h"

RCSID("$Id: send.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");

static int _send_message_to_client(IRCD_CLIENT *cptr, char *message, int msglen);
static int _internal_send_to_client(IRCD_CLIENT *cptr, char *prefix, char *format, va_list ap);

static int _send_message_to_client(IRCD_CLIENT *cptr, char *message, int msglen)
{
	char	*addptr = message;
	int		ret = 0;
	int		len = msglen;

	if (!msglen)
		return 0;

	if ((cptr == meptr) || !IsLocal(cptr))
		return msglen;

	if (IsDead(cptr) || !cptr->c_connptr)
		return -1;

	if ((cptr->c_maxsendq != -1) &&
			(cptr->c_queue_out.queue_size + msglen > cptr->c_maxsendq))
		return -1;

	cptr->c_send_msg++;
	cptr->c_send_bytes += msglen;
	if (cptr->c_send_bytes >= 1024)
	{
		cptr->c_send_kbytes += cptr->c_send_bytes / 1024;
		cptr->c_send_bytes %= 1024;
	}

	if (
		IsServer(cptr) &&
		!cptr->c_queue_out.queue_size &&
		!(cptr->c_flags & CLIENT_FLAGS_DOINGCONNECT))
	{
		ret = write(cptr->c_connptr->fdlistinfo.fd, addptr, len);
		if (ret > 0)
		{
			if (IsClient(cptr))
				Globalinfo->bw_c_outgoing_bytes += ret;
			else if (IsServer(cptr))
				Globalinfo->bw_s_outgoing_bytes += ret;
			addptr += ret;
			len -= ret;
		}
	}
	if (!len)
		return msglen;
	ret = queue_data_add(&(cptr->c_queue_out), addptr, len);
	if (ret < 0)
	{
		cptr->c_flags |= CLIENT_FLAGS_DEAD;
		return -1;
	}
#if 1
	if (cptr->c_connptr->fdlistinfo.fdlist != Fdlist_main)
	{
		int mfd;

		ret = fdlist_del_fd(cptr->c_connptr->fdlistinfo.fdlist,
						cptr->c_connptr->fdlistinfo.fdlist_pos, &mfd);
		if (mfd >= 0)
		{
			if (GET_FDS(mfd))
				GET_FDS(mfd)->fdlistinfo.fdlist_pos =
						cptr->c_connptr->fdlistinfo.fdlist_pos;
		}
		ret = ircd_open_fd(cptr->c_connptr->fdlistinfo.fd,
					FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE, cptr->c_connptr, Fdlist_main);
	}
	else
#endif
		ret = fdlist_add_event(cptr->c_connptr->fdlistinfo.fdlist,
				cptr->c_connptr->fdlistinfo.fdlist_pos, FDLIST_FLAGS_WRITE);
	if ((cptr->c_queue_out.queue_size >= cptr->c_lastqsize + 4096) &&
			!(cptr->c_flags & CLIENT_FLAGS_DOINGCONNECT))
	{
		char	*buf;
		int		len;

		if (cptr->c_writetries++ == 5)
			cptr->c_writetries = 1;
		if (cptr->c_writetries == 1)
		{
			ret = queue_data_get(&(cptr->c_queue_out), &buf, &len);
			if (ret >= 0)
			{
				cptr->c_lastqsize = cptr->c_queue_out.queue_size;
				ret = write(cptr->c_connptr->fdlistinfo.fd, buf, len);
				if (ret > 0)
				{
					if (IsClient(cptr))
						Globalinfo->bw_c_outgoing_bytes += ret;
					else if (IsServer(cptr))
						Globalinfo->bw_s_outgoing_bytes += ret;
					cptr->c_writetries = 0;
					queue_data_del(&(cptr->c_queue_out), ret);
				}
				else if (ret < 0)
				{
					if ((errno != EAGAIN) && (errno != EINTR) &&
										(errno != EWOULDBLOCK))
					{
						cptr->c_flags |= CLIENT_FLAGS_DEAD;
						return -1;
					}
					if (++cptr->c_writetries == 5)
						cptr->c_writetries = 0;
				}
			}
		}
	}
	return ret >= 0 ? len : ret;
}

static int _internal_send_to_client(IRCD_CLIENT *cptr, char *prefix, char *format, va_list ap)
{
	char	buf[2048];
	char	*ptr = buf;
	int		len = 0;

	if (prefix && *prefix)
	{
		*buf = ':';
		len++;
		for(ptr=buf+1;len < sizeof(buf)-2;)
		{
			*ptr++ = prefix[len-1];
			len++;
			if (!prefix[len-1])
				break;
		}
		*ptr++ = ' ';
		len++;
	}
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(ptr, sizeof(buf)-1-len, format, ap);
#else
	len += vsprintf(ptr, format, ap);
#endif
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	if (!IsLocal(cptr))
		cptr = cptr->c_from;

	return _send_message_to_client(cptr, buf, len+2);
}

int send_to_client(IRCD_CLIENT *pptr, char *format, ...)
{
	va_list		ap;
	int			ret;

	if (!format)
		return -1;
	va_start(ap, format);
	ret = _internal_send_to_client(pptr, NULL, format, ap);
	va_end(ap);
	return ret;
}

int send_prefix_to_client(IRCD_CLIENT *from, IRCD_CLIENT *to, char *format, ...)
{
	va_list		ap;
	char		prefix[NICKLEN+USERLEN+HOSTLEN+5];
	char		*sender = from->name;
	int			ret;

	if (!to || !format)
		return -1;

	if (from && (from->c_from != meptr) && (from->c_from != from) && (to->c_from == from->c_from))
	{
/*
**      This if() checks to make sure we don't send the same
**      message back to the place we received it from.  This
**      occurs when a client becomes ghosted and 2 servers
**      have different information about the client.
**             -- Taner & Comstud
*/
		if (IsServer(from))
		{
			send_to_opers(
			"Send message to %s[%s] from %s would have caused Fake Direction",
				to->name, to->c_from->name, from->name);
		}
		else
		send_to_opers("Send message failed to ghosted %s[%s] from %s[%s]",
			to->name, to->c_from->name, from->name, from->c_from->name);
		if (IsClient(from))
			send_prefix_to_client(meptr, from,
				IRCD_REPLY(ERR_GHOSTEDCLIENT, from->name), to->name);
        return -1;
	}
	if (from && MyClient(to) && IsPerson(from))
	{
		char *curptr = sender = prefix;
		char *ptr;

		for(ptr=from->name;*ptr;)
			*curptr++ = *ptr++;
		*curptr++ = '!';
		for(ptr=from->username;*ptr;)
			*curptr++ = *ptr++;
		*curptr++ = '@';
		for(ptr=from->hostname;*ptr;)
			*curptr++ = *ptr++;
		*curptr = '\0';
	}
	va_start(ap, format);
	ret = _internal_send_to_client(to, sender, format, ap);
	va_end(ap);
	return ret;
}

int send_to_opers(char *format, ...)
{
	va_list		ap;
	static char buf[2048];
	static char *nptr;
	static char *eptr;
	static int	doneonce = 0;
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	if (!doneonce)
	{
		doneonce = 1;
		sprintf(buf, ":%s NOTICE ", meptr->name);
		nptr = buf + strlen(buf);
		eptr = nptr + NICKLEN;
		eptr += strlen(strcpy(eptr, " :*** Notice -- "));
	}
	len = eptr - buf;
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(eptr, sizeof(buf)-1-len, format, ap);
#else
	len += vsprintf(eptr, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=*Opers;cptr;cptr=cptr->c_os_next)
	{
		int		tlen = NICKLEN;
		char	*dptr = nptr;
		char	*sptr = cptr->name;

		if (!IsLocal(cptr))
			continue;

		for(sptr=cptr->name;*sptr;tlen--)
			*dptr++ = *sptr++;
		while(tlen--)
			*dptr++ = ' ';
		_send_message_to_client(cptr, buf, len+2);
	}
	return len;
}

int send_to_flag(int flag, IRCD_CLIENT *one, char *format, ...)
{
	va_list		ap;
	static char buf[2048];
	static char *nptr;
	static char *eptr;
	static int	doneonce = 0;
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	if (!doneonce)
	{
		doneonce = 1;
		sprintf(buf, ":%s NOTICE ", meptr->name);
		nptr = buf + strlen(buf);
		eptr = nptr + NICKLEN;
		eptr += strlen(strcpy(eptr, " :*** Notice -- "));
	}
	len = eptr - buf;
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(eptr, sizeof(buf)-1-len, format, ap);
#else
	len += vsprintf(eptr, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
	{
		int		tlen = NICKLEN;
		char	*dptr = nptr;
		char	*sptr = cptr->name;

		if (!IsClient(cptr) || (cptr == one))
			continue;

		if (cptr->c_nflags & flag)
		{
			for(sptr=cptr->name;*sptr;tlen--)
				*dptr++ = *sptr++;
			while(tlen--)
				*dptr++ = ' ';
			_send_message_to_client(cptr, buf, len+2);
		}
	}
	return len;
}

int send_wallops_to_flag(int flag, IRCD_CLIENT *one, IRCD_CLIENT *from, char *format, ...)
{
	va_list		ap;
	char		buf[2048];
	char		*nptr;
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	if (IsClient(from))
		sprintf(buf, ":%s!%s@%s WALLOPS ",
				from->name, from->username, from->hostname);
	else
		sprintf(buf, ":%s WALLOPS ", from->name);
	nptr = buf + strlen(buf);
	*nptr++ = ':';
	len = nptr - buf;
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(nptr, sizeof(buf)-1-len, format, ap);
#else
	len += vsprintf(nptr, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
	{
		if (!IsClient(cptr) || (cptr == one))
			continue;

		if (cptr->c_nflags & flag)
		{
			_send_message_to_client(cptr, buf, len+2);
		}
	}
	return len;
}

int send_to_servers(IRCD_CLIENT *one, char *format, ...)
{
	va_list		ap;
	char		buf[2048];
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len = vsnprintf(buf, sizeof(buf)-1, format, ap);
#else
	len = vsprintf(buf, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=meptr->c_servers;cptr;cptr=cptr->c_u_next)
		if (!one || (cptr != one))
			_send_message_to_client(cptr, buf, len+2);
	return len;
}

int send_to_capab_servers(int flag, IRCD_CLIENT *one, char *format, ...)
{
	va_list		ap;
	char		buf[2048];
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len = vsnprintf(buf, sizeof(buf)-1, format, ap);
#else
	len = vsprintf(buf, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=meptr->c_servers;cptr;cptr=cptr->c_u_next)
		if ((!one || (cptr != one)) && (cptr->c_capab_flags & flag))
			_send_message_to_client(cptr, buf, len+2);
	return len;
}

int send_to_notcapab_servers(int flag, IRCD_CLIENT *one, char *format, ...)
{
	va_list		ap;
	char		buf[2048];
	char		*ptr;
	int			len;
	IRCD_CLIENT	*cptr;

	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len = vsnprintf(buf, sizeof(buf)-1, format, ap);
#else
	len = vsprintf(buf, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(ptr=buf+len-1;len;len--,ptr--)
		if (*ptr != '\r' && *ptr != '\n')
			break;
	*(ptr+1) = '\r';
	*(ptr+2) = '\n';
	*(ptr+3) = '\0';

	for(cptr=meptr->c_servers;cptr;cptr=cptr->c_u_next)
		if ((!one || (cptr != one)) && !(cptr->c_capab_flags & flag))
			_send_message_to_client(cptr, buf, len+2);
	return len;
}

int send_to_common_channels(IRCD_CLIENT *pptr, char *format, ...)
{
	va_list				ap;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CLIENTLIST		*users;
	char				clientbuf[2048];
	char				*sptr;
	char				*d1ptr;
	int					len;
	IRCD_CLIENT			*cptr;

	if (!Sentalong_marker || (Sentalong_marker == (unsigned int)-1))
	{
		register unsigned int *ptr;

		for(ptr=Sentalong;ptr<Sentalong_end;ptr++)
			*ptr = Sentalong_marker;
	}
	Sentalong_marker++;
	if (IsLocal(pptr) && pptr->c_connptr)
	{
		assert(pptr->c_connptr->fdlistinfo.fd < Sentalong_size);
		Sentalong[pptr->c_connptr->fdlistinfo.fd] = Sentalong_marker;
	}

	*clientbuf = ':';
	d1ptr = clientbuf+1;
	sptr = pptr->name;
	while((*d1ptr++ = *sptr++) != '\0');
	if (IsClient(pptr))
	{
		*(d1ptr-1) = '!';
		sptr = pptr->username;
		while((*d1ptr++ = *sptr++) != '\0');
		*(d1ptr-1) = '@';
		sptr = pptr->hostname;
		while((*d1ptr++ = *sptr++) != '\0');
	}
	*(d1ptr-1) = ' ';
	len = d1ptr - clientbuf;
	
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(clientbuf+len, sizeof(clientbuf)-1-len, format, ap);
#else
	len += vsprintf(clientbuf+len, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(sptr=clientbuf+len-1;len;len--,sptr--)
		if (*sptr != '\r' && *sptr != '\n')
			break;
	*(sptr+1) = '\r';
	*(sptr+2) = '\n';
	*(sptr+3) = '\0';

	for(chlptr=pptr->c_channels;chlptr;chlptr=chlptr->next)
	{
		for(users=chlptr->chptr->members;users;users=users->next)
		{
			cptr = users->cptr;
			if (!IsLocal(cptr) || !cptr->c_connptr || (Sentalong[cptr->c_connptr->fdlistinfo.fd] == Sentalong_marker))
				continue;
			Sentalong[cptr->c_connptr->fdlistinfo.fd] = Sentalong_marker;
			_send_message_to_client(cptr, clientbuf, len+2);
		}
	}
	
	if (IsLocal(pptr))
		_send_message_to_client(pptr, clientbuf, len+2);

	return 0;
}

int send_to_channel_butone(IRCD_CLIENT *one, int flags, IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *format, ...)
{
	va_list		ap;
	IRCD_CLIENTLIST		*users;
	char				clientbuf[2048];
	char				serverbuf[512];
	char				*sptr;
	char				*d1ptr;
	char				*d2ptr;
	int					len;
	int					slen = 0;
	IRCD_CLIENT			*cptr;

	if (!Sentalong_marker || (Sentalong_marker == (unsigned int)-1))
	{
		register unsigned int *ptr;

		for(ptr=Sentalong;ptr<Sentalong_end;ptr++)
			*ptr = Sentalong_marker;
	}
	Sentalong_marker++;

	*clientbuf = ':';
	d1ptr = clientbuf+1;
	*serverbuf = ':';
	d2ptr = serverbuf+1;
	sptr = pptr->name;
	while((*d1ptr++ = *d2ptr++ = *sptr++) != '\0');
	if (IsClient(pptr))
	{
		*(d1ptr-1) = '!';
		sptr = pptr->username;
		while((*d1ptr++ = *sptr++) != '\0');
		*(d1ptr-1) = '@';
		sptr = pptr->hostname;
		while((*d1ptr++ = *sptr++) != '\0');
	}
	*(d1ptr-1) = ' ';
	*(d2ptr-1) = ' ';
	len = d1ptr - clientbuf;
	
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(d1ptr, sizeof(clientbuf)-1-len, format, ap);
#else
	len += vsprintf(d1ptr, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(sptr=clientbuf+len-1;len;len--,sptr--)
		if (*sptr != '\r' && *sptr != '\n')
			break;
	*(sptr+1) = '\r';
	*(sptr+2) = '\n';
	*(sptr+3) = '\0';
	if (flags & CHANNEL_SEND_SERVERS)
	{
		strcpy(d2ptr, d1ptr);
		slen = len - (d1ptr - clientbuf) + (d2ptr - serverbuf);
	}

	for(users=chptr->members;users;users=users->next)
	{
		if ((flags & CHANNEL_SEND_LOCALONLY) && !IsLocal(users->cptr))
			continue;
		if ((cptr = users->cptr->c_from) == one)
			continue;
		if (IsServer(cptr))
		{
			if (!(flags & CHANNEL_SEND_SERVERS) ||
						(*chptr->name == '&'))
				continue;
		}
		else
		{
			if (!(flags & CHANNEL_SEND_CLIENTS))
				continue;
		}
		assert(cptr->c_connptr->fdlistinfo.fd < Sentalong_size);
		if (!cptr->c_connptr ||
			(Sentalong[cptr->c_connptr->fdlistinfo.fd] == Sentalong_marker))
				continue;
		Sentalong[cptr->c_connptr->fdlistinfo.fd] = Sentalong_marker;
		if (IsServer(cptr))
			_send_message_to_client(cptr, serverbuf, slen+2);
		else
			_send_message_to_client(cptr, clientbuf, len+2);
	}
	return 0;
}

int send_to_channel_butone2(IRCD_CLIENT *one, int flags, IRCD_CHANNEL *chptr, char *prefix, char *format, ...)
{
	va_list		ap;
	IRCD_CLIENTLIST		*users;
	char				clientbuf[2048];
	char				serverbuf[512];
	char				*sptr;
	char				*d1ptr;
	char				*d2ptr;
	int					len;
	int					slen = 0;
	IRCD_CLIENT			*cptr;

	if (!Sentalong_marker || (Sentalong_marker == (unsigned int)-1))
	{
		register unsigned int *ptr;

		for(ptr=Sentalong;ptr<Sentalong_end;ptr++)
			*ptr = Sentalong_marker;
	}
	Sentalong_marker++;

	*clientbuf = ':';
	d1ptr = clientbuf+1;
	*serverbuf = ':';
	d2ptr = serverbuf+1;
	sptr = prefix;
	while((*d1ptr++ = *d2ptr++ = *sptr++) != '\0');
	*(d1ptr-1) = ' ';
	*(d2ptr-1) = ' ';
	len = d1ptr - clientbuf;
	
	va_start(ap, format);
#ifdef HAVE_VSNPRINTF
	len += vsnprintf(d1ptr, sizeof(clientbuf)-1-len, format, ap);
#else
	len += vsprintf(d1ptr, format, ap);
#endif
	va_end(ap);
	if (len > 509)
	{
		len = 509;
	}
	for(sptr=clientbuf+len-1;len;len--,sptr--)
		if (*sptr != '\r' && *sptr != '\n')
			break;
	*(sptr+1) = '\r';
	*(sptr+2) = '\n';
	*(sptr+3) = '\0';
	if (flags & CHANNEL_SEND_SERVERS)
	{
		strcpy(d2ptr, d1ptr);
		slen = len - (d1ptr - clientbuf) + (d2ptr - serverbuf);
	}

	for(users=chptr->members;users;users=users->next)
	{
		if ((flags & CHANNEL_SEND_LOCALONLY) && !IsLocal(users->cptr))
			continue;
		if ((cptr = users->cptr->c_from) == one)
			continue;
		if (IsServer(cptr))
		{
			if (!(flags & CHANNEL_SEND_SERVERS))
				continue;
		}
		else
		{
			if (!(flags & CHANNEL_SEND_CLIENTS))
				continue;
		}
		assert(cptr->c_connptr->fdlistinfo.fd < Sentalong_size);
		if (!cptr->c_connptr ||
			(Sentalong[cptr->c_connptr->fdlistinfo.fd] == Sentalong_marker))
				continue;
		Sentalong[cptr->c_connptr->fdlistinfo.fd] = Sentalong_marker;
		if (IsServer(cptr))
			_send_message_to_client(cptr, serverbuf, slen+2);
		else
			_send_message_to_client(cptr, clientbuf, len+2);
	}
	return 0;
}
