/*
**   IRC - Internet Relay Chat, source/support.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990, 1991 Armin Gruner
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "support.h"

RCSID("$Id: support.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");

int ircd_getpagesize(void)
{
#ifdef HAVE_GETPAGESIZE
	return getpagesize();
#else
#ifdef _SC_PAGE_SIZE
	return sysconf(_SC_PAGE_SIZE);
#else
	return sysconf(_SC_PAGESIZE);
#endif
#endif
	return 4096;
};

void support_free(void *ptr)
{
	if (ptr)
		free(ptr);
}

void support_free_dbg(void *ptr, char *filename, int lineno)
{
	if (ptr)
		free(ptr);
	fprintf(stderr, "MyFree() on %s:%i\n", filename, lineno);
}

void *support_malloc(size_t len)
{
	void *ptr;

	ptr = malloc(len);
	if (ptr)
		support_bzero(ptr, len);
	return ptr;
}


void *support_realloc(void *ptr, size_t len)
{
	return realloc(ptr, len);
}

char *support_strdup(char *string)
{
	return strdup(string);
}

void support_bcopy(const void *src, void *dest, size_t n)
{
	if (n != 0)
	{
		register char *s1 = dest;
		register const char *s2 = src;

		if (s1 <= s2)
		{
			do
			{
				*s1++ = *s2++;
			} while (--n != 0);
		}
		else
		{
			s2 += n;
			s1 += n;
			do
			{
				*--s1 = *--s2;
			} while (--n != 0);
		}
	}
}

int support_bcmp(const void *s1, const void *s2, size_t n)
{
	register const unsigned char *t1 = s1;
	register const unsigned char *t2 = s2;

	while(n && (*(t1++) == *(t2++)))
		n--;
	return n ? 1 : 0;
}

void support_bzero(void *s, register size_t n)
{
	register char *t = s;

	while(n--)
		*(t++) = '\0';
}

char *support_strncpy(char *dest, char *src, int len)
{
	register char *start = dest;
	register char *ending = dest + len - 1;

	while(*src && (dest < ending))
		*dest++ = *src++;
	if (len > 0)
		*dest = '\0';
	return start;
}

char *support_stristr(char *str1, char *str2)
{
    register char	*s1 = str1;
	register char	*s2 = str2;
    register char	*tptr;
    register char	c;

	if (s2 == NULL || *s2 == '\0')
        return str1;

	c = *s2;

	while (*s1)
	{
		if (toupper(*s1++) == toupper(c))
		{
			tptr = s1;
            while ((toupper(c = *++s2) == toupper(*s1++)) && c)
				continue;
			if (!c)
   				return(tptr-1);
			s1 = tptr;
   			s2 = str2;
			c = *s2;
		}
	}
	return NULL;
}

char *support_ctime(int thetime)
{
	static char buffer[64];
	struct tm *tmptr;

	tmptr = localtime((time_t*)&thetime);
	strftime(buffer, sizeof(buffer)-1, "%a %b %e %T %Y %Z", tmptr);
	buffer[sizeof(buffer)-1] = '\0';
	return buffer;
}

char *support_strerror(int errnum)
{
#ifdef HAVE_STRERROR
	return strerror(errnum);
#else
	return (errnum <= 0 || errnum > sys_nerr) ? "Undefined error" : sys_errlist[errnum];
#endif
}

#ifdef USE_SUPPORT_POLL

int support_poll(struct pollfd *pfd, int nfds, int timeout)
{
	struct timeval	tv;
	int				ret;
	int				i;
	fd_set			readset;
	fd_set			writeset;
	int				highfd = -1;

	tv.tv_sec = timeout/1000;
	tv.tv_usec = timeout % 1000;

	FD_ZERO(&readset);
	FD_ZERO(&writeset);

	for(i=0;i<nfds;i++)
	{
		if (pfd[i].events & POLLIN)
		{
			FD_SET(pfd[i].fd, &readset);
			if (pfd[i].fd > highfd)
				highfd = pfd[i].fd;
		}
		if (pfd[i].events & POLLOUT)
		{
			FD_SET(pfd[i].fd, &writeset);
			if (pfd[i].fd > highfd)
				highfd = pfd[i].fd;
		}
	}

	ret = select(highfd + 1, &readset, &writeset, NULL,
					(timeout == -1) ? NULL : &tv);

	if (ret <= 0)
		return ret;

	for(i=0;i<nfds;i++)
	{
		if (FD_ISSET(pfd[i].fd, &readset))
			pfd[i].revents |= POLLIN;
		if (FD_ISSET(pfd[i].fd, &writeset))
			pfd[i].revents |= POLLOUT;
	}
	return ret;
}

#endif
