/*
**   IRC - Internet Relay Chat, source/log.c
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define DEFINE_VARS_DEFINES
#include "ircd_includes.h"
#include "vars.h"

RCSID("$Id: vars.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");

static void _vars_value_escape(char *value);
static int _vars_value_set(IRCD_CLIENT *pptr, IRCD_VAR *vptr, void *value);

static void _vars_value_escape(char *value)
{
#if 1
	return;
#else
    char    *ptr;       
    char    *nptr;      
    int     ws = 0;     
    
    if (*value == '"')  
    {   
        ws = 1;
		for(nptr=value;(*nptr=*(nptr+1));nptr++)
			continue;
    }   
  
    for(ptr=value;*ptr;ptr++)
    {
        if (!ws && ((*ptr == ' ') || (*ptr == '\t')))
		{
            break;
		}
        else if (*ptr == '\\')
        {
            switch(*(ptr+1))
            {
				/* special chars */
                default:
                    for(nptr=ptr;(*nptr=*(nptr+1));nptr++)
						continue;
                    break;
            }
        }   
        else if (ws && (*ptr == '"'))
            break;
    }
    *ptr = '\0';
    return;
#endif
}

static int _vars_value_set(IRCD_CLIENT *pptr, IRCD_VAR *vptr, void *value)
{
	int			vlen;

	if (!vptr)
		return -EINVAL;
	if (vptr->v_functionbefore)
	{
		vlen = vptr->v_functionbefore(pptr, vptr, value);
		if (vlen)
			return vlen;
	}
	switch(vptr->v_type)
	{
		case IRCD_VAR_TYPE_STRING:
			vlen = value ? strlen(value) : 0;
			if ((vptr->v_val_max && (vlen > vptr->v_val_max)) ||
						(vlen < vptr->v_val_min))
			{
				return -EINVAL;
			}

			if (vptr->v_value)
			{
				MyFree(vptr->v_value);
			}
			vptr->v_value = value ? MyStrdup(value) : value;
			_vars_value_escape(vptr->v_value);
			break;

		case IRCD_VAR_TYPE_INT:
			vlen = (int)value;
			if ((vptr->v_val_max && (vlen > vptr->v_val_max)) ||
					(vlen < vptr->v_val_min))
			{
				return -EINVAL;
			}
			/* fall through */
		case IRCD_VAR_TYPE_BOOL:
			vptr->v_value = value;
			break;
		default:
			return -EFAULT;
	}
	if (vptr->v_functionafter)
		vptr->v_functionafter(pptr, vptr);

	return 0;
}

void vars_init(void)
{
	IRCD_VAR	*vptr;

	for(vptr = Vars_Table;vptr->v_name;vptr++)
	{
		vptr->v_name = MyStrdup(vptr->v_name);
		if ((vptr->v_type == IRCD_VAR_TYPE_STRING) && vptr->v_value)
			vptr->v_value = MyStrdup(vptr->v_value);
	}

}

IRCD_VAR *vars_find_byname(char *name)
{
	IRCD_VAR	*vptr;

	for(vptr=Vars_Table;vptr->v_name;vptr++)
		if (!strcasecmp(vptr->v_name, name))
			return vptr;
	return NULL;
}

void *vars_value_get_byname(char *name)
{
	IRCD_VAR	*vptr = vars_find_byname(name);

	if (!vptr)
	{
		return NULL;
	}

	return vptr->v_value;
}

int vars_value_set_byname(IRCD_CLIENT *pptr, char *name, void *value)
{
	IRCD_VAR	*vptr = vars_find_byname(name);

	if (!vptr)
	{
		return -ENOENT;
	}
	return _vars_value_set(pptr, vptr, value);
}

int vars_value_set_bynum(IRCD_CLIENT *pptr, int num, void *value)
{
	return _vars_value_set(pptr, &(Vars_Table[num]), value);
}

IRCD_VAR *vars_value_getall(void)
{
	if (!Vars_Table[0].v_name)
		return NULL;
	return Vars_Table;
}

IRCD_VAR *vars_value_getnext(IRCD_VAR *v)
{
	if (!((v+1)->v_name))
		return NULL;
	return v+1;
}

int vars_logfilename_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value)
{
	int len;

	if (!value || ((len = strlen(value)) < 5) ||
			((*value == '/') && (strncmp(value, "/var/", 5) ||
								(strstr(value, "..") != NULL))) ||
			((*value == '.') && (*(value+1) == '.')) ||
			(strcmp(value+len-4, ".log")))
	{
		if (!pptr)
			syslog(LOG_NOTICE, "invalid motd filename in conf.");
		else
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Filename must end in \".log\".  The file is allowed to reside in /var.", pptr->name);
		return -EINVAL;
	}
	return 0;
}

int vars_motdfilename_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value)
{
	int len;

	if (!value || ((len = strlen(value)) < 6) ||
			(*value == '/') || ((*value == '.') && (*(value+1) == '.')) ||
			(strcmp(value+len-5, ".motd")))
	{
		if (!pptr)
			syslog(LOG_NOTICE, "invalid motd filename in conf.");
		else
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Filename must end in \".motd\" and not start with a period or slash.",
				pptr->name);
		return -EINVAL;
	}
	return 0;
}

int vars_klinesfilename_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value)
{
	int len;

	if (!value || ((len = strlen(value)) < 6) ||
			(*value == '/') || ((*value == '.') && (*(value+1) == '.')) ||
			(strcmp(value+len-5, ".conf")))
	{
		if (!pptr)
			syslog(LOG_NOTICE, "invalid klines filename in conf.");
		else
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Filename must end in \".conf\" and not start with a period or slash.",
				pptr->name);
		return -EINVAL;
	}
	return 0;
}

int vars_whowas_change(IRCD_CLIENT *pptr, IRCD_VAR *v)
{
	return whowas_init();
}

int vars_logfile_change(IRCD_CLIENT *pptr, IRCD_VAR *v)
{
	logs_init();
	return 0;
}

int vars_init_reload(void *varsptr)
{
	IRCD_VAR	*vptr;
	IRCD_VAR	*vptr2;

	for(vptr = Vars_Table;vptr->v_name;vptr++)
	{
		vptr->v_name = MyStrdup(vptr->v_name);
		if ((vptr->v_type == IRCD_VAR_TYPE_STRING) && vptr->v_value)
			vptr->v_value = MyStrdup(vptr->v_value);
	}

	for(vptr=(IRCD_VAR *)varsptr;vptr->v_name;vptr++)
	{
		for(vptr2=(IRCD_VAR *)Vars_Table;vptr2->v_name;vptr2++)
		{
			if (!strcmp(vptr2->v_name, vptr->v_name))
				break;
		}
		if (!vptr2->v_name)
		{
			if (vptr->v_value && (vptr->v_type == IRCD_VAR_TYPE_STRING))
			{
				MyFree(vptr->v_value);
			}
			MyFree(vptr->v_name);
			continue;
		}
		if (vptr2->v_value && (vptr2->v_type == IRCD_VAR_TYPE_STRING))
			MyFree(vptr2->v_value);
		vptr2->v_val_min = vptr->v_val_min;
		vptr2->v_val_max = vptr->v_val_max;
		vptr2->v_type = vptr->v_type;
		vptr2->v_value = vptr->v_value;
		MyFree(vptr->v_name);
	}
	MyFree(varsptr);
	return 0;
}

int vars_deinit_reload(void **varsptr)
{
	*varsptr = (IRCD_VAR *)MyMalloc(sizeof(Vars_Table));
	BCOPY(Vars_Table, *varsptr, sizeof(Vars_Table));
	return 0;
}
