/*
**   IRC - Internet Relay Chat, source/whowas.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Markku Savela
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "whowas.h"
#include "commands.h"

RCSID("$Id: whowas.c,v 1.19 2001/09/30 11:02:15 cbehrens Exp $");
 
static IRCD_WHOWAS *Whowas = NULL;
static IRCD_WHOWAS **Whowashash = NULL;

static int _whowas_size = 0; /* default is in whowas_init */
static int _whowas_next = 0;

static void whowas_add_to_list(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas);
static void whowas_del_from_list(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas);
static void whowas_add_to_clist(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas);
static void whowas_del_from_clist(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas);
static void whowas_free(IRCD_WHOWAS *new_whowas);

static void whowas_add_to_list(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas)
{
	whowas->prev = NULL;
	if ((whowas->next = *bucket) != NULL)
		whowas->next->prev = whowas;
	*bucket = whowas;
}

static void whowas_del_from_list(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas)
{
	if (whowas->prev)
		whowas->prev->next = whowas->next;
	else
		*bucket = whowas->next;
	if (whowas->next)
		whowas->next->prev = whowas->prev;
}

static void whowas_add_to_clist(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas)
{
    whowas->cprev = NULL;
    if ((whowas->cnext = *bucket) != NULL)
        whowas->cnext->cprev = whowas;
    *bucket = whowas;
}

static void whowas_del_from_clist(IRCD_WHOWAS **bucket, IRCD_WHOWAS *whowas)
{
    if (whowas->cprev)
        whowas->cprev->cnext = whowas->cnext;
    else
        *bucket = whowas->cnext;
    if (whowas->cnext)
        whowas->cnext->cprev = whowas->cprev;
}

/* only call if new_whowas->hashv != -1 */
static void whowas_free(IRCD_WHOWAS *new_whowas)
{
	if (new_whowas->online)
		whowas_del_from_clist(&(new_whowas->online->c_whowas),
			new_whowas);
	whowas_del_from_list(&Whowashash[new_whowas->hashv % HASH_WHOWAS_SIZE], new_whowas);
	if (new_whowas->name)
		MyFree(new_whowas->name);
	if (new_whowas->username)
		MyFree(new_whowas->username);
	if (new_whowas->servername)
		MyFree(new_whowas->servername);
	if (new_whowas->hostname)
		MyFree(new_whowas->hostname);
	if (new_whowas->realname)
		MyFree(new_whowas->realname);
}

void whowas_add_history(IRCD_CLIENT *cptr, int online)
{
	IRCD_WHOWAS			*new_whowas;
	unsigned int	hv;

	new_whowas = &Whowas[_whowas_next];
	if (new_whowas->hashv != -1)
	{
		whowas_free(new_whowas);
	}
	new_whowas->hashv = hash_compute_whowas(cptr->name);
	hv = new_whowas->hashv % HASH_WHOWAS_SIZE;
	new_whowas->logoff = Now;
	new_whowas->name = MyStrdup(cptr->name);
	new_whowas->username = MyStrdup(cptr->username);
	new_whowas->hostname = MyStrdup(cptr->hostname);
	new_whowas->realname = MyStrdup(cptr->ircname);
	new_whowas->servername = MyStrdup(cptr->servername);
	if (online)
	{
		new_whowas->online = cptr;
		whowas_add_to_clist(&(cptr->c_whowas), new_whowas);
	}
	else
		new_whowas->online = NULL;
	whowas_add_to_list(&Whowashash[hv], new_whowas);
	if (++_whowas_next == _whowas_size)
		_whowas_next = 0;
}

void whowas_off_history(IRCD_CLIENT *cptr)
{
	IRCD_WHOWAS *temp, *next;

	for(temp=cptr->c_whowas;temp;temp=next)
	{
		next = temp->cnext;
		temp->online = NULL;
		whowas_del_from_clist(&(cptr->c_whowas), temp);
	}
	cptr->c_whowas = NULL;
}

IRCD_CLIENT *whowas_get_history(char *nick, int timelimit)
{
	IRCD_WHOWAS			*temp;
	unsigned int	hashv;

	if (timelimit)
		timelimit = Now - timelimit;
	hashv = hash_compute_whowas(nick);
	temp = Whowashash[hashv % HASH_WHOWAS_SIZE];
	for(;temp;temp=temp->next)
	{
		if ((temp->hashv != hashv) || mycmp(nick, temp->name))
			continue;
		if (temp->logoff < timelimit)
			continue;
		return temp->online;
	}
	return NULL;
}

void whowas_count_memory(int *wwu, unsigned int *wwum, unsigned int *wwas)
{
	IRCD_WHOWAS	*tmp;
	int		i;
	int		u = 0;
	unsigned int	um = 0;

	for (i = 0, tmp = &Whowas[0]; i < _whowas_size; i++, tmp++)
		if (tmp->hashv != -1)
		{
			u++;
			um += (strlen(tmp->name) + 1);
			um += (strlen(tmp->username) + 1);
			um += (strlen(tmp->hostname) + 1);
			um += (strlen(tmp->servername) + 1);
			um += (strlen(tmp->realname) + 1);
		}
	*wwu = u;
	*wwum = um;
	*wwas = sizeof(Whowashash) + (_whowas_size * sizeof(IRCD_WHOWAS));
	return;
}

/*
** whowas_command
**	  parv[0] = sender prefix
**	  parv[1] = nickname queried
*/
int whowas_command(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_WHOWAS		*temp;
	unsigned int	hashv;
	int				cur = 0;
	int				max = -1;
	char			*nick;
	char			argssave[512];
	char			*maxa;

	if (args)
		strcpy(argssave, args);

	nick = ircd_nextarg(&args, 0);
	maxa = ircd_nextarg(&args, 0);

	if (!nick || !*nick)
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NONICKNAMEGIVEN, pptr->name));
		return 3;
	}

	if (args && *args)
		if (_hunt_server(lptr, pptr, ":%s WHOWAS %s %s :%s", 3, prefix, argssave))
			return 3;

	if (maxa)
		max = atoi(maxa);

	hashv = hash_compute_whowas(nick);
	temp = Whowashash[hashv % HASH_WHOWAS_SIZE];
	for(;temp;temp=temp->next)
	{
		if ((hashv == temp->hashv) && !mycmp(nick, temp->name))
		{
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_WHOWASUSER, pptr->name),
					temp->name, temp->username, temp->hostname, temp->realname);
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_WHOISSERVER, pptr->name),
					temp->name, temp->servername, support_ctime(temp->logoff));
			if ((++cur >= max) && (max > 0))
				break;
		}
	}
	if (!cur)
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_WASNOSUCHNICK, pptr->name), nick);
	send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_ENDOFWHOWAS, pptr->name), nick);
	return 3;
}

int whowas_init(void)
{
	int			i;
	int			newsize = 8*1024; /* resonable default */
	IRCD_WHOWAS	*nww;

	if (!Whowashash)
	{
		Whowashash = (IRCD_WHOWAS **)MyMalloc(sizeof(IRCD_WHOWAS *)*HASH_WHOWAS_SIZE);
	}

	i = (int)vars_value_get_bynum(IRCD_VAR_WHOWAS_SIZE);
	if (i > 0)
		newsize = i;

	if (newsize != _whowas_size) /* will be true for 1st _init, too */
	{
		nww = (IRCD_WHOWAS *)MyMalloc(sizeof(IRCD_WHOWAS) * newsize);
		if (Whowas)
		{
			if (_whowas_size < newsize) /* expanding */
			{
				BCOPY((char *)Whowas, nww, sizeof(IRCD_WHOWAS) * _whowas_size);
				for(i=_whowas_size;i < newsize;i++)
				{
					BZERO((char *)(nww + i), sizeof(IRCD_WHOWAS));
					nww[i].hashv = -1;
				}
			}
			else /* shrinking */
			{
				for(i=newsize;i<_whowas_size;i++)
				{
					if (Whowas[i].hashv != -1)
						whowas_free(&Whowas[i]);
				}
				BCOPY((char *)Whowas, nww, sizeof(IRCD_WHOWAS) * newsize);
			}
			MyFree(Whowas);
		}
		else
		{
			for(i=0;i<newsize;i++)
			{
				BZERO((char *)(nww + i), sizeof(IRCD_WHOWAS));
				nww[i].hashv = -1;
			}
		}
		if (_whowas_size == 0) /* first _init */
		{
			for (i=0;i<HASH_WHOWAS_SIZE;i++)
				Whowashash[i] = NULL;	
		}
		Whowas = nww;
		_whowas_size = newsize;
		if (_whowas_next >= _whowas_size) /* need this check if shrinking */
			_whowas_next = 0;
	}
	return 0;

	return 0;
}

int whowas_init_reload(void *whowasarr, int whowas_size, void **whowashash, int whowas_hash_size, int whowas_next)
{
	IRCD_WHOWAS		*tmp;
	int				i;

	Whowas = (IRCD_WHOWAS *)whowasarr;
	_whowas_size = whowas_size;
	_whowas_next = whowas_next;

	if (whowas_hash_size == HASH_WHOWAS_SIZE)
	{
		Whowashash = (IRCD_WHOWAS **)whowashash;
		return 0;
	}

	Whowashash = (IRCD_WHOWAS **)MyMalloc(sizeof(IRCD_WHOWAS *)*HASH_WHOWAS_SIZE);

	for (i = 0, tmp = &Whowas[0]; i < _whowas_size; i++, tmp++)
	{
		if (tmp->hashv != -1)
			whowas_add_to_list(&Whowashash[tmp->hashv % HASH_WHOWAS_SIZE],tmp);
	}

	MyFree(whowashash);

	return 0;
}

int whowas_deinit_reload(void **whowasarr, int *whowas_size, void ***whowashash, int *whowas_hash_size, int *whowas_next)
{
	if (whowasarr)
		*whowasarr = (void *)Whowas;
	if (whowas_size)
		*whowas_size = _whowas_size;
	if (whowashash)
		*whowashash = (void **)Whowashash;
	if (whowas_hash_size)
		*whowas_hash_size = HASH_WHOWAS_SIZE;
	if (whowas_next)
		*whowas_next = _whowas_next;
	return 0;
}

