/*
**   Malloc debugging routines (dmalloc.c)
**   Copyright (C) 1999 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "dmalloc.h"


static int *_hashtab;
static int _dmalloc_inited = 0;
static int _dmalloc_fbsize;
static int _dmalloc_phashsize;
static int _dmalloc_fhashsize;
static DMALLOC_FPTR	**DMALLOC_FHASH;
static DMALLOC_PPTR **DMALLOC_PHASH;

/* for free cache */
static int _dmalloc_cache_frees = 0;
static DMALLOC_FPTR	**DMALLOC_FFHASH;
static DMALLOC_PPTR **DMALLOC_PFHASH;
static DMALLOC_PPTR	*Freelist_tail = NULL;
static int _dmalloc_freelist_num = 0;


static int _dmalloc_file_get_hnum = 0;
static DMALLOC_FPTR *_dmalloc_file_get_ptr = NULL;


static void add_fptr_to_list(DMALLOC_FPTR **bucket, DMALLOC_FPTR *fptr);
static void add_dmalloc_to_flist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc);
static void del_dmalloc_from_flist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc);
static void add_dmalloc_to_plist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc);
static void del_dmalloc_from_plist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc);
static unsigned int _dmalloc_file_hash(char *file, int line);
static DMALLOC_FPTR *_dmalloc_file_find(DMALLOC_FPTR **bucket, char *file, int line);
static DMALLOC_PPTR *_dmalloc_pptr_find(char *ptr);
static DMALLOC_PPTR *_dmalloc_pptr_find_freed(char *ptr);
static DMALLOC_FPTR	*_dmalloc_file_create(DMALLOC_FPTR **bucket, char *file, int line);
static void _dmalloc_add(char *ptr, int size, char *file, int line);
static void _dmalloc_check(DMALLOC_PPTR *pptr);
static void __dmalloc_free(DMALLOC_PPTR *pptr);
static void _dmalloc_free(DMALLOC_PPTR *pptr);
static void _dmalloc_cache_free(DMALLOC_PPTR *pptr, char *file, int line);
static void __dmalloc_realfree(DMALLOC_PPTR *pptr);
static void _dmalloc_realfree(DMALLOC_PPTR *pptr);

static void add_fptr_to_list(DMALLOC_FPTR **bucket, DMALLOC_FPTR *fptr)
{
	fptr->prev = NULL;
	if ((fptr->next = *bucket) != NULL)
		fptr->next->prev = fptr;
	*bucket = fptr;
}

static void add_dmalloc_to_flist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc)
{
    dmalloc->f_prev = NULL;
    if ((dmalloc->f_next = *bucket) != NULL)
        dmalloc->f_next->f_prev = dmalloc;
    *bucket = dmalloc;
}

static void del_dmalloc_from_flist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc)
{
    if (dmalloc->f_prev)
        dmalloc->f_prev->f_next = dmalloc->f_next;
    else
        *bucket = dmalloc->f_next;
    if (dmalloc->f_next)
        dmalloc->f_next->f_prev = dmalloc->f_prev;
}

static void add_dmalloc_to_plist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc)
{
    dmalloc->p_prev = NULL;
    if ((dmalloc->p_next = *bucket) != NULL)
        dmalloc->p_next->p_prev = dmalloc;
    *bucket = dmalloc;
}

static void del_dmalloc_from_plist(DMALLOC_PPTR **bucket, DMALLOC_PPTR *dmalloc)
{
    if (dmalloc->p_prev)
        dmalloc->p_prev->p_next = dmalloc->p_next;
    else
        *bucket = dmalloc->p_next;
    if (dmalloc->p_next)
        dmalloc->p_next->p_prev = dmalloc->p_prev;
}

static unsigned int _dmalloc_file_hash(char *file, int line)
{
	char	buffer[128];
	unsigned char	*name = (unsigned char *)buffer;
	unsigned int	hash = 1;

	sprintf(buffer, "%s:%i", file, line);
	for(;*name;name++)
	{
		hash <<= 1;
		hash += _hashtab[(unsigned int)*name];
	}
	return hash;
}

static DMALLOC_FPTR *_dmalloc_file_find(DMALLOC_FPTR **bucket, char *file, int line)
{
	unsigned int 	hashv;
	DMALLOC_FPTR	*fptr;

	hashv = _dmalloc_file_hash(file, line);
	for(fptr=bucket[hashv % _dmalloc_fhashsize];fptr;fptr=fptr->next)
	{
		if ((fptr->hashv == hashv) && !strcmp(file, fptr->file) &&
				(fptr->line == line))
			break;
	}
	return fptr;
}

static DMALLOC_PPTR *_dmalloc_pptr_find(char *ptr)
{
	DMALLOC_PPTR	*pptr;

	for(pptr=DMALLOC_PHASH[((unsigned int)ptr)%_dmalloc_phashsize];pptr;pptr=pptr->p_next)
	{
		if (pptr->ptr == ptr)
			break;
	}
	return pptr;
}

static DMALLOC_PPTR *_dmalloc_pptr_find_freed(char *ptr)
{
	DMALLOC_PPTR	*pptr;

	for(pptr=DMALLOC_PFHASH[((unsigned int)ptr)%_dmalloc_phashsize];pptr;pptr=pptr->p_next)
	{
		if (pptr->ptr == ptr)
			break;
	}
	return pptr;
}

static DMALLOC_FPTR	*_dmalloc_file_create(DMALLOC_FPTR **bucket, char *file, int line)
{
	DMALLOC_FPTR	*fptr;

	fptr = (DMALLOC_FPTR *)malloc(sizeof(DMALLOC_FPTR));
	memset(fptr, 0, sizeof(DMALLOC_FPTR));
	fptr->hashv = _dmalloc_file_hash(file, line);
	fptr->file = file;
	fptr->line = line;
	add_fptr_to_list(&(bucket[fptr->hashv % _dmalloc_fhashsize]), fptr);
	return fptr;
}

static void _dmalloc_add(char *ptr, int size, char *file, int line)
{
	DMALLOC_FPTR	*fptr;
	DMALLOC_PPTR	*pptr;

	memset(ptr, 'A', _dmalloc_fbsize);
	memset(ptr+size+_dmalloc_fbsize, 'Z', _dmalloc_fbsize);
	pptr = (DMALLOC_PPTR *)malloc(sizeof(DMALLOC_PPTR));
	memset(pptr, 0, sizeof(DMALLOC_PPTR));
	pptr->flags = 0;
	pptr->ptr = ptr;
	pptr->size = size;
	fptr = _dmalloc_file_find(DMALLOC_FHASH, file, line);
	if (!fptr)
		fptr = _dmalloc_file_create(DMALLOC_FHASH, file, line);
	pptr->fptr = fptr;
	fptr->nmallocs++;
	add_dmalloc_to_flist(&(fptr->mallocs), pptr);
	add_dmalloc_to_plist(&(DMALLOC_PHASH[((unsigned int)ptr) % _dmalloc_phashsize]), pptr);
}

static void _dmalloc_check(DMALLOC_PPTR *pptr)
{
	int		i;
	char	*ptr = pptr->ptr;

	for(i=0;i<_dmalloc_fbsize;i++)
		assert(ptr[i] == 'A');
	ptr = pptr->ptr + pptr->size + _dmalloc_fbsize;
	for(i=0;i<_dmalloc_fbsize;i++)
		assert(ptr[i] == 'Z');
}

static void __dmalloc_free(DMALLOC_PPTR *pptr)
{
	free(pptr->ptr);
	del_dmalloc_from_plist(&(DMALLOC_PHASH[((unsigned int)pptr->ptr) % _dmalloc_phashsize]), pptr);
	pptr->fptr->nmallocs--;
	del_dmalloc_from_flist(&(pptr->fptr->mallocs), pptr);
}

static void _dmalloc_free(DMALLOC_PPTR *pptr)
{
	__dmalloc_free(pptr);
	free(pptr);
}

static void _dmalloc_cache_free(DMALLOC_PPTR *pptr, char *file, int line)
{
	DMALLOC_PPTR	*pptrblah;

	pptrblah = _dmalloc_pptr_find_freed(pptr->ptr);
	__dmalloc_free(pptr);

	if (pptrblah)
	{
		if (pptrblah == Freelist_tail)
			Freelist_tail = pptrblah->p_prev;
		_dmalloc_realfree(pptrblah);
	}

	pptr->fptr = _dmalloc_file_find(DMALLOC_FFHASH, file, line);
	if (!pptr->fptr)
		pptr->fptr = _dmalloc_file_create(DMALLOC_FFHASH, file, line);
	add_dmalloc_to_flist(&(pptr->fptr->mallocs), pptr);
	pptr->fptr->nmallocs++;
	add_dmalloc_to_plist(&(DMALLOC_PFHASH[((unsigned int)pptr->ptr) % _dmalloc_phashsize]), pptr);
	if (!Freelist_tail)
		Freelist_tail = pptr;
	if (_dmalloc_freelist_num == _dmalloc_cache_frees)
	{
		DMALLOC_PPTR	*temp = Freelist_tail->p_prev;
		_dmalloc_realfree(Freelist_tail);
		Freelist_tail = temp;
	}
	else
		_dmalloc_freelist_num++;
}

static void __dmalloc_realfree(DMALLOC_PPTR *pptr)
{
	pptr->fptr->nmallocs--;
	del_dmalloc_from_flist(&(pptr->fptr->mallocs), pptr);
	del_dmalloc_from_plist(&(DMALLOC_PFHASH[((unsigned int)pptr->ptr) % _dmalloc_phashsize]), pptr);
}

static void _dmalloc_realfree(DMALLOC_PPTR *pptr)
{
	__dmalloc_realfree(pptr);
	free(pptr);
}

void *dmalloc_malloc(int size, char *file, int line)
{
	char	*ptr;

	if (!_dmalloc_inited || !size)
		return NULL;
	ptr = (char *)malloc(size + (_dmalloc_fbsize*2));
	if (ptr)
	{
		memset(ptr+_dmalloc_fbsize, 0, size);
		_dmalloc_add(ptr, size, file, line);
		ptr += _dmalloc_fbsize;
	}
	return ptr;
}

char *dmalloc_strdup(char *string, char *file, int line)
{
	char	*ptr;

	if (!string)
		return NULL;
	ptr = dmalloc_malloc(strlen(string)+1, file, line);
	if (ptr)
		strcpy(ptr, string);
	return ptr;
}

void *dmalloc_realloc(void *string, int size, char *file, int line)
{
	DMALLOC_PPTR	*pptr;
	char			*ptr;

	if (!string)
		return dmalloc_malloc(size, file, line);
	pptr = _dmalloc_pptr_find(string - _dmalloc_fbsize);
	assert(pptr);
	if (size == pptr->size)
		return string;
	ptr = dmalloc_malloc(size, file, line);
	if (!ptr)
		return ptr;
	if (pptr->size < size)
		memcpy(ptr, string, pptr->size);
	else
		memcpy(ptr, string, size);
	if (_dmalloc_cache_frees)
		_dmalloc_cache_free(pptr, file, line);
	else
		dmalloc_free(string, file, line);
	return ptr;
}

void dmalloc_free(void *ptr, char *file, int line)
{
	DMALLOC_PPTR	*pptr;

	if (!ptr)
		return;

	pptr = _dmalloc_pptr_find(ptr - _dmalloc_fbsize);

	if (!pptr)
	{
		if (!_dmalloc_cache_frees)
		{
			fprintf(stderr, "pointer trying to be freed at %s:%i which was not malloced or was already freed\n", file, line);
			assert(pptr != NULL);
		}
		else
		{
			pptr = _dmalloc_pptr_find_freed(ptr - _dmalloc_fbsize);
			if (pptr)
			{
				fprintf(stderr, "pointer trying to be freed at %s:%i which was already freed at %s:%i\n", file, line, pptr->fptr->file, pptr->fptr->line);
			}
		}
		assert(0);
	}
	else
		_dmalloc_check(pptr);

	if (_dmalloc_cache_frees)
		_dmalloc_cache_free(pptr, file, line);
	else
		_dmalloc_free(pptr);
}

DMALLOC_FPTR *dmalloc_file_find(char *file, int line)
{
	return _dmalloc_file_find(DMALLOC_FHASH, file, line);
}

int dmalloc_init(int phashsize, int fhashsize, int fbsize, int freecache)
{
	int i;

	if ((fbsize < 0) || (phashsize <= 0) || (fhashsize <= 0))
		return -EINVAL;
	if (_dmalloc_inited)
		return -EALREADY;
	_dmalloc_inited++;
	if (fbsize % 8)
		fbsize += (fbsize % 8);
	_dmalloc_fbsize = fbsize >= 0 ? fbsize : 8;
	_dmalloc_phashsize = phashsize;
	_dmalloc_fhashsize = fhashsize;
	DMALLOC_FHASH = (DMALLOC_FPTR **)malloc(fhashsize*sizeof(DMALLOC_FPTR *));
	for(i=0;i<fhashsize;i++)
		DMALLOC_FHASH[i] = NULL;
	DMALLOC_PHASH = (DMALLOC_PPTR **)malloc(phashsize*sizeof(DMALLOC_PPTR *));
	for(i=0;i<phashsize;i++)
		DMALLOC_PHASH[i] = NULL;
	if (freecache > 0)
	{
		_dmalloc_cache_frees = freecache;
		DMALLOC_FFHASH = (DMALLOC_FPTR **)malloc(fhashsize*sizeof(DMALLOC_FPTR *));
		for(i=0;i<fhashsize;i++)
			DMALLOC_FFHASH[i] = NULL;
		DMALLOC_PFHASH = (DMALLOC_PPTR **)malloc(phashsize*sizeof(DMALLOC_PPTR *));
		for(i=0;i<phashsize;i++)
			DMALLOC_PFHASH[i] = NULL;
	}
	_hashtab = (unsigned int *)malloc(256 * sizeof(unsigned int));
	for (i = 0; i < 256; i++)
		_hashtab[i] = ((char)i) * 109;
	return 0;
}

void dmalloc_file_get_start(void)
{
	_dmalloc_file_get_hnum = 0;
	_dmalloc_file_get_ptr = NULL;
}

DMALLOC_FPTR *dmalloc_file_get_next(void)
{
	int				i;

	if (_dmalloc_file_get_ptr)
	{
		for(;;)
		{
			if ((_dmalloc_file_get_ptr =
							_dmalloc_file_get_ptr->next) != NULL)
				return _dmalloc_file_get_ptr;
			else if (!_dmalloc_file_get_ptr)
				break;
		}
	}
	for(i=_dmalloc_file_get_hnum;i<_dmalloc_fhashsize;i++)
	{
		if (DMALLOC_FHASH[i])
		{
			for(;;)
			{
				_dmalloc_file_get_ptr = DMALLOC_FHASH[i];
				_dmalloc_file_get_hnum = i+1;
				return _dmalloc_file_get_ptr;
			}
		}
	}
	dmalloc_file_get_start();
	return NULL;
}

int dmalloc_check(void)
{
	int				i;
	DMALLOC_PPTR	*pptr;

	for(i=0;i<_dmalloc_phashsize;i++)
		for(pptr=DMALLOC_PHASH[i];pptr;pptr=pptr->p_next)
			_dmalloc_check(pptr);
	return 0;
}

int dmalloc_hash_info(void)
{
	int num;
	int max = 0;
	int i;
	DMALLOC_PPTR	*pptr;

	if (!_dmalloc_inited)
		return -EFAULT;

	for(i=0;i<_dmalloc_phashsize;i++)
	{
		num = 0;
		for(pptr=DMALLOC_PHASH[i];pptr;pptr=pptr->p_next)
			num++;
		if (num > max)
			max = num;
		if (num)
			fprintf(stderr, "hash bucket %i: %i entries\n", i, num);
	}
	fprintf(stderr, "max entries in bucket: %i\n", max);
	return 0;
}

int dmalloc_pptr_exists(char *ptr, u_int ssize)
{
	DMALLOC_PPTR	*pptr;

	pptr = _dmalloc_pptr_find(ptr - _dmalloc_fbsize);
	if (!pptr)
		return 0;
	if (ssize != -1)
		assert(ssize == pptr->size);
	return 1;
}

int dmalloc_pptr_exists_freed(char *ptr)
{
	DMALLOC_PPTR	*pptr;

	pptr = _dmalloc_pptr_find_freed(ptr - _dmalloc_fbsize);
	if (!pptr)
		return 0;
	return 1;
}
