/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
#include "../confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");


#ifdef TS5
#define TS_CURRENT  5   /* current TS protocol version */
#else
#define TS_CURRENT  3   /* current TS protocol version */
#endif

#ifdef TS5_ONLY
#define TS_MIN      5   /* minimum supported TS protocol version */
#else
#define TS_MIN      3   /* minimum supported TS protocol version */
#endif


static char *months[] =
{
	"January",
	"February",
	"March",
	"April",
    "May",
	"June",
	"July",
	"August",
    "September",
	"October",
	"November",
	"December"
};

static char *weekdays[] =
{
	"Sunday",
	"Monday",
	"Tuesday",
	"Wednesday",
    "Thursday",
	"Friday",
	"Saturday"
};

typedef struct _ircd_mode		IRCD_MODE;
typedef struct _ircd_capab		IRCD_CAPAB;

struct _ircd_mode
{
	char				mode;
	char				plusminus;
	IRCD_CLIENT			*cptr;
	char				*banstring;

	IRCD_MODE			*next;
};

struct _ircd_capab
{
	int					value;
	char				*name;
	int					flags;
};

static int nickname_check(char *nick);
static int _register_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr);
static int m_message(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int not);
static int _do_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *username, char *hostname, char *servername, char *ircname);
static int _lusers_send(IRCD_CLIENT *pptr, char *pattern, int sendcounter);
static int _motd_send(IRCD_CLIENT *pptr);
static void _sendnick(IRCD_CLIENT *cptr, IRCD_CLIENT *acptr);
static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf);
static void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr);
static void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask);
static void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask);
static int _channel_canjoin(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr, char *key);
static IRCD_CHANNELLIST *_channel_ismember(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
static void _channel_add_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, int flag);
static void _channel_del_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, IRCD_CHANNELLIST *chlptr);
static int _channel_cansend(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr);
static void _channel_invite_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *target, char *prefix, char *args);
static IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode);
static void _ircdmode_free(IRCD_MODE **bucket);
static int _channel_change_opflag(IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr, IRCD_MODE *opmodeptr, int mine);
static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr);
static int _channel_ban_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
static int _channel_ban_del(IRCD_CHANNEL *chptr, char *str);
static void do_who(IRCD_CLIENT *pptr, IRCD_CLIENT *acptr, IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr);
static char *datestr(int thetime, char *buf);
static char *_cluster(char *hostname, int *ip_or_host);
static void _send_var(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
static void _send_var_to_opers(IRCD_CLIENT *pptr, IRCD_VAR *vptr);
static void _send_servers_recurs(IRCD_CLIENT *lptr, IRCD_CLIENT *cptr);
static void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers);
static void send_capabilities(IRCD_CLIENT *cptr, int send_zip);
static char *_reason_fix(char *reason);
static void _channelname_clean(unsigned char *chname);

static int _user_modes[] =
{
	CLIENT_UMODES_OPER,			'o',
	CLIENT_UMODES_LOCOP,		'O',
	CLIENT_UMODES_FMODE,		'f',
	CLIENT_UMODES_IMODE,		'i',
	CLIENT_UMODES_KMODE,		'k',
	CLIENT_UMODES_SMODE,		's',
	CLIENT_UMODES_WMODE,		'w',
	CLIENT_UMODES_ZMODE,		'z',
	CLIENT_UMODES_YMODE,		'y',
	CLIENT_UMODES_CMODE,		'c',
	CLIENT_UMODES_NMODE,		'n',
	0, 0
};

static int _channel_modes[] =
{
	CHANNEL_MODES_PRIVATE,		'p',
	CHANNEL_MODES_SECRET,		's',
	CHANNEL_MODES_MODERATED,	'm',
	CHANNEL_MODES_NOPRIVMSGS,	'n',
	CHANNEL_MODES_TOPICLIMIT,	't',
	CHANNEL_MODES_INVITEONLY,	'i',
	0, 0
};

IRCD_USERFLAG _user_flags[] =
{
	{ CLIENT_NFLAGS_OWALLOPS,		"OWALLOPS",		1	},
	{ CLIENT_NFLAGS_SWALLOPS,		"SWALLOPS",		0	},
	{ CLIENT_NFLAGS_STATSNOTICES,	"STATSNOTICES",	1	},
	{ CLIENT_NFLAGS_OKILLS,			"OKILLS",		0	},
	{ CLIENT_NFLAGS_SKILLS,			"SKILLS",		0	},
	{ CLIENT_NFLAGS_SNOTICES,		"SNOTICES",		0	},
	{ CLIENT_NFLAGS_CLICONNECTS,	"CLICONNECTS",	1	},
	{ CLIENT_NFLAGS_CLIDISCONNECTS,	"CLIDISCONNECTS",1	},
	{ CLIENT_NFLAGS_THROTTLES,		"THROTTLES",	1	},
	{ CLIENT_NFLAGS_NICKCHANGES,	"NICKCHANGES",	1	},
	{ CLIENT_NFLAGS_IPMISMATCHES,	"IPMISMATCHES", 1	},
	{ CLIENT_NFLAGS_LWALLOPS,		"LWALLOPS",		1	},
	{ CLIENT_NFLAGS_CONNECTS,		"CONNECTS",		1	},
	{ CLIENT_NFLAGS_SQUITS,			"SQUITS",		1	},
	{ 0, NULL, 0 }
};

static IRCD_CAPAB _ircd_capabs[] =
{
#if 0
	{ IRCD_CAPAB_VALUE_IPADDR,		"IPADDR",	0 },
	{ IRCD_CAPAB_VALUE_ZIP,			"ZIP",		0 },
	{ IRCD_CAPAB_VALUE_EMODE,		"EX",		0 },
#endif
	{ IRCD_CAPAB_VALUE_NOQUITSTORM,	"QS",		0 },
	{ 0, NULL, 0 }
};

int hostname_check(char *hostname)
{
	char *tptr = hostname;
	char *dotptr = NULL;

	for(;*tptr;tptr++)
	{
		if ((*tptr <= ' ') || (*tptr > 'z') || (*tptr == ':') ||
				(*tptr == '*') || (*tptr == '%'))
			break;
		if (*tptr == '.')
			dotptr = tptr;
	}
	if (*tptr || !*hostname || !dotptr || (strlen(dotptr) > 4))
		return 0;

	return 1;
}

static int nickname_check(char *nick)
{
	char	*ch;

	if ((*nick == '-') || isdigit(*nick))
		return 0;
	for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++)
		if (!isvalid(*ch) || isspace(*ch))
			break;
	*ch = '\0';
	return (ch - nick);
}

static int _register_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr)
{
	int					err;
	char				exitbuf[512];

	pptr->c_flags |= CLIENT_FLAGS_REG;
	add_client_to_oslist(Clients, pptr);
	if (pptr->c_umodes & CLIENT_UMODES_IMODE)
		Globalinfo->g_invis++;
	if (pptr->c_umodes & CLIENT_UMODES_OPER) /* can happen if remote client */
		Globalinfo->g_opers++;
	if (++Globalinfo->g_clients > Globalinfo->g_maxclients)
	{
		Globalinfo->g_maxclients = Globalinfo->g_clients;
		Globalinfo->g_maxclientstime = Now;
	}
	if (IsLocal(pptr))
	{
		int					ret;
		CONFHOSTIPENTRY		*entryptr;
		char				*ptr1 = NULL;
		char				*ptr2;
		char				*uchrs;
		int					num;

		*exitbuf = '\0';

		if (++Globalinfo->m_clients > Globalinfo->m_maxclients)
		{
			Globalinfo->m_maxclients = Globalinfo->m_clients;
			Globalinfo->m_maxclientstime = Now;
		}
		if (Globalinfo->m_clients + Globalinfo->m_servers > Globalinfo->m_maxconn)
		{
			Globalinfo->m_maxconn = Globalinfo->m_clients + Globalinfo->m_servers;
			Globalinfo->m_maxconntime = Now;
			if (!(Globalinfo->m_maxconn % 10))
				send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
					"New highest connections: %d (%d clients)",
					Globalinfo->m_maxconn, Globalinfo->m_clients);
		}
		if (pptr->c_umodes & CLIENT_UMODES_IMODE)
			Globalinfo->m_invis++;
		if (pptr->c_umodes & CLIENT_UMODES_OPER) /* can't happen */
			Globalinfo->m_opers++;

		/* check username */
		uchrs = (char *)vars_value_get_bynum(IRCD_VAR_USERNAMECHARS);

		if (*pptr->username)
		{
			for(ptr1=pptr->username;*ptr1;ptr1++)
			{
				if (isalnum(*ptr1) || (*ptr1 == '~'))
					continue;
				if (!uchrs)
					break;
				for(ptr2=uchrs;*ptr2;ptr2++)
					if (*ptr1 == *ptr2)
						break;
				if (!*ptr2)
					break;
			}
		}
		if (!*pptr->username || *ptr1)
		{
			client_del(meptr, pptr, "Invalid username");
			return -1;
		}

		if (!hostname_check(pptr->hostname))
		{
			client_del(meptr, pptr, "Invalid hostname");
			return -1;
		}

		ret = ircd_checkallow(pptr, &err);
		if (ret < 0)
		{

			*exitbuf = '\0';
			switch(ret)
			{
				case	-1: /* no auth */
					sprintf(exitbuf, "No authorization: %i", err);
					break;
				case	-2: /* no auth */
					sprintf(exitbuf, "Class error: %i", err);
					break;
				case	-3: /* class full */
					strcpy(exitbuf, "Class is full");
					break;
				case	-4: /* host limit exceeded */
					strcpy(exitbuf, "Class host limit exceeded");
					break;
				case	-5: /* host limit exceeded */
					strcpy(exitbuf, "Class userhost limit exceeded");
					break;
				case	-6: /* host limit exceeded */
					strcpy(exitbuf, "Class vhost limit exceeded");
					break;
				case	-7: /* host limit exceeded */
					strcpy(exitbuf, "Class uservhost limit exceeded");
					break;
				default:
					sprintf(exitbuf, "Unknown error: %i", err);
					break;
			}
			client_del(meptr, pptr, exitbuf);
			return -1;
		}

		if (pptr->c_res_req)
		{
			if (pptr->c_res_req->rl_forward)
				MyFree(pptr->c_res_req->rl_forward);
			if (pptr->c_res_req->rl_reverse)
				MyFree(pptr->c_res_req->rl_reverse);
			MyFree(pptr->c_res_req);
			pptr->c_res_req = NULL;
		}

		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSIRCNAMEDENY))
		{
			STRINGDENY	*sd;

			if ((sd = stringdeny_match(&(_ircd_conf.ircname_denies),
							pptr->ircname, Now)) != NULL)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					sd->reason);
				client_del(meptr, pptr, "Invalid ircname");
				return -1;
			}
		}

		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSLIMITS))
		{
			num = (int)vars_value_get_bynum(IRCD_VAR_MAXCLIENTS);
			if ((num > 0) && (Globalinfo->m_clients >= num))
			{
				client_del(meptr, pptr,
					"Server is full -- reached client maximum");
				return -1;
			}
		}
		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSDENY))
		{
			ret = access_match_ipaddr(&(_ircd_conf.deny),
                pptr->username, pptr->c_ipstring,
                pptr->c_rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
                &entryptr);
			if (ret < 0)
				ret = access_match_hostname(&(_ircd_conf.deny),
	                pptr->username, pptr->hostname, ACCESS_FLAGS_BESTMATCH,
					&entryptr);
			if (ret >= 0) /* banned */
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					entryptr->data ? entryptr->data : "No reason specified");
				sprintf(exitbuf, "Banned: %s",
					entryptr->data ? entryptr->data : "No reason specified");
				client_del(meptr, pptr, exitbuf);
				return -1;
			}
		}
		if (!(pptr->c_flags & CLIENT_FLAGS_GOTAUTH) &&
					(pptr->c_class->flags & CLASS_FLAGS_IDENTONLY))
		{
			client_del(meptr, pptr,
				"Sorry, you must have ident to connect");
			return -1;
		}
		if (pptr->c_class->flags & CLASS_FLAGS_NOTILDE)
		{
			if (!(pptr->c_flags & CLIENT_FLAGS_GOTAUTH))
			{
				char *tptr = pptr->username;
				while((*tptr = *(tptr+1)))
					tptr++;
			}
		}
	}

	if (iphash_handle_client(pptr)) /* client_del is called in here */
		return -1;

	pptr->c_servptr = hash_find_client_incl_hm(pptr->servername);
	if (!pptr->c_servptr || !IsServer(pptr->c_servptr))
	{
		send_to_opers("Ignoring new client %s from %s -- unknown server: %s",
				pptr->name, lptr->name, pptr->servername);
		sprintf(exitbuf, "Unknown server: %s", pptr->servername);
		client_del(meptr, pptr, exitbuf);
		return -1;
	}
	if (IsLocal(pptr))
		if (vars_value_get_bynum(IRCD_VAR_CLIENTCONNECT))
			send_to_flag(CLIENT_NFLAGS_CLICONNECTS, NULL,
				"Client connecting: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, pptr->c_ipstring);
	pptr->c_type = CLIENT_TYPE_CLIENT;
	add_client_to_ulist(&(pptr->c_servptr->c_users), pptr);

	if (IsLocal(pptr))
	{
		char	cnamebuf[CLIENTNAMESIZE];

		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WELCOME, pptr->name),
				pptr->name);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_YOURHOST, pptr->name),
				ircd_clientname(meptr, 0, cnamebuf), VERSION);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_CREATED, pptr->name),
				CREATION);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MYINFO, pptr->name),
				meptr->name, VERSION);
		Globalinfo->m_totclients++;
		_lusers_send(pptr, NULL, 0);
		if (vars_value_get_bynum(IRCD_VAR_SHOW_CLIENTCOUNTER))
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are client #%d to connect since %s",
				pptr->name, Globalinfo->m_totclients,
				support_ctime(Globalinfo->m_totclientstime));
		_motd_send(pptr);
	}

	if (IsLocal(pptr)) /* NICK sent to servers in m_nick, otherwise */
	{
		char ubuf[64];

		_umode_build(pptr, ubuf, SEND_UMODES);
		send_to_servers(lptr, "NICK %s %d %i %s %s %s %s :%s",
				pptr->name, pptr->c_hopcount+1, pptr->c_creation,
				ubuf, pptr->username, pptr->hostname, pptr->servername,
				pptr->ircname);
		_umode_build(pptr, ubuf, -1);
		if (ubuf[1])
			send_to_client(pptr, ":%s MODE %s :%s",
					pptr->name, pptr->name, ubuf);
	}

	

	return 3;
}

static char *_reason_fix(char *reason)
{
	static char	reasonbuf[401];	/* reason can only be 200 long, so.. */
	char *dptr = reasonbuf;
	char *sptr = reason;

	for(;*sptr;)
	{
		if (*sptr == '\"')
			*dptr++ = '\\';
		*dptr++ = *sptr++;
	}
	*dptr = '\0';
	return reasonbuf;
}

#ifdef CONCENTRIC
static void services_klines(char *prefix, char *args);
static void services_tklines(char *prefix, char *args);

static void services_klines(char *prefix, char *args)
{
	u_int	ipaddr;
	u_int	ipmask;
	char	*user;
	char	*host;
	char	*reason;
	int		ret;
	int		ip_or_host;
	CONFHOSTIPENTRY		*entryptr;

	if (strcasecmp(prefix, "services.us"))
		return;
	if (!args || !*args || !(host = strchr(args, '@')))
			return;
	user = args;
	*host++ = '\0';
	reason = strchr(host, ' ');
	if (reason)
		*reason++ = '\0';
	if (!reason || !*reason)
		return;
	if (strlen(reason) > 200)
		reason[200] = '\0';
	if (!match(user, "........") && !match(host, "........."))
		return;
	if (strstr(host, "concentric.net") || strstr(host, "cris.com"))
		return;

	if (strcmp(host, "*") && (1 == cidr_ip2netmask(host, &ipaddr, &ipmask)))
		ip_or_host = 0;
	else
		ip_or_host = 1;

	if (!ip_or_host)
	{
		ircd_uint32_t	net, mask;

		ret = cidr_ip2netmask(host, &net, &mask);
		if (ret == 1)
		{
			ret = access_match_ipaddr(&(_ircd_conf.deny),
				user, host, net, 0, &entryptr);
			if (ret >= 0)
				return;
		}
		ret = access_add(&(_ircd_conf.deny), user, NULL, host,
			NULL, reason, -1, 0, 0, NULL, "from services.us");
	}
	else
	{
		ret = access_match_hostname(&(_ircd_conf.deny),
			user, host, 0, &entryptr);
		if (ret >= 0)
			return;
		ret = access_add(&(_ircd_conf.deny), user, host, NULL,
			NULL, reason, -1, 0, 0, NULL, "from services.us");
	}
	if (!ret)
	{
		char *filename;
		int fd;
		char	buffer[1024];

		Rehashed = 1;
		send_to_opers("Added K-Line from %s for [%s@%s]  %s",
                prefix, user, host, reason);
		filename = vars_value_get_bynum(IRCD_VAR_NEWKLINES_FILE);
		if (!filename)
			filename = Conffile;

		fd = open(filename, O_RDWR|O_APPEND|O_CREAT, 0660);
		if (fd == -1)
			return;
		sprintf(buffer, "deny { \n\tusername \"%s\";\n", user);
		write(fd, buffer, strlen(buffer));
		if (ip_or_host)
			sprintf(buffer, "\thostname \"%s\";\n", host);
		else
			sprintf(buffer, "\tipaddr \"%s\";\n", host);
		write(fd, buffer, strlen(buffer));
		sprintf(buffer, "\treason \"%s\";\n", _reason_fix(reason));
		write(fd, buffer, strlen(buffer));
		write(fd, "};\n\n", 4);
		close(fd);
    }
}

static void services_tklines(char *prefix, char *args)
{
	u_int	ipaddr;
	u_int	ipmask;
	char	*thetime;
	int		expiretime;
	char	*user;
	char	*host;
	char	*reason;
	int		ret;
	int		ip_or_host;
	CONFHOSTIPENTRY		*entryptr;

	if (strcasecmp(prefix, "services.us"))
		return;

	if (!args || !*args)
		return;

	thetime = args;

	user = strchr(thetime, ' ');
	if (!user)
		return;
	*user++ = '\0';

	if (!*user || !(host = strchr(user, '@')))
			return;
	*host++ = '\0';
	reason = strchr(host, ' ');
	if (!reason)
		return;
	*reason++ = '\0';
	if (!*reason)
		return;
	if (strlen(reason) > 200)
		reason[200] = '\0';
	if (!match(user, "........") && !match(host, "........."))
		return;
	if (strstr(host, "concentric.net") || strstr(host, "cris.com"))
		return;

	expiretime = time(NULL) + atoi(thetime);

	if (strcmp(host, "*") && (1 == cidr_ip2netmask(host, &ipaddr, &ipmask)))
		ip_or_host = 0;
	else
		ip_or_host = 1;

	if (!ip_or_host)
	{
		ircd_uint32_t	net, mask;

		ret = cidr_ip2netmask(host, &net, &mask);
		if (ret == 1)
		{
			ret = access_match_ipaddr(&(_ircd_conf.deny),
				user, host, net, 0, &entryptr);
			if (ret >= 0)
				return;
		}
		ret = access_add(&(_ircd_conf.deny), user, NULL, host,
			NULL, reason, expiretime, 0, 0, NULL, "from services.us");
	}
	else
	{
		ret = access_match_hostname(&(_ircd_conf.deny),
			user, host, 0, &entryptr);
		if (ret >= 0)
			return;
		ret = access_add(&(_ircd_conf.deny), user, host, NULL,
			NULL, reason, expiretime, 0, 0, NULL, "from services.us");
	}
	if (!ret)
	{
		char *filename;
		int fd;
		char	buffer[1024];

		Rehashed = 1;
		send_to_opers("Added K-Line from %s for [%s@%s] expiring at %s: %s",
                prefix, user, host, support_ctime(expiretime), reason);
		filename = vars_value_get_bynum(IRCD_VAR_NEWKLINES_FILE);
		if (!filename)
			filename = Conffile;

		fd = open(filename, O_RDWR|O_APPEND|O_CREAT, 0660);
		if (fd == -1)
			return;
		sprintf(buffer, "deny { \n\tusername \"%s\";\n", user);
		write(fd, buffer, strlen(buffer));
		if (ip_or_host)
			sprintf(buffer, "\thostname \"%s\";\n", host);
		else
			sprintf(buffer, "\tipaddr \"%s\";\n", host);
		write(fd, buffer, strlen(buffer));
		sprintf(buffer, "\texpire %i;\n", expiretime);
		write(fd, buffer, strlen(buffer));
		sprintf(buffer, "\treason \"%s\";\n", _reason_fix(reason));
		write(fd, buffer, strlen(buffer));
		write(fd, "};\n\n", 4);
		close(fd);
    }
}

#endif

static int m_message(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int not)
{
#define MAX_MESSAGE_NICKS	50
	char	*tostr = ircd_nextarg(&args, 0);
	char	*message = ircd_nextarg(&args, 0);
	char	*cmd;
	char	*sendtos[100];
	char	**st = NULL;
	char	**std;
	char	*nick;
	char	*server;
	int		resetidle = 0;
	IRCD_CLIENT	*acptr;
	int		numtried = 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (IsServer(pptr) && !not)
		return 3;
	cmd = not ? "NOTICE" : "PRIVMSG";
	if (!tostr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NORECIPIENT, pptr->name), cmd);
		return 3;
	}
	if (!message || !*message)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTEXTTOSEND, pptr->name));
		return 3;
	}
	if (IsLocal(pptr))
	{
		int		num = 0;

		st = sendtos;
		for(;(nick = ircd_next_subarg_comma(&tostr)) != NULL;)
		{
			if (!*nick)
				continue;
			if (++num > MAX_MESSAGE_NICKS)
			{
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Too many recipients", pptr->name);
				return 3;
			}
			*st = nick;
			st++;
		}
		*st = NULL;
		st = sendtos;
	}
	for(;;)
	{
		if (st)
		{
			nick = *st;
			st++;
			if (nick)/* check for already sent */
			{
				for(std=sendtos;*std != nick;std++)
					if (!strcmp(*std, nick))
						break;
				if (*std != nick)
				{
					DEBUG(DEBUG_ALL, ("got dup %s for message\n", nick));
					continue;
				}
			}
		}
		else
			nick = ircd_next_subarg_comma(&tostr);
		if (!nick)
			break;
		numtried++;
		if (IsChannel(nick))
		{
			IRCD_CHANNEL	*chptr = hash_find_channel(nick);

			if (chptr)
			{
				if (_channel_cansend(pptr, chptr) == 0)
				{
					if (IsLocal(pptr) && ((chptr->numusers > 1) ||
							(chptr->members && 
							(chptr->members->cptr != pptr))))
					{
						if (!not)
							resetidle = 1;
					}
					send_to_channel_butone(lptr,
							CHANNEL_SEND_SERVERS|CHANNEL_SEND_CLIENTS,
							chptr, pptr, "%s %s :%s",
							cmd, nick, message);
				}
				else if (!not)
				{
					send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CANNOTSENDTOCHAN, pptr->name), nick);
				}
				continue;
			}
		}
		else
		{
			IRCD_CLIENT		*nptr = hash_find_client(nick);

			if (nptr && IsClient(nptr))
			{
				if (IsClient(pptr))
					send_prefix_to_client(pptr, nptr, "%s %s :%s",
						cmd, nick, message);
				else
					send_to_client(nptr, ":%s %s %s :%s",
						prefix, cmd, nick, message);
				if (nptr != pptr)
					resetidle = 1;
				continue;
			}
		}
		if (IsAnOper(pptr) && ((*nick == '$') || (*nick == '#')))
		{
			char *s;

			if ((s = strrchr(nick, '.')) == NULL)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTOPLEVEL, pptr->name), nick);
				continue;
			}
			while(*++s)
			{
				if (*s == '*' || *s == '?')
					break;
			}
			if (*s)
			{
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_WILDTOPLEVEL,
						pptr->name), nick);
				continue;
			}
			/* sendto_match_butone */
			for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
			{
				if ((IsServer(acptr) && (acptr != lptr)) ||
					(IsClient(acptr) && 
					(((*nick == '$') && !match(nick+1, acptr->servername)) ||
					((*nick == '#') && !match(nick+1, acptr->hostname)))))
				{
					if (IsClient(pptr))
						send_prefix_to_client(pptr, acptr, "%s %s :%s",
							cmd, nick, message); 
					else
						send_to_client(acptr, ":%s %s %s :%s",
							prefix, cmd, nick, message); 
				}
			}
			continue;
		}
		if (((server = strchr(nick, '@')) != NULL) &&
				((acptr = hash_find_client_incl_hm(server+1)) != NULL))
		{
			if (IsServer(acptr))
			{
				if (!IsMe(acptr))
				{
					send_to_client(acptr, ":%s %s %s :%s",
						prefix, cmd, nick, message);
					continue;
				}
				*server = '\0';
				if (!strcasecmp(nick, "OPERS") && IsAnOper(pptr))
				{
#if 0
					send_to_opers("To opers: From %s: %s",
						pptr->name, message);
#else
					for(acptr=*Opers;acptr;acptr=acptr->c_os_next)
    				{
						if (!IsLocal(acptr))
							continue;
						send_prefix_to_client(pptr, acptr, "%s %s :** TO LOCAL OPERS ** %s",
							cmd, acptr->name, message); 
					}
#endif
					continue;
				}
#ifdef CONCENTRIC
				if (!strcasecmp(nick, "KLINES"))
				{
					services_klines(prefix, message);
					continue;
				}
				if (!strcasecmp(nick, "TKLINES"))
				{
					services_tklines(prefix, message);
					continue;
				}
#endif
		
			}
		}
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK, pptr->name),
					nick);
	}
	if (resetidle && !not && IsLocal(pptr))
		pptr->c_idletime = Now;
	numtried /= 5;
	return numtried > 3 ? numtried : 3;
}

static int _do_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *username, char *hostname, char *servername, char *ircname)
{
	if (IsLocal(pptr))
	{
#ifndef NO_DEFAULT_INVISIBLE
		if (!(pptr->c_umodes & CLIENT_UMODES_IMODE))
		{
			pptr->c_umodes |= CLIENT_UMODES_IMODE;
		}
#endif
		if (!(pptr->c_flags & CLIENT_FLAGS_GOTAUTH))
		{
			*pptr->username = '~';
			support_strncpy(pptr->username+1, username, sizeof(pptr->username)-1);
		}
		support_strncpy(pptr->servername, meptr->name, sizeof(pptr->servername));
	}
	else
	{
		support_strncpy(pptr->username, username, sizeof(pptr->username));
		support_strncpy(pptr->hostname, hostname, sizeof(pptr->hostname));
		support_strncpy(pptr->servername, servername, sizeof(pptr->servername));
	}
	support_strncpy(pptr->ircname, ircname, sizeof(pptr->ircname));
	pptr->c_flags |= CLIENT_FLAGS_GOTUSER;
	if (pptr->name[0])
		return _register_user(lptr, pptr);
	return 3;
}

static int _lusers_send(IRCD_CLIENT *pptr, char *pattern, int sendcounter)
{
	int		g_opers = 0;
	int		g_clients = 0;
	int		g_invis = 0;
	int		g_servers = 0;
	int		m_servers = 0;
	int		m_clients = 0;
	int		m_unknown = 0;
	int		channels = 0;

	channels = Globalinfo->channels;
	if (!pattern || ((*pattern == '*') && (*(pattern+1) == '\0')))
	{
		g_opers = Globalinfo->g_opers;
		g_clients = Globalinfo->g_clients;
		g_invis = Globalinfo->g_invis;
		g_servers = Globalinfo->g_servers;
		m_servers = Globalinfo->m_servers;
		m_clients = Globalinfo->m_clients;
		m_unknown = Globalinfo->m_unknown;
	}
	else
	{
		int 		i = 0;
		IRCD_CLIENT	**list = Opers;
		IRCD_CLIENT	*acptr;

		for(;i<3;i++)
		{
			if (i==1)
				list = Clients;
			else if (i==2)
				list = Servers;
			for(acptr=*list;acptr;acptr=acptr->c_os_next)
			{
				if (((list != Servers) &&
							match(pattern, acptr->servername)) ||
					((list == Servers) && match(pattern, acptr->name)))
					continue;
				if ((list == Opers) || IsOper(acptr))
					g_opers++;
				if (list == Servers)
				{
					if (IsLocal(acptr))
						m_servers++;
					g_servers++;
				}
				else
				{
					if (IsLocal(acptr))
						m_clients++;
					g_clients++;
					if (IsInvisible(acptr))
						g_invis++;
				}
			}
		}
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERCLIENT, pptr->name),
			g_clients - g_invis, g_invis, g_servers);
	if (g_opers)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSEROP, pptr->name),
			g_opers);
	if (m_unknown)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERUNKNOWN, pptr->name),
			m_unknown);
	if (channels)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERCHANNELS, pptr->name),
			channels);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERME, pptr->name),
		m_clients, m_servers);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_STATSCONN, pptr->name),
		Globalinfo->m_maxconn, Globalinfo->m_maxclients);
	if (Globalinfo->g_serversbursting)
		send_prefix_to_client(meptr, pptr, "NOTICE %s :%i server%s still bursting.",
			pptr->name, Globalinfo->g_serversbursting,
			Globalinfo->g_serversbursting == 1 ? "" : "s");
	if (sendcounter && vars_value_get_bynum(IRCD_VAR_SHOW_CLIENTCOUNTER))
		send_prefix_to_client(meptr, pptr, "NOTICE %s :%i client%s ha%s connected since %s",
			pptr->name, Globalinfo->m_totclients,
			Globalinfo->m_totclients == 1 ? "" : "s",
			Globalinfo->m_totclients == 1 ? "s" : "ve",
			support_ctime(Globalinfo->m_totclientstime));

	return 0;
}

static int _motd_send(IRCD_CLIENT *pptr)
{
	struct tm	tmbuf;
	char		**lines;

	if (Motd == NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOMOTD, pptr->name));
		return 0;
	}
	tmbuf = motd_time(Motd);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MOTDSTART, pptr->name),
				meptr->name);
	send_prefix_to_client(meptr, pptr, "%d %s :- %d/%d/%d %d:%02d",
		RPL_MOTD, pptr->name, tmbuf.tm_mday, tmbuf.tm_mon+1, 1900+tmbuf.tm_year,
		tmbuf.tm_hour, tmbuf.tm_min);
	for(lines=motd_lines(Motd);*lines;lines++)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MOTD, pptr->name),
			*lines);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFMOTD, pptr->name));
	return 0;
}

int _hunt_server(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *format, int server, char *prefix, char *argptr)
{
	char		theargs[512];
	char		*args = theargs;
	char		*argss[10];
	int			numargs;
	IRCD_CLIENT	*acptr;

	if ((--server < 0) || (server > 7))
		return HUNTED_ISME;

	support_strncpy(theargs, argptr, sizeof(theargs));

	for(numargs=0;numargs < 8;numargs++)
		if (!(argss[numargs] = ircd_nextarg(&args, 0)))
		{
			while(numargs < 8)
				argss[numargs++] = NULL;
			break;
		}

	if (!argss[server] || !match(meptr->name, argss[server]) ||
			!match(argss[server], meptr->name))
		return HUNTED_ISME;
	if ((acptr = hash_find_client_incl_hm(argss[server])) != NULL)
		if (acptr->c_from == pptr->c_from && !IsLocal(acptr))
			acptr = NULL;
	if (!acptr)
	{
		for(acptr=*Servers;acptr;acptr=acptr->c_os_next)
		{
			if ((acptr->c_from == pptr->c_from) && IsLocal(acptr))
				continue;
			if (!match(acptr->name, argss[server]) ||
					!match(argss[server], acptr->name))
				break;
		}
	}
	if (acptr)
	{
		if (IsMe(acptr) || MyClient(acptr))
			return HUNTED_ISME;
		if (match(acptr->name, argss[server]))
			argss[server] = acptr->name;
		send_to_client(acptr, format, prefix, argss[0],
				argss[1], argss[2], argss[3], argss[4],
				argss[5], argss[6], argss[7]);
		return HUNTED_PASS;
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name),
			argss[server]);
	return HUNTED_NOSUCH;
}

static void _channel_del_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, IRCD_CHANNELLIST *chlptr)
{
	IRCD_CLIENTLIST		*clptr;

	if (chlptr == NULL)
		if (!(chlptr = IsMember(chptr, pptr)))
			return;
	pptr->c_numchannels--;
	clptr=chlptr->clptr;
	chlptr->clptr = NULL;
	del_clientlist_from_list(&(chptr->members), clptr);
	if (!--chptr->numusers)
	{
		/* remove channel */
		_channel_ban_del(chptr, NULL);
		assert(!chptr->nbans);
		hash_del_channel(chptr);
		if (chptr->topic)
			MyFree(chptr->topic);
		if (chptr->topic_who)
			MyFree(chptr->topic_who);
		if (chptr->key)
			MyFree(chptr->key);
		MyFree(chptr);
		Globalinfo->channels--;
	}
	MyFree(clptr);
}

static void _channel_add_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, int flag)
{
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CLIENTLIST		*clptr;

	chlptr = (IRCD_CHANNELLIST *)MyMalloc(sizeof(IRCD_CHANNELLIST));
	clptr = (IRCD_CLIENTLIST *)MyMalloc(sizeof(IRCD_CLIENTLIST));
	chlptr->chptr = chptr;
	clptr->cptr = pptr;
	clptr->modes = flag;
	add_channellist_to_list(&(pptr->c_channels), chlptr);
	add_clientlist_to_list(&(chptr->members), clptr);
	chlptr->clptr = clptr;
	clptr->chlptr = chlptr;
	pptr->c_numchannels++;
	chptr->numusers++;
}

static void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask)
{
	int		*s;
	int		flag;
	char	plusminus = '\0';
	char	*modeptr = buf;

	if ((oldmodes & sendmask) != (pptr->c_umodes & sendmask))
	{
		for(s=_user_modes;(flag=*s);s+=2)
		{
			if (MyClient(pptr) && !(flag & sendmask))
				continue;
			if ((flag & oldmodes) && !(pptr->c_umodes & flag))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
			else if (!(flag & oldmodes) && (pptr->c_umodes & flag))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
		}
	}
	*modeptr = '\0';
}

static void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask)
{
	int		*s;
	int		flag;

	*buf++ = '+';
	for(s=_user_modes;(flag=*s);s+=2)
	{
		if (!(flag & mask))
			continue;
		if (pptr->c_umodes & flag)
			*buf++ = *(s+1);
	}
	*buf = '\0';
}

static void _sendnick(IRCD_CLIENT *cptr, IRCD_CLIENT *acptr)
{
	static	char	ubuf[64];

	if (IsReg(acptr))
	{
		_umode_build(acptr, ubuf, SEND_UMODES);
		send_to_client(cptr, "NICK %s %d %ld %s %s %s %s :%s",
			acptr->name, acptr->c_hopcount + 1, acptr->c_creation, ubuf,
			acptr->username, acptr->hostname,
			acptr->servername, acptr->ircname);
    }
}

static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf)
{
	int		ismem = 0;

	if (!IsServer(cptr) && (chptr->limit || chptr->key))
		ismem = (IsMember(chptr, cptr) != NULL);
	*mbuf++ = '+';
	if (chptr->modes & CHANNEL_MODES_SECRET)
		*mbuf++ = 's';
	/* else */
	if (chptr->modes & CHANNEL_MODES_PRIVATE)
		*mbuf++ = 'p';
	if (chptr->modes & CHANNEL_MODES_MODERATED)
		*mbuf++ = 'm';
	if (chptr->modes & CHANNEL_MODES_TOPICLIMIT)
		*mbuf++ = 't';
	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
		*mbuf++ = 'i';
	if (chptr->modes & CHANNEL_MODES_NOPRIVMSGS)
		*mbuf++ = 'n';
	if (chptr->limit)
	{
		*mbuf++ = 'l';
		if (IsServer(cptr) || ismem)
			pbuf += sprintf(pbuf, "%d ", chptr->limit);
	}
	if (chptr->key)
	{
		*mbuf++ = 'k';
		if (IsServer(cptr) || ismem)
			pbuf += strlen(strcpy(pbuf, chptr->key));
	}
	*mbuf = '\0';
	*pbuf = '\0';
}

static void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr)
{
	char				modebuf[64];
	char				parabuf[512];
	char				buf[512];
	char				*t;
	IRCD_CLIENTLIST		*clptr;
	IRCD_CLIENTLIST		*anop = NULL;
	IRCD_CLIENTLIST		*skip = NULL;
	IRCD_BAN			*bans;
	int					n = 0;
	int					paralen;
	int					namlen;
	char				*cp;
	int					count;
	int					send;

	if (*chptr->name != '#') /* don't send '&' channels */
		return;
	_build_channel_modes(chptr, cptr, modebuf, parabuf);

	t = buf + sprintf(buf, ":%s SJOIN %u %s %s %s%s:",
			meptr->name, chptr->creation, chptr->name, modebuf,
			parabuf, *parabuf ? " " : "");

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			anop = clptr;
			break;
		}
	}
	clptr = NULL;
	for(;;)
	{
		if (anop)
		{
			clptr = skip = anop;
			anop = NULL;
		}
		else
		{
			if (clptr == NULL || clptr == skip)
				clptr = chptr->members;
			else
				clptr = clptr->next;
			if (clptr && clptr == skip)
				clptr = clptr->next;
			if (clptr == NULL)
				break;
		}
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*t++ = '@';
		if (clptr->modes & CHANNEL_MODES_VOICE)
			*t++ = '+';
		strcpy(t, clptr->cptr->name);
		t += strlen(t);
		*t++ = ' ';
		n++;
		if (t - buf > 512 - HOSTLEN - 3)
		{
			*t++ = '\0';
			if (t[-1] == ' ')
				t[-1] = '\0';
			send_to_client(cptr, "%s", buf);
			t = buf + sprintf(buf, ":%s SJOIN %u %s 0 :",
					meptr->name, chptr->creation, chptr->name);
			n = 0;
		}
	}
	if (n)
	{
		*t++ = '\0';
		if (t[-1] == ' ')
			t[-1] = '\0';
		send_to_client(cptr, "%s", buf);
	}
	*parabuf = '\0';
	*modebuf = '+';
	cp = modebuf + 1;
	count = 0;
	paralen = 0;
	send = 0;
		
	/* send bans */

	for(bans=chptr->bans;bans;bans=bans->next)
	{
		if (paralen + (namlen = strlen(bans->banstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, bans->banstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'b';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, bans->banstring);
				paralen = namlen;
				*cp++ = 'b';
			}
			count = 1;
		}
	}
	if (*parabuf)
	{
		*cp = '\0';
		send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
				chptr->name, modebuf, parabuf);
	}
}

static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args)
{
	char			*modes = ircd_nextarg(&args, 0);
	char			plusminus = '+';
	IRCD_CLIENT		*acptr;
	char			ubuf[64];
	int				oldmodes;
	char			*ptr;
	int				*i;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!nick || !*nick) /* should always be false, checked in m_mode() */
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "UMODE");
		return 3;
	}

	acptr = hash_find_client(nick);
	if (!acptr || IsServer(acptr))
	{
		if (IsLocal(pptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHCHANNEL, pptr->name),
				nick);
		}
		return 3;
	}

	if (acptr != pptr)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_USERSDONTMATCH, pptr->name));
		return 3;
	}

	if (!modes || !*modes)
	{
		_umode_build(acptr, ubuf, -1);
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_UMODEIS, pptr->name), ubuf);
		return 3;
	}

	oldmodes = pptr->c_umodes;

	for(ptr=modes;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			default:
				for(i=_user_modes;*i;i+=2)
					if (*ptr == *(i+1))
						break;
				if (!*i)
				{
					send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_UMODEUNKNOWNFLAG, pptr->name), *ptr);
					continue;
				}
				if (plusminus == '+')
					pptr->c_umodes |= *i;
				else
					pptr->c_umodes &= ~(*i);
				break;
		}
	}

	if (pptr->c_umodes == oldmodes)
		return 3;

	if (!(oldmodes & CLIENT_UMODES_OPER) &&
					(pptr->c_umodes & CLIENT_UMODES_OPER))
	{
		if (IsServer(lptr))
			Globalinfo->g_opers++;
		else
			pptr->c_umodes &= ~CLIENT_UMODES_OPER;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_OPER) &&
				(oldmodes & CLIENT_UMODES_OPER))
	{
		Globalinfo->g_opers--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);
	}

/* can't get this one from remote server...*/
	if (!(oldmodes & CLIENT_UMODES_LOCOP) &&
					(pptr->c_umodes & CLIENT_UMODES_LOCOP))
	{
		if (!IsServer(lptr))
			pptr->c_umodes &= ~CLIENT_UMODES_LOCOP;
	}

	if (!(oldmodes & CLIENT_UMODES_IMODE) &&
			(pptr->c_umodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
			Globalinfo->m_invis++;
		Globalinfo->g_invis++;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_IMODE) &&
			(oldmodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
			Globalinfo->m_invis--;
		Globalinfo->g_invis--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);
	}

	if (IsLocal(pptr))
	{
		int showflags = 0;

		if (!(oldmodes & CLIENT_UMODES_ZMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_OWALLOPS;
				showflags = 1;
			}
			else
			{
				pptr->c_umodes &= ~CLIENT_UMODES_ZMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_ZMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OWALLOPS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_SMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_SNOTICES;
			showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_SMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SNOTICES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_FMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_SKILLS;
			showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_FMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_KMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_OKILLS;
			showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_KMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_WMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_SWALLOPS;
			showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_WMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SWALLOPS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_CMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS;
				showflags = 1;
			}
			else
				pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
		}
		else if ((oldmodes & CLIENT_UMODES_CMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_NMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_NICKCHANGES;
				showflags = 1;
			}
			else
				pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
		}
		else if ((oldmodes & CLIENT_UMODES_NMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_YMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_STATSNOTICES;
				showflags = 1;
			}
			else
				pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
		}
		else if ((oldmodes & CLIENT_UMODES_YMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
			showflags = 1;
		}

		if (showflags)
			m_flags(lptr, pptr, prefix, NULL);

		if (oldmodes == pptr->c_umodes)
			return 3;
	}

	_umode_build_change(pptr, ubuf, oldmodes, SEND_UMODES);
	if (*ubuf)
		send_to_servers(lptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
	_umode_build_change(pptr, ubuf, oldmodes, -1);
#if 0
	if (MyClient(pptr))
		send_prefix_to_client(pptr, pptr, "MODE %s %s", pptr->name, buf);
/* umodes only send client's nick, not full n!u@h */
#else
	if (MyClient(pptr))
		send_to_client(pptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
#endif

	return 3;
}

static IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode)
{
	IRCD_MODE	*nircdmode;


	nircdmode = (IRCD_MODE *)MyMalloc(sizeof(IRCD_MODE));
	if (!*lastircdmode)
		*ircdmodes = nircdmode;
	else
		(*lastircdmode)->next = nircdmode;
	return (*lastircdmode = nircdmode);
}

static void _ircdmode_free(IRCD_MODE **bucket)
{
	IRCD_MODE	*nextp;

	for(;*bucket;*bucket = nextp)
	{
		nextp = (*bucket)->next;
		MyFree(*bucket);
	}
}

static int _channel_change_opflag(IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr, IRCD_MODE *opmodeptr, int mine)
{
	int	m =
		(opmodeptr->mode == 'o' ? CHANNEL_MODES_CHANOP : CHANNEL_MODES_VOICE);
	int	irm = (int)vars_value_get_bynum(IRCD_VAR_IGNORE_REDUNDANT_MODES);

	if (!clptr)
		return 1;
	if (opmodeptr->plusminus == '+')
	{
		if (irm && mine && (clptr->modes & m))
			return 0;
		clptr->modes |= m;
		if (m == CHANNEL_MODES_CHANOP)
			clptr->modes &= ~CHANNEL_MODES_DEOPPED;
	}
	else
	{
		if (irm && mine && !(clptr->modes & m))
			return 0;
		clptr->modes &= ~m;
	}
	return 1;
}

static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr)
{
	static char	star = '*';
	char	*ptr = buf;
	char	*sptr;
	int		n;

	if (!nick || !*nick)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=nick,n=NICKLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '!';
	if (!user || !*user)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=user,n=USERLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '@';
	if (hostptr)
		*hostptr = ptr;
	if (!host || !*host)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=host,n=HOSTLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr = '\0';
	return buf;
}

static int _channel_ban_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask)
{
	IRCD_BAN		*ban;

	if (MyClient(pptr))
	{
		collapse(str);
		if (chptr->nbans >= MAXBANS)
			return 0;
		hostptr = strchr(str, '@');
		assert(hostptr != NULL);
		hostptr++;
	}
	for(ban=chptr->bans;ban;ban=ban->next)
	{
		if (MyClient(pptr) &&
				(!match(ban->banstring, str) || !match(str, ban->banstring)))
			return 0;
		else if (!mycmp(ban->banstring, str))
			return 0;
	}

	ban = (IRCD_BAN *)MyMalloc(sizeof(IRCD_BAN));
	ban->banstring = MyStrdup(str);
	if (ban->net != -1)
	{
		ban->net = net;
		ban->mask = mask;
		ban->firstpart = (char *)MyMalloc(hostptr - str + 1);
		memcpy(ban->firstpart, str, hostptr - str);
		ban->firstpart[hostptr - str] = '\0';
	}
	ban->bantime = Now;
	if (IsServer(pptr))
		ban->who = MyStrdup(pptr->name);
	else
	{
		ban->who = MyMalloc(strlen(pptr->name)+strlen(pptr->username)+
					strlen(pptr->hostname)+3);
		sprintf(ban->who, "%s!%s@%s", pptr->name, pptr->username,
				pptr->hostname);
	}
	add_ban_to_list(&(chptr->bans), ban);
	chptr->nbans++;

	return 1;
}

static int _channel_ban_del(IRCD_CHANNEL *chptr, char *str)
{
	IRCD_BAN	*ban;
	IRCD_BAN	*nextban;

	for(ban=chptr->bans;ban;ban=nextban)
	{
		nextban = ban->next;
		if (!str || !mycmp(str, ban->banstring))
		{
			MyFree(ban->banstring);
			if (ban->firstpart)
				MyFree(ban->firstpart);
			MyFree(ban->who);
			del_ban_from_list(&(chptr->bans), ban);
			MyFree(ban);
			chptr->nbans--;
		}
	}
	return 1;
}

static void _channel_invite_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr)
{
	IRCD_INVITE	*invite;

	for(invite=cptr->c_invites_list;invite;invite=invite->next)
		if (!strcasecmp(chptr->name, invite->channel))
			return; /* already exists */
	invite = (IRCD_INVITE *)MyMalloc(sizeof(IRCD_INVITE));
	invite->channel = (char *)MyStrdup(chptr->name);
	add_invite_to_list(&(cptr->c_invites_list), invite);
	if (cptr->c_invites_num == 10) /* remove oldest invite */
	{
		IRCD_INVITE	*i;
		for(i=cptr->c_invites_list;i->next;i=i->next)
			continue;
		del_invite_from_list(&(cptr->c_invites_list), i);
	}
	else
		cptr->c_invites_num++;
	return;
}

static int _channel_canjoin(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr, char *key)
{
	IRCD_INVITE	*i = NULL;

	assert(IsLocal(pptr));

	if (chptr->limit && (chptr->numusers >= chptr->limit))
		return ERR_CHANNELISFULL;
	if (chptr->key && (!key || mycmp(chptr->key, key)))
		return ERR_BADCHANNELKEY;
	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
	{
		for(i=pptr->c_invites_list;i;i=i->next)
			if (!strcasecmp(i->channel, chptr->name))
				break;
		if (!i)
			return ERR_INVITEONLYCHAN;
	}
	if (chptr->bans)
	{
		IRCD_BAN	*ban = chptr->bans;
		char		banstr1[NICKLEN+USERLEN+HOSTLEN+6];
		char		banstr2[NICKLEN+USERLEN+HOSTLEN+6];
		char		banstr3[NICKLEN+USERLEN+HOSTLEN+6];
		char		*hostptr;

		_make_nick_user_host(pptr->name, pptr->username, pptr->hostname,
						banstr1, &hostptr);
		strcpy(banstr2, banstr1);
		strcpy(banstr3, banstr1);

		*hostptr = '\0';
		strcpy(banstr2+(hostptr-banstr1), pptr->c_ipstring);

		/* banstr1 just has nick + user@ */
		/* banstr2 has n!u@IP */
		/* banstr3 has n!u@HOST */

		for(;ban;ban=ban->next)
		{
			if (ban->firstpart)
			{
				if (!cidr_match(ban->net, ban->mask, pptr->c_rem_ipaddr)
						&& !match(ban->firstpart, banstr1))
					return ERR_BANNEDFROMCHAN;
			}
			if (!match(ban->banstring, banstr2) ||
						!match(ban->banstring, banstr3))
				return ERR_BANNEDFROMCHAN;
		}
	}
	if (i)
	{
		del_invite_from_list(&(pptr->c_invites_list), i);
		pptr->c_invites_num--;
		MyFree(i->channel);
		MyFree(i);
	}
	return 0;
}

static IRCD_CHANNELLIST *_channel_ismember(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr)
{
	IRCD_CHANNELLIST	*chlptr;

	for(chlptr=pptr->c_channels;chlptr;chlptr=chlptr->next)
		if (chlptr->chptr == chptr)
			break;
	return chlptr;
}

static int _channel_cansend(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr)
{
	IRCD_CHANNELLIST	*chlptr = NULL;

	if (chptr->modes & CHANNEL_MODES_MODERATED)
	{
		if ((chlptr = IsMember(chptr, pptr)) == NULL)
			return -1;
		if (!(chlptr->clptr->modes &
						(CHANNEL_MODES_CHANOP|CHANNEL_MODES_VOICE)))
			return -1;
	}
	if (chptr->modes & CHANNEL_MODES_NOPRIVMSGS)
	{
		if (!chlptr && (IsMember(chptr, pptr) == NULL))
			return -2;
	}
	return 0;
}

static void do_who(IRCD_CLIENT *pptr, IRCD_CLIENT *acptr, IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr)
{
	char	status[5];
	int		i = 0;

	if (acptr->c_away)
		status[i++] = 'G';
	else
		status[i++] = 'H';
	if (IsAnOper(acptr))
		status[i++] = '*';
	if ((chptr != NULL) && (clptr == NULL)) /* should never happen */
	{
		IRCD_CHANNELLIST	*chlptr = IsMember(chptr, acptr);
		clptr = chlptr ? chlptr->clptr : NULL;
	}
	if (clptr)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			status[i++] = '@';
		else if (clptr->modes & CHANNEL_MODES_VOICE)
			status[i++] = '+';
	}
	status[i] = '\0';
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOREPLY, pptr->name),
			chptr ? chptr->name : "*",
			acptr->username,
			acptr->hostname,
			acptr->servername,
			acptr->name,
			status,
			acptr->c_hopcount,
			acptr->ircname);
}

static char *datestr(int thetime, char *buf)
{
	struct tm	gmbuf;
	struct tm	*tmptr;
	time_t		temptime;
	char		plusminus;
	int			minswest;

	if (!thetime)
		thetime = (int)time(NULL);
	temptime = (time_t)thetime;
	tmptr = gmtime(&temptime);
	if (tmptr)
		memcpy(&gmbuf, tmptr, sizeof(struct tm));
	else
	{
		strcpy(buf, "Error getting date");
		return buf;
	}
	tmptr = localtime(&temptime);

	if (tmptr->tm_yday == gmbuf.tm_yday)
		minswest = (gmbuf.tm_hour - tmptr->tm_hour) * 60 +
			(gmbuf.tm_min - tmptr->tm_min);
	else if (tmptr->tm_yday > gmbuf.tm_yday)
		minswest = (gmbuf.tm_hour - (tmptr->tm_hour + 24)) * 60;
	else
		minswest = ((gmbuf.tm_hour + 24) - tmptr->tm_hour) * 60;

	if (minswest > 0)
		plusminus = '-';
	else
	{
		plusminus = '+';
		minswest = -minswest;
	}

	sprintf(buf, "%s %s %d %04d -- %02d:%02d %c%02d:%02d",
		weekdays[tmptr->tm_wday], months[tmptr->tm_mon], tmptr->tm_mday,
		tmptr->tm_year + 1900, tmptr->tm_hour, tmptr->tm_min,
		plusminus, minswest/60, minswest%60);
	return buf;
}

static char *_cluster(char *hostname, int *ip_or_host)
{
	static char		result[HOSTLEN+10];
	int				len;

	if (!hostname || !*hostname)
		return NULL;

	len = strlen(hostname);

	if (len > HOSTLEN)
		return NULL;

	if (isdigit(*(hostname+len-1)))
	{
		char	*tmp = result + len - 1;

		if (ip_or_host)
			*ip_or_host = 0;

		strcpy(result, hostname);
		for(;tmp >= result;tmp--)
		{
			if (*tmp == '.')
				break;
		}
		if (tmp < result)
			return result;
		*(tmp+1) = '*';
		*(tmp+2) = '\0';
	}
	else
	{
		char	*tmp = len <= 3 ? hostname : hostname + len - 3;
		int     num = 2;
		char	*host = hostname + len - 1;

		if (ip_or_host)
			*ip_or_host = 1;

		if (strcasecmp(tmp, "com") && strcasecmp(tmp, "edu") &&
				(support_stristr(hostname, "com") ||
					support_stristr(hostname, "edu")))
			num = 3;

		for(;host>=hostname;host--)
		{
			if ((*host == '.') && (!--num))
				break;
		}
		if (host < hostname)
			host = hostname;
		*result = '*';
		strcpy(result+1, *host == '.' ? host + 1 : host);
	}
	return result;
}

static void _send_servers_recurs(IRCD_CLIENT *lptr, IRCD_CLIENT *cptr)
{
	char			cnamebuf1[CLIENTNAMESIZE];
	IRCD_CLIENT		*acptr;

	for(acptr=cptr->c_servers;acptr;acptr=acptr->c_u_next)
	{
		if (acptr->c_from == lptr) /* will be true for acptr == lptr too */
			continue;
		if (!match(conf_myname_for_link(meptr->name, lptr->c_hostmask,
					cnamebuf1), acptr->name))
			continue;
		send_prefix_to_client(acptr->c_servptr, lptr, "SERVER %s %d :%s",
			acptr->name, acptr->c_hopcount+1, acptr->ircname);
		_send_servers_recurs(lptr, acptr);
	}
}

static void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers)
{
	IRCD_CLIENT	*u;

	for(u=acptr->c_servers;u;u=u->c_u_next)
	{
		_count_servers_and_users(u, numservers, numusers);
	}
	for(u=acptr->c_users;u;u=u->c_u_next)
		(*numusers)++;
	(*numservers)++;
}

static void send_capabilities(IRCD_CLIENT *lptr, int send_zip)
{
	IRCD_CAPAB	*cptr;
	char		buffer[256];
	char		*ptr;
	char		*sptr;
	char		*dptr;

	ptr = dptr = buffer;
	for(cptr=_ircd_capabs;cptr->name;cptr++)
	{
		if ((cptr->value == IRCD_CAPAB_VALUE_ZIP) && !send_zip)
			continue;
		if (ptr != buffer)
		{
			*ptr++ = ' ';
			dptr++;
		}
		sptr = cptr->name;
		while(((dptr - buffer) < sizeof(buffer)) && (*dptr = *sptr++))
			dptr++;
		if ((dptr - buffer) == sizeof(buffer))
		{
			*(ptr-1) = '\0'; /* -1 to kill the space that was added */
			send_to_client(lptr, "CAPAB :%s", buffer);
			dptr = ptr = buffer;
			sptr = cptr->name;
			while((*dptr = *sptr++))
				dptr++;
		}
		ptr = dptr;
	}
	if (ptr != buffer)
	{
		*ptr = '\0';
		send_to_client(lptr, "CAPAB :%s", buffer);
	}
}

static void _channelname_clean(unsigned char *chname)
{
	for(;*chname;chname++)
	{
		if (*chname == '\007' || (*chname == ' ') ||
				(*chname == ',') ||
#if 0
				(*chname == ':') ||
#endif
				((*chname > 127) && (*chname <= 160)))
		{
			*chname = '\0';
			return;
		}
	}
}

int m_privmsg(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_message(lptr, pptr, prefix, args, 0);
}

int m_notice(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_message(lptr, pptr, prefix, args, 1);
}

int m_who(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char				*mask = ircd_nextarg(&args, 0);
	char				*mask2 = ircd_nextarg(&args, 0);
	IRCD_CHANNEL		*chptr;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CLIENT			*acptr;
	int					operonly = 0;
	int					maxmatches = 500;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!mask || !*mask)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHO, pptr->name),
			mask ? mask : "*");
		return 3;
	}

	if (mask2 && (*mask2 == 'o'))
		operonly = 1;

	collapse(mask);
	if (!operonly && (*mask == '*') && (!*(mask+1)))
	{
		if (!pptr->c_channels)
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_ENDOFWHO, pptr->name), mask);
			return 3;
		}
		chptr = pptr->c_channels->chptr;
		chlptr = pptr->c_channels;
	}
	else
	{
		chptr = hash_find_channel(mask);
		chlptr = NULL;
	}
	if (chptr)
	{
		IRCD_CLIENTLIST		*clptr;

		if (!chlptr)
			chlptr = IsMember(chptr, pptr);
		if (chlptr || !(chptr->modes & CHANNEL_MODES_SECRET))
		{
			for(clptr=chptr->members;clptr;clptr=clptr->next)
			{
				if (operonly && !IsAnOper(clptr->cptr))
					continue;
				if (!chlptr && (clptr->cptr->c_umodes & CLIENT_UMODES_IMODE))
					continue;
				do_who(pptr, clptr->cptr, chptr, clptr);
			}
		}
	}
	else if (IsChannel(mask))
		; /* do nothing */
	else if ((acptr = hash_find_client(mask)) != NULL)
	{
		if (IsPerson(acptr) && (!operonly || IsAnOper(acptr)))
		{
			int					isinvis = acptr->c_umodes & CLIENT_UMODES_IMODE;
			IRCD_CHANNELLIST	*chlptr2;

			for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
			{
				chlptr2 = _channel_ismember(chlptr->chptr, pptr);
				if (isinvis && !chlptr2)
					continue;
				if (chlptr2 || (!isinvis && !(chlptr->chptr->modes &
								(CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE))))
					break;
			}
			do_who(pptr, acptr, chlptr ? chlptr->chptr : NULL,
						chlptr ? chlptr->clptr : NULL);
		}
	}
	else
	{
		int					i = 0;
		IRCD_CLIENT			**list;
		int					showperson;
		int					isinvis;
		IRCD_CHANNELLIST	*chlptr2;

		for(i=0;i<2;i++)
		{
			if (!i)
				list = Opers;
			else
			{
				if (operonly)
					break;
				list = Clients;
			}
			for(acptr=*list;acptr;acptr=acptr->c_os_next)
			{
				showperson = 0;
				isinvis = acptr->c_umodes & CLIENT_UMODES_IMODE;
				if (!(chlptr=acptr->c_channels) && !isinvis)
					showperson = 1;
				else if (chlptr)
				for(chptr=chlptr->chptr;chlptr;chlptr=chlptr->next)
				{
					chlptr2 = IsMember(chptr, pptr);
					if (!chlptr2 && isinvis)
						continue;
					if (chlptr2 || (!isinvis && !(chptr->modes &
								(CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE))))
					{
						showperson = 1;
						break;
					}
					if (!isinvis && (chptr->modes & CHANNEL_MODES_PRIVATE) &&
						!(chptr->modes & CHANNEL_MODES_SECRET))
						showperson = 1;
				}

				if (showperson &&
					(!match(mask, acptr->name) ||
					!match(mask, acptr->username) ||
					!match(mask, acptr->hostname) ||
					!match(mask, acptr->servername) ||
					!match(mask, acptr->ircname)))
				{
					do_who(pptr, acptr, chlptr ? chlptr->chptr : NULL,
								chlptr ? chlptr->clptr : NULL);
					if (!--maxmatches)
						break;
				}
			}
			if (!maxmatches)
				break;
		}
	}

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHO, pptr->name),
			mask);

	return 3;
}

int m_whois(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	argssave[512];
	char	*whostr;
	char	*nick;
	char	*server;
	char	*sendtos[21];
	char	**st;
	char	**std;
	char	*name;
	int		num = 0;
	IRCD_CLIENT	*acptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (args)
		support_strncpy(argssave, args, sizeof(argssave));

	whostr = ircd_nextarg(&args, 0);
	server = ircd_nextarg(&args, 0);

	if (server && *server)
	{
		if (_hunt_server(lptr, pptr, ":%s WHOIS %s :%s", 1, prefix, argssave) != HUNTED_ISME)
			return 3;
		whostr = server;
	}

	if (!whostr || !*whostr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NONICKNAMEGIVEN, pptr->name));
		return 3;
	}

	st = sendtos;
	for(;(nick = ircd_next_subarg_comma(&whostr)) != NULL;)
	{
		if (!*nick)
			continue;
		*st = nick;
		st++;
		if (++num == 20)
			break;
	}
	*st = NULL;
	st = sendtos;

	for(;;)
	{
		nick = *st;
		st++;
		if (!nick)
			break;
		/* already done it ? */
		for(std=sendtos;*std != nick;std++)
			if (!strcmp(*std, nick))
				break;
		if (*std != nick)
			continue;
		acptr = hash_find_client(nick);
		if (!acptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK,
				pptr->name), nick);
			continue;
		}
		else if (!IsClient(acptr))
			continue;

		name = *acptr->name ? acptr->name : "?";

		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISUSER, pptr->name),
				name, acptr->username, acptr->hostname, acptr->ircname);
		/* channels */
		{
			IRCD_CHANNELLIST	*chlptr;
			IRCD_CLIENTLIST		*clptr;
			char				buf[512];
			char				*sptr;
			char				*dptr = buf;

			for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
			{
				if ((!(chlptr->chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET))) || (IsMember(chlptr->chptr, pptr) != NULL))
				{
					clptr = chlptr->clptr;
					if (clptr->modes & CHANNEL_MODES_CHANOP)
						*dptr++ = '@';
					else if (clptr->modes & CHANNEL_MODES_VOICE)
						*dptr++ = '+';
					for(sptr=chlptr->chptr->name;*sptr;)
						*dptr++ = *sptr++;
					if (dptr - buf > 200)
					{
						*dptr = '\0';
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(RPL_WHOISCHANNELS, pptr->name),
							name, buf);
						dptr = buf;
					}
					else
						*dptr++ = ' ';
				}
			}
			if (dptr != buf)
			{
				*dptr = '\0';
				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_WHOISCHANNELS, pptr->name), name, buf);
			}
		}

		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISSERVER,
				pptr->name), name, acptr->servername,
				acptr->c_servptr->ircname);
		if (acptr->c_away)
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_AWAY, pptr->name),
				name, acptr->c_away);
		if (IsAnOper(acptr))
		{
			char *ptr = vars_value_get_bynum(IRCD_VAR_ISOPERSTRING);

			if (!ptr)
				ptr = "is an IRC Operator";
			send_prefix_to_client(meptr, pptr, "%d %s %s :%s",
					RPL_WHOISOPERATOR, pptr->name, acptr->name, ptr);
#if 0
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISOPERATOR, pptr->name),
				name);
#endif
		}
		if (IsLocal(acptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISIDLE, pptr->name),
				name, Now - acptr->c_idletime, acptr->c_firsttime);
	}

	name = strchr(argssave, ' ');
	if (name)
		*name = '\0';

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHOIS, pptr->name),
			argssave);

	return 3;
}

int m_whowas(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return whowas_command(lptr, pptr, prefix, args);
}

/*
** Can only come from local unknown connections now..
*/
int m_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*username = ircd_nextarg(&args, 0);
	char	*hostname = ircd_nextarg(&args, 0);
	char	*servername = ircd_nextarg(&args, 0);
	char	*ircname = ircd_nextarg(&args, 0);
	char	cnamebuf[CLIENTNAMESIZE];

	if (IsServer(lptr))
	{
		send_to_opers("Got USER from %s", ircd_clientname(lptr, 0, cnamebuf));
		return 3;
	}
	if (!ircname)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, *pptr->name ? pptr->name : "*"), "USER");
		return 3;
	}
	if (pptr->c_flags & CLIENT_FLAGS_GOTUSER)
		return 3;
	hostname = pptr->hostname;
	servername = meptr->name;
	return _do_user(lptr, pptr, username, hostname, servername, ircname);
}

/*
** m_nick
**  parv[0] = sender prefix
**  parv[1] = nickname
**  parv[2] = optional hopcount when new user; TS when nick change
**  parv[3] = optional TS
**  parv[4] = optional umode
**  parv[5] = optional username
**  parv[6] = optional hostname
**  parv[7] = optional server
**  parv[8] = optional ircname
*/
int m_nick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	newnick[NICKLEN+1];
	char	*nick = ircd_nextarg(&args, 0);
	char	*hopcount = ircd_nextarg(&args, 0);
	char	*ts = ircd_nextarg(&args, 0);
	char	*umode = ircd_nextarg(&args, 0);
	char	*username = ircd_nextarg(&args, 0);
	char	*hostname = ircd_nextarg(&args, 0);
	char	*servername = ircd_nextarg(&args, 0);
	char	*ircname = ircd_nextarg(&args, 0);
	int		newts = 0;
	int		hc = 0;
	IRCD_CLIENT	*acptr;
	char	cnamebuf[CLIENTNAMESIZE];

	if (!nick)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NONICKNAMEGIVEN,
				*pptr->name ? pptr->name : "*"));
		return 3;
	}

	if (IsServer(lptr))
	{
		if (IsServer(pptr)) /* remote new nick */
		{
			if (!ircname)
			{
				send_to_opers("No ircname for new nick \"%s\" from remote \"%s\"",
					nick, pptr->name);
				return 3;
			}
			send_to_servers(lptr, ":%s NICK %s %d %s %s %s %s %s :%s",
				prefix, nick, atoi(hopcount)+1, ts,
				umode, username, hostname, servername, ircname);
			hc = atol(hopcount);
			newts = atol(ts);
		}
		else /* remote nick change */
		{
			send_to_servers(lptr, ":%s NICK %s %s",
				prefix, nick, hopcount?hopcount:"0");
			if (hopcount)
				newts = atol(hopcount);
			username = pptr->username;
			hostname = pptr->hostname;
		}
	}
	support_strncpy(newnick, nick, sizeof(newnick));
	{
		register char *ptr = newnick;

		while(*ptr && (*ptr != '~'))
			ptr++;
		*ptr = '\0';
	}
	if (!nickname_check(newnick) ||
			(IsServer(lptr) && strcmp(nick, newnick)))
	{
		if (IsServer(lptr))
		{
			/* send kill for new nickname everywhere */
			send_to_opers("Bad Nick: %s From: %s %s",
				nick, pptr->name, ircd_clientname(lptr, 0, cnamebuf));
			return 3;
		}
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ERRONEUSNICKNAME,
					pptr->name), nick);
		return 3;
	}

	/* check to see if nick can be used */

	if (!IsServer(lptr) &&
		(!pptr->c_class ||
			!(pptr->c_class->flags & CLASS_FLAGS_BYPASSNICKNAMEDENY)) &&
		stringdeny_match(&_ircd_conf.nickname_denies, newnick, Now))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_ERRONEUSNICKNAME,
				pptr->name), newnick);
		return 0;
	}

	/* check to see if nick exists */

	acptr = hash_find_client(newnick);
	if (acptr)
	{
		if (acptr == pptr) /* same nick */
		{
			if (!strcmp(acptr->name, newnick))
				return 3; /* trying to change to same nick */
			/* case change */
		}
		else if (IsLocal(acptr) && !IsReg(acptr))
		{
			/* get rid of client and fall through */
			client_del(meptr, acptr, "Overridden");
			acptr = NULL;
		}
		else if (!IsServer(lptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NICKNAMEINUSE,
					*pptr->name ? pptr->name : "*"), newnick);
			return 3;
		}
		else /* from remote */
		{
			int				sameuser;
			IRCD_CLIENT		*toexit;
	
			if (!newts || (newts == acptr->c_creation))
			{
				/* kill both */
				if (IsLocal(acptr))
				{
					send_prefix_to_client(meptr, acptr,
						IRCD_REPLY(ERR_NICKCOLLISION, acptr->name),
						acptr->name);
					acptr->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
					acptr->c_flags |= CLIENT_FLAGS_KILLED;
				}
				client_del(meptr, acptr, "Nick collision");
				return (acptr == lptr) ? -1 : 3;
			}
			sameuser = (!mycmp(acptr->username, username) &&
							!mycmp(acptr->hostname, hostname));
			if ((sameuser && (newts < acptr->c_creation)) ||
				(!sameuser && (newts > acptr->c_creation)))
			{
				if (IsServer(pptr))
				{
					/* ignore the new NICK */
					return 3;
				}
				toexit = pptr;
			}
			else
			{
				toexit = acptr;
			}
			if (sameuser)
			{
				send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
					"Nick collision on %s(%s <- %s)(older killed)",
					acptr->name, acptr->c_from->name,
					ircd_clientname(lptr, 0, cnamebuf));
			}
			else
			{
				send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
					"Nick collision on %s(%s <- %s)(newer killed)",
					acptr->name, acptr->c_from->name,
					ircd_clientname(lptr, 0, cnamebuf));
			}

			/* kill off 'toexit' */
			if (IsLocal(toexit))
			{
				send_prefix_to_client(meptr, toexit,
					IRCD_REPLY(ERR_NICKCOLLISION, toexit->name), toexit->name);
				toexit->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
				toexit->c_flags |= CLIENT_FLAGS_KILLED;
			}
			client_del(meptr, toexit, "Nick collision");
			if (toexit == pptr)
				return 3; /* pptr is not local */
			acptr= NULL;
			/* fall through */
		}
	}

	if (IsServer(pptr)) /* new nick from remote server */
	{
		IRCD_CLIENT	*nptr;
		int		*s;
		int		flag;
		char	*m;

#if 1
		static int	nsize = offsetof(IRCD_CLIENT, c_auth_ptr);
#else
		static int	nsize = sizeof(IRCD_CLIENT);
#endif

		nptr = (IRCD_CLIENT *)MyMalloc(nsize);
		nptr->c_from = lptr;
		nptr->c_rem_ipaddr = -1;
		nptr->c_rem_port = 0;
		nptr->c_hopcount = hc;
		strcpy(nptr->name, newnick);
		support_strncpy(nptr->username, username, sizeof(nptr->username));
		support_strncpy(nptr->hostname, hostname, sizeof(nptr->hostname));
		nptr->c_creation = newts ? newts : (newts = Now);
		nptr->c_lasttime = nptr->c_firsttime = Now;
		m = umode + 1;
		for(;*m;m++)
		{
			for(s=_user_modes;(flag = *s);s+=2)
			{
				if (*m == *(s+1))
				{
					nptr->c_umodes |= flag;
					break;
				}
			}
		}
		hash_add_client(nptr);
		return _do_user(lptr, nptr, username, hostname, servername, ircname);
	}
	else if (pptr->name[0])
	{
		/* nick change */

		pptr->c_creation = newts ? newts : (newts = Now);
		if (IsReg(pptr)) /* send NICK +chan stuff to other servers */
		{
			whowas_add_history(pptr, 1);
			send_to_common_channels(pptr, "NICK :%s", newnick);
			if (IsLocal(pptr))
			{
				if (vars_value_get_bynum(IRCD_VAR_SHOW_NICKCHANGES))
					send_to_flag(CLIENT_NFLAGS_NICKCHANGES, NULL,
						"Nick change: From %s to %s [%s@%s]",
							prefix, newnick, pptr->username,
								pptr->hostname);
				send_to_servers(NULL, ":%s NICK %s :%i",
					pptr->name, newnick, pptr->c_creation);
			}
		}
		hash_del_client(pptr);
		strcpy(pptr->name, newnick);
		hash_add_client(pptr);
		return 3;
	}
	strcpy(pptr->name, newnick);
	pptr->c_creation = Now;
	hash_add_client(pptr);
	if (pptr->c_flags & CLIENT_FLAGS_GOTUSER)
		return _register_user(lptr, pptr);
	return 3;
}


/*
** :prefix SERVER servername serverinfo/hopcount serverinfo
*/
int m_server(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*servername = ircd_nextarg(&args, 0);
	char			*hopcount = ircd_nextarg(&args, 0);
	char			*serverinfo = ircd_nextarg(&args, 0);
	char			cnamebuf1[CLIENTNAMESIZE];
	int				hc = 0;
	IRCD_CLIENT		*acptr;
	SERVLINK		*sl;
	IRCD_CLASSENTRY	*classptr;
	int				ret;

	if (!servername)
	{
		client_del(meptr, lptr, "No servername");
		return -1;
	}
	if (!serverinfo || !hopcount || !(hc = atoi(hopcount)))
	{
		serverinfo = hopcount;
		hopcount = NULL;
	}
	if (!(lptr->c_flags & CLIENT_FLAGS_SCONNECT)
			&& !IsServer(lptr) &&
					(*lptr->name || (lptr->c_flags & CLIENT_FLAGS_GOTUSER)))
	{
		if (*lptr->name)
			send_prefix_to_client(meptr, lptr, IRCD_REPLY(ERR_ALREADYREGISTRED,
					lptr->name));
		return 3;
	}
	else
	{
		char	clean_host[(HOSTLEN*2)+1];
		char	*sptr = servername;
		char	*dptr = clean_host;
		int		gotdot = 0;
		int		bogus = 0;
		int		len = 0;

		while(*sptr)
		{
			if (*sptr < ' ')
			{
				bogus++;
				*dptr++ = '^';
				*dptr++ = (*(sptr++) + 'A');
			}
			else if (*sptr > '~')
			{
				bogus++;
				*dptr++ = '.';
				sptr++;
			}
			else if ((*dptr++ = *sptr++) == '.')
				gotdot++;
			if (++len == HOSTLEN)
			{
				servername[len] = '\0';
				break;
			}
		}
		*dptr = '\0';
		if (!gotdot || bogus)
		{
			send_to_opers("Bogus server name (%s) from %s",
				clean_host, ircd_clientname(lptr, 1, cnamebuf1));
			if (IsUnknown(lptr))
			{
				client_del(meptr, lptr, "Invalid servername");
				return -1;
			}
			return 3;
		}
	}

	if ((acptr = hash_find_client_incl_hm(servername)) != NULL)
	{
		char			cnamebuf2[CLIENTNAMESIZE];
		IRCD_CLIENT		*bcptr;

		bcptr = (lptr->c_firsttime > acptr->c_from->c_firsttime) ?
					lptr : acptr->c_from;
		send_to_client(bcptr, "ERROR :Server %s already exists", servername);
		if (bcptr == lptr)
		{
			send_to_opers("Link %s cancelled, server %s already exists",
					ircd_clientname(bcptr, 1, cnamebuf1), servername);
			client_del(meptr, bcptr, "Server Exists");
			return -1;
		}
		send_to_opers("Link %s cancelled, server %s reintroduced by %s",
				ircd_clientname(bcptr, 1, cnamebuf1), servername,
				ircd_clientname(lptr, 1, cnamebuf2));
		client_del(meptr, bcptr, "Server Exists");
	}

	if (IsServer(lptr))
	{
		IRCD_CLIENT		*bcptr;

		if (!serverinfo)
			serverinfo = "*Unknown*";

		/* check leaf + hub stuff */

		acptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
		acptr->c_type = CLIENT_TYPE_SERVER;
		acptr->c_from = lptr;
		acptr->c_hopcount = hc;
		acptr->c_lasttime = acptr->c_firsttime = Now;
		acptr->c_rem_ipaddr = -1;
		acptr->c_rem_port = 0;
		support_strncpy(acptr->name, servername, sizeof(acptr->name));
		support_strncpy(acptr->ircname, serverinfo, sizeof(acptr->ircname));
		if (++Globalinfo->g_servers > Globalinfo->g_maxservers)
		{
			Globalinfo->g_maxservers = Globalinfo->g_servers;
			Globalinfo->g_maxserverstime = Now;
		}
		hash_add_client(acptr);
		acptr->c_servptr = pptr;
		add_client_to_ulist(&(pptr->c_servers), acptr);
		add_client_to_oslist(Servers, acptr);
		for(bcptr=*Servers;bcptr;bcptr=bcptr->c_os_next)
		{
			if ((bcptr == meptr) || (bcptr == lptr) || !IsLocal(bcptr))
				continue;
			if (!match(conf_myname_for_link(meptr->name,
						bcptr->c_hostmask, cnamebuf1), acptr->name))
				continue;
			send_prefix_to_client(pptr, bcptr, "SERVER %s %d :%s",
					acptr->name, hc+1, acptr->ircname);
		}
		Globalinfo->g_serversbursting++;
		CliSetSentEOB(acptr);
		send_prefix_to_client(meptr, acptr, "PING %s :%s",
						meptr->name, acptr->name);
		send_to_flag(CLIENT_NFLAGS_CONNECTS, NULL,
			"Server %s[%s] being introduced by %s",
				acptr->name, acptr->ircname, pptr->name);
		return 3;
	}

	if (serverinfo == NULL || !*serverinfo)
	{
		char	exitbuf[512];

		sprintf(exitbuf, "No server info specified for %s", servername);
		client_del(meptr, lptr, exitbuf);
		return -1;
	}

#ifndef HUB
	if (Globalinfo->m_servers == 1)
	{
		send_to_opers("Link %s cancelled, I'm not a HUB",
			ircd_clientname(lptr, 1, cnamebuf1));
		client_del(meptr, lptr, "I'm a leaf, not a hub");
		return -1;
	}
#else
	YOU_CANT_RUN_THIS_SERVER_AS_A_HUB_YET();
#endif

	/* new local server */
	if (!*lptr->c_passwd)
	{
		send_to_opers("Link %s cancelled, no password given",
				ircd_clientname(lptr, 1, cnamebuf1));
		client_del(meptr, lptr, "No password given");
		return -1;
	}
	if (!DoesTS(lptr))
	{
		send_to_opers("Link %s dropped, non-TS server",
				ircd_clientname(lptr, 1, cnamebuf1));
		client_del(meptr, lptr, "non-TS server");
		return -1;
	}

	if (lptr->c_flags & CLIENT_FLAGS_SCONNECT)
	{
		if (strcasecmp(lptr->name, servername))
		{
			send_to_opers("Link %s dropped, name mismatch -- %s",
				ircd_clientname(lptr, 1, cnamebuf1), servername);
			client_del(meptr, lptr, "servername mismatch");
			return -1;
		}
	}
	else
		strcpy(lptr->name, servername);

	lptr->c_hopcount = hc;

	sl = servlink_find(servername, lptr);
	if (!sl)
	{
		send_to_opers("Received unauthorized connection from %s.",
				ircd_clientname(lptr, 1, cnamebuf1));
		client_del(meptr, lptr, "No C/N conf lines");
		return -1;
	}

	if (!sl->npasswd || !*sl->npasswd || !*lptr->c_passwd ||
				strcmp(sl->npasswd, lptr->c_passwd))
	{
		send_to_opers("Access denied (passwd mismatch) %s",
			ircd_clientname(lptr, 1, cnamebuf1));
		client_del(meptr, lptr, "Access denied (passwd mismatch)");
		sl->tryconnecttime = 0;
		return -1;
	}

	if (sl->classptr)
		classptr = (IRCD_CLASSENTRY *)sl->classptr;
	else
	{
		ret = class_find(&(_ircd_conf.classes), sl->classname, &classptr);
		if (ret < 0) /* hrm...*/
		{
			send_to_opers("Class %s not found in conf -- denying %s",
				sl->classname, ircd_clientname(lptr, 1, cnamebuf1));
			client_del(meptr, lptr, "Configuration Error -- Class not found");
			sl->tryconnecttime = 0;
			return -1;
	    }
		sl->classptr = (void *)classptr;
	}

	sl->tryconnecttime = 0;
	sl->lastconnect = Now;

    add_client_to_cllist(&(classptr->clients), lptr);
	classptr->num_on++;
    lptr->c_class = classptr;
    lptr->c_maxsendq = -1; /*  we'll reset this below */
    lptr->c_pingfreq = classptr->pingfreq;

	BZERO(lptr->c_passwd, sizeof(lptr->c_passwd));

	lptr->c_hostmask = sl->hostmask;

	if (!(lptr->c_flags & CLIENT_FLAGS_SCONNECT))
	{
		if (sl->cpasswd && *sl->cpasswd)
			send_to_client(lptr, "PASS %s :TS", sl->cpasswd);
		send_capabilities(lptr, 0);
		send_to_client(lptr, "SERVER %s 1 :%s",
			conf_myname_for_link(meptr->name, lptr->c_hostmask, cnamebuf1),
			meptr->ircname);
	}


	send_to_client(lptr, "SVINFO %d %d %d :%d", TS_CURRENT, TS_MIN,
			Globalinfo->m_servers ? 1 : 0, Now);

	lptr->c_type = CLIENT_TYPE_SERVER;
	support_strncpy(lptr->ircname, serverinfo, sizeof(lptr->ircname));
	queue_resize(&(lptr->c_queue_in), 32*1024);
	queue_resize(&(lptr->c_queue_out), 32*1024);
	add_client_to_oslist(Servers, lptr);
	hash_add_client(lptr);
	lptr->c_servptr = meptr;
#if 0
	add_client_to_slist(&(meptr->c_servers), lptr);
#else
	add_client_to_ulist(&(meptr->c_servers), lptr);
#endif

	if (!lptr->c_connectby)
	{
		lptr->c_connectby = MyMalloc(CLIENTNAMESIZE);
		sprintf(lptr->c_connectby, "Remote!Connect@%s", meptr->name);
	}

	send_to_opers("Link with %s (TS) established.",
			ircd_clientname(lptr, 1, cnamebuf1));

	send_to_flag(CLIENT_NFLAGS_CONNECTS, NULL,
			"Server %s being introduced locally", lptr->name);

	ret = Recvbuffersize;
	setsockopt(lptr->c_connptr->fdlistinfo.fd, SOL_SOCKET, SO_RCVBUF, (char *)&ret, sizeof(ret));
	ret = Sendbuffersize;
	setsockopt(lptr->c_connptr->fdlistinfo.fd, SOL_SOCKET, SO_SNDBUF, (char *)&ret, sizeof(ret));

	if (++Globalinfo->g_servers > Globalinfo->g_maxservers)
	{
		Globalinfo->g_maxservers = Globalinfo->g_servers;
		Globalinfo->g_maxserverstime = Now;
	}
	if (++Globalinfo->m_servers > Globalinfo->m_maxservers)
	{
		Globalinfo->m_maxservers = Globalinfo->m_servers;
		Globalinfo->m_maxserverstime = Now;
	}
	if (Globalinfo->m_clients + Globalinfo->m_servers > Globalinfo->m_maxconn)
	{
		Globalinfo->m_maxconn = Globalinfo->m_clients + Globalinfo->m_servers;
		Globalinfo->m_maxconntime = Now;
	}

	/* send new server to my connected servers. */
	for(acptr=meptr->c_servers;acptr;acptr=acptr->c_u_next)
	{
		if (acptr == lptr)
			continue;
		if (!match(conf_myname_for_link(meptr->name,
						acptr->c_hostmask, cnamebuf1), lptr->name))
			continue;
		send_prefix_to_client(meptr, acptr, "SERVER %s 2 :%s",
				lptr->name, lptr->ircname);
	}

	/* send existing servers to new server. */
	_send_servers_recurs(lptr, meptr);

#if 0
	for(acptr=meptr->c_servers;acptr;acptr=acptr->c_u_next)
	{
		if (acptr->c_from == lptr)
			continue;
		if (!match(conf_myname_for_link(meptr->name, lptr->c_hostmask,
					cnamebuf1), acptr->name))
		{
			_send_servers_recurs(lptr, acptr);
			continue;
		}
		send_prefix_to_client(acptr->c_servptr, lptr, "SERVER %s %d :%s",
				acptr->name, acptr->c_hopcount+1, acptr->ircname);
		_send_servers_recurs(lptr, acptr);
	}
#endif

	{
		static char		nickissent = 1;
		IRCD_CHANNEL	*chptr;
		IRCD_CLIENTLIST	*clptr;

		nickissent = 3 - nickissent;

		for(chptr=Channels;chptr;chptr=chptr->next)
		{
			for(clptr=chptr->members;clptr;clptr=clptr->next)
			{
				if (clptr->cptr->c_nicksent != nickissent)
				{
					clptr->cptr->c_nicksent = nickissent;
					if (clptr->cptr->c_from != lptr)
						_sendnick(lptr, clptr->cptr);
				}
			}
			_send_channel_modes(lptr, chptr);
		}

		/*
		** also send out those that are not on any channel
		*/
		for (acptr=*Clients;acptr;acptr=acptr->c_os_next)
		{
            if (acptr->c_nicksent != nickissent)
            {
                acptr->c_nicksent = nickissent;
                if (acptr->c_from != lptr)
                    _sendnick(lptr, acptr);
            }
		}
		/* Local Opers aren't in the above Clients list */
		for (acptr=*Opers;acptr;acptr=acptr->c_os_next)
		{
            if (acptr->c_nicksent != nickissent)
            {
                acptr->c_nicksent = nickissent;
                if (acptr->c_from != lptr)
                    _sendnick(lptr, acptr);
            }
		}

		Globalinfo->g_serversbursting++;
        CliSetSentEOB(lptr);
		send_prefix_to_client(meptr, lptr, "PING %s :%s",
						meptr->name, lptr->name);
	}

#if 0
/*
	for when we allow unlimited sendq during burst...
      but, we'll also need to somehow check that a remote
     server is bursting, before this stuff will work
*/
     
	if ((lptr->c_queue_out.queue_size + 1000000) > lptr->c_class->maxsendq)
		lptr->c_maxsendq = lptr->c_queue_out.queue_size + 1000000;
	else
		lptr->c_maxsendq = lptr->c_class->maxsendq;

#else
		lptr->c_maxsendq = lptr->c_class->maxsendq;
#endif

	return 3;
}

int m_list(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*channels = ircd_nextarg(&args, 0);
	char			*channel;
	IRCD_CHANNEL	*chptr;
	int				sc;

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LISTSTART, pptr->name));

	if (!channels || !*channels)
	{
		IRCD_CHANNELLIST	*chlptr;

		for(chptr=Channels;chptr;chptr=chptr->next)
		{
			if (!(chlptr = IsMember(chptr, pptr)) &&
						(chptr->modes & CHANNEL_MODES_SECRET))
				continue;
			sc = (chlptr || !(chptr->modes & (CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE)));
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_LIST, pptr->name),
					sc ? chptr->name : "*", chptr->numusers,
					sc ? (chptr->topic ? chptr->topic : "[No topic set]") : "[Topic hidden for non-member]");
		}
	}
	else for(;(channel = ircd_next_subarg_comma(&channels)) != NULL;)
	{
		chptr = hash_find_channel(channel);
		if (chptr &&
			(!(chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET)) ||
					IsMember(chptr, pptr)))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_LIST, pptr->name), channel,
				chptr->numusers, chptr->topic ? chptr->topic : "[No topic set]");
		}
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LISTEND, pptr->name));
	return 3;
}

int m_topic(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
#define MAX_TOPIC_CHANS	10
	char				*channels = ircd_nextarg(&args, 0);
	char				*topic = ircd_nextarg(&args, 0);
	char				*name;
	IRCD_CHANNEL		*chptr;
	IRCD_CHANNELLIST	*chlptr;
	char				*sendtos[MAX_TOPIC_CHANS+1];
	char				**st = NULL;
	char				**std;
	int					numtried;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!channels || !*channels)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "TOPIC");
		return 3;
	}
	if (topic && (strlen(topic) > TOPICLEN))
		topic[TOPICLEN] = '\0';

	if (IsLocal(pptr))
	{
		numtried = 0;
		st = sendtos;
		for(;(name = ircd_next_subarg_comma(&channels)) != NULL;)
		{
			if (!*name)
				continue;
			if (++numtried > MAX_TOPIC_CHANS)
			{
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Too many recipients", pptr->name);
				return 3;
			}
			*st = name;
			st++;
		}
		*st = NULL;
		st = sendtos;
	}

	numtried = 0;

	for(;;)
	{
		if (st)
		{
			name = *st;
			st++;
			if (name)/* check for already sent */
			{
				for(std=sendtos;*std != name;std++)
					if (!strcmp(*std, name))
						break;
				if (*std != name)
				{
					DEBUG(DEBUG_ALL, ("got dup %s for topic\n", name));
					continue;
				}
			}
		}
		else
			name = ircd_next_subarg_comma(&channels);
		if (!name)
			break;
		numtried++;
		chptr = hash_find_channel(name);
		if (!chptr || ((chlptr = IsMember(chptr, pptr)) == NULL))
		{
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
					name);
			continue;
		}
		if (!topic)
		{
			if (!chptr->topic)
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_NOTOPIC, pptr->name), chptr->name);
			else
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPIC, pptr->name),
					chptr->name, chptr->topic);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPICWHOTIME, pptr->name),
					chptr->name, chptr->topic_who, chptr->topic_time);
			}
		}
		else
		{
			IRCD_CLIENTLIST	*clptr = chlptr->clptr;

			if (!IsServer(lptr) && (chptr->modes & CHANNEL_MODES_TOPICLIMIT))
			{
				if (!(clptr->modes & CHANNEL_MODES_CHANOP))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					continue;
				}
			}
			if (chptr->topic)
			{
				MyFree(chptr->topic);
				MyFree(chptr->topic_who);
				chptr->topic = NULL;
				chptr->topic_who = NULL;
			}
			if (*topic)
			{
				chptr->topic = MyStrdup(topic);
				chptr->topic_who = MyStrdup(pptr->name);
				chptr->topic_time = Now;
			}
			send_to_channel_butone(IsServer(lptr) ? lptr : NULL,
					CHANNEL_SEND_SERVERS|CHANNEL_SEND_CLIENTS,
					chptr, pptr, "TOPIC %s :%s", chptr->name, topic);
		}
	}
	numtried /= 5;
	return numtried > 3 ? numtried : 3;
}

void _channel_invite_del(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr)
{
	IRCD_INVITE	*inext;
	IRCD_INVITE	*invite = cptr->c_invites_list;

	for(;invite;invite=inext)
	{
		inext = invite->next;
		if (!chptr || !strcasecmp(chptr->name, invite->channel))
		{
			del_invite_from_list(&(cptr->c_invites_list), invite);
			cptr->c_invites_num--;
			MyFree(invite->channel);
			MyFree(invite);
		}
	}
}

int m_invite(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*user = ircd_nextarg(&args, 0);
	char			*channel = ircd_nextarg(&args, 0);
	int				doit = 0;
	IRCD_CLIENT		*acptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CHANNELLIST	*chlptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "INVITE");
		return 3;
	}

	if ((!(acptr = hash_find_client(user))) || !IsClient(acptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK,
				pptr->name), user);
		return 3;
	}

	if (!(chptr = hash_find_channel(channel)))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
			pptr->name), channel);
#if 0
		send_prefix_to_client(pptr, acptr, "INVITE %s :%s",
				user, channel);
#endif
		return 3;
	}
	if ((chlptr = IsMember(chptr, pptr)) == NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
			pptr->name), channel);
		return 3;
	}
	if (IsMember(chptr, acptr) != NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_USERONCHANNEL,
			pptr->name), user, channel);
		return 3;
	}

	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
	{
		IRCD_CLIENTLIST	*clptr = chlptr->clptr;

		if (!(clptr->modes & CHANNEL_MODES_CHANOP))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED,
				pptr->name), chptr->name);
			return 3;
		}
		doit = 1;
	}

	if (IsLocal(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INVITING,
				pptr->name), acptr->name, chptr->name);
		if (acptr->c_away)
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_AWAY,
					pptr->name), acptr->name, acptr->c_away);
	}

	if (IsLocal(acptr) && doit)
		_channel_invite_add(chptr, acptr);
	send_prefix_to_client(pptr, acptr, "INVITE %s :%s",
			acptr->name, chptr->name);
	return 3;
}

int m_version(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsServer(lptr) && !IsReg(lptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_VERSION,
				pptr->name), VERSION, "", meptr->name, "http://www.codestud.com/ircd");
		client_del(pptr, pptr, "Leaving");
		return -1;
	}

	if (_hunt_server(lptr, pptr, ":%s VERSION :%s", 1, prefix, args) == HUNTED_ISME)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_VERSION,
				pptr->name), VERSION, "", meptr->name, "http://www.codestud.com/ircd");
	return 3;
}

int m_quit(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*comment = ircd_nextarg(&args, 0);

	if (IsServer(pptr))
		return 3;
	if (!comment ||
		((lptr == pptr) && (!strncmp("Local Kill", comment, 10) ||
				!strncmp(comment, "Killed", 6))))
			comment = lptr->name;
	if (strlen(comment) > QUITLEN)
		comment[QUITLEN] = '\0';
	client_del(pptr, pptr, comment);
	if (pptr == lptr)
		return -1;
	return 3;
}

int m_squit(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*server = ircd_nextarg(&args, 0);
	char			*comment = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char			cnamebuf[CLIENTNAMESIZE];
	IRCD_CLIENT		*acptr;

	if (!IsAnOper(pptr) && !IsServer(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (server && *server)
	{
		if ((*server == '*') && IsServer(lptr))
		{
			if (!mycmp(server, conf_myname_for_link(meptr->name,
						lptr->c_hostmask, cnamebuf)))
				server = lptr->name;
		}

		for(acptr=*Servers;acptr;acptr=acptr->c_os_next)
			if (!match(server, acptr->name))
				break;
		if (acptr && IsMe(acptr))
		{
			acptr = lptr;
			server = lptr->name;
		}
	}
	else
	{
		server = lptr->name;
		acptr = lptr;
	}

	if (!acptr || IsClient(acptr))
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name), server);
		return 3;
	}

	if ((pptr->c_umodes & CLIENT_UMODES_LOCOP) && !IsLocal(acptr))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!comment || !*comment)
		comment = lptr->name;

	if (IsLocal(acptr) && IsServer(lptr))
	{
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
			"Received SQUIT %s from %s (%s)",
			server, ircd_clientname(pptr, 0, cnamebuf), comment);
		send_to_servers(NULL, ":%s WALLOPS :Received SQUIT %s from %s (%s)",
			meptr->name,
			server, ircd_clientname(pptr, 0, cnamebuf), comment);
	}
	else if (IsLocal(acptr))
	{
		send_to_opers("Received SQUIT %s from %s (%s)",
			acptr->name, ircd_clientname(pptr, 0, cnamebuf), comment);
	}
	acptr->c_flags |= CLIENT_FLAGS_KILLED;
	client_del(pptr, acptr, comment);
	return (acptr == lptr) ? -1 : 3;
}

int m_kill(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	cnamebuf[CLIENTNAMESIZE];
	char	*inpath = ircd_clientname(lptr, 0, cnamebuf);
	char	*victim = ircd_nextarg(&args, 0);
	char	*comment = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char	*commentsave;
	char	buf[512];
	char	buf2[512];
	IRCD_CLIENT	*acptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!victim || !*victim)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "KILL");
		return 3;
	}

	if (!comment || !*comment)
		comment = pptr->name;

	commentsave = comment;

	if (!(IsServer(lptr) || IsAnOper(lptr)))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (IsAnOper(lptr))
	{
		if (strlen(comment) > KILLLEN)
			comment[KILLLEN] = '\0';
	}

	if (!(acptr = hash_find_chasing(victim)))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOSUCHNICK, pptr->name), victim);
		return 3;
	}

	if (!IsLocal(acptr) && (lptr->c_umodes & CLIENT_UMODES_LOCOP))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (IsServer(acptr) || (acptr == meptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_CANTKILLSERVER, pptr->name));
		return 3;
	}

	if (MyClient(pptr))
	{
		if (IsLocal(acptr) &&
					!(pptr->c_oline_flags & OLINE_FLAGS_CANLOCALKILL))
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are not allowed to LOCAL KILL", pptr->name);
			return 3;
		}
		if (!IsLocal(acptr) &&
					!(pptr->c_oline_flags & OLINE_FLAGS_CANGLOBALKILL))
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are not allowed to GLOBAL KILL", pptr->name);
			return 3;
		}
	}
	if (!IsServer(lptr))
	{
		/* kill came from oper on this server */
		inpath = lptr->hostname;
		sprintf(buf, "%s%s (%s)",
			lptr->name, IsOper(pptr) ? "" : "(L)", comment);
		comment = buf;
	}

	if (IsServer(pptr))
	{
		send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
				"Received KILL message for %s. From %s Path: %s!%s",
				acptr->name, pptr->name, inpath, comment);
	}
	else
	{
		send_to_flag(CLIENT_NFLAGS_OKILLS, NULL,
				"Received KILL message for %s. From %s Path: %s!%s",
				acptr->name, pptr->name, inpath, comment);
	}

	if (!IsLocal(acptr) || !IsLocal(pptr) || !IsAnOper(pptr))
	{
		send_to_servers(lptr, ":%s KILL %s :%s!%s",
			prefix, acptr->name, inpath, comment);
		acptr->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
	}

	if (IsLocal(acptr))
	{
		acptr->c_flags |= CLIENT_FLAGS_KILLED;

		if (IsClient(pptr))
			send_prefix_to_client(pptr, acptr, "KILL %s :%s!%s",
				acptr->name, inpath, comment);
		else
			send_to_client(acptr, ":%s KILL %s :%s!%s",
				prefix, acptr->name, inpath, comment);
	}

	if (IsLocal(acptr) && IsLocal(pptr) && IsAnOper(pptr))
		sprintf(buf2, "Local kill by %s (%s)", pptr->name, commentsave);
	else
	{
		char	*killer;

		if ((killer = strchr(comment, ' ')) != NULL)
		{
			while (*killer && (*killer != '!'))
				killer--;
			if (!*killer)
				killer = comment;
			else
				killer++;
		}
		else
			killer = comment;
		sprintf(buf2, "Killed (%s)", killer);
	}
/* above already
	acptr->c_flags |= CLIENT_FLAGS_KILLED;
*/
	client_del(pptr, acptr, buf2);
	return (acptr == lptr) ? -1 : 3;
}

int m_info(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	static int	ratenum = 0;
	static int	ratetime = 0;
	char		numbuf[256];

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s INFO :%s", 1, prefix, args) != HUNTED_ISME)
	{
		return 6; /* remote info requests from local clients */
	}

	if (!IsLocal(pptr) && !IsAnOper(pptr) && ircd_ratelimit(pptr, &ratenum, &ratetime, (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_INFO_NUM), (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_TIME)))
		return 0;

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Welcome to csircd (CodeStud Ircd) -- http://www.codestud.com/ircd");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "This server was mostly written by Chris Behrens <cbehrens@codestud.com>");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "between 1998 and 1999 to create a more efficient client server for EFnet.");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Other contributors include:");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Thomas Helvey");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Fred Jacobs");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Richard Steenbergen");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "...and I'm sure others that I've forgotten.");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	sprintf(numbuf, "Current TS Version: %d", TS_CURRENT);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), numbuf);
	sprintf(numbuf, "Minimum TS Version: %d", TS_MIN);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), numbuf);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "Please send comments, bug reports, patches, etc to ircd@codestud.com.");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), " - Chris");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INFO, pptr->name), "");
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFINFO, pptr->name));

	return 3;
}

/*
** m_links
**  parv[0] = sender prefix
**  parv[1] = servername mask
** or
**  parv[0] = sender prefix
**  parv[1] = server to query
**      parv[2] = servername mask
*/
int m_links(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	argssave[512];
	char	*servername;
	char	*mask;
	IRCD_CLIENT	*acptr;
	
	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (args)
		support_strncpy(argssave, args, sizeof(argssave));

	mask = ircd_nextarg(&args, 0);
	servername = ircd_nextarg(&args, 0);
#if 0
	if (servername && *servername)
	{
		if (_hunt_server(lptr, pptr, ":%s LINKS %s :%s", 1, prefix, argssave) != HUNTED_ISME)
			return 3;
		mask = servername;
	}
#endif

	if (!mask || !*mask)
		mask = NULL;

	for(acptr=*Servers;acptr;acptr=acptr->c_os_next)
	{
		if (!mask || !match(mask, acptr->name))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LINKS,
					pptr->name), acptr->name, acptr->c_servptr->name,
					acptr->c_hopcount, *acptr->ircname ? acptr->ircname :
					"(Unknown Location)");
		}
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFLINKS, pptr->name),
			mask ? mask : "*");
	return 3;
}

int m_stats(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CLIENT		*sptr;
	IRCD_CLIENT		*acptr;
	IRCD_LISTENER	*liptr;
	char			cnamebuf[CLIENTNAMESIZE];
	char			*thing;
	char			*server;
	char			*rest;
	OLINE			*oline;
	SERVLINK		*servlink;
	int				now;
	CONFHOSTIPENTRY	*chentry;
	IRCD_CLASSENTRY	**ceptr;
	int				i;
	static char 	LFormat[] = "%d %s %s %u %u %u %u %u :%u";
	static int		ratenum = 0;
	static int		ratetime = 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s STATS %s :%s", 2, prefix, args) != HUNTED_ISME)
	{
		return 6; /* remote queries are penalized more */
	}

	thing = ircd_nextarg(&args, 0);
	server = ircd_nextarg(&args, 0);
	rest = ircd_nextarg(&args, 0);

	if (!thing || !*thing)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "STATS");
		return 3;
	}

	send_to_flag(CLIENT_NFLAGS_STATSNOTICES, NULL,
			"STATS %s%c requested by %s (%s@%s) [%s]",
			*thing < ' ' ? "^" : "",
			(*thing) < ' ' ? (*thing)+0x40 : *thing,
			pptr->name, pptr->username, pptr->hostname,
			pptr->servername);

#define Meg 1024
#define Gig (Meg*1024)
#define Tera (Gig*1024)
#define IsMeg(x) (x>Meg)
#define IsGig(x) (x>Gig)
#define IsTera(x) (x>Tera)
#define ByteString(x) (IsTera(x) ? "Tb " : \
            IsGig(x) ? "Gig" : \
            IsMeg(x) ? "Meg" : \
            "KB ")
#define ByteConvert(x) (IsTera(x) ? (float)((float)x/(float)Tera) : \
            IsGig(x) ? (float)((float)x/(float)Gig) : \
            IsMeg(x) ? (float)((float)x/(float)Meg) : \
            (float)x)

	acptr = NULL;
	if (server && *server)
	{
		acptr = hash_find_chasing(server);
	}

	if ((*thing == 'L') || (*thing == 'l'))
	{
		if (!IsAnOper(pptr) && (!acptr || !IsClient(acptr)))
			*thing = '?';
	}

	if ((*thing != '?') && !IsLocal(pptr) && !IsAnOper(pptr) && ircd_ratelimit(pptr, &ratenum, &ratetime, (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_STATS_NUM), (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_TIME)))
		return 0;

	if (!strcasecmp(thing, "rate"))
	{
		int itotal = Globalinfo->bw_c_incoming_rate + Globalinfo->bw_s_incoming_rate;
		int ototal = Globalinfo->bw_c_outgoing_rate + Globalinfo->bw_s_outgoing_rate;

		send_prefix_to_client(meptr, pptr,
				"%d %s :Current Bandwidth Statistics (in bytes/second):",
					RPL_STATSDEBUG, pptr->name);
		send_prefix_to_client(meptr, pptr,
				"%d %s :Clients --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name, Globalinfo->bw_c_incoming_rate,
					Globalinfo->bw_c_outgoing_rate);
		send_prefix_to_client(meptr, pptr,
				"%d %s :Servers --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name,
					Globalinfo->bw_s_incoming_rate,
					Globalinfo->bw_s_outgoing_rate);
		send_prefix_to_client(meptr, pptr,
				"%d %s :  Total --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name, itotal, ototal);
	}
	else switch(*thing)
	{
		case '?':
			for(sptr=*Servers;sptr;sptr=sptr->c_os_next)
			{
				if (!IsLocal(sptr))
					continue;
				send_prefix_to_client(meptr, pptr, LFormat,
						RPL_STATSLINKINFO, pptr->name,
						IsAnOper(pptr) ? ircd_clientname(sptr, 1, cnamebuf) :
						ircd_clientname(sptr, 0, cnamebuf),
						sptr->c_queue_out.queue_size,
						sptr->c_send_msg,
						sptr->c_send_kbytes,
						sptr->c_recv_msg,
						sptr->c_recv_kbytes,
						Now - sptr->c_firsttime);
			}
			break;
		case 'C':
		case 'c':
			for(servlink=_ircd_conf.servers;servlink;servlink=servlink->next)
			{
				static char hidden[] = "127.0.0.1";
				char		flagsbuf[64];
				char		*ptr;
				char		*hname;

				ptr = flagsbuf;

				if (servlink->flags & SERVLINK_FLAGS_AUTOCONN)
				{
					if (ptr != flagsbuf)
						*ptr++ = ' ';
					ptr += sprintf(ptr, "autoconn");
				}
#if 0
				else if (servlink->flags & ... )
				{
					if (ptr != flagsbuf)
						*ptr++ = ' ';
					ptr += sprintf(ptr, "newflag");
				}
#endif

				*ptr = '\0';

				if ( /* !IsAnOper(pptr) && */
						(servlink->flags & SERVLINK_FLAGS_HIDEHOSTNAME))
					hname = hidden;
				else
					hname = servlink->hostname;

				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSCLINE, pptr->name),
						'C', servlink->username ? servlink->username : "*",
						hname, servlink->servername,
						servlink->port, servlink->classname, flagsbuf);
				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSNLINE, pptr->name),
						'N', servlink->username ? servlink->username : "*",
						hname, servlink->servername,
						servlink->hostmask, servlink->classname, flagsbuf);
			}
			break;
		case 'H':
		case 'h':
			for(servlink=_ircd_conf.servers;servlink;servlink=servlink->next)
			{
				IRCD_STRINGLIST *stptr;

				for(stptr=servlink->hubs;stptr;stptr=stptr->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSHLINE, pptr->name),
						'H', stptr->string, servlink->servername, 0, -1);
				}
				for(stptr=servlink->leafs;stptr;stptr=stptr->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSLLINE, pptr->name),
						'L', stptr->string, servlink->servername, 0, -1);
				}
			}
			break;
		case 'I':
		case 'i':
			for(chentry=_ircd_conf.allow.conf_all;chentry;chentry=chentry->ch_next)
			{
				char *iname = NULL;
				char *hname = NULL;
				IRCD_CLASSENTRY	*classptr;

				if (chentry->classptr)
					classptr = (IRCD_CLASSENTRY *)chentry->classptr;
				else
				{
					int ret = class_find(&(_ircd_conf.classes),
							chentry->classname, &classptr);
					if (ret < 0)
						classptr = NULL;
					else
						chentry->classptr = (void *)classptr;
				}

				if (classptr && (classptr->flags & CLASS_FLAGS_HIDEHOSTNAME))
					continue;

				if (chentry->ipstring)
				{
					if (chentry->username)
					{
						iname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->ipstring) + 2);
						sprintf(iname, "%s@%s", chentry->username,
										chentry->ipstring);
					}
					else
						iname = MyStrdup(chentry->ipstring);
				}
				if (chentry->hostname)
				{
					if (chentry->username)
					{
						hname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->hostname) + 2);
						sprintf(hname, "%s@%s", chentry->username,
										chentry->hostname);
					}
					else
						hname = MyStrdup(chentry->hostname);
				}

				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSILINE, pptr->name),
					'I', iname ? iname : "X", hname ? hname : "X",
					0, chentry->classname);

				if (iname)
					MyFree(iname);
				if (hname)
					MyFree(hname);
			}
			break;
		case 'K':
		case 'k':
		{
			char *user = NULL;
			char *host = NULL;
			static char _star[] = "*";
			int	numfound = 0;

			if (!IsOper(pptr) && (!rest || !*rest))
			{
				user = pptr->username;
				host = pptr->hostname;
			}
			else if (rest && *rest)
			{
				acptr = hash_find_client(rest);
				if (!acptr || IsServer(acptr))
				{
					user = host = rest;
					while(*host && (*host != '@'))
						host++;
					if (!*host)
					{
						host = rest;
						user = _star;
					}
					else
					{
						*host++ = '\0';
						if (!*host)
							host = _star;
					}
					if (!*user)
						user = _star;
				}
				else
				{
					user = acptr->username;
					host = acptr->hostname;
				}
			}
			for(chentry=_ircd_conf.deny.conf_all;chentry;chentry=chentry->ch_next)
			{
				char *iname = NULL;
				char *hname = NULL;

				if (chentry->expires && (chentry->expires < Now))
					continue;

				if (user)
				{
					if (chentry->username &&
							match(chentry->username, user) &&
							match(user, chentry->username))
						continue;
					if (!(((chentry->ipstring &&
						(!match(chentry->ipstring, host) ||
								!match(host, chentry->ipstring))) ||
						(chentry->hostname &&
						(!match(chentry->hostname, host) ||
								!match(host, chentry->hostname))))))
						continue;
				}
				if (chentry->ipstring)
				{
					if (chentry->username)
					{
						iname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->ipstring) + 2);
						sprintf(iname, "%s@%s", chentry->username,
										chentry->ipstring);
					}
					else
						iname = MyStrdup(chentry->ipstring);
				}
				if (chentry->hostname)
				{
					if (chentry->username)
					{
						hname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->hostname) + 2);
						sprintf(hname, "%s@%s", chentry->username,
										chentry->hostname);
					}
					else
						hname = MyStrdup(chentry->hostname);
				}

				numfound++;
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSKLINE, pptr->name),
					'K', iname ? iname : "X",
					hname ? hname : "X", chentry->expires,
					chentry->flags,
					chentry->data ? chentry->data : "*");

				if (iname)
					MyFree(iname);
				if (hname)
					MyFree(hname);

				if (user && (numfound == 100))
				{
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Maximum of 100 bans reached -- stopping request.",
						pptr->name);
					break;
				}
			}
			if (!numfound)
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :No bans %s%s%s%s%s%sfound",
					pptr->name,
					user ? "matching" : "", user ? " " : "",
					user ? user : "", user ? "@": "",
					user ? host : "", user ? " " : "");
			}
		}
			break;
		case 'L':
		case 'l':
			if (acptr && IsClient(acptr))
			{
				if (IsLocal(acptr))
					send_prefix_to_client(meptr, pptr, LFormat,
						RPL_STATSLINKINFO, pptr->name,
						*thing == 'L' ? ircd_clientname(acptr, 1, cnamebuf):
						ircd_clientname(acptr, 0, cnamebuf),
						acptr->c_queue_out.queue_size,
						acptr->c_send_msg, acptr->c_send_kbytes,
						acptr->c_recv_msg, acptr->c_recv_kbytes,
						Now - acptr->c_firsttime);
				break;
			}
			/* IsAnOper(pptr) is TRUE because of check above switch() */

			for(liptr=_ircd_conf.listeners;liptr;liptr=liptr->l_next)
			{
				sprintf(cnamebuf, "%s[*@%s.%d]",
						meptr->name,
						liptr->l_ipstring, liptr->l_port);
				send_prefix_to_client(meptr, pptr, LFormat,
					RPL_STATSLINKINFO, pptr->name, cnamebuf, 0, 0, 0, 0, 0,
					Now - liptr->l_firsttime);
			}
			for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
			{
				if (!IsClient(acptr) && !IsServer(acptr))
					continue;
				send_prefix_to_client(meptr, pptr, LFormat,
					RPL_STATSLINKINFO, pptr->name,
					*thing == 'L' ? ircd_clientname(acptr, 1, cnamebuf):
					ircd_clientname(acptr, 0, cnamebuf),
					acptr->c_queue_out.queue_size,
					acptr->c_send_msg, acptr->c_send_kbytes,
					acptr->c_recv_msg, acptr->c_recv_kbytes,
					Now - acptr->c_firsttime);
			}

			break;
		case 'O':
		case 'o':
			for(oline=_ircd_conf.olines;oline;oline=oline->next)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSOLINE, pptr->name),
					oline->flags & OLINE_FLAGS_LOCAL ? 'o' : 'O', 
#if 0
					oline->username ? oline->username : "*",
					oline->hostname,
#else
					"*", "127.0.0.1",
#endif
					oline->name, 0, oline->classname);
			}
			break;
		case 'U':
		case 'u':
			now = Now - meptr->c_firsttime;
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_STATSUPTIME,
					pptr->name), now/86400, (now/3600)%24, (now/60)%60,
					now % 60);
			break;

		case 'Y':
		case 'y':
			for(ceptr=_ircd_conf.classes.entries,i=_ircd_conf.classes.num;i--;ceptr++)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSYLINE, pptr->name),
					'Y', (*ceptr)->classname, (*ceptr)->pingfreq,
					(*ceptr)->connfreq, (*ceptr)->maxclients,
					(*ceptr)->maxsendq, (*ceptr)->uhlimit,
					(*ceptr)->hlimit,
					(*ceptr)->floodnum, (*ceptr)->floodtime,
					(*ceptr)->floodexit, (*ceptr)->ipaddr_rate_num,
					(*ceptr)->ipaddr_rate_time);
			}
			break;
		case 'p':
		case 'P':
			i = 0;
			for(acptr=*Opers;acptr;acptr=acptr->c_os_next)
			{
				if (!IsLocal(acptr))
					continue;
				i++;
				send_prefix_to_client(meptr, pptr,
						"%d %s :[%c] %s (%s@%s) idle %is",
						RPL_STATSDEBUG, pptr->name,
						IsOper(acptr) ? 'O' : 'o',
						acptr->name, acptr->username,
						acptr->hostname, Now - acptr->c_idletime);
			}
			if (!i)
				send_prefix_to_client(meptr, pptr, "%d %s :No opers found",
						RPL_STATSDEBUG, pptr->name);
			break;
		case 'Q':
		case 'q':
			if (IsAnOper(pptr))
			{
				STRINGDENY	*sd;

				for(sd=_ircd_conf.nickname_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSQLINE, pptr->name),
						sd->name, "-", "-", sd->expires,
						sd->flags, sd->reason);
				}
				for(sd=_ircd_conf.channel_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSQLINE, pptr->name),
						sd->name, "-", "-", sd->expires,
						sd->flags, sd->reason);
				}
			}
			break;
		case 'X':
		case 'x':
			if (IsAnOper(pptr))
			{
				STRINGDENY	*sd;

				for(sd=_ircd_conf.ircname_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSXLINE, pptr->name),
						sd->name, sd->expires,
						sd->flags, sd->reason);
				}
			}
			break;
		case 'Z':
		case 'z':
		{
			unsigned int total_usage = 0;
			unsigned int iphash_num_nethash;
			unsigned int iphash_num_ips;
			unsigned int iphash_size_nethash;
			unsigned int whowas_num;
			unsigned int whowas_memory;
			unsigned int whowas_array_size;
			int i;
			IRCD_CLIENT	*list = Locals;
			IRCD_CLIENT *nextacptr;
			unsigned int lo_num = 0;
			unsigned int cl_num = 0;
			unsigned int extra_mem = 0;
			unsigned int ch_num = 0;
			unsigned int ch_mem = 0;
			unsigned int ban_num = 0;
			unsigned int ban_mem = 0;
			IRCD_CHANNEL	*chptr;

			iphash_count_memory(&iphash_num_nethash, &iphash_num_ips,
				&iphash_size_nethash);
			total_usage += iphash_num_nethash * sizeof(IRCD_NETHASH);
			total_usage += iphash_num_ips * sizeof(IRCD_IPENTRY);
			total_usage += iphash_size_nethash;

			send_prefix_to_client(meptr, pptr,
				"%d %s :IPHASH: Array: %i NetEntries: %i(%i) IPEntries: %i(%i)",
				RPL_STATSDEBUG, pptr->name,
				iphash_size_nethash, iphash_num_nethash,
				iphash_num_nethash * sizeof(IRCD_NETHASH),
				iphash_num_ips, iphash_num_ips * sizeof(IRCD_IPENTRY));
			whowas_count_memory(&whowas_num, &whowas_memory, &whowas_array_size);
			total_usage += whowas_memory + whowas_array_size;

			send_prefix_to_client(meptr, pptr,
				"%d %s :Whowas: Array: %i Entries: %i(%i)",
				RPL_STATSDEBUG, pptr->name,
				whowas_array_size, whowas_num, whowas_memory);

			for(i=0;i<4;i++)
			{
				if (i==1)
					list = *Clients;
				else if (i==2)
					list = *Opers;
				else if (i==3)
					list = *Servers;
				for(acptr=list;acptr;acptr=nextacptr)
				{
					if (list == Locals)
					{
						nextacptr = acptr->c_lo_next;
						lo_num++;
					}
					else
					{
						nextacptr = acptr->c_os_next;
						if (IsLocal(acptr))
							continue; /* already done */
						if (IsClient(acptr))
							cl_num++;
						else
							lo_num++;
					}
					if (acptr->c_away)
						extra_mem += strlen(acptr->c_away)+1;
					if (IsLocal(acptr) && acptr->c_connectby)
						extra_mem += strlen(acptr->c_connectby)+1;
					if (acptr->c_invites_list)
					{
						IRCD_INVITE	*iptr;
						for(iptr=acptr->c_invites_list;iptr;iptr=iptr->next)
							extra_mem += sizeof(IRCD_INVITE) + strlen(iptr->channel) + 1;
					}
					if (acptr->c_channels)
					{
						IRCD_CHANNELLIST	*chlptr;
						for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
							extra_mem += sizeof(IRCD_CHANNELLIST) + sizeof(IRCD_CLIENTLIST);
					}
				}
			}
			total_usage += lo_num * sizeof(IRCD_CLIENT);
			total_usage += cl_num * offsetof(IRCD_CLIENT, c_auth_ptr);
			total_usage += extra_mem;
			total_usage += sizeof(Fds);
			send_prefix_to_client(meptr, pptr,
				"%d %s :Clients: Local: %i(%i) Remote: %i(%i) Extra: %i  Fds: %i",
				RPL_STATSDEBUG, pptr->name,
				lo_num, lo_num*sizeof(IRCD_CLIENT), cl_num,
				cl_num * offsetof(IRCD_CLIENT, c_auth_ptr),
				extra_mem, sizeof(Fds));
			for(chptr=Channels;chptr;chptr=chptr->next)
			{
				ch_num++;
				ch_mem += strlen(chptr->name);
				if (chptr->key)
					extra_mem += strlen(chptr->key) + 1;
				if (chptr->topic)
					extra_mem += strlen(chptr->topic) + 1;
				if (chptr->topic_who)
					extra_mem += strlen(chptr->topic_who) + 1;
				if (chptr->bans)
				{
					IRCD_BAN	*b;
					for(b=chptr->bans;b;b=b->next)
					{
						ban_num++;
						if (b->firstpart)
							ban_mem += strlen(b->firstpart) + 1;
						if (b->banstring)
							ban_mem += strlen(b->banstring) + 1;
						if (b->who)
							ban_mem += strlen(b->who) + 1;
					}
				}
			}
			total_usage += ch_mem + ban_mem + (ban_num * sizeof(IRCD_BAN));
			total_usage += extra_mem;
			send_prefix_to_client(meptr, pptr,
				"%d %s :Channels: %i(%i) Bans: %i(%i) Extra: %i",
				RPL_STATSDEBUG, pptr->name, ch_num,
				ch_mem, ban_num, ban_mem + (ban_num * sizeof(IRCD_BAN)),
				extra_mem);
			send_prefix_to_client(meptr, pptr,
				"%d %s :Total Usage: %i",
				RPL_STATSDEBUG, pptr->name, total_usage);

		}
			break;
	}

	send_prefix_to_client(meptr, pptr,
		IRCD_REPLY(RPL_ENDOFSTATS, pptr->name), *thing);

	return 3;
}

int m_help(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	return 3;
}

int m_error(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (args && *args &&
			(IsServer(lptr) || (lptr->c_flags & CLIENT_FLAGS_SCONNECT)))
	{
		if (*args == ':')
			args++;
		send_to_opers("ERROR from %s: %s", lptr->name, args);
		return 3;
	}
	/* ignore ERROR */
	return 3;
}

int m_away(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*away = ircd_nextarg(&args, 0);

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (away && !*away)
		away = NULL;
	else if (away && (strlen(away) > AWAYLEN))
		away[AWAYLEN] = '\0';
		
	if (pptr->c_away)
		MyFree(pptr->c_away);
	pptr->c_away = away ? MyStrdup(away) : away;
	if (IsLocal(pptr))
	{
		if (!away)
			send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_UNAWAY, pptr->name));
		else
			send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_NOWAWAY, pptr->name));
	}

	if (!away)
		send_to_servers(lptr, ":%s AWAY", pptr->name);
#if 0
	else
		send_to_servers(lptr, ":%s AWAY :%s", pptr->name, away);
#endif
	return 3;
}

int _do_connect(IRCD_CLIENT *pptr, SERVLINK *sptr, int portnum)
{
	int				port = 6667;
	IRCD_CLIENT		*acptr;
	int				afd;
	int				ret;
	char			cnamebuf[CLIENTNAMESIZE];
	ircd_uint32_t	ipaddr = -1;
	struct sockaddr_in	loc_sin;
	struct sockaddr_in	rem_sin;
	int				sin_size;

	if (portnum > 0)
		port = portnum;
	else if (sptr->port > 0)
		port = sptr->port;

	if (pptr)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :*** Connecting to %s[%s].",
			pptr->name, sptr->servername, sptr->hostname);
	}
	else
	{
		send_to_opers("Trying to connect to %s[%s].",
			sptr->servername, sptr->hostname);
	}

	if (sptr->inaddr.s_addr != -1)
		ipaddr = sptr->inaddr.s_addr;
	else
	{
		/*
		** TODO  -- do DNS lookup and continue to try to connect when
		**  have the answer 
		*/
		if (pptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Couldn't connect to %s: Unknown host -- try /REHASH to try to resolve again",
				pptr->name, sptr->servername);
		}
		else
		{
			send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
				"Couldn't connect to %s: Unknown host.", sptr->hostname);
		}
		sptr->tryconnecttime = 0;
        return -1;
	}
    afd = mysocket_connectip(ipaddr, port,
			sptr->bindipaddr == -1 ? INADDR_ANY : sptr->bindipaddr,
			MYSOCKET_FLAGS_REUSEADDR|MYSOCKET_FLAGS_NONBLOCKING);
	if (afd < 0)   
	{
		if (pptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Couldn't connect to %s: %i",
				pptr->name, sptr->servername, afd);
		}
		else
		{
			send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
				"Couldn't connect to %s: %s.", sptr->hostname,
				support_strerror(-afd));
		}
		sptr->tryconnecttime = 0;
		return -1;
	}

	acptr = (IRCD_CLIENT *)MyMalloc(sizeof(IRCD_CLIENT));
	acptr->c_connptr = (IRCD_CONN *)MyMalloc(sizeof(IRCD_CONN));
	acptr->c_connptr->conn_type = IRCD_CONN_TYPE_CLIENT;
	acptr->c_connptr->conn_ptr = (void *)acptr;

	support_strncpy(acptr->name, sptr->servername, sizeof(acptr->name));

	acptr->c_lasttime = acptr->c_firsttime = Now;

	acptr->c_connectby = MyMalloc(CLIENTNAMESIZE);
	sprintf(acptr->c_connectby, "AutoConn!*@%s", meptr->name);

	sin_size = sizeof(loc_sin);
	ret = getsockname(afd, (struct sockaddr *)(&loc_sin), &sin_size);
	if (ret >= 0)
	{
		acptr->c_loc_ipaddr = ntohl(loc_sin.sin_addr.s_addr);
		acptr->c_loc_port = ntohs(loc_sin.sin_port);
	}
	acptr->c_rem_ipaddr = ipaddr;
	acptr->c_rem_port = port;
	rem_sin.sin_addr.s_addr = htonl(ipaddr);
	strcpy(acptr->c_ipstring, inet_ntoa(rem_sin.sin_addr));
	acptr->c_type = CLIENT_TYPE_UNKNOWN;
	acptr->c_flags |= CLIENT_FLAGS_DOINGCONNECT|CLIENT_FLAGS_SCONNECT;
	acptr->c_from = acptr;
	ret = queue_init(&(acptr->c_queue_in), QUEUE_SIZE, QUEUEINIT_FLAGS);
	if (ret)
	{
		close(afd);
		MyFree(acptr->c_connectby);
		MyFree(acptr->c_connptr);
		MyFree(acptr);
		if (pptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Couldn't connect to %s: queue_init(): %i",
				pptr->name, sptr->servername, ret);
		}
		else
		{
			send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
				"Couldn't connect to %s: queue_init(): %s", sptr->servername,
				support_strerror(-ret));
		}
		sptr->tryconnecttime = 0;
		return -1;
	}
	ret = queue_init(&(acptr->c_queue_out), QUEUE_SIZE, QUEUEINIT_FLAGS);
	if (ret)
	{
		queue_delete(&(acptr->c_queue_in));
		MyFree(acptr->c_connectby);
		MyFree(acptr->c_connptr);
		MyFree(acptr);
		close(afd);
		if (pptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Couldn't connect to %s: queue_init(): %i",
				pptr->name, sptr->servername, ret);
		}
		else
		{
			send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
				"Couldn't connect to %s: queue_init(): %s", sptr->servername,
				support_strerror(-ret));
		}
		sptr->tryconnecttime = 0;
		return -1;
	}

	if (ircd_open_fd(afd, FDLIST_FLAGS_READ|FDLIST_FLAGS_WRITE, acptr->c_connptr, NULL) < 0)
	{
		queue_delete(&(acptr->c_queue_out));
		queue_delete(&(acptr->c_queue_in));
   		close(afd);
		MyFree(acptr->c_connectby);
		MyFree(acptr->c_connptr);
		MyFree(acptr);
		sptr->tryconnecttime = 0;
		return -1;
	}
	acptr->c_flags |= CLIENT_FLAGS_LOCAL;
	add_client_to_lolist(&Locals, acptr);

	acptr->c_hostmask = sptr->hostmask;

	acptr->c_maxsendq = 1000000; /* this will get reset in m_server */
	acptr->c_pingfreq = 300;

	if (sptr->cpasswd && *sptr->cpasswd)
		send_to_client(acptr, "PASS %s :TS", sptr->cpasswd);
	send_capabilities(acptr, 0);
	send_to_client(acptr, "SERVER %s 1 :%s",
		conf_myname_for_link(meptr->name, acptr->c_hostmask, cnamebuf),
		meptr->ircname);

	sptr->tryconnecttime = Now;

	return 0;
}

/*
** m_connect
**  parv[0] = sender prefix
**  parv[1] = servername
**  parv[2] = port number
**  parv[3] = remote server
*/
int m_connect(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*server = NULL;
	char			*portnum = NULL;
	char			*remserver;
	int				port = -1;
	IRCD_CLIENT		*acptr;
	SERVLINK		*sptr;
	char			cnamebuf[CLIENTNAMESIZE];
	int				already = 0;
	char			argssave[512];

	if (!IsServer(lptr) && !IsAnOper(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));  
		return 3;
	}

	if (MyClient(pptr) && !(pptr->c_oline_flags & OLINE_FLAGS_CANGLOBALROUTE) &&
			args && *args)
	{
		char *argptr = argssave;

		support_strncpy(argptr, args, sizeof(argssave));

		server = ircd_nextarg(&argptr, 0);
		portnum = ircd_nextarg(&argptr, 0);
		remserver = ircd_nextarg(&argptr, 0);
		if (remserver && *remserver)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are not allowed to GLOBAL ROUTE", pptr->name);
			return 3;
		}
		already = 1;
	}

	if (_hunt_server(lptr, pptr, ":%s CONNECT %s %s :%s", 3, prefix,
				args) != HUNTED_ISME)
		return 3;

	if (MyClient(pptr) && !(pptr->c_oline_flags & OLINE_FLAGS_CANLOCALROUTE))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to LOCAL ROUTE", pptr->name);
		return 3;
	}

	if (!already)
	{
		server = ircd_nextarg(&args, 0);
		portnum = ircd_nextarg(&args, 0);
		remserver = ircd_nextarg(&args, 0);
	}

	if (!server || !*server)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
					pptr->name), "CONNECT");
		return 3;
	}

	for(sptr=_ircd_conf.servers;sptr;sptr=sptr->next)
		if (!match(server, sptr->hostname))
			break;
	if (!sptr)
	{
		for(sptr=_ircd_conf.servers;sptr;sptr=sptr->next)
			if (!match(server, sptr->servername))
				break;
	}

	if (!sptr)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Connect: Host %s not listed in config file",
			pptr->name, server);
		return 3;
	}

	if (sptr)
	{
		if (Now < (sptr->tryconnecttime + 60))
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Connect: Already trying to connect to %s",
				pptr->name, server);
			return 3;
		}
	}

	if (!*sptr->cpasswd || !*sptr->npasswd)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Connect: Missing cpasswd or npasswd from conf file for %s",
				pptr->name, server);
		return 3;
	}

	acptr = hash_find_client_incl_hm(sptr->servername);
	if (acptr)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Connect: Server %s already exists from %s.",
			pptr->name, sptr->servername, acptr->c_from->name);
		return 3;
	}

	if (portnum && *portnum)
	{
		int p = atoi(portnum);

		if (p <= 0)
		{
			if (IsClient(pptr))
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Connect: Invalid port.", pptr->name);
			return 3;
		}
		port = p;
	}

	if (IsServer(lptr))
	{
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr,
			"Remote CONNECT %s %i from %s",
			server, port, ircd_clientname(pptr, 0, cnamebuf));
		send_to_servers(lptr, ":%s WALLOPS :Remote CONNECT %s %i from %s",
			meptr->name, server, port, ircd_clientname(pptr, 0, cnamebuf));
	}

	(void)_do_connect(pptr, sptr, port);

	return 3;
}

int m_ping(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*origin = ircd_nextarg(&args, 0);
	char			*destination = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;

	if (!origin || !*origin)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOORIGIN, pptr->name));
		return 3;
	}
	acptr = hash_find_client_incl_hm(origin);
	if (acptr && acptr != pptr)
		origin = lptr->name;
	if (destination && mycmp(destination, meptr->name))
	{
		acptr = hash_find_client_incl_hm(destination);
		if (!acptr || !IsServer(acptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name), destination);
		}
		else
		{
			if (IsServer(pptr))
				send_to_client(acptr, ":%s PING %s :%s",
					prefix, origin, destination);
			else
				send_prefix_to_client(pptr, acptr, "PING %s :%s",
					origin, destination);
		}
	}
	else
		send_prefix_to_client(meptr, pptr, "PONG %s :%s", 
			(destination) ? destination : meptr->name, origin);
	return 3;
}

int m_pong(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*origin = ircd_nextarg(&args, 0);
	char			*destination = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;

	if (!origin || !*origin)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOORIGIN, pptr->name));
		return 3;
	}

	/* ping reset in ircd_parse() */

	if (!IsServer(lptr) && !IsReg(lptr))
		return 3;

	if (IsServer(pptr) && CliSentEOB(pptr) && !CliGotEOB(pptr))
	{
		if (pptr == lptr)
		{
			send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
				"End-of-Burst acknowledged from %s",
				lptr->name);
		}
		Globalinfo->g_serversbursting--;
		CliSetGotEOB(pptr);
	}

	if (destination && mycmp(destination, meptr->name))
	{
		acptr = hash_find_client_incl_hm(destination);
		if (acptr)
		{
			if (IsClient(pptr))
				send_prefix_to_client(pptr, acptr, "PONG %s %s",
					origin, destination);
			else
				send_to_client(acptr, ":%s PONG %s %s",
					prefix, origin, destination);
		}
		else
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name),
				destination);
		}
	}
	return 3;
}

int m_oper(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*name = ircd_nextarg(&args, 0);
	char	*password = ircd_nextarg(&args, 0);
	OLINE	*oline;
	int		oldmodes;
	int 	ret;
	char	 *encr;
	char	buf[64];
	IRCD_CLASSENTRY	*classptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (IsServer(pptr) || IsServer(lptr)) /* eh ? */
		return 3;

	if (!password || !*password)
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "OPER");
		return 3;
	}

	if (IsOper(pptr))
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_YOUREOPER, pptr->name));
		return 3;
	}

	oldmodes = pptr->c_umodes;

	ret = oline_find(&(_ircd_conf.olines), name, pptr->username,
			pptr->hostname, pptr->c_rem_ipaddr, &oline);
	if (ret)
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOOPERHOST, pptr->name));
		return 3;
	}

#ifdef CRYPT_OPER_PASSWORD

	if (((encr = crypt(password, oline->passwd)) == NULL) ||
			strcmp(encr, oline->passwd))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_PASSWDMISMATCH, pptr->name));
		send_to_opers("Failed OPER attempt: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, name);

		if (vars_value_get_bynum(IRCD_VAR_BADOPER_LOG))
			log_printf(_ircd_conf.log_badoper, "%s %s!%s@%s tried [%s:%s]\n",
                    support_ctime(Now), pptr->name, pptr->username,
					pptr->hostname, oline->name, password);

		return 3;
	}
#else
	if (strcmp(password, oline->c_passwd))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_PASSWDMISMATCH, pptr->name));
		send_to_opers("Failed OPER attempt: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, name);
		if (vars_value_get_bynum(IRCD_VAR_BADOPER_LOG))
			log_printf(_ircd_conf.log_badoper, "%s %s!%s@%s tried [%s:%s]\n",
                    support_ctime(Now), pptr->name, pptr->username,
					pptr->hostname, oline->name, password);
		return 3;
	}

#endif
	if (oline->flags & OLINE_FLAGS_LOCAL)
	{
		pptr->c_umodes |= CLIENT_UMODES_LOCOP;
	}
	else
	{
		pptr->c_umodes |= CLIENT_UMODES_OPER;
		Globalinfo->m_opers++;
		Globalinfo->g_opers++;
	}

	pptr->c_oline_flags = oline->flags;

	assert(pptr->c_oslist);
	del_client_from_oslist(pptr->c_oslist, pptr);
	add_client_to_oslist(Opers, pptr);

	if (oline->userflags)
		pptr->c_nflags = oline->userflags;
	else
	{
		pptr->c_nflags |= CLIENT_NFLAGS_SNOTICES;
		pptr->c_nflags |= CLIENT_NFLAGS_SWALLOPS;
		pptr->c_nflags |= CLIENT_NFLAGS_OKILLS;
	}
	if (pptr->c_nflags & CLIENT_NFLAGS_SNOTICES)
		pptr->c_umodes |= CLIENT_UMODES_SMODE;
	if (pptr->c_nflags & CLIENT_NFLAGS_SWALLOPS)
		pptr->c_umodes |= CLIENT_UMODES_WMODE;
	if (pptr->c_nflags & CLIENT_NFLAGS_OKILLS)
		pptr->c_umodes |= CLIENT_UMODES_KMODE;

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_YOUREOPER, pptr->name));
	send_to_opers("%s (%s@%s) is now operator (%c)", pptr->name,
			pptr->username, pptr->hostname, IsOper(pptr) ? 'O' : 'o');
	/*														^^^^^^^^^
	**										these should really be reversed
	**										but this is how old code is.
	*/
	if (vars_value_get_bynum(IRCD_VAR_GOODOPER_LOG))
		log_printf(_ircd_conf.log_goodoper, "%s %s!%s@%s [%s] opered\n",
			support_ctime(Now), pptr->name, pptr->username,
			pptr->hostname, oline->name);

	_umode_build_change(pptr, buf, oldmodes, SEND_UMODES);
	send_to_servers(NULL, ":%s MODE %s %s", prefix, pptr->name, buf);
	_umode_build_change(pptr, buf, oldmodes, -1);
#if 0
	send_prefix_to_client(pptr, pptr, "MODE %s %s", pptr->name, buf);
/* umodes only send client's nick, not full n!u@h */
#else
	send_to_client(pptr, ":%s MODE %s %s", prefix, pptr->name, buf);
#endif

	m_flags(lptr, pptr, prefix, NULL);

	ret = class_find(&(_ircd_conf.classes), oline->classname, &classptr);
	if (ret < 0)
		return 3;

	del_client_from_cllist(&(pptr->c_class->clients), pptr);
	if (!--pptr->c_class->num_on &&
		(pptr->c_class->flags & CLASS_FLAGS_ILLEGAL))
	{           
		class_del(&(_ircd_conf.classes), pptr->c_class);
	}

    add_client_to_cllist(&(classptr->clients), pptr);
    classptr->num_on++;
	pptr->c_class = classptr;
	pptr->c_maxsendq = classptr->maxsendq;
	pptr->c_pingfreq = classptr->pingfreq;
	pptr->c_rate_msg_limit = classptr->floodnum * 3;
	pptr->c_rate_time_limit = classptr->floodtime;

	return 3;
}

int m_pass(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*password = ircd_nextarg(&args, 0);

	if (!password || !*password)
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				*pptr->name ? pptr->name : "*"), "PASS");
		return 3;
	}
	if (!IsLocal(pptr) || !IsUnknown(lptr))
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ALREADYREGISTRED,
				pptr->name));
		return 3;
	}
	support_strncpy(pptr->c_passwd, password, sizeof(pptr->c_passwd));
	if (args && *args)
	{
		int	l = strlen(args);

		if (l < 2)
			return 3;
		if (!strcmp(args+l-2, "TS"))
			pptr->c_flags |= CLIENT_FLAGS_DOESTS;
	}
	return 3;
}

int m_wallops(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*message = ircd_nextarg(&args, 0);

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!message || !*message)
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "WALLOPS");
		return 3;
	}

	if (IsClient(pptr))
	{
		if (MyClient(pptr) && !IsAnOper(pptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
			return 3;
		}
		send_wallops_to_flag(CLIENT_NFLAGS_OWALLOPS, NULL, pptr, "%s",
				message);
	}
	else
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, pptr, "%s",
				message);

	if (IsServer(pptr))
		send_to_servers(lptr, ":%s WALLOPS :%s", prefix, message);
	else
		send_to_servers(lptr, ":%s OPERWALL :%s", prefix, message);

	return 3;
}

int m_locops(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*message = ircd_nextarg(&args, 0);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!message || !*message)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "LOCOPS");
		return 3;
	}

	if (!IsAnOper(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}
	send_wallops_to_flag(CLIENT_NFLAGS_LWALLOPS, NULL, pptr, "LOCOPS - %s", message);
	return 3;
}

int m_time(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	datebuf[128];

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s TIME :%s", 1, prefix, args) == HUNTED_ISME)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_TIME, pptr->name),
			meptr->name, datestr(0, datebuf));
	return 3;
}

int m_names(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*channel;
	IRCD_CHANNEL	*acptr;
	char			buf[512];
	char			*dptr;
	char			*saveptr;
	int				ismem;
	int				flag;
	int				mylen;
	IRCD_CLIENTLIST	*clptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(lptr))
		return 3; /* drop it */

	channel = ircd_nextarg(&args, 0);

	if (!channel || !*channel)
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "NAMES");
		return 3;
	}

	acptr = hash_find_channel(channel);
	if (!acptr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
				pptr->name), channel);
		return 3;
	}

	ismem = (IsMember(acptr, pptr) != NULL);
	if (!ismem &&
			(acptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET)))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_ENDOFNAMES, pptr->name), channel);
		return 3;
	}
	mylen = strlen(meptr->name);
	dptr = buf;
	*dptr++ = '*';
	*dptr++ = ' ';
	dptr += strlen(strcpy(dptr, acptr->name));
	*dptr++ = ' ';
	*dptr++ = ':';
	saveptr = dptr;
	if (acptr->modes & CHANNEL_MODES_SECRET)
		*buf = '@';
	else if (!(acptr->modes & CHANNEL_MODES_PRIVATE))
		*buf = '=';
	flag = 1;
	for(clptr=acptr->members;clptr;clptr=clptr->next)
	{
		if (!ismem && (clptr->cptr->c_umodes & CLIENT_UMODES_IMODE))
			continue;
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*dptr++ = '@';
		else if (clptr->modes & CHANNEL_MODES_VOICE)
			*dptr++ = '+';
		dptr += strlen(strcpy(dptr, clptr->cptr->name));
		*dptr++ = ' ';
		flag = 1;
		if (mylen + (dptr - buf) > 475)
		{
			*(dptr-1) = '\0';
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_NAMREPLY,
					pptr->name), buf);
			dptr = saveptr;
			flag = 0;
		}
	}
	if (flag)
	{
		if (*(dptr-1) == ' ')
			*(dptr-1) = '\0';
		else
			*dptr = '\0';
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_NAMREPLY,
				pptr->name), buf);
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFNAMES, pptr->name),
		channel);

	return 3;
}

int m_admin(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s ADMIN :%s", 1, prefix, args) != HUNTED_ISME)
		return 3;

	if (_ircd_conf.admin && *_ircd_conf.admin)
	{
		char	**lines;
		int		i;
		int		num;

		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_ADMINME, pptr->name), meptr->name);
		for(lines=_ircd_conf.admin,i=0;*lines;i++,lines++)
		{
			if (!i)
				num = RPL_ADMINLOC1;
			else if (i == 1)
				num = RPL_ADMINLOC2;
			else
				num = RPL_ADMINEMAIL;
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(num, pptr->name), *lines);
		}
	}
	else
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOADMININFO, pptr->name), meptr->name);

	return 3;
}

int m_trace(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*arg1;
	int				doall;
	IRCD_CLIENT		*acptr;
	char			cnamebuf[CLIENTNAMESIZE];
	int				numservers;
	int				numusers;
	int				i;
	IRCD_CLASSENTRY	**ceptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (args && *args && (_hunt_server(lptr, pptr,
					":%s TRACE %s :%s", 2, prefix, args) != HUNTED_ISME))
	{
			return 3;
	}

	switch (_hunt_server(lptr, pptr, ":%s TRACE :%s", 1, prefix, args))
	{
		case HUNTED_PASS:
		{
			char *name = 	ircd_nextarg(&args, 0);
			IRCD_CLIENT		*ac2ptr;

			if (!name || !*name)
				name = meptr->name;

			ac2ptr = hash_find_client_incl_hm(name);

			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACELINK, pptr->name),
					VERSION, "", name, ac2ptr ? ac2ptr->c_from->name : "unknown");
		
			return 3;
		}
		case HUNTED_ISME:
			break;
		default:
			return 3;
	}

	arg1 = ircd_nextarg(&args, 0);

	if (!arg1 || !*arg1)
	{
		arg1 = meptr->name;
		doall = 1;
	}
	else
		doall = !match(arg1, meptr->name);

	if (!doall)
	{
		acptr = hash_find_client(arg1);
		if (acptr && IsLocal(acptr))
		{
			if (IsClient(acptr))
			{
				if (IsAnOper(acptr))
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEOPERATOR, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				else
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEUSER, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
			}
			else if (IsServer(acptr))
			{
				numservers = numusers = 0;
				_count_servers_and_users(acptr, &numservers, &numusers);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACESERVER, pptr->name),
					acptr->c_class->classname, numservers, numusers,
					ircd_clientname(acptr, 0, cnamebuf), acptr->c_connectby);
			}
		}
	}
	else
	{
		for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
		{
			if (IsServer(acptr))
			{
				numservers = numusers = 0;
				_count_servers_and_users(acptr, &numservers, &numusers);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACESERVER, pptr->name),
					acptr->c_class->classname, numservers, numusers,
					ircd_clientname(acptr, 0, cnamebuf), acptr->c_connectby);
			}
			else if (IsClient(acptr))
			{
				if (IsAnOper(acptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEOPERATOR, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				}
				else if (IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEUSER, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				}
			}
			else if (IsUnknown(acptr) && IsAnOper(pptr))
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACEUNKNOWN, pptr->name),
					acptr->c_class ? acptr->c_class->classname : "unknown", ircd_clientname(acptr, 0, cnamebuf));
			}
		}
		if (IsAnOper(pptr))

			for(ceptr=_ircd_conf.classes.entries,i=_ircd_conf.classes.num;i--;ceptr++)
		{
			if ((*ceptr)->num_on > 0)
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACECLASS, pptr->name),
					(*ceptr)->classname, (*ceptr)->num_on);
		}

	}

	send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_ENDOFTRACE, pptr->name), arg1);

	return 3;
}


void _user_del_channels(IRCD_CLIENT *pptr)
{
	IRCD_CHANNELLIST		*nextch;

	for(;pptr->c_channels;pptr->c_channels=nextch)
	{
		nextch = pptr->c_channels->next;
		_channel_del_user(pptr->c_channels->chptr, pptr, pptr->c_channels);
		MyFree(pptr->c_channels);
	}
}


int m_join(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*channel = ircd_nextarg(&args, 0);
	char	*key = ircd_nextarg(&args, 0);
	char	*ptr;
	char	*k;
	char	*name;
	int		chlen;
	int		flag;
	int		i;
	IRCD_CHANNEL	*chptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
			"JOIN");
		return 3;
	}

	for(ptr=channel;;)
	{
		if ((*ptr == ',') && (*(ptr+1) == '0') && !atoi(ptr+1))
		{
			channel = ptr+1;
			ptr++;
		}
		if (!*ptr++)
			break;
	}

	k = ircd_next_subarg_comma(&key);
	for(;(name=ircd_next_subarg_comma(&channel)) != NULL;
							k=ircd_next_subarg_comma(&key))
	{
		_channelname_clean(name);

		if (!*name)
			continue;

		if (*name == '0' && !atoi(name))
		{
			IRCD_CHANNELLIST		*nextch;

			for(;pptr->c_channels;pptr->c_channels=nextch)
			{
				nextch = pptr->c_channels->next;
				send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						pptr->c_channels->chptr,
						pptr, "PART %s", pptr->c_channels->chptr->name);
				_channel_del_user(pptr->c_channels->chptr, pptr,
						pptr->c_channels);
				MyFree(pptr->c_channels);
			}
			send_to_servers(lptr, ":%s JOIN 0", pptr->name);
			continue;
		}
		if (IsServer(lptr) && (*name == '&'))
			continue;
		if ((chlen=strlen(name)) > CHANNELLEN)
		{
			name[CHANNELLEN] = '\0';
			chlen = CHANNELLEN;
		}
		if (!IsChannel(name))
		{
			if (MyClient(pptr))
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
						pptr->name), name);
			continue;
		}
		if (IsLocal(pptr))
		{
			int		maxchannels = pptr->c_class->maxchannels;

			if (maxchannels < 0)
				maxchannels = (int)vars_value_get_bynum(IRCD_VAR_DEFAULT_MAXCHANNELS);

			if (pptr->c_numchannels >= maxchannels)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_TOOMANYCHANNELS, pptr->name), name);
				continue;
			}

			if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSCHANNELDENY))
			{
				STRINGDENY		*sd;

				sd = stringdeny_match(&(_ircd_conf.channel_denies),
							name, Now);
				if (sd)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_UNAVAILRESOURCE, pptr->name),
						name);
					continue;
				}
			}
		}
		flag = 0;
		chptr = hash_find_channel(name);
		if (!chptr)
		{
			if (!IsLocal(pptr))
			{
				send_to_opers("User %s on %s remotely JOINing new channel",
						pptr->name, pptr->servername);
			}
			else
			{
				int chanopsonsplit = (int)vars_value_get_bynum(IRCD_VAR_CHANOPSONSPLIT);

				if (IsAnOper(pptr) &&
					vars_value_get_bynum(IRCD_VAR_OPERSCANJOINONSPLIT))
				{
					if (chanopsonsplit && (*name != '&') &&
						((Globalinfo->g_servers -
							Globalinfo->g_serversbursting) <
												chanopsonsplit))
					{
						send_to_opers("%s is creating channel %s during split",
							pptr->name, name);
					}
					flag |= CHANNEL_MODES_CHANOP;
				}
				else if (!chanopsonsplit || (*name == '&') ||
						((Globalinfo->g_servers -
							Globalinfo->g_serversbursting) >=
												chanopsonsplit))
				{
					flag |= CHANNEL_MODES_CHANOP;
				}
			}
			chptr = (IRCD_CHANNEL *)MyMalloc(sizeof(IRCD_CHANNEL)+chlen);
			strcpy(chptr->name, name);
			hash_add_channel(chptr);
			if (++Globalinfo->channels > Globalinfo->maxchannels)
			{
				Globalinfo->maxchannels = Globalinfo->channels;
				Globalinfo->maxchannelstime = Now;
			}
			chptr->creation = Now;
		}

		if (IsMember(chptr, pptr))
		{
			if (IsLocal(pptr))
			{
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_USERONCHANNEL,
					pptr->name), pptr->name, name);
			}
			continue;
		}

		if (IsClient(lptr) && (i = _channel_canjoin(pptr, chptr, k)))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(i, pptr->name),
					name);
			continue;
		}

		_channel_add_user(chptr, pptr, flag);

		if (*name != '&')
		{
			if (IsClient(lptr) && (flag & CHANNEL_MODES_CHANOP))
			{
				send_to_servers(NULL, ":%s SJOIN %u %s + :@%s",
					meptr->name, chptr->creation, name, pptr->name);
			}
			else if (IsClient(lptr))
				send_to_servers(NULL, ":%s SJOIN %u %s + :%s",
					meptr->name, chptr->creation, name, pptr->name);
			else
			{
				/* this won't happen with all TS servers */
				send_to_servers(lptr, ":%s JOIN :%s", pptr->name, name);
			}
		}

		send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "JOIN :%s", name);

		if (MyClient(pptr))
		{
			if ((chptr->numusers == 1) && !(flag & CHANNEL_MODES_CHANOP))
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :*** Channel ops on %s are unavailable during split ***",
						pptr->name, name);
#if 0
			_channel_invite_del(chptr, pptr);
/* already done in _canjoin() */
#endif
			if (chptr->topic)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPIC, pptr->name),
					name, chptr->topic);
				if (chptr->topic_who)
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TOPICWHOTIME, pptr->name),
						name, chptr->topic_who, chptr->topic_time);
			}

			/* do m_names */
			m_names(lptr, pptr, prefix, name);
		}
	}

	return 3;
}

int m_part(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CHANNEL	*chptr;
	char			*channels = ircd_nextarg(&args, 0);
	char			*name;
	IRCD_CHANNELLIST	*chlptr;
	char			buf[512];
	char			*dptr = buf;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (!channels || !*channels)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "PART");
		return 3;
	}

	for(;(name = ircd_next_subarg_comma(&channels)) != NULL;)
	{
		chptr = hash_find_channel(name);
		if (!chptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
					pptr->name), name);
			continue;
		}
		if ((chlptr = IsMember(chptr, pptr)) == NULL)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
				pptr->name), name);
			continue;
		}

		if (*name != '&')
		{
			if (dptr != buf)
				*dptr++ = ',';
			dptr += strlen(strcpy(dptr, name));
		}

		send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "PART %s", name);
		_channel_del_user(chptr, pptr, chlptr);
		del_channellist_from_list(&(pptr->c_channels), chlptr);
		MyFree(chlptr);
	}
	if (dptr != buf)
	{
		*dptr = '\0';
		send_to_servers(lptr, ":%s PART %s", pptr->name, buf);
	}
	return 3;
}

int m_lusers(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*pattern;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s LUSERS %s :%s", 2, prefix, args) != HUNTED_ISME)
		return 3;

	pattern = ircd_nextarg(&args, 0);
	if (pattern && !*pattern)
		pattern = NULL;

	_lusers_send(pptr, pattern, 1);

	return 3;
}

int m_motd(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	static int ratenum = 0;
	static int ratetime = 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (_hunt_server(lptr, pptr, ":%s MOTD :%s", 1, prefix, args) != HUNTED_ISME)
	{
		return 6; /* remote info requests from local clients */
	}
	if (!IsLocal(pptr) && !IsAnOper(pptr) && ircd_ratelimit(pptr, &ratenum, &ratetime, (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_MOTD_NUM), (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_TIME)))
		return 0;
	_motd_send(pptr);
	return 3;
}


int m_mode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CLIENT		*tptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CHANNELLIST	*chlptr2;
	char			*channel = ircd_nextarg(&args, 0);
	char			*modes = ircd_nextarg(&args, 0);
	char			*modeargs = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char			*ptr;
	char			plusminus = '+';
	char			newban[NICKLEN+USERLEN+HOSTLEN+6];
	IRCD_MODE		*opmodes = NULL;
	IRCD_MODE		*lastopmode = NULL;
	IRCD_MODE		*banmodes = NULL;
	IRCD_MODE		*lastbanmode = NULL;
	IRCD_MODE		*opmodeptr;
	char			parabuf[512];
	char			modebuf[512];
	char			modeargbuf[512];
	char			*modeptr = modebuf;
	char			*modeargptr = modeargbuf;
	char			*newkey = NULL;
	int				limitchange = 0;
	int				limit = 0;
	int				tlimit;
	int				keychange = 0;
	int				opcnt = 0;
	int				*i;
	int				newmodes;
	int				oldmodes;
	char			*sptr;
	char			*target;
	int				isok;
	int				ischop;
	int				isdeop;
	int				got_ovk = 0;
	IRCD_BAN		*ban;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "MODE");
		return 3;
	}
	chptr = hash_find_channel(channel);
	if (!chptr)
		return m_umode(lptr, pptr, channel, prefix, modes);
	if (!modes || !*modes)
	{
		_build_channel_modes(chptr, pptr, modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CHANNELMODEIS, pptr->name), chptr->name,
			modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CREATIONTIME, pptr->name), chptr->name,
			chptr->creation);
		return 3;
	}

	chlptr = IsMember(chptr, pptr);

	ischop = IsServer(pptr) ||
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_CHANOP));
	isdeop = !ischop && !IsServer(pptr) &&
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_DEOPPED));
	isok = ischop || (!isdeop && IsServer(lptr) && chptr->creation);

	oldmodes = newmodes = chptr->modes;

	for(ptr=modes;ptr && *ptr;)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			case 'o':
			case 'v':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
#if 0
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
						"MODE-ov");
#endif
					break;
				}
				if (IsClient(lptr))
				{
					if (!chlptr)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
							chptr->name);
						break;
					}
					if (!ischop)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
							chptr->name);
						break;
					}

					if (opcnt++ >= MAXMODEPARAMS)
						break;
				}
				tptr = hash_find_chasing(target);
				if (!tptr)
					break;
				if (!IsMember(chptr, tptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_USERNOTINCHANNEL, pptr->name),
						tptr->name, chptr->name);
					break;
				}
				if ((tptr == lptr) && (plusminus == '+') && (*ptr == 'o'))
				{
					break;
				}
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (got_ovk != *ptr))
					break;
				/* keep track of 'tptr' and what mode change */
				opmodeptr = _ircdmode_create(&opmodes, &lastopmode);
				opmodeptr->cptr = tptr;
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				break;
			case 'k':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					if (MyClient(pptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-k");
					break;
				}
				if (!IsServer(pptr) && !chlptr)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (keychange)
					break;
				else
				{
					register u_char		*s;

					if (IsClient(lptr))
					{
						register u_char		*d;
						register u_char		ch;

						for(d=s=(u_char *)target;(ch=*s);s++)
						{
							ch &= 0x7f;
							if ((ch != ':') && (ch > 0x20))
							{
								*d++ = ch;
							}
						}
						*d = '\0';
					}
					else
					{
						for(s=(u_char *)target;*s;s++)
						{
							*s &= 0x7f;
							if ((*s == ' ') || (*s == '\t') || (*s == '\n'))
							{
								*s = '\0';
								break;
							}
						}
					}
				}
				if (!*target || ((IsClient(lptr) &&
									(opcnt++ >= MAXMODEPARAMS))))
					break;
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (*ptr != got_ovk))
					break;
				keychange = 1;
				newkey = chptr->key;
				if (plusminus == '+')
				{
					if (chptr->key && IsClient(lptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_KEYSET, pptr->name),
							chptr->name);
					else if (!chptr->key || IsServer(lptr))
						newkey = target;
				}
				else
				{
					if (IsServer(lptr) ||
							(chptr->key && !mycmp(chptr->key, target)))
					{
						newkey = NULL;
					}
				}
				break;
			case 'b':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					/* return banlist */
					for(ban=chptr->bans;ban;ban=ban->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_BANLIST, pptr->name),
							chptr->name,
							ban->banstring,
							ban->who,
							ban->bantime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFBANLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&banmodes, &lastbanmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->banstring = target;
				break;
			case 'l':
				tlimit = 0;
				if (plusminus == '+')
				{
					target = ircd_next_subarg_space(&modeargs);
					if (!target || !*target || !(tlimit = atoi(target)))
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-l");
						break;
					}
				}
				if (!IsServer(pptr) && !chlptr)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (limitchange++)
					break;
				limit = tlimit;
				break;
			case 'i':
			case 's':
			case 'p':
			case 't':
			case 'n':
			case 'm':
				if (!IsServer(pptr) && !chlptr)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				for(i=_channel_modes;*i;i+=2)
				{
					if (*(i+1) != *ptr)
						continue;
					if (plusminus == '+')
						newmodes |= *i;
					else
						newmodes &= ~(*i);
				}
				break;
			default:
				if (IsClient(pptr))
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_UNKNOWNMODE, pptr->name), *ptr);
				break;
		}
		if (!*++ptr)
			ptr = ircd_next_subarg_space(&modeargs);
	}

    /* if it's a non-deopped non-op on a TS channel from a TS
    ** server, allow only ops and deops and ignore everything
    ** else (to avoid having desynched ops) -orabidoo
    */
	if (!ischop && isok)
	{
		newmodes = oldmodes;
		limitchange = 0;
		keychange = 0;
		_ircdmode_free(&banmodes);
	}

	plusminus = '\0';

	if (newmodes != oldmodes)
	{
		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & newmodes) && !(*i & oldmodes))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes |= *i;
			}
		}

		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & oldmodes) && !(*i & newmodes))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes &= ~(*i);
			}
		}
	}

	if (limitchange)
	{
		if (limit && (chptr->limit != limit))
		{
			char	buf[20];

			if (plusminus != '+')
			{
				plusminus = '+';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			sprintf(buf, "%-15d", limit);
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=buf;*sptr && *sptr != ' ';)
					*modeargptr++ = *sptr++;
			chptr->limit = limit;
		}
		if (!limit && chptr->limit)
		{
			if (plusminus != '-')
			{
				plusminus = '-';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			chptr->limit = 0;
		}
	}

	/* build +ov */

	for(opmodeptr=opmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		if ((modeargptr-modeargbuf) +
				strlen(opmodeptr->cptr->name) + 2 >= sizeof(modeargbuf))
			break;
		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		chlptr2 = IsMember(chptr, opmodeptr->cptr);
		if (opmodeptr->mode == 'o')
		{
			if (IsServer(pptr))
			{
				send_to_opers("Server %s doing mode %co and removing TS on %s",
					pptr->name, plusminus, chptr->name);
				chptr->creation = 0;
			}

			if (isdeop && (plusminus == '+'))
			{
				if (chlptr2 && !(chlptr2->clptr->modes & CHANNEL_MODES_CHANOP))
					chlptr2->clptr->modes |= CHANNEL_MODES_DEOPPED;
			}
		}
		if (isok)
		{
			isok = _channel_change_opflag(chptr, chlptr2->clptr,
							opmodeptr, IsClient(lptr));
			if (isok || IsServer(lptr))
			{
				*modeptr++ = opmodeptr->mode;
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=opmodeptr->cptr->name;*sptr;)
					*modeargptr++ = *sptr++;
			}
		}
	}
	_ircdmode_free(&opmodes);

	/* +k */

	if (keychange)
	{
		if (chptr->key && !newkey)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(chptr->key) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=chptr->key;*sptr;)
					*modeargptr++ = *sptr++;
				MyFree(chptr->key);
				chptr->key = NULL;
			}
		}
		else if (newkey && newkey != chptr->key)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(newkey) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=newkey;*sptr;)
					*modeargptr++ = *sptr++;
				if (chptr->key)
					MyFree(chptr->key);
				chptr->key = MyStrdup(newkey);
			}
		}
	}

	/* bans */
	for(opmodeptr=banmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->banstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->banstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->banstring,
								user, host, newban, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_ban_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_ban_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&banmodes);

	*modeptr = '\0';
	*modeargptr = '\0';

	if (*modebuf && isok)
	{
		if (IsClient(pptr))
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
		else
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);

		if (*chptr->name != '&')
			send_to_servers(IsServer(lptr) ? lptr : NULL, ":%s MODE %s %s %s",
				prefix, chptr->name, modebuf, modeargbuf);
	}

	return 3;
}

int m_kick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char				*channels = ircd_nextarg(&args, 0);
	char				*nickskeep = ircd_nextarg(&args, 0);
	char				*comment = ircd_nextarg(&args, 0);
	char				*channel;
	char				*nick;
	IRCD_CHANNEL		*chptr;
	char				buf[512];
	char				buf2[512];
	char				*dptr = buf;
	char				*dptr2 = buf2;
	IRCD_CLIENT			*acptr;
	char				nicksbuf[512];
	char				*nicks = nickskeep;
	char				*tptr;
	IRCD_CHANNELLIST	*chlptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "KICK");
		return 3;
	}

	if (!comment)
		comment = pptr->name;

	if (IsServer(pptr))
		send_to_opers("KICK from %s for %s %s", pptr->name,
					channels, nicks);

	for(;;)
	{
		if (!IsServer(lptr))
		{
			if ((channel = channels) == NULL)
				break;
			channels = NULL;
			tptr = strchr(channel, ',');
			if (tptr)
				*tptr = '\0';
		}
		else if ((channel = ircd_next_subarg_comma(&channels)) == NULL)
			break;
		chptr = hash_find_channel(channel);
		if (!chptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
					pptr->name), channel);
			continue;
		}
		if (!IsServer(pptr))
		{
			chlptr = IsMember(chptr, pptr);
			if (!chlptr)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
					chptr->name);
				continue;
			}
			else if (!(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
					chptr->name);
				continue;
			}
		}

		if (*channel != '&')
		{
			if (dptr != buf)
				*dptr++ = ',';
			else
				dptr += strlen(strcpy(dptr, channel));
		}

		support_strncpy(nicksbuf, nickskeep, sizeof(nicksbuf));
		nicks = nicksbuf;

		for(;;)
		{
			if (!IsServer(lptr))
			{
				if ((nick = nicks) == NULL)
					break;
				nicks = NULL;
				tptr = strchr(nick, ',');
				if (tptr)
					*tptr = '\0';
			}
			else if ((nick = ircd_next_subarg_comma(&nicks)) == NULL)
				break;

			if (!(acptr = hash_find_chasing(nick)))
				continue;
			if ((chlptr = IsMember(chptr, acptr)) != NULL)
			{
				if (*channel != '&')
				{
					if (dptr2 != buf2)
						*dptr2++ = ',';
					else
						dptr2 += strlen(strcpy(dptr2, acptr->name));
				}
				send_to_channel_butone(NULL,
					CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
					chptr, pptr, "KICK %s %s :%s", channel,
					acptr->name, comment);
				_channel_del_user(chptr, acptr, chlptr);
				del_channellist_from_list(&(acptr->c_channels), chlptr);
				MyFree(chlptr);
			}
			else
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_USERNOTINCHANNEL, pptr->name),
					nick, channel);
		}
	}

	if ((dptr != buf) && (dptr2 != buf2))
	{
		*dptr = '\0';
		*dptr2 = '\0';
		send_to_servers(lptr, ":%s KICK %s %s :%s",
			pptr->name, buf, buf2, comment);
	}

	return 3;
}

int m_userhost(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*nicks = ircd_nextarg(&args, 0);
	int			maxuhost = 20;
	char		buf[475];
	IRCD_CLIENT *acptr;
	char		*end_ptr;
	char		*this_start_ptr;
	char		*sptr;
	char		*dptr = buf;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 1;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "USERHOST");
		return 1;
	}

	end_ptr = buf + sizeof(buf) - 1;
	*end_ptr = '\0';

	for(;maxuhost-- && nicks && *nicks;nicks=ircd_next_subarg_space(&args))
	{
		if ((acptr = hash_find_client(nicks)) != NULL)
		{
			if (IsClient(acptr))
			{
				this_start_ptr = dptr;

				if (dptr != buf)
					*dptr++ = ' ';

				/*
				** each section is: nick*=+~username@hostname 
				**  no * if not oper, - instead of + if away
				**
				** used to have snprintf/sprintf, but this is more efficient:
				*/

				for(sptr=acptr->name;dptr<end_ptr;sptr++,dptr++)
					if (!(*dptr=*sptr))
						break;

				if (dptr+4 >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}

				if (IsAnOper(acptr))
					*dptr++ = '*';
				*dptr++ = '=';
				*dptr++ = acptr->c_away ? '-' : '+';

				for(sptr=acptr->username;dptr<end_ptr;sptr++,dptr++)
					if (!(*dptr=*sptr))
						break;

				if (dptr+2 >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}

				*dptr++ = '@';

				for(sptr=acptr->hostname;dptr<end_ptr;sptr++,dptr++)
					if (!(*dptr=*sptr))
						break;

				if (dptr >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}
			}
		}
	}

	*dptr = '\0';

	send_prefix_to_client(meptr, pptr,
		IRCD_REPLY(RPL_USERHOST, pptr->name), buf);

	return 1;
}

int m_ison(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*nicks = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;
	char			buf[475];
	char			*dptr = buf;
	char			*sptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 1;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "ISON");
		return 1;
	}

	for(;nicks && *nicks;nicks=ircd_next_subarg_space(&args))
	{
		if ((acptr = hash_find_client(nicks)) != NULL)
		{
			int len = strlen(acptr->name) + 1;

			if (dptr + len < buf + sizeof(buf))
			{
				if (dptr != buf)
					*dptr++ = ' ';
				for(sptr=acptr->name;*sptr;)
					*dptr++ = *sptr++;
			}
		}
	}
	*dptr = '\0';
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ISON, pptr->name), buf);
	return 1;
}

int m_rehash(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!(pptr->c_oline_flags & OLINE_FLAGS_CANREHASH))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to REHASH", pptr->name);
		return 3;
	}

	if (!strcasecmp(args, "counter"))
	{
		Globalinfo->m_totclients = 0;
		Globalinfo->m_totclientstime = Now;
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Counter has been reset.",
				pptr->name);
		return 3;
	}
	else if (!strcasecmp(args, "motd"))
	{
		char	*motd_file = vars_value_get_bynum(IRCD_VAR_MOTD_FILE);

		if (motd_file)
		{
			int ret;

			motd_unload(&Motd);
			ret = motd_load(motd_file, &Motd);
			if (ret < 0)
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Couldn't load %s: %s", pptr->name,
					motd_file, support_strerror(-ret));
			else
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Motd has been reloaded.", pptr->name);
		}
		else
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :No MOTD_FILE specified in /set",
				pptr->name);
		}
		return 3;
	}

	send_prefix_to_client(meptr, pptr, "NOTICE %s :Rehash has been queued.",
				pptr->name);

	if (!RehashWho)
	{
		RehashWho = MyMalloc(CLIENTNAMESIZE);
		ircd_clientname(pptr, 0, RehashWho);
	}
	DoRehash = 1;

	return 3;
}

int m_restart(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_close(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_svinfo(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*ts_cur = ircd_nextarg(&args, 0);
	char	*ts_min = ircd_nextarg(&args, 0);
	char	*standalone = ircd_nextarg(&args, 0);
	char	*thetime = ircd_nextarg(&args, 0);
	char	cnamebuf[CLIENTNAMESIZE];

	if (!IsServer(pptr) || !IsLocal(pptr) || !thetime || !*thetime)
		return 3;

	standalone = NULL; /* standalone is not used, and this stops gcc warning */

	/* pptr == lptr */

	if (!DoesTS(lptr))
	{
		send_to_opers("Link %s dropped, non-TS server",
				ircd_clientname(lptr, 1, cnamebuf));
		client_del(meptr, lptr, "non-TS server");
		return -1;
	}

	if (TS_CURRENT < atoi(ts_min) || atoi(ts_cur) < TS_MIN)
	{
		send_to_opers("Link %s dropped, wrong TS protocol version (%s,%s)",
				ircd_clientname(lptr, 1, cnamebuf), ts_cur, ts_min);
		/* lptr == lptr */
		client_del(meptr, lptr, "wrong TS protocol");
		return -1;
	}
	return 3;
}

/*
** m_sjoin
** parv[0] - sender
** parv[1] - TS
** parv[2] - channel
** parv[3] - modes + n arguments (key and/or limit)
** parv[4+n] - flags+nick list (all in one parameter)
*/
int m_sjoin(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*ts = ircd_nextarg(&args, 0);
	char	*channel = ircd_nextarg(&args, 0);
	char	*modes = ircd_nextarg(&args, 0);
	char	*modeargs = args;
	char	*newkey = NULL;
	int		newlimit = 0;
	char	*tempptr;
	char	*ptr;
	int		newts;
	int		oldts;
	int		isnew = 0;
	int		doesop = 0;
	int		haveops = 0;
	int		keepourmodes = 1;
	int		keepnewmodes = 1;
	int		newmodes;
	int		oldmodes;
	int		tstosend;
	char	plusminus;
	char	modebuf[512];
	char	modeargbuf[512];
	char	*modeptr = modebuf;
	char	*modeargptr = modeargbuf;
	char	largebuf[1024];
	int		people;
	IRCD_CLIENT	*acptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CLIENTLIST	*clptr;
	char	*name;
	int		flags;
	char	*sptr;
	int		pargs;
	int		*i;

	if (!IsServer(pptr))
	{
		if (IsLocal(pptr))
			;
		return 3;
	}

	if (!modeargs || !*modeargs)
		return 3;

	if (*channel == '&')
		return 3;

	newts = atoi(ts);
	chptr = hash_find_channel(channel);

	newmodes = 0;

	for(ptr=modes;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case 'i':
				newmodes |= CHANNEL_MODES_INVITEONLY;
				break;
			case 'm':
				newmodes |= CHANNEL_MODES_MODERATED;
				break;
			case 'n':
				newmodes |= CHANNEL_MODES_NOPRIVMSGS;
				break;
			case 'p':
				newmodes |= CHANNEL_MODES_PRIVATE;
				break;
			case 's':
				newmodes |= CHANNEL_MODES_SECRET;
				break;
			case 't':
				newmodes |= CHANNEL_MODES_TOPICLIMIT;
				break;
			case 'k':
				newkey = ircd_next_subarg_space(&modeargs);
				if (newkey && !*newkey)
					newkey = NULL;
				break;
			case 'l':
				tempptr = ircd_next_subarg_space(&modeargs);
				if (tempptr)
					newlimit = atoi(tempptr);
				break;
		}
	}
			
	if (!chptr)
	{
		isnew = 1;
		chptr = (IRCD_CHANNEL *)MyMalloc(sizeof(IRCD_CHANNEL)+strlen(channel));
		strcpy(chptr->name, channel);
		hash_add_channel(chptr);
		if (++Globalinfo->channels > Globalinfo->maxchannels)
		{
			Globalinfo->maxchannels = Globalinfo->channels;
			Globalinfo->maxchannelstime = Now;
		}
	}
	oldts = chptr->creation;

	if (modeargs && (*modeargs == ':'))
		modeargs++;

	if (modeargs && !isnew && (newts > 0) && (oldts > 0) && (newts != oldts))
	{
		char	*start = modeargs;
		char	save;

		for(ptr=start;;ptr++)
		{
			if (*ptr && *ptr != ' ')
				continue;
			save = *ptr;
			*ptr = '\0';
			if ((*start == '@') || (*(start+1) == '@'))
			{
				if ((*++start == '@') || (*start == '+'))
					start++;
				if (((acptr = hash_find_chasing(start)) != NULL) &&
						(acptr->c_from == lptr))
				{
					*ptr = save;
					doesop = 1;
					break;
				}
			
			}
			start = ptr+1;
			if (save == '\0')
				break;
			else
				*ptr = save;
		}
	}

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			haveops = 1;
			break;
		}
	}

	oldmodes = chptr->modes;

	if (isnew)
		chptr->creation = tstosend = newts;
	else if (!newts || !oldts)
		chptr->creation = tstosend = 0;
	else if (newts == oldts)
		tstosend = oldts;
	else if (newts < oldts)
	{
#ifdef TS5
		keepourmodes = 0;
		chptr->creation = tstosend = newts;
#else
		if (doesop)
		{
			keepourmodes = 0;
			chptr->creation = tstosend = newts;
		}
		else if (haveops)	/* !doesop and we have ops */
		{
			tstosend = oldts;
		}
		else /* !doesop and we don't have ops */
		{
			chptr->creation = tstosend = newts;
		}
#endif
	}
	else
	{
#ifdef TS5
		keepnewmodes = 0;
		tstosend = oldts;
#else
		if (haveops)
		{
			keepnewmodes = 0;
			tstosend = oldts;
		}
		else if (doesop) /* !haveops && doesop */
		{
			chptr->creation = tstosend = newts;
			send_to_opers("Hacked ops on opless channel: %s from %s",
					chptr->name, pptr->name);
		}
		else /* !haveops && !doesop */
			tstosend = oldts;
#endif
	}

	if (!keepnewmodes)
	{
		newmodes = oldmodes;
		newkey = chptr->key;
		newlimit = chptr->limit;
	}
	else if (keepourmodes)
	{
		newmodes |= oldmodes;
		if (chptr->limit > newlimit)
			newlimit = chptr->limit;
		if (strcmp(newkey ? newkey : "", chptr->key ? chptr->key : "") < 0)
			newkey = chptr->key;
	}

	plusminus = '\0';

	if (newmodes != oldmodes)
	{
		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & newmodes) && !(*i & oldmodes))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
			}
		}

		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & oldmodes) && !(*i & newmodes))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
			}
		}
	}

	if (chptr->limit && !newlimit)
	{
		if (plusminus != '-')
		{
			plusminus = '-';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'l';
	}

	if (chptr->key && !newkey)
	{
		if (plusminus != '-')
		{
			plusminus = '-';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'k';
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=chptr->key;*sptr;)
			*modeargptr++ = *sptr++;
	}

	if (newlimit && (newlimit != chptr->limit))
	{
		char buf[20];

		if (plusminus != '+')
		{
			plusminus = '+';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'l';
		sprintf(buf, "%-15d", newlimit);
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=buf;*sptr && (*sptr != ' ');)
			*modeargptr++ = *sptr++;
	}

	if (newkey && (!chptr->key || mycmp(chptr->key, newkey)))
	{
		if (plusminus != '+')
		{
			plusminus = '+';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'k';
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=newkey;*sptr;)
			*modeargptr++ = *sptr++;
	}

	chptr->modes = newmodes;

	if (newkey != chptr->key)
	{
		if (chptr->key)
			MyFree(chptr->key);
		chptr->key = !newkey ? newkey : MyStrdup(newkey);
	}
	if (newlimit != chptr->limit)
		chptr->limit = newlimit;

	if (!keepourmodes)
	{
		char	meplusminus = '\0';
		int		mpargs = 0;
		char	memodebuf[512];
		char	memodeargbuf[512];
		char	*memodeptr = memodebuf;
		char	*memodeargptr = memodeargbuf;

		for(clptr=chptr->members;clptr;clptr=clptr->next)
		{
			if (clptr->modes & CHANNEL_MODES_CHANOP)
			{
				if (meplusminus != '-')
				{
					meplusminus = '-';
					*memodeptr++ = meplusminus;
				}
				*memodeptr++ = 'o';
				if (memodeargptr != memodeargbuf)
					*memodeargptr++ = ' ';
				for(sptr=clptr->cptr->name;*sptr;)
					*memodeargptr++ = *sptr++;
				if (++mpargs == MAXMODEPARAMS)
				{
					*memodeptr = '\0';
					*memodeargptr = '\0';
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, meptr, "MODE %s %s %s", chptr->name,
						memodebuf, memodeargbuf);
					memodeptr = memodebuf;
					memodeargptr = memodeargbuf;
					meplusminus = '\0';
					mpargs = 0;
				}
				clptr->modes &= ~CHANNEL_MODES_CHANOP;
			}
			if (clptr->modes & CHANNEL_MODES_VOICE)
			{
				if (meplusminus != '-')
				{
					meplusminus = '-';
					*memodeptr++ = meplusminus;
				}
				*memodeptr++ = 'v';
				if (memodeargptr != memodeargbuf)
					*memodeargptr++ = ' ';
				for(sptr=clptr->cptr->name;*sptr;)
					*memodeargptr++ = *sptr++;
				if (++mpargs == MAXMODEPARAMS)
				{
					*memodeptr = '\0';
					*memodeargptr = '\0';
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, meptr, "MODE %s %s %s", chptr->name,
						memodebuf, memodeargbuf);
					memodeptr = memodebuf;
					memodeargptr = memodeargbuf;
					meplusminus = '\0';
					mpargs = 0;
				}
				clptr->modes &= ~CHANNEL_MODES_VOICE;
			}
		} /* for() */
		if (memodeptr != memodebuf)
		{
			*memodeptr = '\0';
			*memodeargptr = '\0';
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, meptr, "MODE %s %s %s", chptr->name,
				memodebuf, memodeargbuf);
		}
	} /* (!keepourmodes) */

	if (modeptr != modebuf)
	{
		*modeptr = '\0';
		*modeargptr = '\0';
		if (IsClient(pptr))
		{
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name,
				modebuf, modeargbuf);
		}
		else
		{
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name,
				modebuf, modeargbuf);
		}
		modeptr = modebuf;
		modeargptr = modeargbuf;
	}

	if ((*modes != '0') && keepnewmodes)
		_build_channel_modes(chptr, pptr, modebuf, modeargbuf);
	else
	{
		modebuf[0] = '0';
		modebuf[1] = '\0';
		modeargbuf[0] = '\0';
	}

	tempptr = largebuf;
	tempptr += sprintf(tempptr, ":%s SJOIN %d %s %s %s :",
				prefix, tstosend, chptr->name,
				modebuf, modeargbuf);

	*modeptr++ = '+';

	people = 0;
	pargs = 0;


	for(name = ircd_next_subarg_space(&modeargs);name;
					name = ircd_next_subarg_space(&modeargs))
	{
		char *savename = name;

		flags = 0;
		if (*name == '@' || *(name+1) == '@')
			flags |= CHANNEL_MODES_CHANOP;
		if (*name == '+' || *(name+1) == '+')
			flags |= CHANNEL_MODES_VOICE;

		if (flags && ((*++name == '@') || (*name == '+')))
			name++;

		if (!keepnewmodes)
		{
			if (flags & CHANNEL_MODES_CHANOP)
				flags = CHANNEL_MODES_DEOPPED;
			else
				flags = 0;
		}

		if (!(acptr = hash_find_chasing(name)) || (acptr->c_from != lptr))
			continue;
		people++;

		if (!IsMember(chptr, acptr))
		{
			_channel_add_user(chptr, acptr, flags);
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, acptr, "JOIN :%s", chptr->name);
		}
		if (keepnewmodes)
			strcpy(tempptr, savename);
		else
			strcpy(tempptr, name);
		tempptr += strlen(tempptr);
		*tempptr++ = ' ';
		if (flags & CHANNEL_MODES_CHANOP)
		{
			*modeptr++ = 'o';
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=name;*sptr;)
				*modeargptr++ = *sptr++;
			if (++pargs == MAXMODEPARAMS)
			{
				*modeptr = '\0';
				*modeargptr = '\0';
				if (IsClient(pptr))
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, pptr, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				else
					send_to_channel_butone2(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, prefix, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				pargs = 0;
				modeptr = modebuf;
				modeargptr = modeargbuf;
				*modeptr++ = '+';
			}
		}
		if (flags & CHANNEL_MODES_VOICE)
		{
			*modeptr++ = 'v';
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=name;*sptr;)
				*modeargptr++ = *sptr++;
			if (++pargs == MAXMODEPARAMS)
			{
				*modeptr = '\0';
				*modeargptr = '\0';
				if (IsClient(pptr))
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, pptr, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				else
					send_to_channel_butone2(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, prefix, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				pargs = 0;
				modeptr = modebuf;
				modeargptr = modeargbuf;
				*modeptr++ = '+';
			}
		}
	}

	if (modeptr != (modebuf+1))
	{
		*modeptr = '\0';
		*modeargptr = '\0';
		if (IsClient(pptr))
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
		else
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
	}

	if (people)
	{
		if (*(tempptr-1) == ' ')
			tempptr--;
		*tempptr = '\0';
		send_to_servers(lptr, "%s", largebuf);
	}

	return 3;
}

int m_die(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!(pptr->c_oline_flags & OLINE_FLAGS_CANDIE))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to DIE", pptr->name);
		return 3;
	}

	exit(0);
	return 0; /* so some gccs don't complain */
}

int m_dns(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_gline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsAnOper(pptr) && !IsServer(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (MyClient(pptr) && !(pptr->c_oline_flags & OLINE_FLAGS_CANGLINE))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to GLINE", pptr->name);
		return 3;
	}


	return 3;
}

int m_capab(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*capab = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char		*ptr;
	IRCD_CAPAB	*cptr;

	if (!capab || !*capab)
		return 0;
	while((ptr = ircd_next_subarg_space(&capab)) != NULL)
	{
		for(cptr=_ircd_capabs;cptr->name;cptr++)
			if (!strcmp(cptr->name, ptr))
				lptr->c_capab_flags |= cptr->value;
	}
	return 0;
}

int m_ext(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_operwall(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_wallops(lptr, pptr, prefix, args);
}

int m_kline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*mask;
	char	*user;
	char	*host = NULL;
	char	*reason;
	char	*timelimit = NULL;
	char	*ptr;
	char	tempuser[USERLEN+1];
	int		ret;
	int		expires = -1;
	int		ip_or_host;
	char	buffer[1024];
	char	*filename;
	int		fd;
	CONFHOSTIPENTRY		*entryptr;
	ircd_uint32_t	ipaddr;
	ircd_uint32_t	ipmask;
	int		failed = 1;

	if (!IsAnOper(pptr) && !IsServer(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (MyClient(pptr) && !(pptr->c_oline_flags & OLINE_FLAGS_CANKLINE))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to KLINE", pptr->name);
		return 3;
	}

	mask = ircd_nextarg(&args, 0);

	if (mask && isdigit(*mask))
	{
		char *ptr = mask;

		while(*mask && isdigit(*mask))
			mask++;
		if (!*mask || !*(mask+1))
		{
			timelimit = ptr;
			mask = ircd_nextarg(&args, 0);
		}
		else
			mask = ptr;
	}

	if (!mask || !*mask)
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "KLINE");
		return 3;
	}

	user = mask;

	if (MyClient(pptr))
	{
		reason = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	}
	else
	{
		host = ircd_nextarg(&args, 0);

		if (!host || !*host)
			return 3;

		reason = ircd_nextarg(&args, 0);
	}

	if (!host)
	{
		for(ptr=user;*ptr;ptr++)
			if ((*ptr == '@') || (*ptr == '*') || (*ptr == '.') || (*ptr == '/'))
				break;

		if (*ptr)
		{
			while(*ptr && (*ptr != '@'))
				ptr++;
			if (*ptr)
			{
				*ptr++ = '\0';
				host = ptr;
			}
			else
			{
				host = user;
				user = "*";
			}
			if (!host || !*host)
				host = "*";

			if (strcmp(host, "*") && (1 == cidr_ip2netmask(host, &ipaddr, &ipmask)))
				ip_or_host = 0;
			else
				ip_or_host = 1;
		}
		else
		{
			IRCD_CLIENT		*acptr = hash_find_client(mask);

			if (!acptr)
				acptr = whowas_get_history(mask, 0);

			if (!acptr)
			{
				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOSUCHNICK, pptr->name), mask);
				return 3;
			}

			if (!IsPerson(acptr) || !IsReg(acptr))
			{
				send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Can't KLINE an unregistered connection or a server",
					pptr->name);
				return 3;
			}

			tempuser[0] = '*';
			if (*acptr->username == '~')
				strcpy(tempuser + 1, acptr->username+1);
			else
				strcpy(tempuser + 1, acptr->username);
			user = tempuser;
			host = _cluster(acptr->hostname, &ip_or_host);
		}
	}

	if (!reason || !*reason)
		reason = "No reason specified";

	if (strlen(reason) > 200)
		reason[200] = '\0';

	if (!match(user, ".....") && !match(host, "...ba.."))
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s : Can't K-Line *@*",
			pptr->name);
		return 3;
	}

	if (vars_value_get_bynum(IRCD_VAR_PASS_KLINES))
	{
		send_to_servers(IsServer(lptr) ? lptr : NULL,
				":%s KLINE %s %s :%s", prefix, user, host, reason);
	}

	if (!MyClient(pptr))
	{
		return 3;
	}

	if (timelimit)
	{
		int num = atoi(timelimit);

		if (num <= 0)
		{
			if (MyClient(pptr))
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Invalid expire time given", pptr->name);
			return 3;
		}
		else if (num > 0)
		{
			char *ptr = timelimit;

			while(*ptr && isdigit(*ptr))
				ptr++;
			switch(*ptr)
			{
				case	'd':
				case	'D':
					expires = Now + (num * 86400);
					break;
				case	'm':
				case	'M':
					expires = Now + (num * 60);
					break;
				case	'h':
				case	'H':
					expires = Now + (num * 3600);
					break;
				case	'y':
				case	'Y':
					expires = Now + (num * 86400 * 365);
					break;
				case	'w':
				case	'W':
					expires = Now + (num * 86400 * 7);
					break;
				case	's':
				case	'S':
				default:
					expires = Now + (num * 60);
					break;
			}
		}
	}

	if (IsClient(pptr))
		sprintf(buffer, "kline by %s!%s@%s at %s",
			pptr->name, pptr->username, pptr->hostname,
			support_ctime(Now));
	else
		sprintf(buffer, "kline by %s at %s",
			pptr->name, support_ctime(Now));

	if (ip_or_host)
	{
		ret = access_match_hostname(&(_ircd_conf.deny),
				user, host, 0, &entryptr);
		if (ret >= 0)
		{
			if (IsClient(pptr))
				send_prefix_to_client(meptr, pptr,
				"NOTICE %s :KLINE \"%s@%s\" not added.  Matched by: \"%s@%s\"",
					pptr->name, user, host,
					entryptr->username ? entryptr->username : "*",
					entryptr->hostname ? entryptr->hostname : "*");
			return 3;
		}
	}
	else
	{
		ret = access_match_ipaddr(&(_ircd_conf.deny),
			user, host, ipaddr, 0, &entryptr);
		if (ret >= 0)
		{
			if (IsClient(pptr))
				send_prefix_to_client(meptr, pptr,
			"NOTICE %s :KLINE \"%s@%s\" not added.  Matched by: \"%s@%s\"",
				pptr->name, user, host,
				entryptr->username ? entryptr->username : "*",
				entryptr->ipstring ? entryptr->ipstring : "*");
			return 3;
		}
	}

	filename = vars_value_get_bynum(IRCD_VAR_NEWKLINES_FILE);
	if (!filename)
		filename = Conffile;

	fd = open(filename, O_RDWR|O_APPEND|O_CREAT, 0660);
	if (fd == -1)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Problem opening \"%s\" to write new kline",
			pptr->name, filename);
	}
	else
	{
		static char *_start = "deny { \n\tcomment \"";

		write(fd, _start, strlen(_start));
		write(fd, buffer, strlen(buffer));
		write(fd, "\";\n", 3);

		sprintf(buffer, "\n\tusername \"%s\";\n", user);
		write(fd, buffer, strlen(buffer));
		if (ip_or_host)
			sprintf(buffer, "\thostname \"%s\";\n", host);
		else
			sprintf(buffer, "\tipaddr \"%s\";\n", host);
		write(fd, buffer, strlen(buffer));
		if (expires != -1)
		{
			sprintf(buffer, "\texpire %i;\n", expires);
			write(fd, buffer, strlen(buffer));
		}
		sprintf(buffer, "\treason \"%s\";\n", _reason_fix(reason));
		write(fd, buffer, strlen(buffer));
		if (write(fd, "};\n\n", 4) != 4)
		{
			if (MyClient(pptr))
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Problem writing to \"%s\"",
					pptr->name, filename);
			}
		}
		else
			failed = 0;
		close(fd);
	}

	if (ip_or_host)
		ret = access_add(&(_ircd_conf.deny), user, host, NULL,
			NULL, reason, expires, 0, 0, NULL, buffer);
	else
		ret = access_add(&(_ircd_conf.deny), user, NULL, host,
			NULL, reason, expires, 0, 0, NULL, buffer);

	if (ret)
	{
		if (MyClient(pptr))
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Adding of K-Line to memory failed: %i", pptr->name, ret);
	}
	else
	{
		Rehashed = 1;
		if (MyClient(pptr) || IsServer(pptr))
			send_to_opers("%s added K-Line for [%s@%s]%s%s: %s",
				pptr->name, user, host,
				(expires == -1) ? "" : " expiring at ",
				(expires == -1) ? "" : support_ctime(expires),
				reason);
		else if (IsClient(pptr)) /* should always be true */
			send_to_opers("%s from %s added K-Line for [%s@%s]%s%s: %s",
				pptr->name, pptr->servername, user, host,
				(expires == -1) ? "" : " expiring at ",
				(expires == -1) ? "" : support_ctime(expires),
				reason);
	}

	if (!failed)
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Added K-Line [%s@%s] to configfile",
			pptr->name, user, host);

	return 3;
}

int m_unkline(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*mask;
	IRCD_CLIENT		*acptr;
	char			*user;
	char			*host = NULL;
	static char		_star[] = "*";
	CONFHOSTIPENTRY	*chentry;
	CONFHOSTIPENTRY	*nextchentry;
	CONFIGFILE		*cfptr;
	CONFIGFILE		*cfp;
	CONFIGENTRY		*ce;
	CONFIGENTRY		*ce2;
	char			*username;
	char			*hostname;
	char			*ipstring;
	char			*mmaparea;
	struct stat		sb;
	int				numfound1;
	int				numfound;
	int				exact = 0;
	int				fd;
	struct _tempstuff
	{
		int		startpos;
		int		len;
		struct _tempstuff *next;
	} *tempstuff, *tptr, *nexttptr;

	tempstuff = NULL;

	if (!IsAnOper(pptr) && !IsServer(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (MyClient(pptr) && !(pptr->c_oline_flags & OLINE_FLAGS_CANUNKLINE))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to UNKLINE", pptr->name);
		return 3;
	}

	if (IsServer(lptr))
		return 3;

	mask = ircd_nextarg(&args, 0);

	if (!mask || !*mask)
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "UNKLINE");
		return 3;
	}
	acptr = hash_find_client(mask);
	if (!acptr || IsServer(acptr))
	{
		user = host = mask;
		while(*host && (*host != '@'))
				host++;
		if (!*host)
		{
			host = mask;
			user = _star;
		}
		else
		{
			*host++ = '\0';
			if (!*host)
				host = _star;
		}
		if (!*user)
			user = _star;
	}
	else
	{
		user = acptr->username;
		host = acptr->hostname;
	}

	numfound = 0;
	for(chentry=_ircd_conf.deny.conf_all;chentry;chentry=nextchentry)
	{
		nextchentry = chentry->ch_next;
		if (chentry->username &&
				!strcmp(chentry->username, user) &&
			((chentry->hostname && !strcmp(chentry->hostname, host)) ||
				(chentry->ipstring && !strcmp(chentry->ipstring, host))))
		{
			numfound++;
			exact = 1;
			break;
		}
		if (chentry->username &&
					match(chentry->username, user) &&
					match(user, chentry->username))
			continue;
		if (!(((chentry->ipstring &&
				(!match(chentry->ipstring, host) ||
						!match(host, chentry->ipstring))) ||
				(chentry->hostname &&
				(!match(chentry->hostname, host) ||
						!match(host, chentry->hostname))))))
			continue;
		numfound++;
	}

	if (numfound > (int)vars_value_get_bynum(IRCD_VAR_MAXUNKLINES))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Too many bans matched", pptr->name);
		return 3;
	}

	numfound1 = 0;
	for(chentry=_ircd_conf.deny.conf_all;chentry;chentry=nextchentry)
	{
		nextchentry = chentry->ch_next;
		if (exact)
		{
			if (chentry->username &&
				!strcmp(chentry->username, user) &&
			((chentry->hostname && !strcmp(chentry->hostname, host)) ||
				(chentry->ipstring && !strcmp(chentry->ipstring, host))))
			{
				numfound1++;
				access_del(&_ircd_conf.deny, chentry, 0);
			}
			continue;
		}
		if (chentry->username &&
					match(chentry->username, user) &&
					match(user, chentry->username))
			continue;
		if (!(((chentry->ipstring &&
				(!match(chentry->ipstring, host) ||
						!match(host, chentry->ipstring))) ||
				(chentry->hostname &&
				(!match(chentry->hostname, host) ||
						!match(host, chentry->hostname))))))
			continue;
		numfound1++;
		access_del(&_ircd_conf.deny, chentry, 0);
	}

	if (numfound1)
	{
		if (exact)
			send_to_opers("%s removed %i K-Line%s from memory with exact match of %s@%s",
				pptr->name, numfound1, numfound1 == 1 ? "" : "s", user, host);
		else
			send_to_opers("%s removed %i K-Line%s from memory that match%s %s@%s",
				pptr->name, numfound1, numfound1 == 1 ? "" : "s",
				numfound1 == 1 ? "es" : "", user, host);
	}

	cfptr = cfp = config_load(Conffile);

	if (!cfptr)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :%d ban%s removed from memory",
			pptr->name, numfound1, numfound1 == 1 ? "" : "s");
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Couldn't figure out .conf files to check",
			pptr->name);
		return 3;
	}

	numfound = 0;

    for(;cfptr;cfptr=cfptr->cf_next)
	{
		fd = open(cfptr->cf_filename, O_RDWR);
		if (fd < 0)
			continue;

		if (fstat(fd,  &sb) == -1)
		{
			close(fd);
			continue;
		}

		mmaparea = (char *)mmap(NULL, sb.st_size+1, PROT_READ|PROT_WRITE,
				MAP_SHARED, fd, 0);

		if (mmaparea == (char *)-1)
		{
			close(fd);
			continue;
		}

		tempstuff = NULL;

		for(ce=cfptr->cf_entries;ce;ce=ce->ce_next)
		{
			if (!ce->ce_varname)
				continue;
			if (strcasecmp("deny", ce->ce_varname))
				continue;

			username = hostname = ipstring = NULL;

			for(ce2=ce->ce_entries;ce2;ce2=ce2->ce_next)
			{
				if (!ce2->ce_varname ||
						!ce2->ce_vardata || !*ce2->ce_vardata)
					continue;
				if (!strcasecmp(ce2->ce_varname, "hostname"))
					hostname = ce2->ce_vardata;
				if (!strcasecmp(ce2->ce_varname, "username"))
					username = ce2->ce_vardata;
				if (!strcasecmp(ce2->ce_varname, "ipaddr"))
					ipstring = ce2->ce_vardata;
			}

			if (exact && !(username && !strcmp(username, user) &&
				((hostname && !strcmp(hostname, host)) ||
					(ipstring && !strcmp(ipstring, host)))))
				continue;

			if (username &&
					match(username, user) &&
					match(user, username))
				continue;
			if (!(((ipstring &&
					(!match(ipstring, host) ||
							!match(host, ipstring))) ||
				(hostname &&
					(!match(hostname, host) ||
						!match(host, hostname))))))
				continue;

			if (ce->ce_fileposend <= ce->ce_fileposstart)
			{
				/* sanity */
				continue;
			}

			if (ce->ce_fileposend >= sb.st_size)
				continue;

			if (strncasecmp(mmaparea+ce->ce_fileposstart, "deny", 4)
					|| (mmaparea[ce->ce_fileposend] != ';')) /* odd */
				continue;

			numfound++;

			tptr = (struct _tempstuff *)MyMalloc(sizeof(struct _tempstuff));
			tptr->startpos = ce->ce_fileposstart;
			tptr->len = ce->ce_fileposend - ce->ce_fileposstart + 1;

			while(mmaparea[ce->ce_fileposstart + tptr->len ] == '\n')
				tptr->len++;

			tptr->next = tempstuff;
			tempstuff = tptr;
		}

		if (tempstuff)
		{
			int todelete = 0;

			for(tptr=tempstuff;tptr;tptr=nexttptr)
			{
				nexttptr = tptr->next;

				if (sb.st_size - (tptr->startpos+tptr->len))
					BCOPY(mmaparea + tptr->startpos + tptr->len,
						mmaparea + tptr->startpos,
						sb.st_size - (tptr->startpos + tptr->len));

				todelete += tptr->len;

				MyFree(tptr);

			}

			ftruncate(fd, sb.st_size - todelete);

		}

		if (mmaparea != (char *)-1)
			munmap((void *)mmaparea, sb.st_size+1);
		close(fd);
	}

	config_free(cfp);

	if (numfound)
	{
		if (exact)
			send_to_opers("%s removed %i K-Line%s from disk with exact match of %s@%s",
				pptr->name, numfound, numfound == 1 ? "" : "s", user, host);
		else
			send_to_opers("%s removed %i K-Line%s from disk that match%s %s@%s",
				pptr->name, numfound, numfound == 1 ? "" : "s",
				numfound == 1 ? "es" : "", user, host);
	}

	send_prefix_to_client(meptr, pptr,
		"NOTICE %s :%d ban%s removed from memory and %d from .conf files",
		pptr->name, numfound1, numfound1 == 1 ? "" : "s", numfound);

	return 3;
}

int m_note(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_service(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_squery(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_servlist(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_servset(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

static void _send_var(IRCD_CLIENT *pptr, IRCD_VAR *vptr)
{
	switch(vptr->v_type)
	{
		case IRCD_VAR_TYPE_STRING:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :%s set to %s",
				pptr->name, vptr->v_name,
				vptr->v_value ? vptr->v_value : "<NULL>");
			break;
		case IRCD_VAR_TYPE_INT:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :%s set to %i", pptr->name,
				vptr->v_name, (int)vptr->v_value);
			break;
		case IRCD_VAR_TYPE_BOOL:
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :%s set to %s", pptr->name,
				vptr->v_name, vptr->v_value ? "YES" : "NO");
			break;
	}
}

static void _send_var_to_opers(IRCD_CLIENT *pptr, IRCD_VAR *vptr)
{
	switch(vptr->v_type)
	{
		case IRCD_VAR_TYPE_STRING:
			send_to_opers("%s SET %s to %s",
				pptr->name, vptr->v_name,
				vptr->v_value ? vptr->v_value : "<NULL>");
			break;
		case IRCD_VAR_TYPE_INT:
			send_to_opers("%s SET %s to %i", pptr->name,
				vptr->v_name, (int)vptr->v_value);
			break;
		case IRCD_VAR_TYPE_BOOL:
			send_to_opers("%s SET %s to %s", pptr->name,
				vptr->v_name, vptr->v_value ? "YES" : "NO");
			break;
	}
}

int m_set(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*var = ircd_nextarg(&args, 0);
	IRCD_VAR	*vptr;
	char		*value = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	int			ret;

	if (!IsAnOper(lptr))
	{
		if (MyClient(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!var || !*var)
	{
		for(vptr=vars_value_getall();vptr;vptr=vars_value_getnext(vptr))
		{
			_send_var(pptr, vptr);
		}
		send_prefix_to_client(meptr, pptr, "NOTICE %s :*** End of vars",
				pptr->name);
		return 3;
	}

	if (!value || !*value)
	{
		vptr = vars_find_byname(var);
		if (!vptr)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Unknown var: %s", pptr->name, var);
			return 3;
		}
		_send_var(pptr, vptr);
		return 3;
	}

	if (!(pptr->c_oline_flags & OLINE_FLAGS_CANSET))
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :You are not allowed to SET", pptr->name);
		return 3;
	}

	vptr = vars_find_byname(var);
	if (!vptr)
	{
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** Unknown var: %s", pptr->name, var);
		return 3;
	}

	if (vptr->v_type == IRCD_VAR_TYPE_INT)
		value = (void *)atoi(value);
	else if (vptr->v_type == IRCD_VAR_TYPE_BOOL)
	{
		if (!value || !*value || !strncasecmp(value, "yes", 3) || !strncasecmp(value, "on", 2))
			value = (void *)1;
		else if (!strncasecmp(value, "no", 2) || !strncasecmp(value, "off", 3))
			value = (void *)0;
		else
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** SET failed: %s can only be set to YES or NO",
					pptr->name, vptr->v_name);
			return 3;
		}
	}

	ret = vars_value_set_byname(pptr, var, value);

	if (ret)
	{
		send_prefix_to_client(meptr, pptr,
				"NOTICE %s :*** SET failed: %i (%s)", pptr->name, -ret,
				support_strerror(-ret));
		return 3;
	}

	_send_var(pptr, vptr);
	_send_var_to_opers(pptr, vptr);

	return 3;
}

int m_queue(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*command = ircd_nextarg(&args, 0);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!command || !*command)
	{
		int	queue_size;
		int	queue_fl_size;
		int queue_malloced_size;

		queue_get_stats(&queue_size, &queue_fl_size, &queue_malloced_size);
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queue space in use: %i out of %i allocated",
				pptr->name, queue_size, queue_malloced_size - queue_fl_size);
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queue space in freelist: %i",
				pptr->name, queue_fl_size);
		return 3;
	}

	if (!strcasecmp(command, "flush"))
	{
		queue_cleanup();
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queues flushed.",
				pptr->name);
	}
	else
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Unknown queue command.",
				pptr->name);

	return 3;
}

int m_flags(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_USERFLAG		*ufptr;
	IRCD_USERFLAG		*ufptrkeep;
	char				*flag;
	char				plusminus = '+';
	char				buffer[255];
	char				*ptr = buffer;
	int					len = 0;
	int					oldflags = 0;
	int					flaglen;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (!IsClient(lptr))
		return 3;

	oldflags = pptr->c_nflags;

	if (args && *args == ':')
		args++;

	if (args && *args)
	{
		while((flag = ircd_next_subarg_space(&args)) != NULL)
		{
			if ((*flag == '+') || (*flag == '-'))
				plusminus = *flag++;
			if (!strcasecmp(flag, "all"))
			{
				if (plusminus == '-')
				{
					pptr->c_nflags = 0;
					continue;
				}
				for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
					if (IsAnOper(pptr) || !ufptr->opersonly)
						pptr->c_nflags |= ufptr->flag;
				continue;
			}

			ufptrkeep = NULL;
			flaglen = strlen(flag);

			for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
			{
				if (!strncasecmp(ufptr->flagstr, flag, flaglen))
				{
					if (ufptrkeep)
						break;
					ufptrkeep = ufptr;
				}
			}
			if (!ufptrkeep || ufptr->flagstr)
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Unknown flag: %s", pptr->name, flag);
				pptr->c_nflags = oldflags; /* restore flags */
				return 3;
			}
			ufptr = ufptrkeep;
			if (plusminus == '+')
			{
				if (ufptr->opersonly && !IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Need to be Oper to set %s",
						pptr->name, ufptr->flagstr);
					continue;
				}
				pptr->c_nflags |= ufptr->flag;
			}
			else
				pptr->c_nflags &= ~ufptr->flag;
		}
	}

	for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
	{
		if (!(lptr->c_nflags & ufptr->flag))
			continue;
		len = strlen(ufptr->flagstr);
		if (len + 1 + (ptr-buffer) > 250)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current flags: %s", pptr->name, buffer);
			ptr = buffer;
		}
		if (ptr != buffer)
			*ptr++ = ' ';
		strcpy(ptr, ufptr->flagstr);
		ptr += len;
	}
	if (ptr != buffer)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Current flags: %s", pptr->name, buffer);
	}
	ptr = buffer;
	for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
	{
		if ((lptr->c_nflags & ufptr->flag))
			continue;
		if (ufptr->opersonly && !IsAnOper(pptr))
			continue;
		len = strlen(ufptr->flagstr);
		if (len + 1 + (ptr-buffer) > 250)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current missing flags: %s", pptr->name, buffer);
			ptr = buffer;
		}
		if (ptr != buffer)
			*ptr++ = ' ';
		strcpy(ptr, ufptr->flagstr);
		ptr += len;
	}
	if (ptr != buffer)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Current missing flags: %s", pptr->name, buffer);
	}
	return 3;
}

int m_iphash(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return iphash_command(lptr, pptr, args);
}

int m_rejecthold(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_REJECTHOLD	*rptr;
	int				num = (int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if ((rptr = Rejectholds) != NULL)
	{
		for(;rptr;rptr=rptr->rh_next)
		{
			ircd_uint32_t nipaddr = htonl(rptr->rh_rem_ipaddr);
			int e = (num <= 0) ? -1 : rptr->rh_firsttime - Now + num;

			if (e > 0)
				send_prefix_to_client(meptr, pptr, "NOTICE %s :%s put on hold %i second%s ago.  Expires in %i second%s.",
					pptr->name, inet_ntoa(*((struct in_addr *)&nipaddr)),
					Now - rptr->rh_firsttime, Now - rptr->rh_firsttime == 1 ? "" : "s",
					e, e == 1 ? "" : "s");
			else
				send_prefix_to_client(meptr, pptr, "NOTICE %s :%s put on hold %i second%s ago.  Entry expired.",
					pptr->name, inet_ntoa(*((struct in_addr *)&nipaddr)),
					Now - rptr->rh_firsttime, Now - rptr->rh_firsttime == 1 ? "" : "s");
		}
	}
	else
		send_prefix_to_client(meptr, pptr, "NOTICE %s :There are currently no IPs on REJECT_HOLD", pptr->name);
	return 3;
}

